:- module(_, [], [ciaobld(bundlehooks)]).

:- doc(title,  "Bundle Hooks for the Timing Model").
% Timingmodel (a simple WAM for time analysis)

:- use_module(library(bundle/bundle_paths), [bundle_path/3]).
:- use_module(ciaobld(messages_aux), [normal_message/2]).

% TODO: split into two bundles? miniprolog and timingmodel

:- use_module(library(system), [copy_file/2, copy_file/3, file_exists/1]).

'$builder_hook'(prepare_build_bin) :-
	% do_timingmodel % TODO: must be called explicitly using custom_run
	% (needed even if miniprolog is not compiled)
	copy_mp_auto.

% TODO: Show help: "Benchmark and estimate timing model for 'miniprolog'"
%
% TODO: rename by 'generate_model'?
% TODO: (called from 'ciao_builder', called from 'ciaobot' as a test)
'$builder_hook'(custom_run(gen_timingmodel, [])) :- !,
	do_timingmodel,
	copy_mp_auto.

timingmodel_cmd := bench|estimate.

do_timingmodel :-
	normal_message("compiling miniprolog engine", []),
	invoke_gmake_miniprolog(all),
	copy_file(~bundle_path(timingmodel, 'src/miniprolog/bin/timingmodel_auto.pl'),
	          ~bundle_path(timingmodel, 'src/timingmodel_pre.pl'),
		  [overwrite]),
	%
	normal_message("generating timing model for miniprolog", []),
	( % (failure-driven loop)
	  timingmodel_cmd(Cmd),
	    invoke_gmake_timingmodel(Cmd),
	    fail
	; true
	).

% This extra step is to ensure the generation of timingmodel_auto.pl
% even if miniprolog has not been configured
copy_mp_auto :-
	Orig = ~bundle_path(timingmodel, 'src/timingmodel_pre.pl'),
	( file_exists(Orig) ->
	    copy_file(Orig,
	              ~bundle_path(timingmodel, 'src/timingmodel_auto.pl'),
		      [overwrite])
	; true
	).

benchmp :-
	normal_message("running timing model benchmarks", []),
	invoke_gmake_timingmodel('bench').

estimatemp :-
	normal_message("running timing model estimate", []),
	invoke_gmake_timingmodel('estimate').

% TODO: Show help: "Clean results of gen_timingmodel"
% TODO: rename to 'clean_model'?
'$builder_hook'(custom_run(clean_timingmodel, [])) :- !,
	normal_message("cleaning timingmodel", []),
	invoke_gmake_miniprolog(clean),
	invoke_gmake_timingmodel(clean).

:- use_module(engine(system_info), [get_platform/1]).
:- use_module(ciaobld(builder_aux), [invoke_gmake/2]).

invoke_gmake_miniprolog(Cmd) :-
	invoke_gmake(~bundle_path(timingmodel, 'src/miniprolog'),
	             ['-s', '-j1',
		      ~atom_concat('MPARCH=', ~get_platform),
		      Cmd]).

invoke_gmake_timingmodel(Cmd) :-
	invoke_gmake(~bundle_path(timingmodel, 'src'),
	             ['-s', '-j1',
		      ~atom_concat('MPARCH=', ~get_platform),
		      Cmd]).
