:- module(_, [], [ciaobld(bundlehooks)]).

:- doc(title,  "Bundle Hooks for CiaoPP").

% ===========================================================================

:- use_module(library(bundle/bundle_flags), [get_bundle_flag/2]).

% ===========================================================================

% '$builder_hook'(item_nested(ilciao)).
% :- include(.('ilciao.hooks')).
% :- include(.('resources.hooks')).

% ============================================================================

% % Prepare source for build
% % (e.g., for automatically generated code, foreign interfaces, etc.)
% '$builder_hook'(prepare_build_bin) :-
%         create_ppl_interface_file,
% 	create_mathematica_interface_file,
%         % create_llvm_interface_file, % TODO: create conditional compilation flags (i.e., make ciao_llvm dependency optional)
% 	prepare_build_resources.

% top-level and command line
'$builder_hook'(item_nested(cmds)).
'$builder_hook'(cmds:cmd('ciaopp', [main='cmds/ciaoppcl'])).

% ============================================================================

% :- include(.('ppl_interface.hooks')).
% :- include(.('mathematica_interface.hooks')).

% ===========================================================================
:- doc(section, "Tests and Benchmarks").
% TODO: Add bundle defs for unit tests, integration tests, regression
%   tests, etc.

% :- use_module(ciaobld(ciaoc_aux), [runtests_dir/2]).

% % TODO: split resources tests from the rest of tests
% '$builder_hook'(test) :- !,
% 	% TODO: incomplete (should be 'src')
% 	runtests_dir(ciaopp, 'resources'),
% 	runtests_dir(ciaopp, 'infercost'),
% 	% TODO: move as a item_nested for 'ciaopptests'
% 	ciaopptests_hook,
% 	% TODO: move as a item_nested for 'resources'
% 	resources_runtests.

% :- use_module(library(system), [working_directory/2]).

% % (regression tests, see documentation for details)
% ciaopptests_hook :-
% 	working_directory(ThisDir, ~bundle_path(ciaopp, 'tests')),
% 	invoke_ciaosh_batch([
% 	  use_module(ciaopp_tests(generate_benchmarks)),
% 	  generate_benchmarks:generate,
% 	  use_module(ciaopp_tests(ciaopp_test)),
% 	  ciaopp_test:main
% 	]),
% 	working_directory(_, ThisDir).


:- doc(section, "Hooks for caching library assertions").
% This precompilation of the library assertions allows faster module
% loading in CiaoPP (libraries are not checked, the cached data is
% used).

% Warning: This hook only caches the libraries.
% To enable the lib cache execute:
%   ?- use_module(library(persdb/datadir), [ensure_datadir/2]).
%   ?- use_module(ciaopp(p_unit/p_asr), [load_lib_sources/1]).
%   ?- ensure_datadir('ciaopp_lib_cache', _Dir), load_lib_sources(_Dir).

'$builder_hook'(cmds:cmd('gen_lib_cache', [main='cmds/gen_lib_cache'])).
'$builder_hook'(custom_run(cache_libraries, [])) :- !,
	cache_libraries.

:- use_module(library(bundle/bundle_paths)).
:- use_module(ciaobld(config_common), [cmd_path/4]).
:- use_module(ciaobld(cpx_process), [cpx_process_call/3]).

:- use_module(library(persdb/datadir), [ensure_datadir/2]).

db_data_dir(Dir) :-
	ensure_datadir('ciaopp_lib_cache', Dir).

cache_libraries :-
	db_data_dir(Dir),
	bundle_path(ciaopp, 'src/p_unit', LibFakeDir),
	cmd_path(ciaopp, plexe, 'gen_lib_cache', CmdPath),
	cpx_process_call(CmdPath, [Dir], [cwd(LibFakeDir)]).
