% (included file)

:- doc(section, "Resource Analysis Bundle").

:- use_module(ciaobld(messages_aux), [normal_message/2]).

%% % TODO: Add as help for custom_run on this bundle
%%
%% 	help_section("CiaoPP"),
%% 	showcmd(calibrateciaopp, "ciaopp", [
%%           %1_______________________________________________
%%           "Calibrate for resource analysis (for 'ciaopp')"
%%         ]),
%% 	showcmd(save_calibrateciaopp, "ciaopp", [
%%           %1_______________________________________________
%%           "Save the current calibration results (for 'ciaopp')"
%%         ]),
%% 	showcmdsep.

% TODO: create a hook	
prepare_build_resources :-
	time_analysis_auto,
	load_calibration_results.

% ---------------------------------------------------------------------------

:- use_module(ciaobld(ciaoc_aux), [invoke_ciaosh_batch/1]).

time_analysis_auto :-
	normal_message("generating auto created modules for res_wamcode", []),
	%
	invoke_ciaosh_batch([
	  use_module(predefres(res_wamcode/res_wamcode_gen)),
	  res_wamcode_gen:res_wamcode_gen,
	  halt
        ]).

% ---------------------------------------------------------------------------

:- use_module(ciaobld(ciaoc_aux), [invoke_ciaosh_batch/1]).

:- use_module(library(llists), [flatten/2]).
:- use_module(library(lists),  [append/3]).

load_calibration_results :-
	% TODO: 'loading' just copy files here
	normal_message("loading previous profiling results", []),
	In = [
          use_module(ciaopp_tests(resources/resources_tests_utils)),
	  resources_tests_utils:load_profiling_results
	],
	do_calibrate(~gen_load_results_cmd(In)).

% (hook)
'$builder_hook'(custom_run(save_calibrateciaopp, [])) :- !,
	save_calibration_results.

save_calibration_results :-
	normal_message("saving current profiling results", []),
	In = [
          use_module(ciaopp_tests(resources/resources_tests_utils)),
	  resources_tests_utils:save_profiling_results
	],
	do_calibrate(~gen_load_results_cmd(In)).

% (hook)
'$builder_hook'(custom_run(calibrateciaopp, [])) :- !,
	gen_calibrate.

gen_calibrate :-
	normal_message("calibration of execution time analysis", []),
	do_calibrate(~gen_calibrate_cmd).

gen_calibrate_cmd := ~flatten([
		~gen_common_cmd,
		message(note, 'Platform independent step, analyzing calibrators'),
		make_po(profcost(caliresource/caliresource_indep)),
		use_module(profcost(caliresource/caliresource_indep)),
		analyze_calibration_tests,
		make_po(profcost(caliresource/caliresource_cost)),
		message(note, 'Platform dependent step, generating cost models'),
		use_module(profcost(caliresource/caliresource_dep)),
		fill_all_model_params,
		halt
	    ]).

gen_load_results_cmd(In) := ~flatten([
	        In,
		~gen_common_indep_cmd,
		use_module(profcost(caliresource/caliresource_dep_utils)),
		res_exectime_hlm_gen_all,
		halt
	    ]).

gen_data_table_cmd := ~flatten([~gen_common_cmd, halt]).

gen_common_cmd := ~append(~gen_common_indep_cmd, ~gen_common_dep_cmd).

gen_common_indep_cmd([
                use_module(engine(messages_basic), [message/2]),
		message(note, 'Platform independent step, generating calibrators'),
		use_module(profcost(caliresource/caliresource_gen)),
		generate_calibration_tests,
		generate_calibration_measures(5),
		message(note, 'Platform independent step for arithmetics'),
		use_module(predefres(res_arith/res_arith_gen)),
		res_arith_gen_indep
	    ]).

gen_common_dep_cmd([
                use_module(engine(messages_basic), [message/2]),
		message(note, 'Platform dependent step, measuring times'),
		use_module(profcost(caliresource/caliresource_gen_data)),
		% for papers, 250, but to save time set to 100 iterations
		generate_data_table(100, ticks),
		message(note, 'Platform dependent step, timing model for arithmetics'),
		use_module(profcost(calibuil)),
		calibrate_builtins(100, ticks),
		res_arith_gen_dep
	    ]).

do_calibrate(GenCommands) :-
	Cmds = [use_module(engine(runtime_control), [set_prolog_flag/2]),
                set_prolog_flag(read_assertions,      no),
		set_prolog_flag(runtime_checks,       no)|GenCommands],
	invoke_ciaosh_batch(Cmds).

% ---------------------------------------------------------------------------

resources_runtests :-
	% TODO: make sure that those dirs are outside 'src'
	resources_load_profiling_results,
	runtests_dir(ciaopp, 'tests/resources'), % TODO: fails...
	runtests_dir(ciaopp, 'examples/resources').

:- use_module(ciaobld(ciaoc_aux), [invoke_ciaosh_batch/1]).
% Fill data in
%   ./ciaopp/resources/predefres/res_exectime_hlm/auto/
% from
%   ./ciaopp/resources/saved_results
resources_load_profiling_results :-
	invoke_ciaosh_batch([
          use_module(ciaopp_tests(resources/resources_tests_utils)),
	  resources_tests_utils:load_profiling_results
	]).

