% (included file)
% ---------------------------------------------------------------------------
% Auto-interface for Java programs

:- export(customize_java/1).
:- export(customize_and_preprocess_java/1).
% :- export(again_java/1).

% ---------------------------------------------------------------------------
% PLEASE READ: The following predicates generate a menu for preprocessing
% of Java programs loaded directly from Emacs. For now, the only option is
% 'analyze' since few Java analyses and options are available. The menu is
% still very naive and the predicates: customize_java/1,
% customize_and_preprocess_java/1, exec_auto_java/2, again_java/2,
% auto_java_analyze/1, auto_java_analyze/2 are almost duplicates of their
% 'ciao' counterparts but I prefer to maintain these copies to avoid
% side-effects and other problems. -JNL
% ---------------------------------------------------------------------------

:- pred customize_java(X)

# "Customize is used for change the values of a set of flags in case of
  java analysis. So far, the value of @var{X} is only @tt{'all'}.".

customize_java(all) :- !,
	ask_use_config(USE_CONFIG, Bool),
	( USE_CONFIG == none ->
	    menu(javall, Bool),
	    ask_save_menu
	; note_message("Restoring ~w Menu Configuration...", [USE_CONFIG]),
	  restore_menu_config(USE_CONFIG)
	).

customize_java(A) :-
	error_message("Option ~w not customizable", [A]).

:- pred customize_and_preprocess_java(File)
   # "It is like doing @tt{customize_java(all)}, and then calling with
      @var{File} as argument.".

customize_and_preprocess_java(File) :-
	atom(File),
	customize_java(all),
	set_last_file(File),
	again_java.

:- pred again_java/0
   # "Performs the last actions done by
      @pred{customize_and_preprocess_java/1}, on the last file
      previously processed.".

again_java :-
	get_last_file(File),
	get_menu_flag(javall, inter_javall, NM),
	exec_auto_java(NM, File).

exec_auto_java(analyze, F) :- !,
	auto_interface:auto_java_analyze(F).

:- push_prolog_flag(multi_arity_warnings, off).

:- pred auto_java_analyze(F).

auto_java_analyze(File) :-
	auto_interface:auto_java_analyze(File, _).

:- pred auto_java_analyze(F, OFile)

# "Same as @pred{auto_java_analyze/1} but the output file will be generated
  with @var{OFile} name.".

auto_java_analyze(File, OFile) :-
	with_menu_flags(javanal, auto_java_analyze_(File, OFile)).

auto_java_analyze_(File, OFile) :-
	% NOTE: flags should be set from the menu before loading the
	% module since the transformation reads the flag
	% java_analysis_level. -MKL
	module(File),
	get_menu_flag(javanal, inter_javana, LIST),
	exec_analysis_list(LIST, ana),
	% 	get_menu_flag( ana , vers , VERS ) ,
	% 	(VERS == on -> transform( vers ) ; true ),
	do_output(OFile),
	set_last_file(File).

:- pop_prolog_flag(multi_arity_warnings).

javall_tr(analyze, javanal).
javall_tr(A,       A).

javall_menu_branch(A, B) :-
	member(inter_javall = Br,  A),
	member(menu_java_level= L, A),
	javall_tr(Br, BrT),
	menu_branch(A, BrT, ~menu_level_tr(L), B).

javall   , 'Select Menu Level'   # menu_java_level - naive.
javall   , 'Select Action Group' # inter_javall    - analyze :: javall_menu_branch.
javanal  , 'Select Analysis Level'         # java_analysis_level - bytecode <- javana.
% Changed modes/types to shfr/eterms in the following to fix Menu for java.. Quick fix  -UL
%javanal  , 'Select Aliasing-Mode Analysis' # modes       - none      <- javana.
%javanal  , 'Select Shape-Type Analysis'    # types       - none      <- javana.
javanal  , 'Select Aliasing-Mode Analysis' # modes       - shfr      <- javana.
javanal  , 'Select Shape-Type Analysis'    # types       - eterms      <- javana.
javanal  , 'Select Resource Analysis'      # java_ana_cost    - resources <- javana.
javanal  , 'Multivariant Success'          # multi_success    - off       <- javana.
javanal  , 'Print Program Point Info'      # pp_info          - off       <- javana.
javanal  , 'Collapse AI Info'              # collapse_ai_vers - on        <- javana.
javanal(1)  , 'Select Recurrence Solver'   # req_solver - builtin        <- javana.
javanal(1)  , 'Select Algebraic System'    # math_system - builtin        <- javana.
javanal  , 'Select Output Language'        # output_lang - source   <-javana.

guard javana(X) :-
	member(inter_javall=I, X),
	I == analyze.

% 	member( prog_lang = P    , X ),
% 	P == java.
