% (included file)
% ---------------------------------------------------------------------------
% Auto-interface for XC programs

:- export(customize_xc/1).
:- export(customize_and_preprocess_xc/1).
:- export(again_xc/1).

% ---------------------------------------------------------------------------
% PLEASE READ: The following predicates generate a menu for preprocessing
% of XC programs loaded directly from Emacs. For now, the only option is
% 'analyze' since few XC analyses and options are available. The menu is
% still very naive and the predicates: customize_xc/1,
% customize_and_preprocess_xc/1, exec_auto_xc/3, again_xc/1,
% auto_analyze_xc/1, auto_analyze_xc/2 are almost duplicates of their
% 'ciao' counterparts but I prefer to maintain these copies to avoid
% side-effects and other problems.
% ---------------------------------------------------------------------------

:- pred customize_xc(X) 
   # "Customize is used for changing the values of a set of flags for
      xc analysis. So far, the value of @var{X} is only @tt{'all'}.".

customize_xc(all) :- !,
	ask_use_config(USE_CONFIG, Bool),
	( USE_CONFIG == none ->
	    menu(xcll, Bool),
	    ask_save_menu
	; note_message("Restoring ~w Menu Configuration...", [USE_CONFIG]),
	  restore_menu_config(USE_CONFIG)
	).

customize_xc(A) :-
	error_message("Option ~w not customizable", [A]).

:- pred customize_and_preprocess_xc(File)
   # "It is like doing @tt{customize_xc(all)}, and then calling with
      @var{File} as argument.".

customize_and_preprocess_xc(File) :-
	atom(File),
	customize_xc(all),
	set_last_file(File),
	again_xc(_).

:- pred again_xc/1
   # "Performs the last actions done by
      @pred{customize_and_preprocess_xc/1}, on the last file
      previously processed.".

again_xc(Output) :-
	get_last_file(File),
	get_menu_flag(xcll, inter_xcll, NM),
	exec_auto_xc(NM, File, Output).

exec_auto_xc(analyze, F, Output) :- !,
	auto_interface:auto_analyze_xc(F, Output).
exec_auto_xc(check_assertions, F, Output) :- !,
	auto_interface:auto_check_assert_xc(F, Output).
exec_auto_xc(none, F, Output) :- !,
	auto_interface:auto_none_xc(F, Output).

:- push_prolog_flag(multi_arity_warnings, off).

:- pred auto_check_assert_xc(F)
   # "Check the assertions in file @var{F}, with the current options,
      giving errors if assertions are violated (use
      @tt{customize(check_assertions)} to change these options).".

auto_check_assert_xc(File) :-
	auto_check_assert_xc(File, _).

:- pred auto_check_assert_xc(F, OFile)
   # "Same as @pred{auto_check_assrt_xc/1} but the output file will be
      @var{OFile}.".

auto_check_assert_xc(File, OFile) :-
	Flags = [
	    ctchecks_intervals = on,
	    use_check_assrt = off % TODO:[new-resources]
	],
	with_menu_flags(xcnal, with_pp_flags(Flags, auto_check_assert_xc_(File, OFile))).

auto_check_assert_xc_(File, OFile) :-
	% flags should be set form the menu before loading the module
	% since the transformation reads the flag xc_layer. -ULQ
	module(File),
	get_menu_flag(xcnal, inter_xcna, LIST),
	exec_analysis_list(LIST, ana),
	acheck,
	% 	get_menu_flag( ana , vers , VERS ) ,
	% 	(VERS == on -> transform( vers ) ; true ),
	do_output(OFile),
	set_last_file(File).

%exec_analysis_list_acheck(File,LIST,ANYERROR)
:- pred auto_none_xc(F).

auto_none_xc(File) :-
	auto_interface:auto_none_xc(File, _).

:- pred auto_none_xc(F, OFile)
   # "Same as @pred{auto_none_xc/1} but the output file will be
      generated with @var{OFile} name.".

auto_none_xc(File, OFile) :-
	with_menu_flags(xcnal, auto_none_xc_(File, OFile)).

auto_none_xc_(File, OFile) :-
	% flags should be saved before loading the module since the
	% transformation reads the flag xc_layer.
	module(File),
	do_output(OFile),
	set_last_file(File).

:- pred auto_none_xc(F).

auto_analyze_xc(File) :-
	auto_interface:auto_analyze_xc(File, _).

:- pred auto_analyze_xc(F, OFile)
   # "Same as @pred{auto_analyze_xc/1} but the output file will be
      generated with @var{OFile} name.".

auto_analyze_xc(File, OFile) :-
	with_menu_flags(xcnal, auto_analyze_xc_(File, OFile)).

auto_analyze_xc_(File, OFile) :-
	% flags should be saved before loading the module since the
	% transformation reads the flag xc_layer. -ULQ
	module(File),
	get_menu_flag(xcnal, inter_xcna, LIST),
	exec_analysis_list(LIST, ana),
	% 	get_menu_flag( ana , vers , VERS ) ,
	% 	(VERS == on -> transform( vers ) ; true ),
	do_output(OFile),
	set_last_file(File).

:- pop_prolog_flag(multi_arity_warnings).

xcll_tr(analyze,          xcnal).
xcll_tr(check_assertions, xcnal).
xcll_tr(A,                A).

xcll_menu_branch(A, B) :-
	member(inter_xcll = Br,  A),
	member(menu_xc_level= L, A),
	xcll_tr(Br, BrT),
	menu_branch(A, BrT, ~menu_level_tr(L), B).

% all_tr( optimize         , opt   ).
% all_tr( analyze          , ana   ).
% all_tr( check_assertions , check ).
% all_tr( A                , A     ).

% all_menu_branch( A , B ) :-
% 	member( inter_all=Br , A ),
% 	member( menu_level=L , A ),
% 	all_tr( Br, BrT ),
% 	( BrT = check_certificate -> A = B ;
% 	    menu_branch( A , BrT , ~menu_level_tr(L) , B )).

xcll,  'Select Menu Level'        # menu_xc_level - naive.
xcll,  'Select Action Group'      # inter_xcll - analyze :: xcll_menu_branch.
%xcnal  , 'Select Aliasing-Mode Analysis' # modes       - none      <- xcna.
%xcnal  , 'Select Shape-Type Analysis'    # types       - none      <- xcna.
xcnal, 'Select Resource Analysis' # xc_ana_cost - res_plai <- xcna.
%xcnal  , 'Include Energy Model'          # xc_energy_cost - yes      <- xcna.
%xcnal  , 'Multivariant Success'          # multi_success    - off       <- xcna.
%xcnal  , 'Print Program Point Info'      # pp_info          - off       <- xcna.
%xcnal  , 'Collapse AI Info'              # collapse_ai_vers - on        <- xcna.
xcnal(1), 'Select Recurrence Solver' # req_solver - builtin <- xcna.
xcnal(1), 'Select Algebraic System'  # math_system - builtin <- xcna.
xcnal,  'Select Analysis Level'   # xc_layer - isa <-xcna.
% TODO:[new-resources] was 'off' in new-resources (it was changed in master before the merge)
% xcnal  , 'Report Non-Verified Assrts'    # ass_not_stat_eval - off      <-xcna.
xcnal,  'Report Non-Verified Assrts' # ass_not_stat_eval - warning <-xcna.
%xcnal  , 'Select Output Language'        # output_lang - intermediate   <-xcna.
xcnal,  'Select Output Language'  # output_lang - source <-xcna.

guard xcna(X) :-
	member(inter_xcll=I, X),
	I == analyze.
guard xcna(X) :-
	member(inter_xcll=I, X),
	I == check_assertions.
