:- module(_, [main/1], [assertions, hiord, datafacts]).

:- doc(title, "@tt{ciaopp-dump-cmp}: Semantic dump comparator command").

:- doc(author, "Isabel Garcia-Contreras").

:- doc(stability, devel).

:- doc(summary, "This command compares semantically analysis results.").

:- doc(module, "This command compares semantically the results of an
analysis of two given paths. The paths can be either files or
directories. If the paths are directories, each of the files in one directory
will be compared with the files in the other directory with the same
name.

Currently, only the files with extension @tt{.dump}, @tt{.reg} or
@tt{.inc_reg} will be checked, the rest will be omitted.

@subsection{Usage}

@begin{verbatim}
$ ciaopp-dump-cmp <base_path> <test_path> <domain>
@end{verbatim}
").

:- use_module(engine(io_basic)).
:- use_module(library(format), [format/3]).
:- use_module(library(aggregates), [findall/3]).
:- use_module(library(system), [file_exists/1, directory_files/2, file_property/2]).
:- use_module(library(pathnames), [path_concat/3, path_split/3]).
:- use_module(library(sort), [sort/2]).

:- import(ciaopp_bench_manager,
	[directory_dir/3]).
:- import(compare_dump,[
checking_domain/1 ,		 
compare_dumps_auto_detect_db/6 ,	 
print_diff/2 
	]).

main([Dir1, Dir2, AbsInt]) :- !,
	evaluate_results(Dir1, Dir2, AbsInt).
main(_) :-
	display('Usage: ./compare_dumps_dir <base_path> <test_path> <domain>'), nl.

:- pred evaluate_results(Path1, Path2, AbsInt) : atm * atm * atm.
evaluate_results(Path1, Path2, AbsInt) :-
	set_fact(checking_domain(AbsInt)), % TODO: should be written in the dump
	( file_property(Path1, type(regular)) ->
	    compare_files(Path1, Path2, AbsInt)
	;
	    compare_dirs(Path1, Path2, AbsInt)
	).

compare_files(F1, F2, AbsInt) :-
	path_split(F1, Dir1, File1),
	path_split(F2, Dir2, File2),
 	dumpfile_checker_loop([File1], [File2], Dir1, Dir2, AbsInt).
compare_dirs(Dir1, Dir2, AbsInt) :-
	directory_files(Dir1, Fs1),
	directory_files(Dir2, Fs2),
	dumpfile_checker_loop(Fs1, Fs2, Dir1, Dir2, AbsInt).

dumpfile_checker_loop([], _, _, _, _).
dumpfile_checker_loop([F1|Fs1], [F1|Fs2], Dir1, Dir2, AbsInt) :- !,
	path_concat(Dir1, F1, DF1),
	path_concat(Dir2, F1, DF2),
	format(user_error, 'Checking ~w', [F1]),
	nl(user_error),
	compare_dumps_auto_detect_db(DF1, DF2, db1, db2, AbsInt, Diff),
	print_diff(Diff, DSumm),
	( DSumm = [] ->
	  format(user_error, 'EQUIVALENT', []) ; true),
	nl(user_error),
	dumpfile_checker_loop(Fs1, Fs2, Dir1, Dir2, AbsInt).
dumpfile_checker_loop([F1|Fs1], [F2|Fs2], Dir1, Dir2, AbsInt) :-
	F1 < F2, !,
	format(user_error, 'Skipping ~w ~w ...~n', [Dir1, F1]),
	dumpfile_checker_loop(Fs1, [F2|Fs2], Dir1, Dir2,  AbsInt).
dumpfile_checker_loop(Fs1, [F2|Fs2], Dir1, Dir2, AbsInt) :-
	format(user_error, 'Skipping ~w ~w ...~n', [Dir2, F2]),
	dumpfile_checker_loop(Fs1, Fs2, Dir1, Dir2, AbsInt).
