:- module(_,[filter_out_irrelevant_vars/2],[argnames]).

:- use_module(library(assoc)).
:- use_module(relevant_vars_dtree_deprecated).
:- use_module(relevant_vars_eqs2fvector_deprecated).
:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(library(lists),[length/2,append/3,nth/3]).

:- use_module(domain(resources/eq_solver/eq_solver_graphs), [varmember/2]).
:- use_module(library(terms_vars)).
:- include(relevant_vars_generated_dtree_deprecated).

%:- export(remove_uneeded_vars/3).
remove_uneeded_vars(Eqs,NEqs,Vars):-
	remove_uneeded_vars_(Eqs,Res),
	select_vars(Eqs,Res,Vars),
	remove_vars(Res,Eqs,NEqs),!.

%:- export(select_vars/3).
select_vars([equation(_F,Vars,_,_)|_],Res,Chosen):-
	select_vars_(Vars,Res,Chosen).

select_vars_([],_,[]).
select_vars_([X|T],[no|R],[X|Rest]):-
	select_vars_(T,R,Rest).
select_vars_([_X|T],[yes|R],Rest):-
	select_vars_(T,R,Rest).

%:- export(remove_vars/3).
remove_vars(_,[],[]).
remove_vars([],E,E).
remove_vars(Res,[equation(F,Args,Expr,Cons)|T],[equation(F,NArgs,NExpr,Cons)|TT]):-
	select_vars_(Args,Res,NArgs),
	remove_vars_in_exp(Res,F,Expr,NExpr),
	remove_vars(Res,T,TT).
%:- export(remove_vars_in_exp/4).
remove_vars_in_exp(_Res,_F,X,X):-
	var(X),!.
remove_vars_in_exp(Res,F,fun(F,Call),fun(F,NCall)):-
	select_vars_(Call,Res,NCall).
remove_vars_in_exp(Res,F,X,NX):-
	X =.. [Op|Args],
	remove_vars_in_exp_args(Res,F,Args,NArgs),
	NX =.. [Op|NArgs].

remove_vars_in_exp_args(_Res,_F,[],[]).
remove_vars_in_exp_args(Res,F,[A|T],[B|TT]):-
	remove_vars_in_exp(Res,F,A,B),
	remove_vars_in_exp_args(Res,F,T,TT).

%:- export(remove_uneeded_vars_/2).
remove_uneeded_vars_([],[]).
remove_uneeded_vars_([equation(F,Args,Expr,C)|T],Result):-
	remove_uneeded_vars_one_eq(Args,F,Expr,C,(Sol,_IsRec)),!,
	remove_uneeded_vars_(T,RR),
	or_list(Sol,RR,Result).
remove_uneeded_vars_([_|T],RR):-
	%remove_uneeded_vars_one_eq(Args,F,Expr,C,),
	remove_uneeded_vars_(T,RR).

or_list([],A,A).
or_list(A,[],A).
or_list([A|A1],[B|B1],[AB|AB1]):-
	or_(A,B,AB),
	or_list(A1,B1,AB1).

or_(false,V,V).
or_(true,_,true).


%:- export(remove_uneeded_vars_one_eq/5).
remove_uneeded_vars_one_eq(Args,F,Expr,Cons,(Ut,R)):-
	is_recursive(F,Expr,R),
	label_args(F,Args,Expr,Cons,Ut),!.

:- export(is_recursive/3).
is_recursive(F,Expr,T):-
	contains_function(F,Expr),!,T=true.
is_recursive(_F,_Expr,false).


contains_function(_F,V):-
	var(V),!,
	fail.
contains_function(F,fun(F,_)).
contains_function(F,Exp):-
	Exp =.. [_|ExpArgs],
	contains_function_args(F,ExpArgs).

contains_function_args(F,[E|_T]):-
	contains_function(F,E),!.
contains_function_args(F,[_|T]):-
	contains_function_args(F,T).

%:- export(label_args/5).
label_args(_F,[],_,_,[]).
label_args(F,[X|T],E,Cons,[true|TT]):-
	nonvar(X),!,
	label_args(F,T,E,Cons,TT).
label_args(F,[X|T],Expr,Cons,[true|TT]):-
	take_off_recursive_calls(F,Expr,NRExpr,_Recs),
	varset(NRExpr,Args0),
	varset(Cons,ArgsCons),
	append(Args0,ArgsCons,Args),
	varmember(X,Args),!,
	label_args(F,T,Expr,Cons,TT).
label_args(F,[X|T],Expr,Cons,[false|TT]):-
	take_off_recursive_calls(F,Expr,_NRExpr,Recs),
	used_in_rec_calls(X,Recs,true),!,
	label_args(F,T,Expr,Cons,TT).	
label_args(F,[_X|T],Expr,Cons,[false|TT]):-
	label_args(F,T,Expr,Cons,TT).

take_off_recursive_calls(F,Exp,[],[Exp]):-
	nonvar(Exp),
	Exp = fun(F,_),!.
take_off_recursive_calls(_F,Exp,[Exp],[]):-
	var(Exp),!.
take_off_recursive_calls(_F,Exp,[Exp],[]):-
	num(Exp),!.
take_off_recursive_calls(F,Exp,NRExp,Recs):-
	nonvar(Exp),
	Exp =.. [_|Args],
	take_off_recursive_calls_args(F,Args,NRExp,Recs).

take_off_recursive_calls_args(_,[],[],[]).
take_off_recursive_calls_args(F,[A|Args],NRExp,Recs):-
	take_off_recursive_calls(F,A,NRA,RecA),
	take_off_recursive_calls_args(F,Args,NRR,RR),
	append(NRR,NRA,NRExp),
	append(RR,RecA,Recs).

%:- export(used_in_rec_calls/3).
used_in_rec_calls(X,[fun(_,Args)|_T],true):-
	used_in_rec_call(X,Args,true),!.
used_in_rec_calls(X,[fun(_,_Args)|T],V):-
	used_in_rec_calls(X,T,V).
used_in_rec_calls(_X,[],false).

used_in_rec_call(_X,[],false).
used_in_rec_call(X,[D|_T],true):-
	nonvar(D),
	X \== D,
	varset(D,Vars),
	varmember(X,Vars),!.
used_in_rec_call(X,[_D|T],V):-
	used_in_rec_call(X,T,V).

analyze_relevant_vars(Eqs,Vars):-
	generate_fvectors(Eqs,FVectors),
	load_dtree(Tree),
	execute_dtree_(Tree,FVectors,Vars).

execute_dtree_(_Tree,[],[]).
execute_dtree_(Tree,[(_V,F)|R],[Res|T]):-
	execute_dtree(Tree,F,Res),
	execute_dtree_(Tree,R,T).


filter_out_irrelevant_vars(Eqs,NEqs):-
	analyze_relevant_vars(Eqs,FVectors), 
	remove_vars(FVectors,Eqs,NEqs).




