:- module(_,[execute_dtree/3,generate_dtree/1],[argnames]).

:- use_module(library(assoc)).
:- use_module(library(lists)).
:- use_module(library(aggregates)).

:- include(relevant_vars_learning_set_deprecated).



execute_dtree(if(Key=Value,BTrue,_BFalse),FVector,Answer):-
	%FVector${Key=>Value},!,
	get_assoc(Key,FVector,Value),!,
	execute_dtree(BTrue,FVector,Answer).
execute_dtree(if(_Key=_Value,_BTrue,BFalse),FVector,Answer):-
	%get_assoc(Key,FVector,Value),!,
	execute_dtree(BFalse,FVector,Answer).
execute_dtree(G,_FVector,G):-
	G \= if(_,_,_).


dtlearn(Goal, Examples, _ , Val) :-
   all_examples_agree(Goal, Examples, Val).

dtlearn(Goal, Examples, Atts, if(Cond=YesVal,YT,NT)) :-
   \+ all_examples_agree(Goal, Examples, _),
   select_split(Goal, Examples, Atts, Cond, Rem_Atts),
   split(Examples, Cond, YesVal, Yes, No),
   dtlearn(Goal, Yes, Rem_Atts, YT),
   dtlearn(Goal, No, Rem_Atts, NT).

% all_examples_agree(Goal, Examples, Val) is true if all examples agree that
%      Val is the value of attribute Goal.
all_examples_agree(_,[],_).
all_examples_agree(Att,[Obj|Rest],Val) :-
   val(Obj,Att,Val),
   all_examples_agree(Att,Rest,Val).


% split(Examples,Att,YesVal,T,F) is true if T is
% the examples in Examples with attribute Att
% having value YesVal and F is the examples with
% attribute Att having the other value.
split([],_,_,[],[]).
split([Obj|Rest],Cond,YesVal,[Obj|Yes],No) :-
   val(Obj,Cond,YesVal),
   split(Rest,Cond,YesVal,	Yes,No).
split([Obj|Rest],Cond,YesVal,Yes,[Obj|No]) :-
   val(Obj,Cond,NoVal),
   \+ NoVal = YesVal,
   split(Rest,Cond,YesVal,Yes,No).

select_split(Goal, Examples, [A1|RAtts], Cond, Rem_Atts) :-
%   here is where we really want maximum information, but it
%   is too early in the book. This is defined using max information here. 
    split_info(Goal,Examples,A1,I1),
    select_max_info_gain(Goal, Examples, RAtts, A1, I1, Cond, [], Rem_Atts).

% select_max_info_gain(Goal, Examples, Atts, BestC, BestIG, 
%       Cond, Rej_Atts, Rem_Atts)
%  This selects the attribute with the minimum expected information after the split.
select_max_info_gain(_, _, [], BestC, _, BestC, Rem_Atts, Rem_Atts).
select_max_info_gain(Goal, Exs, [Att|R], BestC, BestIG,Cond, Rej_Atts, Rem_Atts) :-
   split_info(Goal,Exs,Att,Info),
   ( Info > BestIG
     -> select_max_info_gain(Goal, Exs, R, BestC, BestIG,Cond, [Att|Rej_Atts], Rem_Atts)
     ; select_max_info_gain(Goal, Exs, R, Att, Info ,Cond, [BestC|Rej_Atts], Rem_Atts)
   ).

% split_info(Goal,Examples,Att,Info) is true if Info is the expected 
% information about Goal in Examples given a test for Att
split_info(Goal,Examples,Att,Info) :-
   split(Examples,Att,_,T,F),
   information(Goal,T,IT),
   information(Goal,F,IF),
   length(T,NT),
   length(F,NF),
   Info is (NT*IT + NF*IF)/(NT+NF).


% information(Goal,Examples,I) is true if I is the amount of information
%    in Examples with respect to Goal
information(Goal,Examples,I) :-
    count(Goal,_,Examples,True,False),
   ( (True=0 ; False=0) -> I=0
   ;
    Total is True + False,
    I is - True/Total * log(True/Total)/log(2) - False/Total * log(False/Total)/log(2)
   ).

% count(Att,TrueVal,Examples,T,F) is true if there are T examples in Examples
%   with attribute Att true and F examples with attribute Att false.
count(_,_,[],0,0).
count(Att,TrueVal,[Ex|R],T1,F) :-
   val(Ex,Att,TrueVal),
   count(Att,TrueVal,R,T,F),
   T1 is T+1.
count(Att,TrueVal,[Ex|R],T,F1) :-
   val(Ex,Att,FalseVal),
   \+ TrueVal=FalseVal,
   count(Att,TrueVal,R,T,F),
   F1 is F+1.



generate_dtree(Tree):-
	findall(Id,learning_set(Id,_,_,_,_,_,_,_,_,_),Examples), dtlearn(remove,Examples, [in_basecase_expr,
	in_basecase_cond,
	in_rec_case_cond,
	in_rec_case_expr_nr,
	in_rec_case_expr_samepos_changed,
	in_rec_case_expr_samepos_unchanged,
	in_rec_case_expr_diffpos_changed,
	in_rec_case_expr_diffpos_unchanged],Tree
).

% :- export(remove_uneeded_vars/3).
% remove_uneeded_vars(Eqs,NEqs,Vars):-
% 	remove_uneeded_vars_(Eqs,Res),
% 	select_vars(Eqs,Res,Vars),
% 	remove_vars(Res,Eqs,NEqs),!.

% :- export(select_vars/3).
% select_vars([equation(_F,Vars,_,_)|_],Res,Chosen):-
% 	select_vars_(Vars,Res,Chosen).

% select_vars_([],_,[]).
% select_vars_([X|T],[true|R],[X|Rest]):-
% 	select_vars_(T,R,Rest).
% select_vars_([_X|T],[false|R],Rest):-
% 	select_vars_(T,R,Rest).

% :- export(remove_vars/3).
% remove_vars(_,[],[]).
% remove_vars([],E,E).
% remove_vars(Res,[equation(F,Args,Expr,Cons)|T],[equation(F,NArgs,NExpr,Cons)|TT]):-
% 	select_vars_(Args,Res,NArgs),
% 	remove_vars_in_exp(Res,F,Expr,NExpr),
% 	remove_vars(Res,T,TT).
% :- export(remove_vars_in_exp/4).
% remove_vars_in_exp(_Res,_F,X,X):-
% 	var(X),!.
% remove_vars_in_exp(Res,F,fun(F,Call),fun(F,NCall)):-
% 	select_vars_(Call,Res,NCall).
% remove_vars_in_exp(Res,F,X,NX):-
% 	X =.. [Op|Args],
% 	remove_vars_in_exp_args(Res,F,Args,NArgs),
% 	NX =.. [Op|NArgs].

% remove_vars_in_exp_args(_Res,_F,[],[]).
% remove_vars_in_exp_args(Res,F,[A|T],[B|TT]):-
% 	remove_vars_in_exp(Res,F,A,B),
% 	remove_vars_in_exp_args(Res,F,T,TT).




% :- export(remove_uneeded_vars_/2).
% remove_uneeded_vars_([],[]).
% remove_uneeded_vars_([equation(F,Args,Expr,C)|T],Result):-
% 	remove_uneeded_vars_one_eq(Args,F,Expr,C,(Sol,_IsRec)),!,
% 	remove_uneeded_vars_(T,RR),
% 	or_list(Sol,RR,Result).
% remove_uneeded_vars_([_|T],RR):-
% 	%remove_uneeded_vars_one_eq(Args,F,Expr,C,),
% 	remove_uneeded_vars_(T,RR).

% or_list([],A,A).
% or_list(A,[],A).
% or_list([A|A1],[B|B1],[AB|AB1]):-
% 	or_(A,B,AB),
% 	or_list(A1,B1,AB1).

% or_(false,V,V).
% or_(true,_,true).


% :- export(remove_uneeded_vars_one_eq/5).
% remove_uneeded_vars_one_eq(Args,F,Expr,Cons,(Ut,R)):-
% 	is_recursive(F,Expr,R),
% 	label_args(F,Args,Expr,Cons,Ut),!.

% :- export(is_recursive/3).
% is_recursive(F,Expr,T):-
% 	contains_function(F,Expr),!,T=true.
% is_recursive(_F,_Expr,false).


% contains_function(_F,V):-
% 	var(V),!,
% 	fail.
% contains_function(F,fun(F,_)).
% contains_function(F,Exp):-
% 	Exp =.. [_|ExpArgs],
% 	contains_function_args(F,ExpArgs).

% contains_function_args(F,[E|_T]):-
% 	contains_function(F,E),!.
% contains_function_args(F,[_|T]):-
% 	contains_function_args(F,T).

% :- export(label_args/5).
% label_args(_F,[],_,_,[]).
% label_args(F,[X|T],E,Cons,[true|TT]):-
% 	nonvar(X),!,
% 	label_args(F,T,E,Cons,TT).
% label_args(F,[X|T],Expr,Cons,[true|TT]):-
% 	take_off_recursive_calls(F,Expr,NRExpr,_Recs),
% 	varset(NRExpr,Args0),
% 	varset(Cons,ArgsCons),
% 	append(Args0,ArgsCons,Args),
% 	varmember(X,Args),!,
% 	label_args(F,T,Expr,Cons,TT).
% label_args(F,[X|T],Expr,Cons,[false|TT]):-
% 	take_off_recursive_calls(F,Expr,_NRExpr,Recs),
% 	used_in_rec_calls(X,Recs,true),!,
% 	label_args(F,T,Expr,Cons,TT).	
% label_args(F,[_X|T],Expr,Cons,[false|TT]):-
% 	label_args(F,T,Expr,Cons,TT).

% take_off_recursive_calls(F,Exp,[],[Exp]):-
% 	nonvar(Exp),
% 	Exp = fun(F,_),!.
% take_off_recursive_calls(_F,Exp,[Exp],[]):-
% 	var(Exp),!.
% take_off_recursive_calls(_F,Exp,[Exp],[]):-
% 	num(Exp),!.
% take_off_recursive_calls(F,Exp,NRExp,Recs):-
% 	nonvar(Exp),
% 	Exp =.. [_|Args],
% 	take_off_recursive_calls_args(F,Args,NRExp,Recs).

% take_off_recursive_calls_args(_,[],[],[]).
% take_off_recursive_calls_args(F,[A|Args],NRExp,Recs):-
% 	take_off_recursive_calls(F,A,NRA,RecA),
% 	take_off_recursive_calls_args(F,Args,NRR,RR),
% 	append(NRR,NRA,NRExp),
% 	append(RR,RecA,Recs).




% :- export(used_in_rec_calls/3).
% used_in_rec_calls(X,[fun(_,Args)|_T],true):-
% 	used_in_rec_call(X,Args,true),!.
% used_in_rec_calls(X,[fun(_,_Args)|T],V):-
% 	used_in_rec_calls(X,T,V).
% used_in_rec_calls(_X,[],false).

% used_in_rec_call(_X,[],false).
% used_in_rec_call(X,[D|_T],true):-
% 	nonvar(D),
% 	X \== D,
% 	varset(D,Vars),
% 	varmember(X,Vars),!.
% used_in_rec_call(X,[_D|T],V):-
% 	used_in_rec_call(X,T,V).
