:- module(_,[generate_fvectors/2],[argnames,assertions,
	 nativeprops,
	 regtypes]).


% TODO: A predicate for normalize equations (e.g. equation(f,[1,X],1,[]) -> equation(f,[Y,X],1,[Y=1])).
  
:- use_module(library(assoc)).
:- use_module(library(lists),[length/2,append/3,nth/3]).
:- use_module(domain(resources/eq_solver/eq_solver_graphs), [varmember/2]).
:- use_module(library(terms_vars)).
:- use_module(library(terms_check),[instance/2]).

is_recursive_eq(equation(Name,_Args,Exp,_Cons)):-
	contains_sub_exp(fun(Name,_),Exp).

contains_sub_exp(Sub,Exp):-
	var(Exp),!,
	Sub == Exp.
contains_sub_exp(Sub,Exp):-
	instance(Exp,Sub).
contains_sub_exp(Sub,Exp):-
	nonvar(Exp),!,
	Exp =.. [_|Args],
	contains_sub_exp_list(Sub,Args).

contains_sub_exp_list(Sub,[Exp|_R]):-
	contains_sub_exp(Sub,Exp),!.

contains_sub_exp_list(Sub,[_|R]):-
	contains_sub_exp_list(Sub,R).

split_equations([],[],[]).
split_equations([E|R],[E|RR],NR):-
	is_recursive_eq(E),!,
	split_equations(R,RR,NR).
split_equations([E|R],RR,[E|NR]):-
	split_equations(R,RR,NR).

in_cond(X,[equation(_,_,_,Cond)]):-
	varset(Cond,SetV),
	varmember(X,SetV).
in_cond(X,[_|R]):-
	in_cond(X,R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

in_basecase_expr(_X,_,[],0).
in_basecase_expr(X,_,[equation(_,_,Exp,_)|_RR],1):-
	varset(Exp,Vars),
	varmember(X,Vars).
in_basecase_expr(X,_,[_|RR],V):-
	in_basecase_expr(X,_,RR,V).

in_basecase_cond(X,_R,NR,1):-
	in_cond(X,NR),!.
in_basecase_cond(_,_R,_NR,0).

in_rec_case_cond(X,R,_NR,1):-
	in_cond(X,R),!.
in_rec_case_cond(_,_R,_NR,0).

in_rec_case_expr_nr(X,[equation(N,_Args,Exp,_Cons)|_RR],_,1):-
	obtain_non_rec_part(N,Exp,Exp1),
	varset(Exp1,Vars),
	varmember(X,Vars),!.	 
in_rec_case_expr_nr(X,[_|RR],_,V):-
	in_rec_case_expr_nr(X,RR,_,V).
in_rec_case_expr_nr(_,[],_,0).

in_rec_case_expr_samepos_changed(X,Pos,R,_,V):-
	obtain_rec_changes(R,[],Chs),
	check_samepos_changed(X,Pos,Chs,V).

in_rec_case_expr_samepos_unchanged(X,Pos,R,_,V):-
	obtain_rec_changes(R,[],Chs),
	check_samepos_unchanged(X,Pos,Chs,V).

in_rec_case_expr_diffpos_changed(X,Pos,NArgs,R,_,V):-
	generate_check_possitions(Pos,NArgs,LPos),
	in_rec_case_expr_diffpos_changed_(X,LPos,R,_,V).

in_rec_case_expr_diffpos_unchanged(X,Pos,NArgs,R,_,V):-
	generate_check_possitions(Pos,NArgs,LPos),
	in_rec_case_expr_diffpos_unchanged_(X,LPos,R,_,V).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Generate the feature vector for each variable in the equations
generate_fvectors([equation(Name,Args,Expr,Cond)|T],FVectors):-
	length(Args,NArgs),
	split_equations([equation(Name,Args,Expr,Cond)|T],Recs,NRecs),
	generate_fvectors_(Recs,NRecs,0,Args,NArgs,FVectors).

generate_fvectors_(_,_,_,[],_,[]).
generate_fvectors_(Recs,NRecs,CPos,[Var|R],NArgs,[(Var,Assoc)|T]):-
	CPos1 is CPos + 1,
	in_basecase_expr(Var,Recs,NRecs,X1),
	in_basecase_cond(Var,Recs,NRecs,X2),
	in_rec_case_cond(Var,Recs,NRecs,X3),
	in_rec_case_expr_nr(Var,Recs,NRecs,X4),
	in_rec_case_expr_samepos_changed(Var,CPos1,Recs,NRecs,X5),
	in_rec_case_expr_samepos_unchanged(Var,CPos1,Recs,NRecs,X6),
	in_rec_case_expr_diffpos_changed(Var,CPos1,NArgs,Recs,NRecs,X7),
	in_rec_case_expr_diffpos_unchanged(Var,CPos1,NArgs,Recs,NRecs,X8),
	list_to_assoc([in_basecase_expr-X1,in_basecase_cond-X2,
	in_rec_case_cond-X3,
	in_rec_case_expr_nr-X4,
	in_rec_case_expr_samepos_changed-X5,
	in_rec_case_expr_samepos_unchanged-X6,
	in_rec_case_expr_diffpos_changed-X7,
	in_rec_case_expr_diffpos_unchanged-X8],Assoc),
	generate_fvectors_(Recs,NRecs,CPos1,R,NArgs,T).

generate_check_possitions(_,0,[]).
generate_check_possitions(P,P,R):-
	P1 is P -1,
	generate_check_possitions(P,P1,R).
generate_check_possitions(P,N,[N|R]):-
	P1 is N -1,
	generate_check_possitions(P,P1,R).

in_rec_case_expr_diffpos_changed_(_X,[],_,_,0).
in_rec_case_expr_diffpos_changed_(X,[P|_T],R,_,1):-
	in_rec_case_expr_samepos_changed(X,P,R,_,1),!.
in_rec_case_expr_diffpos_changed_(X,[_P|T],R,_,V):-
	in_rec_case_expr_diffpos_changed_(X,T,R,_,V),!.

in_rec_case_expr_diffpos_unchanged_(_X,[],_,_,0).
in_rec_case_expr_diffpos_unchanged_(X,[P|_T],R,_,1):-
	in_rec_case_expr_samepos_unchanged(X,P,R,_,1),!.
in_rec_case_expr_diffpos_unchanged_(X,[_P|T],R,_,V):-
	in_rec_case_expr_diffpos_unchanged_(X,T,R,_,V),!.

check_samepos_changed(_,_,[],0).
check_samepos_changed(X,Pos,[A|_R],1):-
	nth(Pos,A,XP),
	varset(XP,Vars),
	varmember(X,Vars),!,
	X \== XP.
check_samepos_changed(X,Pos,[_A|R],V):-
	check_samepos_changed(X,Pos,R,V).

check_samepos_unchanged(_,_,[],0).
check_samepos_unchanged(X,Pos,[A|_R],1):-
	nth(Pos,A,XP),
	varset(XP,Vars),
	varmember(X,Vars),!,
	X == XP.
check_samepos_unchanged(X,Pos,[_A|R],V):-
	check_samepos_unchanged(X,Pos,R,V).

obtain_rec_changes([],L,L).
obtain_rec_changes([equation(N,_A,Exp,_)|R],L,LRec):-
	obtain_recs(N,Exp,LLRec),
	append(L,LLRec,LLL),
	obtain_rec_changes(R,LLL,LRec).

obtain_recs(N,Exp,LRec):-
	obtain_recs_(N,Exp,[],LRec).

obtain_recs_(_N,Exp,L,L):-
	var(Exp),!.
obtain_recs_(N,Exp,L,[Args|L]):-
	Exp = fun(N,Args).
obtain_recs_(N,Exp,L,RL):-
	nonvar(Exp),!,
	Exp =.. [_|Args],
	obtain_recs_args(N,Args,LArgs),
	append(L,LArgs,RL).

obtain_recs_args(_N,[],[]).
obtain_recs_args(N,[Eq|NEqs],Res):-
	obtain_recs(N,Eq,A),
	obtain_recs_args(N,NEqs,RA),
	append(A,RA,Res).

obtain_non_rec_part(_Name,Exp,Exp):-
	var(Exp),!.
obtain_non_rec_part(Name,Exp,Exp1):-
	nonvar(Exp),!,
	Exp \= fun(Name,_),
	Exp =.. [Op|Args],
	%length(Args,Ar),
	obtain_non_rec_part_list(Name,Args,NArgs),
	%length(NArgs,ArN),
	Exp1 =.. [Op|NArgs].

obtain_non_rec_part_list(_Name,[],[]).
obtain_non_rec_part_list(Name,[Exp|T],[NExp|TT]):-
	obtain_non_rec_part(Name,Exp,NExp),!,
	obtain_non_rec_part_list(Name,T,TT).
obtain_non_rec_part_list(Name,[_Exp|T],TT):-
	%obtain_non_rec_part(Name,Exp,NExp),!,
	obtain_non_rec_part_list(Name,T,TT).


% :- export(obtain_parts/7).
% obtain_parts(RecRels,Vars,VarsP,ConstraintsRec,ConstraintsBase,Cost,Base):-
% 	obtain_recursive_and_base_case(RecRels,equation(Name,Vars,Exp,ConstraintsRec),equation(Name,_,Base,ConstraintsBase)),
% 	obtain_recursive_call(Exp,Name,fun(Name,VarsP)),
% 	obtain_node_cost(Exp, Cost),!.

% obtain_recursive_call(Exp,_Name,_Exp):-
% 	var(Exp),!,fail.
% obtain_recursive_call(Exp,Name,Exp):-
% 	nonvar(Exp),
% 	Exp = fun(Name,_Args),!.
% obtain_recursive_call(Exp,Name,RExp):-
% 	Exp =.. [F|Args],
% 	F \== fun,
% 	obtain_recursive_call_args(Args,Name,RExp).

% obtain_recursive_call_args([X|_N],Name,Exp):-
% 	obtain_recursive_call(X,Name,Exp),!.
% obtain_recursive_call_args([_X|N],Name,Exp):-
% 	obtain_recursive_call_args(N,Name,Exp),!.

% obtain_recursive_and_base_case([equation(Name,Args,Exp,Cons),Base],equation(Name,Args,Exp,Cons),Base):-
% 	obtain_recursive_call(Exp,Name,_),!.
% obtain_recursive_and_base_case([Base,equation(Name,Args,Exp,Cons)],equation(Name,Args,Exp,Cons),Base):-
% 	obtain_recursive_call(Exp,Name,_),!.

% obtain_node_cost(Exp,Cost):-
% 	nonvar(Exp),
% 	Exp =.. [Op|[Call,Cost]],
% 	(Op = '+';Op = '-'),
% 	instance(Call,fun(_,_)),
% 	copy_term(Cost,C),
% %	nonvar(Cost),
% 	arithexpression(C),!.
% obtain_node_cost(Exp,0):-
% 	nonvar(Exp),
% 	Exp =.. [fun|[_Call,_Cost]],!.
% obtain_node_cost(Exp,Cost):-
% 	nonvar(Exp),
% 	Exp =.. [Op|[Cost,Call]],
% 	(Op = '+';Op = '-'),
% 	instance(Call,fun(_,_)),
% 	arithexpression(Cost).

% is_recursive_eq(equation(N,Args,Exp,Cons)):-
% 	is_recursive_exp(N,Exp).


% obtain_recursive_call(Exp,_Name,_Exp):-
% 	var(Exp),!,fail.
% obtain_recursive_call(Exp,Name,Exp):-
% 	nonvar(Exp),
% 	Exp = fun(Name,_Args).
% obtain_recursive_call(Exp,Name,RExp):-
% 	Exp =.. [F|Args],
% 	F \== fun,
% 	obtain_recursive_call_args(Args,Name,RExp).

% obtain_recursive_call_args([X|_N],Name,Exp):-
% 	obtain_recursive_call(X,Name,Exp).
% obtain_recursive_call_args([_X|N],Name,Exp):-
% 	obtain_recursive_call_args(N,Name,Exp).

% analyze_relevant_vars(Eqs,Vars):-
% 	generate_fvectors(Eqs,FVectors),
% 	generate_dtree(Tree),
% 	execute_dtree_(Tree,FVectors,Vars).

% execute_dtree_(_Tree,[],[]).
% execute_dtree_(Tree,[(_V,F)|R],[Res|T]):-
% 	execute_dtree(Tree,F,Res),
% 	execute_dtree_(Tree,R,T).

% generate_dtree(Tree):-
% 	findall(Id,learning_set(Id,_,_,_,_,_,_,_,_,_),Examples), dtlearn(remove,Examples, [in_basecase_expr,
% 	in_basecase_cond,
% 	in_rec_case_cond,
% 	in_rec_case_expr_nr,
% 	in_rec_case_expr_samepos_changed,
% 	in_rec_case_expr_samepos_unchanged,
% 	in_rec_case_expr_diffpos_changed,
% 	in_rec_case_expr_diffpos_unchanged],Tree
% ).
