:- module(_,_,
	% [translate/4],
	 [assertions,
	  nativeprops,fsyntax,
	  regtypes]).

:- use_module(domain(resources/term_utilities)).
:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    algebraic_expression/1,
	    backend_solver/1
	]).

% % TODO: Fix translation to/from languages without explicitly translate
% % to algebraic form


% :- pred translate(From,To,Term,Result)
% 	:(backend_solver(From),backend_solver(To),term(Term),var(Result))
%  =>(backend_solver(From),backend_solver(To),term(Term),term(Result)) #
%  "Translates the expression @var{Term} from the syntax of the backend
%  solver @var{From} to the syntax of the backend solver @var{To}, if
%  possible.".

translate(A,B):-
	operates(A,~all,[mapt(translate_mathematica2algebraic)],update,B).


translate_mathematica2algebraic(A,B):-
	translate_functor(mathematica,algebraic,A,B).
translate_algebraic2mathematica(A,B):-
	translate_functor(algebraic,mathematica,A,B).


% translate(From,To,Input,Output):-
% 	replace_subterm_cond(Input,to_be_translated(From,To),translate_functor(From,To),Output),!.
% translate(X,Y,Input,Output):-
% 	translate(X,Z,Input,IntOutput),
% 	translate(Z,Y,IntOutput,Output).


% :- pred translate_functor(Pos,From,To,Term,Result)
% 	:(list(Pos,num),backend_solver(From),backend_solver(To),term(Term),var(Result))
%  =>(list(Pos,num),backend_solver(From),backend_solver(To),term(Term),term(Result)).

translate_functor(mathematica,_,A,B):-
	var(A),var(B),!,fail.
translate_functor(algebraic,mathematica,A,B):-
	translate_functor(mathematica,algebraic,B,A).
translate_functor(mathematica,algebraic,'Power'(A,B),A**B):-!.
translate_functor(mathematica,algebraic,'Gamma'(A),B):- var(B),B = fact(A-1),!.
translate_functor(mathematica,algebraic,'Sqrt'(A),sqrt(A)):-!.
translate_functor(mathematica,algebraic,'Exp'(A),exp(A)):-!.
translate_functor(mathematica,algebraic,'Factorial'(A),fact(A)):-!.
translate_functor(mathematica,algebraic,'LucasL'(A),lucasl(A)):-!.
translate_functor(mathematica,algebraic,'Fibonacci'(A),fibo(A)):-!.
translate_functor(mathematica,algebraic,'Log'(A),log(A)):-!.
translate_functor(mathematica,algebraic,'Log'(A,B),log(A,B)):-!.
translate_functor(mathematica,algebraic,'Infinity',inf):-!.
translate_functor(mathematica,algebraic,'DirectedInfinity'(1),inf):-!.
translate_functor(mathematica,algebraic,'Max'(A,B),max(A,B)):-!.
translate_functor(mathematica,algebraic,'Min'(A,B),min(A,B)):-!.
translate_functor(mathematica,algebraic,MaxMinN,MaxMin2):-
	((Op = 'Max',OpA=max);(Op='Min',OpA=min)),
	functor(MaxMinN,Op,N),
	N > 2,
	MaxMinN =.. [Op|Args],
	generate_op_pairwise(OpA,Args,MaxMin2),!.
translate_functor(mathematica,algebraic,F,WF):-
	nonvar(F),
	F =.. [Op|_Args],
	functor(F,Op,N),
	functor(FE,Op,N),
	\+ algebraic_expression(FE),
	Op \= fun,
	N > 0,
	wrap_user_defined_function(F,WF),!.
% translate_functor(mathematica,algebraic,F,WF):-
% 	var(F),nonvar(WF),
% 	wrap_user_defined_function(F,WF),!.

% translate_functor(_,builtin,_,A,B):-
% 	var(A),var(B),!,fail.
% translate_functor(_,algebraic,builtin,A,B):-
% 	translate_functor(_,builtin,algebraic,B,A).
% translate_functor(_,builtin,algebraic,exp(A,B),A**B):-!.
% %translate_functor(_,builtin,algebraic,log(A,B),log(A,B)):-!.
% translate_functor(_,builtin,algebraic,F,WF):-
% 	nonvar(F),
% 	F =.. [Op|_Args],
% 	functor(F,Op,N),
% 	functor(FE,Op,N),
% 	\+ algebraic_expression(FE),
% 	Op \= fun,
% 	N > 0,
% 	wrap_user_defined_function(F,WF),!.
% translate_functor(_,builtin,algebraic,F,WF):-
% 	var(F),nonvar(WF),
% 	wrap_user_defined_function(F,WF),!.

wrap_user_defined_function(F,fun(Op,Args)):-
	nonvar(F),
	F =.. [Op|Args].
wrap_user_defined_function(F,fun(Op,Args)):-
	var(F), 
	F =.. [Op|Args].

generate_op_pairwise(Op,[X,Y],OpXY):-
	OpXY =.. [Op|[X,Y]].
generate_op_pairwise(Op,[X,Y|T],OpXY):-
	OpXY0 =.. [Op|[X,Y]],
	generate_op_pairwise(Op,[OpXY0|T],OpXY).

% to_be_translated(Term,From,To):-
% 	translate_functor(_,From,To,Term,_).
% to_be_translated(Term,From,To):-
% 	translate_functor(_,From,Z,Term,R),
% 	translate_functor(_,Z,To,R,_).

