:- module(ppl_ops,_,[regtypes,assertions]).

%:- use_package(regtypes).
:- import(selected_ppl_interface_auto,[
ppl_Polyhedron_get_minimized_constraints/2 ,		
ppl_new_NNC_Polyhedron_from_constraints/2 ,		
ppl_new_NNC_Polyhedron_from_NNC_Polyhedron/2 ,		
ppl_new_NNC_Polyhedron_from_constraints/2 ,		
ppl_Polyhedron_is_empty/1 ,				
ppl_Polyhedron_remove_space_dimensions/2 ,		
ppl_Polyhedron_equals_Polyhedron/2 ,			
ppl_Polyhedron_is_disjoint_from_Polyhedron/2 ,		
ppl_Polyhedron_is_universe/1 ,				
ppl_Polyhedron_contains_Polyhedron/2 ,			
ppl_Polyhedron_intersection_assign/2 ,			
ppl_Polyhedron_is_empty/1 ,				
ppl_Polyhedron_poly_hull_assign/2 ,			
ppl_Polyhedron_H79_widening_assign/2 ,			
ppl_Polyhedron_BHRZ03_widening_assign/2 ,		
ppl_Polyhedron_bounded_H79_extrapolation_assign/3 ,	
ppl_Polyhedron_bounded_BHRZ03_extrapolation_assign/3 ,	
ppl_Polyhedron_is_empty/1 ,				
ppl_Polyhedron_map_space_dimensions/2 ,			
ppl_initialize/0 ,					
ppl_version/1 ,						
ppl_finalize/0 ,						
 ppl_initialize/0 ,					
 ppl_new_NNC_Polyhedron_from_constraints/2 ,		
 ppl_one_affine_ranking_function_PR_C_Polyhedron/2 ,	
 ppl_finalize/0 
	]).
:- use_module(domain(resources/plai2infer), [replace_term/3]).
:- use_module(library(write),[numbervars/3]).
:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(library(lists),[length/2,append/3]).
:- use_module(engine(messages_basic),[message/2]).

% TODO: simplify this code (see other uses of domain(selected_ppl_interface_auto))
:- compilation_fact(ppl_present).

:- if(defined(ppl_present)).

% Constraint relations using PPL

getConstraint(H,Cs0) :-
	ppl_Polyhedron_get_minimized_constraints(H,Cs0).
	
makePolyhedron(Cs,H1)  :-
	ppl_new_NNC_Polyhedron_from_constraints(Cs,H1).
	
copyPolyhedron(H1,H2)  :-
	ppl_new_NNC_Polyhedron_from_NNC_Polyhedron(H1,H2).
	
satisfiable(Cs,H1) :-
	ppl_new_NNC_Polyhedron_from_constraints(Cs,H1),
	\+ ppl_Polyhedron_is_empty(H1).
			
project(H,Zs,H) :-
	ppl_Polyhedron_remove_space_dimensions(H,Zs).
	
equivalent(H0,H1) :-
	ppl_Polyhedron_equals_Polyhedron(H0,H1).
	
disjointFrom(H0,H1) :-
	ppl_Polyhedron_is_disjoint_from_Polyhedron(H0,H1).

entails(H0,_) :-
	ppl_Polyhedron_is_universe(H0),
	!.
entails(H0,H1) :-
	ppl_Polyhedron_contains_Polyhedron(H0,H1).
	
consistent(H0,H1) :-
	copyPolyhedron(H0,H2),
	ppl_Polyhedron_intersection_assign(H2,H1),
	\+ ppl_Polyhedron_is_empty(H2).
convhull(empty,H1,H1) :-
	!.
convhull(H1,empty,H1) :-
	!.
convhull(H0,H1,H2) :-
	ppl_Polyhedron_poly_hull_assign(H1,H0),
	H2 = H1.
	
widenPolyhedraH79(H0,H1) :-
	ppl_Polyhedron_H79_widening_assign(H0,H1).
	
widenPolyhedraBHRZ03(H0,H1) :-
	ppl_Polyhedron_BHRZ03_widening_assign(H0,H1).
	
widenUptoH79(H0,H1,Cs) :-
	ppl_Polyhedron_bounded_H79_extrapolation_assign(H0,H1,Cs).
	
widenUptoBHRZ03(H0,H1,Cs) :-
	ppl_Polyhedron_bounded_BHRZ03_extrapolation_assign(H0,H1,Cs).
	
isEmpty(H) :-
	ppl_Polyhedron_is_empty(H).

mapCoords(H,Map) :-
	ppl_Polyhedron_map_space_dimensions(H,Map).

start_ppl :-
	ppl_initialize,
	ppl_version(Pv),
	message(user, ['PPL version used: ', ~~(Pv)]).
	
end_ppl :-
	ppl_finalize.

% TODO: The types used in this module should be defined outside plai/domain/resources.
:- pred ppl_find_linear_rank_function(Vars,PrimedVars,Constraints,Function):(list(Vars,var),list(PrimedVars,var),list(Constraints,test),var(Function)) => (point_rf(Function))
	# "Given the list @var{Vars} of variables of a loop; the list of primed variables; and the set of (linear) constraints that @var{Vars} U vars(@var{PrimedVars}) verify inside the loop (@var{Constraints}), this predicates returns in @var{Function} the (linear) ranking function of the loop, if it exists. Note that this predicate is limited to linear expressions.".

ppl_find_linear_rank_function(Vars,PrimedVars,Constraints,Function):-
	append(PrimedVars,Vars,AllVars),
	copy_term(AllVars,NAllVars),
	numbervars(NAllVars,0,_),
	build_map(AllVars,NAllVars,MappingF),
	build_map(NAllVars,AllVars,MappingB),
	build_map(PrimedVars,Vars,VarMap),
	replace_term(Constraints, MappingF,Cs0),
	ppl_initialize,
	ppl_new_NNC_Polyhedron_from_constraints(Cs0,H0),
	%	ppl_one_affine_ranking_function_MS_C_Polyhedron(H0,H1),
	ppl_one_affine_ranking_function_PR_C_Polyhedron(H0,H1),
	replace_term(H1,MappingB,Function0),
	replace_term(Function0,VarMap,Function),
	ppl_finalize.

build_map([],[],[]).
build_map([X|X1],[Y|Y1],[(X,Y)|Z]):-
	build_map(X1,Y1,Z).


:- else.

getConstraint(_H,_Cs0):-
	throw(error(not_implemented, 'ppl_ops:getConstraint'/2-0)).
	
makePolyhedron(_Cs,_H1)  :-
		throw(error(not_implemented, 'ppl_ops:makePolyhedron'/2-0)).
	
copyPolyhedron(_H1,_H2)  :-
		throw(error(not_implemented, 'ppl_ops:copyPolyhedron'/2-0)).
	
satisfiable(_Cs,_H1) :-
		throw(error(not_implemented, 'ppl_ops:satisfiable'/2-0)).
			
project(_H,_Zs,_H) :-
		throw(error(not_implemented, 'ppl_ops:project'/3-0)).
	
equivalent(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:equivalent'/2-0)).
	
disjointFrom(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:disjointFrom'/2-0)).

entails(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:entails'/2-0)).

consistent(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:consistent'/2-0)).


convhull(_H0,_H1,_H2) :-
		throw(error(not_implemented, 'ppl_ops:convhull'/3-0)).
	
widenPolyhedraH79(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:widenPolyhedraH79'/2-0)).
	
widenPolyhedraBHRZ03(_H0,_H1) :-
		throw(error(not_implemented, 'ppl_ops:widenPolyhedraBHRZ03'/2-0)).
	
widenUptoH79(_H0,_H1,_Cs) :-
		throw(error(not_implemented, 'ppl_ops:widenUptoH79'/3-0)).
	
widenUptoBHRZ03(_H0,_H1,_Cs) :-
		throw(error(not_implemented, 'ppl_ops:widenUptoBHRZ03'/3-0)).
	
isEmpty(_H) :-
		throw(error(not_implemented, 'ppl_ops:isEmpty'/1-0)).

mapCoords(_H,_Map) :-
		throw(error(not_implemented, 'ppl_ops:mapCoords'/2-0)).

start_ppl :-
		throw(error(not_implemented, 'ppl_ops:start_ppl'/0-0)).

end_ppl :-
		throw(error(not_implemented, 'ppl_ops:end_ppl'/0-0)).

ppl_find_linear_rank_function(_Vars,_PrimedVars,_Constraints,_Function):-
		throw(error(not_implemented, 'ppl_ops:ppl_find_linear_rank_function'/4-0)).

:- endif.

:- regtype point_rf(RF).

point_rf(point(Exp)):-
	arithexpression(Exp).
point_rf(point(Exp,C)):-
	arithexpression(Exp),
	num(C).
