:- module(solver, [
	init_solver/0,available_solver/1
   ],[assertions, rtchecks, datafacts]).

% :- use_module(library(messages), 
% 	[warning_message/2, debug_message/2, error_message/2]).

:- multifile issue_debug_messages/1.
:- data issue_debug_messages/1.
:- data available_solver/1.
%%%%%%%%%%%%%%%%%%%%Uncomment for show debug messages%%%%%%%%%%%%%%%%%%%%%
% issue_debug_messages(solver).
%issue_debug_messages(solver_utils).
% issue_debug_messages(solver_strategies).
% issue_debug_messages(solver_strategy_chain).
% issue_debug_messages(solver_utils_mathematica).
% issue_debug_messages(solver_utils_builtin).
% issue_debug_messages(solver_utils_ppl).
%issue_debug_messages(solver_messages).
:- import(selected_mathematica_interface_auto, [evaluate/2]).
:- use_module(domain(resources/eq_solver/solver/ppl_ops),[start_ppl/0]).
:- use_module(library(messages),[debug_message/2]).
:- use_module(domain(resources/eq_solver/eq_solver_types),[backend_solver/1]).
:- reexport(domain(resources/eq_solver/solver/solver_strategies), [
	rsolve/5,
	rsolve/4
]).

:- reexport(domain(resources/eq_solver/solver/solver_utils), [
	simplify/2,simplify/3,
	expand/2,expand/3,
	solve/3,solve/4,
	greater_than/2,greater_than/3,
	greater_or_equal_than/2,greater_or_equal_than/3,
	equal_than/2,equal_than/3,clean_solver_cache/0
]).


:- pred init_solver # "Initializes the list of solvers that are
currently available. This list can be retrieved from the dynamic
predicate @tt{available_solver/1}, obtaining all its solutions by
backtracking".  

init_solver:- 
	set_available_backend_solvers,
	clean_solver_cache.


set_available_backend_solvers:-
	retractall_fact(available_solver(_)),
	set_available_solvers.


set_available_solvers:-
	backend_solver(S),
	is_active(S),
	assertz_fact(available_solver(S)),
	fail.
set_available_solvers.


% % todo: Define these checks in each backend solver module. The check to know whether mathematica is available or not
% %should not be done catching the exception. 
:- pred is_active(Solver):(backend_solver(Solver))
	# "Checks if the solver @var{Solver} is currently available.".
is_active(builtin).
is_active(ppl).
% is_active(ppl):-
% 	catch(start_ppl,error(not_implemented,_),fail).
is_active(mathematica):-
	catch(
		 evaluate("1+1",_),
		 _Ex,
		 (
		     debug_message("Mathematica is not currently available, therefore it's not going to be used for solve equations in chain strategy.",[]),
		     fail
		 )
	     ).
