:- module(solver_messages,[show_user_message/1],[assertions,regtypes]).

:- use_module(library(messages)).
:- use_module(library(format),[format_to_string/3]).
:- use_module(domain(resources/eq_solver/eq_solver_types),[solver_exception/1,solver_operation/1,backend_exception/1]).

:- doc(title, "Solver - Messages").
:- doc(author, "Maximiliano Klemen").
:- doc(module, "This module provides predicates for showing messages to the user regarding solver's exceptions").


:- pred show_user_message(ME): (solver_exception(ME)) # " It shows a
	proper message corresponding to the exception @var{ME}. It
	could be a warning, debug, or error message.".  

:- pred	show_user_message(ME): (backend_exception(ME)) # " It shows a
	proper message corresponding to the backend solver exception
	@var{ME}. It could be a warning, debug, or error message.".

show_user_message(not_implemented(Op,Solver,Exp)):-
	can_continue_op(Op),!,
	debug_message("~q is not implemented for the back-end ~q. The input (~q) is returned unchanged.",[Op,Solver,Exp]).

show_user_message(not_implemented(Op,Solver,Exp)):-
	error_message("~q (~q) is not implemented for the back-end ~q.",[Op,Exp,Solver]).
show_user_message(backend_exception(Op,Solver,Args,BEx)):-
	backend_exception_text(BEx,TBEx),
	error_message("Exception thrown by ~q solver during ~q operation (args: ~q). The exception was: ~q",[Solver,Op,Args,TBEx]).
% show_user_message(not_solved_by_strategy(St,RecRel,Vars,Bound)):-
% 	debug_message("~q strategy could not solve (~q, ~q, ~q) RR.",[St,RecRel,Vars,Bound]).
% show_user_message(not_solved(Op,Solver,RecRel,Vars,Bound)):-
% 	debug_message("~q could not solve (~q, ~q, ~q) RR during ~q",[Solver,RecRel,Vars,Bound,Op]).
show_user_message(BEx):-
	backend_exception_text(BEx,TBEx),
	error_message(TBEx,[]).




:- pred can_continue_op(Op): 
	(solver_operation(Op)).
can_continue_op(simplify).
can_continue_op(expand).

:- pred backend_exception_text(Ex,T):
	(backend_exception(Ex),var(T)) => (string(T)).
backend_exception_text(error(mathematica_opening_link,_),"Unable to connect to mathematica"):-!.
backend_exception_text(Ex,TEx):-
	format_to_string("~q",[Ex],TEx).
