:- module(_, _, [assertions,regtypes,basicmodes,nativeprops]).

% :- doc(module,
% 	"Small utilities used in argument size analysis.").

% :- use_module(ciaopp(plai/plai_db)).
% :- use_module(term_utilities,[update_term/4,get_subterm_cond/3, term_at_position/3]).
% :- use_module(library(aggregates),[findall/3, setof/3]).
:- use_module(library(lists),[append/3, length/2,nth/3,delete/3,select/3]).
% :- use_module(domain(s_eqs), [peel/4]).
:- use_module(library(terms_vars)).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).
:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(domain(resources/eq_solver/solver/solver)).
:- use_module(library(terms_check)).
% :- use_module(.(eq_solver/solver/solver)).
% :- use_module(ciaopp(preprocess_flags)).
% :- use_module(library(sort)).
% :- use_module(library(write)).
% :- use_module(.(eq_solver/eq_solver),[solve_special_equation/3]). %This should be in solver.


:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(sets)).
%:- use_module(library(terms_check)).
%:- use_module(library(terms_vars)).
:- use_module(library(iso_misc)).
:- use_module(ciaopp(preprocess_flags), [current_pp_flag/2] ).
:- use_module(library(system),[time/1]).
:- use_module(library(hiordlib),[maplist/2]).
:- use_module(library(messages)).
:- use_module(domain(resources/eq_solver/solver/solver_messages),[show_user_message/1]).



%:- export(perm/2).
perm([],[]).
perm([X|Xs],[R|Rs]) :-
	select(R,[X|Xs],Y),
	perm(Y,Rs).

% select(X,[X|Xs],Xs).
% select(X,[Y|Ys],[Y|Zs]) :-
% 	select(X,Ys,Zs).

% BEGIN Code to solve equations with max and min and some with two arguments
:- pred solve_special_equation(RecurrenceRelations,Bound,Solution)
	: recurrence_relation * bound * var
       => recurrence_relation * bound * resolution.
:- doc(solve_special_equation/2,
	"Tries to solve a small set of possible
         special @var{RecurrenceRelations}.").
:- export(solve_special_equation/3).

solve_special_equation(Eqs,_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[Eqs,_B]),
	solve_fact_eq1(Eqs,Solved).
solve_special_equation(Eqs,_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[Eqs,_B]),
	solve_nondet1(Eqs,Solved).
solve_special_equation(Eqs,_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[Eqs,_B]),
	solve_invariant_result(Eqs,Solved).
%	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B],Bound,Solved):-
% 	solve_basecase_unknown_size([A,B],Bound,Solved),!.
% solve_special_equation([A,B],Bound,Solved):-
% 	solve_basecase_unknown_size([B,A],Bound,Solved).
% solve_special_equation([A,B,C],Bound,Solved):-
% 	%length(Eqs,3),!,
% %	error_message("RED: ~q (~q)",[Eqs,_B]),
% 	solve_extra_var([A,B,C],Bound,Solved).
% %	error_message("SOLV: ~q",[Solved]).
solve_special_equation(Eqs,_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[Eqs,_B]),
	solve_min_max_equation(Eqs,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
	solve_min_max_equation([B,A],Solved).
%	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B],_B,Solved) :-
% %	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
% 	solve_argument_fixed_1([A,B],_B,Solved),!.
% %	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B],_B,Solved) :-
% %	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
% 	solve_argument_fixed_1([B,A],_B,Solved),!.
% %	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B,C],Bound,Solved) :-
% %	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
% 	solve_nondet_equation_1([A,B,C],Bound,Solved).
% %solve_special_equation([A,B,C],Bound,Solved) :-
% %	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
% %	solve_two_lists_equation_2([A,B,C],Bound,Solved).
% solve_special_equation(Eqs,B,Solved) :-
% % %	error_message("2LISTS: ~q (~q)",[Eqs,B]),
% 	solve_two_lists_equation(Eqs,B,Solved).
% % %	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B],Bound,Solved) :-
% %	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
% 	solve_two_lists_equation([B,A],Bound,Solved).
%	error_message("SOLV: ~q",[Solved]).
% % solve_special_equation(Eqs,_B,Solved):-
% % 	length(Eqs,3),!,
% % %	error_message("RED: ~q (~q)",[Eqs,_B]),
% % 	solve_redundant_equation(Eqs,Solved).
% % %	error_message("SOLV: ~q",[Solved]).
% solve_special_equation(Eqs,B,Solved):-
% 	%length(Eqs,3),!,
% %	error_message("RED: ~q (~q)",[Eqs,_B]),
% 	solve_variable_basecase(Eqs,B,Solved).
% %	error_message("SOLV: ~q",[Solved]).
% solve_special_equation([A,B],Bound,Solved):-
% 	%length(Eqs,3),!,
% %	error_message("RED: ~q (~q)",[Eqs,_B]),
% 	solve_variable_basecase([B,A],Bound,Solved).
% %	error_message("SOLV: ~q",[Solved]).


%%Infinite number of equal results (assuming termination)
%h:upper=======================
%[equation(h,[_77089],_77089 + C,[]),equation(h,[_77089],fun(h,[_77089-1])+1,[])]
%[equation(f2,[_A],_A,[]),equation(f2,[_A],fun(f2,[_A]),[])] 
solve_invariant_result([A,B],Sol):-
	solve_invariant_result_([A,B],Sol).
solve_invariant_result([B,A],Sol):-
	solve_invariant_result_([A,B],Sol).
solve_invariant_result_([A,B],Sol):-
	A = equation(F,Args1,Exp1,[]),
	B = equation(F,Args2,Exp2,[]),
	Args1 == Args2,
	Args1 = [Arg],
	Exp1 == Arg,
	nonvar(Exp2),
	(Exp2 = fun(F,[MArg]) + C;Exp2 = C + fun(F,[MArg])),
	(MArg == Arg - C;MArg == -C+Arg),!,
	 Sol = Arg.

%h:upper=======================
%[equation(h,[_77089],_77089 + C,[]),equation(h,[_77089],fun(h,[_77089-1])+1,[])], being C a constant
solve_nondet1([A,B],Sol):-
	solve_nondet1_([A,B],Sol).
solve_nondet1([B,A],Sol):-
	solve_nondet1_([A,B],Sol).
solve_nondet1_([A,B],Sol):-
	A = equation(F,Args1,Exp1,[]),
	B = equation(F,Args2,Exp2,[]),
	Args1 == Args2,
	Args1 = [Arg],
	match_nondet1(Exp1,Arg),
	%Exp1 == Arg,
	nonvar(Exp2),
	(unify_with_occurs_check(Exp2,fun(F,[MArg]) + C);unify_with_occurs_check(Exp2,C + fun(F,[MArg]))),
	C==1,
	(MArg == Arg - 1;MArg == -1+Arg),!,
	 Sol = Exp1.
solve_nondet1_([A,B],Sol):-
	A = equation(F,Args1,Exp1,[]),
	B = equation(F,Args2,Exp2,[]),
	Args1 == Args2,
	Args1 = [Arg],
	nonvar(Exp2),
	unify_with_occurs_check(Exp2,fun(F,[MArg])),
%	Exp2 = fun(F,[MArg]),
	MArg == Arg,!,
	Sol = Exp1.


match_nondet1(Exp,Arg):-
	nonvar(Exp),
	((Exp = A + C);(Exp=C+A);(Exp=A-C);(Exp= -C+A)),
	  A == Arg,
	  nonvar(C),
	  num(C).


% [equation(s,[_A],_A*fun(s,[_A-1])+2*_A+1,[]),equation(s,[1],1,[])]
% Sol = -2*fact(n)+sum(a,1,n,2*fact(a)**(-1)*a+factorial(a)**(-1))*factorial(n) (from PURRS)
solve_fact_eq1([A,B],Sol):-
	solve_fact_eq1_([A,B],Sol).
solve_fact_eq1([B,A],Sol):-
	solve_fact_eq1_([A,B],Sol).
solve_fact_eq1_([A,B],Sol):-
	A = equation(F,Args1,Exp1,[]),
	B = equation(F,Args2,Exp2,[]),
	Args2 == [1],
	Args1 = [Arg],
	nonvar(Exp1),
	Exp1 == Arg*fun(s,[Arg-1])+2*Arg+1,
	nonvar(Exp2),
	Exp2 ==1,!,
	Sol = -2*fact(Arg)+sum(X,1,Arg,2*fact(X)**(-1)*X+fact(X)**(-1))*fact(Arg).

% % [equation(s,[_A],_A*fun(s,[_A-1])+C*_A**K+D,[]),equation(s,[1],1,[])], being C, D and K integers
% % Sol = -2*fact(n)+sum(a,1,n,2*fact(a)**(-1)*a+factorial(a)**(-1))*factorial(n) (from PURRS)
% solve_fact_eq1([A,B],Sol):-
% 	solve_fact_eq1_([A,B],Sol).
% solve_fact_eq1([B,A],Sol):-
% 	solve_fact_eq1_([A,B],Sol).
% solve_fact_eq1_([A,B],Sol):-
% 	A = equation(F,Args1,Exp1,[]),
% 	B = equation(F,Args2,Exp2,[]),
% 	Args2 == [1],
% 	nonvar(Exp2),
% 	Exp2 ==1,
% 	Args1 = [Arg],
% 	nonvar(Exp1),
% 	Exp1 = FacPart + VarPart + ConsPart
% %	obtain_terms(Exp1,Terms),
% 	perm(Terms,[FacPart,VarPart,ConsPar]),
% 	FacPart = Var1*fun(F,[ArVar1]),
% 	Var1 == Arg, (ArVar1 == Arg -1;ArVar1 == -1 + Arg),
% 	nonvar(ConsPart), num(ConsPart),
% 	obtain_varparts(VarPart, Cons1, K1),!,
% 	Sol = -(Cons1 + ConsPart - 1)*fact(Arg) + fact(Arg)*sum(X,1,Arg,(Cons1*(X**K1) + ConsPart)/fact(X)).




% %eqs([h(_C,_D,_O)==(g(_O+ -1*_D+1)*h(_C-1,_D,_O)+i(_O+ -1*_D+1),[]), h(1,?,_O)==(0,[])]))).
% solve_basecase_unknown_size([A,B],_Bound,Solved):-
% 	A = equation(F,[G, S1, S2],GE,[]),        %recursive
% 	B = equation(F,[V, V1, V2],BaseVal,[]),  %base
% 	V==1,V1=='?',V2==S2,
% 	ground(BaseVal),
% 	BaseVal==0,
% 	var(G),var(S1),var(S2),nonvar(GE),
% 	(
% 	    GE = C1*fun(F,[G2, S11, S22])+C2,C3=0
% 	    ;
% 	    GE = fun(F,[G2, S11, S22])+C2,C1=1,C3=0
% 	    ;
% 	    GE = fun(F,[G2, S11, S22])+C2-C3, C1=1
% 	),
% %	C1==1,
% 	G2 == G-1,S1==S11,S2==S22,
% 	Solved = (G-1)*(C2-C3).

% % eqs([l(_C,_E)==(l(_C-1,_E)*g(_E)+m(_E),[]),l(1,?)==(0,[])])
% solve_basecase_unknown_size([A,B],_Bound,Solved):-
% 	A = equation(F,[G, S1],GE,[]),        %recursive
% 	B = equation(F,[V, V1],BaseVal,[]),  %base
% 	V==1,V1=='?',
% 	ground(BaseVal),
% 	BaseVal==0,
% 	var(G),var(S1),nonvar(GE),
% 	(
% 	    GE = C1*fun(F,[G2, S11])+C2,C3=0
% 	    ;
% 	    GE = fun(F,[G2, S11])+C2,C1=1,C3=0
% 	    ;
% 	    GE = fun(F,[G2, S11])+C2-C3, C1=1
% 	),
% %	C1==1,
% 	G2 == G-1,S1==S11,
% 	Solved = (G-1)*(C2-C3).

% % eqs([j(_D,_H)==(fun(j,[_D-1,_H-1]),[]),j(1,_H)==(V,[]),j(1,1)==(V,[])]
% % eqs([j(_D,_H)==(fun(j,[_D-1,_H-1]) + 1,[]),j(1,_H)==(V,[]),j(1,1)==(V,[])]
% solve_extra_var([A,B,C],_Bound,Solved):-
% 	A = equation(F,[G,Extra],GE,[]),
% 	B = equation(F,[1,Extra1],BaseVal,[]),
% 	ground(BaseVal),
% 	C = equation(F,[1,1],BaseVal0,[]),
% 	BaseVal == BaseVal0,
% 	var(G),nonvar(GE),
% 	Extra == Extra1,
% 	GE = fun(F,[G2,Extra2]),
% 	G2 == G-1,
% 	Extra2 == Extra - 1,!,
% 	Solved = BaseVal.
% solve_extra_var([A,B,C],_Bound,Solved):-
% 	A = equation(F,[G,Extra],GE,[]),
% 	B = equation(F,[1,Extra1],BaseVal,[]),
% 	ground(BaseVal),
% 	C = equation(F,[1,1],BaseVal0,[]),
% 	BaseVal == BaseVal0,
% 	var(G),nonvar(GE),
% 	Extra == Extra1,
% 	GE = fun(F,[G2,Extra2]) + 1,
% 	G2 == G-1,
% 	Extra2 == Extra - 1,!,
% 	Solved = G - 1 + BaseVal.

% % solve_extra_var([A,B],_Bound,Solved):-
% % 	A = equation(F,[G,_Extra],GE,[]),
% % 	B = equation(F,[GV,Extra1],RecCase,[]),
% % 	ground(GE),
% % 	ground(G),
% % 	RecCase = fun(F,[GM,ExtraR]) + 4*Extra1 + 2,
% % 	Extra1 == ExtraR,
% % 	GM == GV - 1,
% % 	Solved = 2*GV*(2*Extra1 + 1).


% %eqs([f(A)==(1,[]),f(A)==(fun(f,[A-1])+1,[])])
% solve_variable_basecase(Eqs,Bound,Solved):-
% 	%perm(Eqs,LEqs),
% 	Eqs = [ equation(F,[G],GE,[]), equation(F,[V],VE,[])],
% 	var(G), ground(GE), var(V),G==V,nonvar(VE),VE = fun(F,[XX]) + 1,XX == G - 1,!,
% 	(Bound == upper -> Solved = max(G,GE); Solved = min(GE,G)).

% % eqs([f(C,A)==(fun(f,[C-1,A])-A**(C-1)+5*(A**C*C)-2*A**C+1,[C>0]),f(0,A)==(1,[])])
% solve_argument_fixed_1(Eqs,_Bound,SolutionExpr) :-
% 	Eqs = [equation(F,[A1,A2],Expr, T1), equation(F,[A3,A4],BaseCase,[])],
% 	ground(A3),A3=0,var(A1),var(A2),var(A4),A2==A4,Expr = fun(F,[A7,A6])-A6**(A5-1)+5*(A6**A5*A5)-2*A6**A5+1,
% 	var(A5),var(A6),nonvar(A7),A7 == A5 - 1,A1==A5,A2==A6,T1 == [A1 > 0], ground(BaseCase),BaseCase=1,
% 	SolutionExpr = ((5*(A6-1)*A6*A5-2*A6*(A6+2)+1)*A6**A5+A6*(A6*(A5+3)-2*A5+2)+A5)/(A6-1)**2.
% solve_argument_fixed_1(Eqs,_Bound,SolutionExpr) :-
% 	Eqs = [equation(F,[A1,A2],Expr, T1), equation(F,[A3,A4],BaseCase,[])],
% 	ground(A3),A3=0,var(A1),var(A2),var(A4),A2==A4,Expr = fun(F,[A7,A6])-A6**(A5-1)+2*(A6**A5*A5) + A6**A5+1,
% 	var(A5),var(A6),nonvar(A7),A7 == A5 - 1,A1==A5,A2==A6,T1 == [A1 > 0], ground(BaseCase),BaseCase=1,
% 	SolutionExpr = ((5*(A6-1)*A6*A5-2*A6*(A6+2)+1)*A6**A5+A6*(A6*(A5+3)-2*A5+2)+A5)/(A6-1)**2.

% %Solve the following case:
% %eqs([f(1)==(1,[]),f(_A)==(_A,[_A>1]),f(_A)==(fun(f,[_A-1]),[_A>1])])
% solve_nondet_equation_1(Eqs,Bound,SolutionExpr) :-
% 	perm(Eqs,LEqs),
% 	LEqs = [ equation(F,[A1],E1,[]), equation(F,[A2],E2,T2), equation(F,[A3],E3,T3)],
% 	A1==1,E1==1,var(A2),var(A3), A2 == A3, var(E2),E2==A2, nonvar(E3), E3 = fun(F,[A3 - 1]),T2 == T3,!,
% 	(Bound = lower -> SolutionExpr = 0; SolutionExpr = A3).


%Solve the following case:
%g(1)==(1,[]),g(_E)==(_E,[]),g(_E)==(fun(g,[_E-1])+1,[]) ==> _E
% solve_redundant_equation(Eqs,SolutionExpr) :-
% 	perm(Eqs,LEqs),
% 	LEqs = [ equation(F,[G],GE,_T1), equation(F,[V],VE,_T2), equation(F,[X],XE,_T3)],
% 	G==1,GE==1,var(V), (VE == V; VE == V - 1 ), var(X), X == V, (XE == fun(F,[X - 1]) + 1;XE == 1 + fun(F,[X - 1])),!,
% 	SolutionExpr = X.
% solve_two_lists_equation(Eqs,Bound,SolutionExpr) :-
% 	Eqs = [ equation(F,[Ground,Var2],GroundExpr,_T1), equation(F,[Var1,Var2_2],_,_T2) ],
% 	ground(Ground),ground(GroundExpr),var(Var1),Var2_2 == Var2,
% 	get_input_positions(Eqs,[ Yes, _ ]),
% 	keep_only_needed_vars_(Eqs,[Yes],2,[Var1],SimpEqs),
% 	get_input_positions(SimpEqs,Pos),
% 	get_mappings(SimpEqs,1,Pos,Vars,_),
% 	%MODIFIED 10-06-14: Switch to builtin solver.
% 	solve_rec_with_solver(SimpEqs,Vars,Bound,SolutionExpr).
% 	%solve_rec_rel_with_mathematica(SimpEqs,Vars,SolutionExpr).

% % eqs([g(1,?)==(0,[]),g(_C,_E)==(_E,[]),g(_C,_E)==(fun(g,[_C-1,_E]),[])])
% solve_two_lists_equation_2(Eqs,_Bound,SolutionExpr) :-
% 	Eqs = [ equation(F,[Ground,Ques],GroundExpr,_T1), equation(F,[Var1,Var2_2],VarExp,_T2), equation(F,[Var1,Var2_2],Exp3,_T2)],
% 	ground(Ground),ground(GroundExpr),Ques == '?',var(Var1),var(Var2_2),Exp3 = fun(F,[Var1-1,Var2_2]),
% 	GroundExpr = 0,VarExp == Var2_2,!,
% 	SolutionExpr = Var2_2.
% %eqs([f(1,?)==(?,[]),f(_B,_D)==(_D,[]),f(_B,_D)==(min(fun(f,[_B-1,_D]),_D),[])])
% solve_two_lists_equation_2(Eqs,_Bound,SolutionExpr) :-
% 	Eqs = [equation(F,[Ground,Ques],GroundExpr,_T1), equation(F,[Var1,Var2_2],VarExp,_T2), equation(F,[Var1,Var2_2],Exp3,_T2)],
% 	ground(Ground),ground(GroundExpr),Ques == '?',var(Var1),var(Var2_2),Exp3 = min(fun(F,[Var1-1,Var2_2]),Var2_2),
% 	GroundExpr = '?',VarExp == Var2_2,!,
% 	SolutionExpr = Var2_2.

:- pred solve_min_max_equation(RecurrenceRelations,Solution)
	: recurrence_relation * var
       => recurrence_relation * resolution.
:- doc(solve_min_max_equation/2,
	"Tries to solve a small set of possible
         @var{RecurrenceRelations} involving
         maximum and minimum operators.").

% Assume all functions and results are increasingly monotonic

% | f(c) = ?
% | f(n) = min(f(n-1),g(n))
% => f(n) = g(c+1)
solve_min_max_equation([ equation(F,[GroundParam],Q1,_T1), equation(F,[Var],Min,_T2) ], Result) :-
	ground(GroundParam), var(Var), Q1 == '?',
	( Min = min(L,GofN) ; Min = min(GofN,L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[Var-1])),!,
%	L = fun(F,[Var-1]),!,
%	L =.. [F,Var-1], !,
	( num(GroundParam)
	-> ResultP is GroundParam + 1
	 ; ResultP = GroundParam + 1
	),
	substitute_vars(GofN,[Var],[ResultP],SubstGofN),
	simplify(SubstGofN,Result).

%simplify_in_mathematica(SubstGofN,Result).
% | f(c) = ?
% | f(n) = max(f(n-1),g(n))
% => f(n) = g(n)
solve_min_max_equation([ equation(F,[GroundParam],Q1,_T1), equation(F,[Var],Max,_T2) ], GofN):-
	ground(GroundParam), var(Var), Q1 == '?',
	( Max = max(L,GofN) ; Max = max(GofN, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[Var-1])),!.
%	L = fun(F,[Var-1]),!.
%	L =.. [F,Var-1], !.
% | f(c) = ?
% | f(n) = max(f(n-1),g(n))
% => f(n) = g(n)
solve_min_max_equation([ equation(F,[GroundParam],Q1,_T1), equation(F,[Var],Max,_T2) ], Var*GofN):-
	ground(GroundParam), var(Var), Q1 == '?',
	( Max = max(L,GofN) ; Max = max(GofN, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[Var-1]) + GofN),!.
%	L = fun(F,[Var-1]) + GofN,!.
%	L =.. [F,Var-1], !.



% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = min(f(n-1,m),g(m)) or f(n,m) = min(f(c,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Min,_T2) ], GofVarNotR ) :-
	ground(GroundParam), var(VarR), var(Q1), Q2 == '?',
	( Min = min(L, GofVarNotR) ; Min = min(GofVarNotR, L) ),
	nonvar(L),
%	L = fun(F,[WorkVarR,VarNotR]),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = max(f(n-1,m),g(m)) or f(n,m) = max(f(c,m),g(m))
% => f(n,m) = g(m) (independent of n)
%%%%%MK
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Max,_T2) ], GofVarNotR ) :-
	ground(GroundParam), var(VarR), var(Q1) , Q2 == '?',
	( Max = max(L, GofVarNotR) ; Max = max(GofVarNotR, L) ),
	nonvar(L),
%	L = fun(F,[WorkVarR,VarNotR]),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = max(f(n-1,m) + m, g(m)) 
% => f(n,m) = g(m)*n 
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Max,_T2) ], GofVarNotR*VarR ) :-
	ground(GroundParam), var(VarR), var(Q1) , Q2 == '?',
	( Max = max(L, GofVarNotR) ; Max = max(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR]) + GG),
%	L = fun(F,[WorkVarR,VarNotR]) + GG,
	GG == Q1,
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = min(f(n-1,m) + m, g(m)) 
% => f(n,m) = g(m)*n 
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Max,_T2) ], GofVarNotR ) :-
	ground(GroundParam), var(VarR), var(Q1) , Q2 == '?',
	( Max = min(L, GofVarNotR) ; Max = min(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR]) + GG),
%	L = fun(F,[WorkVarR,VarNotR]) + GG,
	GG == Q1,
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ?
% | f(n,m) = f(n-1,m)
% | f(n,m) = min(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ Eq1, Eq2, Eq3 ], GofVarNotR) :-
	Eq1 = equation(F,[GroundParam,Q1],Q2,_T1),
	var(Q1), Q2 == '?',
	( Eq2 = equation(F,[VarR,VarNotR],Min,_T2),
	  nonvar(Min), Min = min(_,_),
	  Eq3 = equation(F,[VarR,VarNotR],L,_T3)
	-> true
	 ; Eq3 = equation(F,[VarR,VarNotR],Min,_T3),
	   nonvar(Min), Min = min(_,_),
	   Eq2 = equation(F,[VarR,VarNotR],L,_T2)
	),
	ground(GroundParam), var(VarR),
	( unify_with_occurs_check(Min,min(L, GofVarNotR))
	; unify_with_occurs_check(Min,min(GofVarNotR, L))
	),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[VarR-1,VarNotR])),
%	L = fun(F,[VarR-1,VarNotR]),
%	L =.. [F,VarR-1,VarNotR],
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ?
% | f(n,m) = f(n-1,m)
% | f(n,m) = max(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ Eq1, Eq2, Eq3 ], GofVarNotR) :-
	Eq1 = equation(F,[GroundParam,Q1],Q2,_T1),
	var(Q1), Q2 == '?',
	( Eq2 = equation(F,[VarR,VarNotR],Max,_T2),
	  nonvar(Max), Max = max(_,_),
	  Eq3 = equation(F,[VarR,VarNotR],L,_T3)
	-> true
	 ; Eq3 = equation(F,[VarR,VarNotR],Max,_T3),
	   nonvar(Max), Max = max(_,_),
	   Eq2 = equation(F,[VarR,VarNotR],L,_T2)
	),
	ground(GroundParam), var(VarR),
	( unify_with_occurs_check(Max,max(L, GofVarNotR))
	; unify_with_occurs_check(Max,max(GofVarNotR, L))
	),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[VarR-1,VarNotR])),
%	L = fun(F,[VarR-1,VarNotR]),
%	L =.. [F,VarR-1,VarNotR],
	varset(GofVarNotR,[VarNotR]), !.

% | f(c,?,m) = g(m)
% | f(x,n,m) = min(f(x-1,n,m),g(n))
% => g(min(n,m))
:- pred solve_min_max_equation(I,O):(term(I),var(O)) => (atom(O)).
solve_min_max_equation([ equation(F,[GroundParam,Q,Var2],GofVar2,_T1), equation(F,[Var,Var1,Var2],Min,_T2) ], Result ) :-
	ground(GroundParam), var(Var), var(Q),%Q == '?',
	( Min = min(L, GofVar1) ; Min = min(GofVar1, L) ),
	nonvar(L),
	L == fun(F,[Var-1,Var1,Var2]),
%	L =.. [F,Var-1,Var1,Var2],
	varset(GofVar2,[Var2]),
	varset(GofVar1,[Var1]),
	unifiable(GofVar1,GofVar2,_), !,
	min_(Var1,Var2,MinVars),
	substitute_vars(GofVar1,[Var1],[MinVars],SubstG),
	simplify(SubstG,Result).
	%simplify_in_mathematica(SubstG,Result).
% | f(c,?,m) = g(m)
% | f(x,n,m) = max(f(x-1,n,m),g(n))
% => g(max(n,m))
solve_min_max_equation([ equation(F,[GroundParam,Q,Var2],GofVar2,_T1), equation(F,[Var,Var1,Var2],Max,_T2) ], Result ) :-
	ground(GroundParam), var(Var), var(Q),%Q == '?',
	( Max = max(L, GofVar1) ; Max = max(GofVar1, L) ),
	nonvar(L),
	L == fun(F, [Var-1,Var1,Var2]),
%	L =.. [F,Var-1,Var1,Var2],
	varset(GofVar2,[Var2]),
	varset(GofVar1,[Var1]),
	unifiable(GofVar1,GofVar2,_), !,
	max_(Var1,Var2,MaxVars),
	substitute_vars(GofVar1,[Var1],[MaxVars],SubstG),
	simplify(SubstG,Result).
	%simplify_in_mathematica(SubstG,Result).

% | f(n,c,?) = ?
% | f(n,x,m) = min(f(n,x-1,m),g(n,m))
% => g(n,m)
solve_min_max_equation([ equation(F,[N,G,Q1],Q2,_T1), equation(F,[N,X,M],Min,_T2) ], Result) :-
	ground(G), var(N), var(M), var(Q1), Q2 == '?',
	( Min = min(L,GOfNM) ; Min = min(GOfNM, L) ),
	nonvar(L),
	L == fun(F,[N,X-1,M]),
%	L =.. [F,N,X-1,M],
	( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

% %%%%
% %([g(1,?,_J)==(?,[]),g(_D,_F,_J)==(max(_F*_J,fun(g,[_D-1,_F,_J])),[])])
% % | f(c,?,n) = ?
% % | f(x,n,m) = min(f(x-1,n,m),g(n,m))
% % => g(n,m)
solve_min_max_equation([ equation(F,[G,Q1,N],Q2,_T1), equation(F,[X,M,N],Min,_T2) ], Result) :-
	ground(G), var(N), var(M), var(Q1), Q2 == '?',
	( Min = min(L,GOfNM) ; Min = min(GOfNM, L) ),
	nonvar(L),
	L == fun(F,[X-1,M,N]),
%	L =.. [F,N,X-1,M],
	varset(GOfNM,VSet),included_vars(VSet,[M,N]),!,
	%( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  Result = GOfNM.
	  %simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

% %([g(1,?,_J)==(?,[]),g(_D,_F,_J)==(max(_F*_J,fun(g,[_D-1,_F,_J])),[])])
% % | f(c,?,n) = ?
% % | f(x,n,m) = max(f(x-1,n,m),g(n,m))
% % => g(n,m)
solve_min_max_equation([ equation(F,[G,Q1,N],Q2,_T1), equation(F,[X,M,N],Max,_T2) ], Result) :-
	ground(G), var(N), var(M), var(Q1), Q2 == '?',
	( Max = max(L,GOfNM) ; Max = max(GOfNM, L) ),
	nonvar(L),
	L == fun(F,[X-1,M,N]),
%	L =.. [F,N,X-1,M],
	varset(GOfNM,VSet),included_vars(VSet,[M,N]),!,
	% ( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  Result = GOfNM.
	  %simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).



% % | f(n,c,?) = ?
% % | f(n,x,m) = max(f(n,x-1,m),g(n,m))
% % => g(n,m)
solve_min_max_equation([ equation(F,[N,G,Q1],Q2,_T1), equation(F,[N,X,M],Max,_T2) ], Result) :-
	ground(G), var(N), var(M), var(Q1), Q2 == '?',
	( Max = max(L,GOfNM) ; Max = max(GOfNM, L) ),
	nonvar(L),
	L == fun(F, [N,X-1,M]),
%	L =.. [F,N,X-1,M],
	( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

%:- export(included_vars/2).
included_vars([],_).
included_vars([X|Y],T):-
	included_var(X,T),
	included_vars(Y,T).

included_var(X,[D|_T]):-
	D == X,!.
included_var(X,[_|T]):-
	included_var(X,T).


% BEGIN Code to work with symbolic expressions
:- pred max_(X,Y,Z) : term * term * var => term * term * term.
:- doc(max_/3,
	"@var{Z} is an expression representing the
         maximum between @var{X} and @var{Y}.").
% Special cases
max_(A,B,A) :- A == B, !.
max_(Q,A,A) :- Q == '?', !.
max_(A,Q,A) :- Q == '?', !.
max_(I,_,inf) :- I == 'inf', !.
max_(_,I,inf) :- I == 'inf',!.
max_(Z,A,A) :- Z == 0, !.
max_(A,Z,A) :- Z == 0, !.
% Simplification
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
% Variables
max_(N,M,max(N,M)) :- var(N), var(M), N @=< M, !.
max_(N,M,max(M,N)) :- var(N), var(M), M @=< N, !.
% Syntactic
max_(N,M,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(M,N,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(N,M,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
max_(M,N,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
% Not ground elements
max_(N,M,N):-
	greater_or_equal_than(N,M),!.
max_(N,M,M):-
	greater_or_equal_than(M,N),!.


max_(N,M,max(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
max_(N,M,max(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
max_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N > M -> O = N ; O = M ).
max_(N,M,max(M,N)) :- N @>= M, !.
max_(N,M,max(N,M)) :- N @< M.

:- pred min_(X,Y,Z) : term * term * var => term * term * term.
:- doc(min_/3,
	"@var{Z} is an expression representing the
         minimum between @var{X} and @var{Y}.").
% Special cases
min_(A,B,A) :- A == B, !.
min_(Q,A,A) :- Q == '?', !.
min_(A,Q,A) :- Q == '?', !.
min_(I,A,A) :- I == 'inf', !.
min_(A,I,A) :- I == 'inf', !.
min_(Z,_,0) :- Z == 0, !.
min_(_,Z,0) :- Z == 0, !.
% Simplification
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
% Variables
min_(N,M,min(N,M)) :- var(N), var(M), M @=< N, !.
min_(N,M,min(M,N)) :- var(N), var(M), N @=< M, !.
% Syntactic
min_(N,M,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(M,N,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(N,M,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
min_(M,N,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
% Not ground elements
% Using solver
min_(N,M,N):-
	greater_or_equal_than(M,N),!.
min_(N,M,M):-
	greater_or_equal_than(N,M),!.
min_(N,M,min(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
min_(N,M,min(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
min_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N < M -> O = N ; O = M ).
min_(N,M,N) :- N @=< M, !.
min_(N,M,M) :- N @> M.


:- pred get_input_positions(Equations,Positions)
	: list(equation) * var
       => list(equation) * list(list(num)).
:- doc(get_input_positions/2,
	"Finds the @var{Positions} of the variables that
         are used as input for the @var{Equations}.").
get_input_positions([],[]).
get_input_positions([equation(_,IArgs,_,_T)|EqR],Pos) :-
	leaves_position(IArgs,PosEq),
	% varfilter(PosEqPlusEmpty,[],PosEq),
	get_input_positions(EqR,PosR),
	ord_union(PosEq,PosR,Pos).
