:- module(solver_strategies,[rsolve/5,rsolve/4],[assertions,nativeprops]).

% TODO: REMOVE WARNINGS ABOUT MULTI ARITY.
% TODO: qualified predicate calls to predicates from solver_strategy_chain.

:- use_module(library(aggregates)).
:- use_module(solver_utils, 
	[
	    solve_rec_rel/5,
	    simplify/3,
	    simplify/2,
	    expand/3,
	    expand/2,
	    solve/4,
	    solve/3
	]
).
:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    backend_solver/1,
	    recurrence_relation/1,
	    solver_strategy/1,
%	    bound/1,
	    algebraic_expression/1
	]
).
:- use_module(ciaopp(preprocess_flags)).
:- use_module(library(messages), 
	[warning_message/2, debug_message/2, error_message/2]).
:- use_module(solver_strategy_chain).

%NOTE: rsolve/4 is limited to one function only, i.e., it is not
%possible solving a system of more than one functions.

:- pred rsolve(Strategy,RecRel,Vars,Approx,Sol):
	(recurrence_relation(RecRel),
	 list(Vars,var),solver_strategy(Strategy), bound(Approx),
	 var(Sol)) => algebraic_expression(Sol) # "Try to solve the
 recurrence equation or to find a safe @var{Approx}
 approximation, following the strategy @var{Strategy}".

% TODO: Take into account that some solvers could not being
% installed. Show Warnings in those cases.
% DONE:%TODO: Define a default strategy, and implement rsolve/3 using 
%rsolve/4 and that strategy.
% TODO: Implement a strategy for transforming the equations taking
% into account the approximation, in order to make it solvable.  (For
% example adding base cases to linear second-order diff. equations
% with less than two base cases).

% rsolve(chain(Solvers),RecRel,Vars,Approx,Sol):-
% 	solver_strategy_chain:rsolve(Solvers,RecRel,Vars,Approx,Sol).
rsolve(chain,RecRel,Vars,Approx,Sol):-
	bagof(Solver,backend_solver(Solver),Solvers),
	solver_strategy_chain:rsolve(Solvers,RecRel,Vars,Approx,Sol).

rsolve(Solver,RecRel,Vars,Approx,Sol):-
	backend_solver(Solver),
	solve_rec_rel(Solver,RecRel,Vars,Approx,Sol), 
	debug_message("Solved using ~q solver",[Solver]).

:- pred rsolve(RecRel,Vars,Approx,Sol): 
	(recurrence_relation(RecRel),list(Vars,var),bound(Approx),var(Sol)) =>
	 algebraic_expression(Sol) #
 "Try to solve the recurrence equation or to find a safe @var{Approx} approximation, following the default strategy".

% TODO: Define the default strategy with a ciaopp flag.
rsolve(RecRel,Vars,Approx,Sol):-
	findall(X,backend_solver(X),Solvers),
	solver_strategy_chain:rsolve(Solvers,RecRel,Vars,Approx,Sol).

% solve_rec_rel_chain(Solvers,RecRel,Vars,Approx,Sol):-
% 	member(Solver,Solvers),
% 	solve_rec_rel(Solver,RecRel,Vars,Approx,Sol),
% 	debug_message("Solved using ~q solver",[Solver]).
% solve_rec_rel_chain(_Solvers,RecRel,_Vars,Approx,_Sol):-
% 	warning_message("Recurrence not solved: ~q (~q)",[RecRel,Approx]),!.

