:- module(solver_strategy_chain,[rsolve/5],[assertions,nativeprops]).
:- use_module(library(aggregates)).
:- use_module(library(lists), [member/2]).
:- use_module(domain(resources/eq_solver/solver),[available_solver/1]).
:- use_module(solver_utils, 
	[
	    solve/4,
	    solve/3,
	    simplify/3,
	    simplify/2,
	    expand/3,
	    expand/2,
	    solve_rec_rel/5,
	    solve_rec_rel/4
	]
).
:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    backend_solver/1,
	    recurrence_relation/1,
	    solver_strategy/1,
%	    bound/1,
	    algebraic_expression/1
	]
).

:- use_module(ciaopp(preprocess_flags)).
:- use_module(library(messages), 
	[warning_message/2, debug_message/2, error_message/2]).




%NOTE: rsolve/4 is limited to one function only, i.e., it is not
%possible solving a system of more than one functions.

:- pred rsolve(Solvers,RecRel,Vars,Approx,Sol):
	(list(Solvers,backend_solver),recurrence_relation(RecRel),list(Vars,var),bound(Approx),var(Sol))
	=> algebraic_expression(Sol) # "Try to solve the recurrence
	equation or to find a safe @var{Approx} approximation,
	with the first solver of the list able to solve it".

% TODO: Take into account that some solvers could not being
% installed. Show Warnings for those cases.
% TODO: Define a default strategy, and implement rsolve/3 using 
%rsolve/4 and that strategy.
% TODO: Implement a strategy for transforming the equations taking
% into account the approximation, in order to make it solvable.  (For
% example adding base cases to linear second-order diff. equations
% with less than two base cases).

rsolve(Solvers,RecRel,Vars,Approx,Sol):-
	%active_solvers(Solvers0,Solvers),!,
	member(Solver,Solvers),
	available_solver(Solver),
	solve_rec_rel_(Solver,RecRel,Vars,Approx,Sol),!,
	debug_message("Closed-form found for RR. ~q using ~q solver",[RecRel,Solver]).
rsolve(Solvers,RecRel,Vars,Approx,_Sol):-
	debug_message("The following RR cold not be solved by chain (~q): (~q, ~q, ~q)",[Solvers,RecRel,Vars,Approx]),!,
	fail.
%	throw(not_solved_by_strategy(chain,RecRel,Vars,Approx)).


% :- pred active_solvers(Solvers,ActiveSolvers): (list(Solvers,backend_solver),var(ActiveSolvers)) => (list(ActiveSolvers,backend_solver))
% 	# "Returns in @var{ActiveSolvers} all the solvers contained in @var{Solvers}
%  that are currently available for being used.".
% active_solvers([],[]).
% active_solvers([X|Y],[X|Z]):-
% 	is_active(X),!,
% 	active_solvers(Y,Z).
% active_solvers([_X|Y],Z):-
% 	active_solvers(Y,Z).

solve_rec_rel_(builtin,RecRel,Vars,Approx,Sol):-
	solve_rec_rel(builtin,RecRel,Vars,Approx,Sol).
solve_rec_rel_(mathematica,RecRel,Vars,A,Sol):-
	solve_rec_rel(mathematica,RecRel,Vars,A,Sol).
solve_rec_rel_(ppl,RecRel,V,A,Sol):-
	catch(solve_rec_rel(ppl,RecRel,V,A,Sol),_,fail).
