:- module(solver_utils_builtin,
	[ 
	    solve_rec_ciaopp_solver/2,
	    simplify_in_ciaopp/2,
	    greater_than_ciaopp/2,
	    greater_or_equal_than_ciaopp/2,
	    equal_than_ciaopp/2,
	    translate_from_algebraic/2,
	    translate_to_algebraic/2,
	    order_of_function/2,
	    trivial_tautologies/1
	],
	[assertions,
	 nativeprops,
	 regtypes,
	 dcg]).

:- use_module(engine(hiord_rt), [call/1]). % TODO: review uses here

:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(sets)).
:- use_module(library(terms_check)).
:- use_module(library(terms_vars)).
:- use_module(library(iso_misc)).
:- use_module(ciaopp(preprocess_flags), [current_pp_flag/2] ).
:- use_module(library(messages)).
:- import(algebraic_res, [order_of_function/2]).
:- import(normal_form_res ,[]).
:- import(general_form_res ,[]).
:- import(diff_equ_res, [
 normal_form/2 ,	   
 diff_equ_type/7 ,	   
 solve_diff_equ/7 ,	   
 general_form/2 ,	   
 normal_form/2 ,	   
 normal_form/2 ,	   
 exp_greater_than/2 ,	   
 exp_greater_eq_than/2 ,   
 exp_eq_than/2 ,	   
 simplification/2
	]).
:- use_module(domain(resources/eq_solver/eq_solver_graphs), [varmember/2]).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).
:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(domain(resources/plai2infer), [replace_term/3]).
:- use_module(typeslib(typeslib), [dz_equivalent_types/2]).
:- use_module(domain(resources/eq_solver/solver/solver_utils_trans), [wrap_math_fun/2, unwrap_fun/2]).

:- doc(title, "Operations of the builtin ciaopp solver").
:- doc(author, "Maximiliano Klemen").
:- doc(module, 
	"It defines an API for communication with CiaoPP builtin solver."). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%CODE ORIGINALLY IN eq_solver MODULE%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% TODO: Revision to remove unused code.
% TODO: The predicate equations_to_general_form and all its relative predicates
% have to be in solver_utils_trans module.

:- pred get_input_positions(Equations,Positions)
	: list(equation) * var
       => list(equation) * list(list(num)).

:- doc(get_input_positions/2,
	"Finds the @var{Positions} of the variables that
         are used as input for the @var{Equations}.").
get_input_positions([],[]).
get_input_positions([equation(_,IArgs,_,_T)|EqR],Pos) :-
	leaves_position(IArgs,PosEq),
	% varfilter(PosEqPlusEmpty,[],PosEq),
	get_input_positions(EqR,PosR),
	ord_union(PosEq,PosR,Pos).

% TODO: Bound is not used. 
:- pred analyze_eq_vars(Positions,Equations,Bound,Information)
	: list(list(num)) * list(equation) *atom * var
       => list(list(num)) * list(equation) *atom * list.
:- doc(analyze_eq_vars/3,
	"Analyzes the variable @var{Positions} in the
         @var{Equations} returning a list of triples
         (corresponding variable, the variable was used
	 in the expression, the variable was used as
	 recurrence argument).").
analyze_eq_vars([],_,_,[]).
analyze_eq_vars([Pos|PosR],Eqs,Bound,[(Pos,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests)|AnR]) :-
	analyze_eq_vars_pos(Eqs,Pos,Bound,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests),
	analyze_eq_vars(PosR,Eqs,Bound,AnR).

analyze_eq_vars_pos([],_,_,false,false,false,true,false).
analyze_eq_vars_pos([equation(Function,IArgs,Expr,Tests)|EqR],Pos,Bound,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests) :-
	analyze_eq_vars_pos(EqR,Pos,Bound,UExprR,UArgR,UUnkR,UAllGR,UInTests), !,
	% Analyze our equation
	term_at_position(IArgs,Pos,TheArg),
	( ground(TheArg)
	-> ThisUsedExpr = false,
	ThisUsedArgInRec = false, % it is not used in the recursive call in principle.
	%ThisUsedArgInRec = true, 
	ThisUsedAllGround = true,
	ThisUsedInTests = false,
	   ( TheArg = '?'
	   -> ThisUsedUnknown = true
	    ; ThisUsedUnknown = false
	   )
	 ; analyze_expr(Expr,Function,TheArg,ThisUsedExpr,ThisUsedArgInRec),
	   ThisUsedUnknown = false, ThisUsedAllGround = false,
	   varset(Tests,Vars), (varmember(TheArg,Vars) -> ThisUsedInTests = true; ThisUsedInTests = false)
	),
	% Combine both
	or(UExprR,ThisUsedExpr,UsedExpr),
	or(UArgR,ThisUsedArgInRec,UsedArgInRec),
	or(UUnkR,ThisUsedUnknown,UsedUnknown),
	or(UInTests,ThisUsedInTests,UsedInTests),
	and(UAllGR,ThisUsedAllGround,UsedAllGround).

analyze_expr(Expr,_,Var,true,false) :-
	var(Expr),
	Expr == Var, !. % Our var is used
analyze_expr(Expr,_,Var,false,false) :-
	var(Expr),
	Expr \== Var, !. % Another var
analyze_expr(Expr,_,_,false,false) :-
	ground(Expr), !. % Ground terms has no vars
analyze_expr(Expr,Function,Var,false,UsedArgInRec) :-
	Expr =.. [Function|CallArgs], !,
	varset(Expr,CallVars),
	% We need it to be in the variables inside
	% but not in first level
	( varmember(Var,CallArgs)
	-> UsedArgInRec = false
	 ; ( varmember(Var,CallVars)
	   -> UsedArgInRec = true
	    ; UsedArgInRec = false
	   )
	).
analyze_expr(Expr,Function,Var,UsedExpr,UsedArgInRec) :-
	functor(Expr,G,N),
	G \== Function,
	analyze_expr_arg_each(Function,Var,Expr,1,N,UsedExpr,UsedArgInRec).

analyze_expr_arg_each(_,_,_,I,N,false,false) :-
	I > N, !.
analyze_expr_arg_each(Function,Var,Expr,I,N,UsedExpr,UsedArgInRec) :-
	I =< N, !,
	% The rest
	I1 is I + 1,
	analyze_expr_arg_each(Function,Var,Expr,I1,N,UsedExprR,UsedArgInRecR),
	% Analyze this argument
	arg(I,Expr,Arg),
	analyze_expr(Arg,Function,Var,UsedExprArg,UsedArgInRecArg),
	% Combine both
	or(UsedExprR,UsedExprArg,UsedExpr),
	or(UsedArgInRecR,UsedArgInRecArg,UsedArgInRec).

or(false,false,false) :- !.
or(true,_,true) :- !.
or(_,true,true) :- !.

and(true,true,true) :- !.
and(false,_,false) :- !.
and(_,false,false) :- !.

not(true,false).
not(false,true).

eval_bool(P,true):-
	call(P),!.
eval_bool(_,false).

:- pred decide_eq_vars(Bound,Info,ChosenVars)
	: list * bound * var
       => list * bound * list(list(num)).
:- doc(decide_eq_vars/3,
	"Decides, taking into account the @var{Info}
         in the analyze_eq_vars phase, which variable
         positions should be included in the
         final equation to be solved.").
decide_eq_vars([],_,[]).
decide_eq_vars([(V1,UEx1,UArg1,UUnk1,_UAllG1,T1),(V2,UEx2,UArg2,UUnk2,_UAllG2,T2)|R],Bound,ChosenVars) :-
	decide_eq_vars(R,Bound,ChosenVarsR), !,
	decide_table(UEx1,UArg1,UUnk1,UEx2,UArg2,UUnk2,Bound,PInclude1,PInclude2),
	eval_bool(Bound=lower,Lb),
	and(Lb,T1,IsLow),
	eval_bool(Bound=upper,Ub),
	and(Ub,T2,IsUpp),
	or(PInclude1,IsLow,Include1),
	or(PInclude2,IsUpp,Include2),
	( Include2 = true
	-> ChosenStep1 = [V2|ChosenVarsR]
	 ; ChosenStep1 = ChosenVarsR
	),
	( Include1 = true
	-> ChosenVars = [V1|ChosenStep1]
	 ; ChosenVars = ChosenStep1
	).

:- pred decide_table(UEx1,UArg1,UUnk1,UEx2,UArg2,UUnk2,Bound,Include1,Include2).
% There are 128 possibilities 

% If both appear in expressions, we must include both
decide_table(true, _, _,  true, _, _,  _, true, true) :- !. % 32 - 32

% Only one is available -> include only that
decide_table(false, false, _, true, _, _,  _, false, true) :- !. % 16 - 48
decide_table(false, false, _, false, true, UUnk2, _, false, I2) :- % 8 - 56
	!, not(UUnk2,I2).

decide_table(true, _, _, false, false, _, _, true, false) :- !. % 16 - 72
decide_table(false, true, UUnk1, false, false, _, _, I1, false) :- % 8 - 80
	!, not(UUnk1,I1).

% None of them are available -> include none
decide_table(false, false, _, false, false, _,  _, false, false) :- !. % 8 - 88

% One in expression, the other in argument -> include both
decide_table(true, false, _, false, true, UUnk2, _, true, I2) :- % 8 - 96
	!, not(UUnk2,I2).
decide_table(false, true, UUnk1, true, false, _, _, I1, true) :- % 8 - 104
	!, not(UUnk1,I1).

% One in both, the other in argument -> depends on bound
decide_table(true, true, _, false, true, _, lower, true, false) :- !. % 4 - 108
decide_table(true, true, _, false, true, UUnk2, upper, true, I2) :- % 4 - 112
	!, not(UUnk2,I2).
decide_table(false, true, UUnk1, true, true, _, lower, I1, true) :- % 4 - 116
	!, not(UUnk1,I1).
decide_table(false, true, _, true, true, _, upper, false, true) :- !. % 4 - 120

% Both in argument -> only the one for our bound
decide_table(false, true, UUnk1, false, true, _, lower, I1, false) :- % 4 - 124
	!, not(UUnk1,I1).
decide_table(false, true, _, false, true, UUnk2, upper, false, I2) :- % 4 - 128
	!, not(UUnk2,I2).

decide_not_ground_vars([],_,[]).
decide_not_ground_vars([V|R],Info,[V|NotGR]) :-
	member((V,_,_,_,false,_),Info), !,
	decide_not_ground_vars(R,Info,NotGR).
decide_not_ground_vars([V|R],Info,NotGR) :-
	member((V,_,_,_,true,_),Info), !,
	decide_not_ground_vars(R,Info,NotGR).

:- pred add_extra_equations(Equations,Exception,InputArgs,AlreadySolved,AllEqs).
:- doc(add_extra_equations/5,
	"Add those equations from the @var{AlreadySolved} ones
         which are referenced from the initial @var{Equations}
         and returns them in @var{AllEqs}, except for @var{Exception}.").
% First taking themselves needed for coupled equations
add_extra_equations(Eqs,Exception,IArgs,Already,AllEqs) :-
	add_extra_eqs_round(Eqs,Exception,IArgs,Already,RoundEqs),
	sort(Eqs,EqsSorted),
	sort(RoundEqs,RoundEqsSorted),
	( ord_subset(RoundEqsSorted,EqsSorted)
	-> AllEqs = Eqs
	 ; ord_subtract(RoundEqsSorted,EqsSorted,NewEqs),
	   append(Eqs,NewEqs,NextStepEqs),
	   add_extra_equations(NextStepEqs,Exception,IArgs,Already,AllEqs)
	).

add_extra_eqs_round([],_,_,_,[]).
add_extra_eqs_round([equation(F,_,Expr,_T)|R],F,IArgs,Already,RoundEqs) :-
	!,
	get_extra_eqs_expr(Expr,Already,ThisEqs),
	choose_with_correct_args(ThisEqs,F,IArgs,ThisEqs2),
	add_extra_eqs_round(R,F,IArgs,Already,REqs),
	append(ThisEqs2,REqs,RoundEqs).
add_extra_eqs_round([equation(G,_,Expr,_T)|R],F,IArgs,Already,RoundEqs) :-
	G \= F,
	get_extra_eqs_expr(Expr,Already,ThisEqs),
	add_extra_eqs_round(R,F,IArgs,Already,REqs),
	append(ThisEqs,REqs,RoundEqs).

choose_with_correct_args([],_,_,[]).
choose_with_correct_args([equation(F,IArgs,E,T)|R],F,CorrectArgs,Out) :-
	!, choose_with_correct_args(R,F,CorrectArgs,OutR),
	( unifiable(IArgs,CorrectArgs,_)
	-> Out = [equation(F,IArgs,E,T)|OutR]
	 ; Out = OutR
	).
choose_with_correct_args([equation(G,IArgs,E,T)|R],F,CorrectArgs,[equation(G,IArgs,E,T)|CR]) :-
	G \= F, !,
	choose_with_correct_args(R,F,CorrectArgs,CR).

get_extra_eqs_expr(Var,_,[]) :-
	var(Var), !.
get_extra_eqs_expr(Funtor,Already,Eqs) :-
	Funtor =.. [F|Args], !,
	get_extra_eqs_arg_each(Args,Already,ArgEqs),
	( find_function(Already,F,FEqs)
	-> append(FEqs,ArgEqs,Eqs)
	 ; Eqs = ArgEqs
	).

get_extra_eqs_arg_each([],_,[]).
get_extra_eqs_arg_each([A|R],Already,Eqs) :-
	get_extra_eqs_expr(A,Already,EqA),
	get_extra_eqs_arg_each(R,Already,EqR),
	append(EqA,EqR,Eqs).

find_function([resolution(_,_,_,Vars)|_],F,Eqs) :-
	find_function_in_vars(Vars,F,Eqs), !.
find_function([_|R],F,Eqs) :-
	find_function(R,F,Eqs).

find_function_in_vars([(_,unsolved_equations(_,Eqs))|_],F,Eqs) :-
	Eqs = [equation(F,_,_,_)|_], !.
find_function_in_vars([_|R],F,Eqs) :-
	find_function_in_vars(R,F,Eqs).

remove_duplicate_eqs([],[]).
remove_duplicate_eqs([E|R],Rem) :-
	remove_duplicate_eqs(R,RemR),
	( member(X,RemR), X == E
	-> Rem = RemR
	 ; Rem = [E|RemR]
	).


:- pred trivial_tautologies(Constrains): (test_set(Constrains)) #
"Checks if the set of constrains contains just trivial tautologies.  A
trivial tautology is a test that with all its arguments grounds and it
is true (e.g. 1=1, 2=<3), or an unification of the form (X=X).".

trivial_tautologies([]).
trivial_tautologies([X|T]):-
	trivial_tautology(X),
	trivial_tautologies(T).

:- pred trivial_tautology(Test): (test(Test)).
trivial_tautology(F):-
	F =.. [_Op|Args],
	list(Args,ground),
	call(F),!.
trivial_tautology(X=Y):-
	X == Y.

%:- export(solve_rec_ciaopp_solver/2).
% BEGIN Code to interface CiaoPP recurrence solver
:- pred solve_rec_ciaopp_solver(Equations,Solution).
:- doc(solve_rec_ciaopp_solver/2,
	"Solves the @var{Equations} using the integrated CiaoPP solver.").

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ADDED SPECIAL TRIVIAL CASE. THIS SPECIAL CASES
% ARE GOING TO BE HANDLED IN A MODULE IN CHARGE
% OF SOLVING TRIVIAL CASES AND PERFORMING 
% TRANSFORMATIONS ON EQUATIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
solve_rec_ciaopp_solver([equation(Name1,[Args1],Exp1,T1),equation(Name2,[Args2],Exp2,T2)],Solution):-
	Name1 == Name2,
	Args1 == Exp1,Args2 == Exp2,
	trivial_tautologies(T1),trivial_tautologies(T2),
	(
	    (ground(Args1),var(Args2),!, Solution = Args2);
	    (ground(Args2),var(Args1),!, Solution = Args1)
	),!.

% solve_rec_ciaopp_solver([equation(Name1,[Args1],Exp1,T1),equation(Name2,[Args2],Exp2,T2)],Solution):-
% 	Name1 == Name2,
% 	var(Args1),
% 	Args2 == 0,
% 	nonvar(Exp2),num(Exp2),
% 	nonvar(Exp1), Exp1 = fun(Name1,[RExp]) + C2,nonvar(RExp),
% 	RExp = C1*Args1 - C1, nonvar(C1), nonvar(C2), ground(C1),ground(C2),
% 	C1 < 1, C3 is 1/C1,   
% 	trivial_tautologies(T1),trivial_tautologies(T2),!,
% 	Solution = (log(Args1+1)/log(C3)) + C2.

% solve_rec_ciaopp_solver([equation(Name2,[Args2],Exp2,T2),equation(Name1,[Args1],Exp1,T1)],Solution):-
% 	Name1 == Name2,
% 	var(Args1),
% 	Args2 == 0,
% 	nonvar(Exp2),num(Exp2),
% 	nonvar(Exp1), Exp1 = fun(Name1,[RExp]) + C2,nonvar(RExp),
% 	RExp = C1*Args1 - C1, nonvar(C1), nonvar(C2), ground(C1),ground(C2),
% 	C1 < 1, C3 is 1/C1,   
% 	trivial_tautologies(T1),trivial_tautologies(T2),!,
% 	Solution = (log(Args1+1)/log(C3)) + C2.

	
	
	
solve_rec_ciaopp_solver(Equations,Solution) :-
	Equations = [equation(F,_,_,_Test)|_],
	% We can only solve 1-parameter 1-equation recurrences
	equations_to_general_form(Equations,(VarM,NumberM),[ChosenVar],[EqNF],EqBoundary),
	solve_general_form_rec_rel(EqNF,F,ChosenVar,EqBoundary,SolutionNF),
	general_form_to_vars(SolutionNF,NumberM,VarM,Solution).

:- pred equations_to_general_form(Equations,VarMappings,ChosenVars,EqNF,EqBoundary).
equations_to_general_form(Equations,(Vars,Numbers),ChosenN,EqNF,EqBoundary) :-
	get_input_positions(Equations,IPos),
	get_mappings(Equations,1,IPos,Vars,Numbers),
	discard_tests(Equations,Equations0), % because by now we are not using them.
	analyze_eq_vars(IPos,Equations0,_,Info),
	decide_general_form_vars(Info,ChosenVars),
	terms_at_positions(Numbers,ChosenVars,ChosenN),
	convert_to_equations(Equations0,ChosenVars,Vars,Numbers,EqNF,EqBoundary).

discard_tests([],[]).
discard_tests([equation(A,B,C,_D)|R], [equation(A,B,C,[])|RR]):-
	discard_tests(R,RR).

get_mappings(_,_,[],[],[]).
get_mappings(Equations,I,[Pos|PosR],[Var|VarR],[$(I)|NumberR]) :-
	get_mappings_(Equations,Pos,Var),
	I1 is I+1,
	get_mappings(Equations,I1,PosR,VarR,NumberR).

get_mappings_([],_,0).
get_mappings_([equation(_,IPos,_,_T)|EqR],Pos,Var) :-
	get_mappings_(EqR,Pos,VarR),
	term_at_position(IPos,Pos,EqVar),
	( var(VarR)
	-> Var = VarR
	 ; Var = EqVar
	).

decide_general_form_vars([],[]).
decide_general_form_vars([(V,_,true,_,_)|VR],[V|ChosenR]) :-
	decide_general_form_vars(VR,ChosenR).
decide_general_form_vars([(_,_,false,_,_)|VR],ChosenR) :-
	decide_general_form_vars(VR,ChosenR).

convert_to_equations([],_,_,_,[],[]).
convert_to_equations([equation(F,IArgs,Expr,_T)|EqR],ChosenVars,Vars,Numbers,EqNF,EqB) :-
	convert_to_equations(EqR,ChosenVars,Vars,Numbers,EqNFR,EqBR),
	terms_at_positions(IArgs,ChosenVars,EqInputArgs),
	( ground(EqInputArgs)
	-> % We have a base case
	   substitute_inside_equation(Expr,F,ChosenVars,Vars,Numbers,BoundExpr),
	   ( EqInputArgs = [Base]
	   -> BoundVal = Base
	    ; BoundVal =.. [','|EqInputArgs]
	   ),
	   EqNF = EqNFR, EqB = [val(BoundVal,BoundExpr)|EqBR]
	 ; substitute_inside_equation(Expr,F,ChosenVars,Vars,Numbers,EqExpr),
	   EqNF = [EqExpr|EqNFR], EqB = EqBR
	).

substitute_inside_equation(Var,_,_,Vars,Numbers,Subst) :-
	var(Var), !,
	substitute_vars(Var,Vars,Numbers,Subst).
substitute_inside_equation(Gr,_,_,_,_,Gr) :-
	ground(Gr), !.
substitute_inside_equation(Functor,F,ChosenVars,Vars,Numbers,Subst) :-
	Functor =.. [F|Args], !,
	terms_at_positions(Args,ChosenVars,ChosenArgs),
	substitute_inside_equation_each(ChosenArgs,F,ChosenVars,Vars,Numbers,SubstArgs),
	Subst =.. [F|SubstArgs].
substitute_inside_equation(Functor,F,ChosenVars,Vars,Numbers,Subst) :-
	Functor =.. [G|Args], G \== F, !,
	substitute_inside_equation_each(Args,F,ChosenVars,Vars,Numbers,SubstArgs),
	Subst =.. [G|SubstArgs].

substitute_inside_equation_each([],_,_,_,_,[]).
substitute_inside_equation_each([Arg|RArg],F,ChosenVars,Vars,Numbers,[SubstArg|SubstR]) :-
	substitute_inside_equation(Arg,F,ChosenVars,Vars,Numbers,SubstArg),
	substitute_inside_equation_each(RArg,F,ChosenVars,Vars,Numbers,SubstR).

:- pred general_form_to_vars(Expr,Numbers,Vars,VarsExpr).
general_form_to_vars($(I),Numbers,Vars,Var) :-
	find_var_by_number($(I),Numbers,Vars,Var),!.
general_form_to_vars(Functor,Numbers,Vars,Expr) :-
	Functor =.. [F|Args],
	sum_or_prod(F),!,
	 Args = [I|_],
	general_form_to_vars_each(Args,[I|Numbers],[_I|Vars],ExprArgs),
	Expr =.. [F|ExprArgs].
general_form_to_vars(Functor,Numbers,Vars,Expr) :-
	Functor =.. [F|Args], !,
	general_form_to_vars_each(Args,Numbers,Vars,ExprArgs),
	Expr =.. [F|ExprArgs].

sum_or_prod(sum).
sum_or_prod(prod).

find_var_by_number(A,[A|_],[Var|_],Var) :- !.
find_var_by_number(A,[_|R],[_|VarR],Var) :-
	find_var_by_number(A,R,VarR,Var).

general_form_to_vars_each([],_,_,[]).
general_form_to_vars_each([Arg|ArgR],Numbers,Vars,[V|VR]) :-
	general_form_to_vars(Arg,Numbers,Vars,V),
	general_form_to_vars_each(ArgR,Numbers,Vars,VR).

:- pred solve_general_form_rec_rel(Equation,F,Var,Boundary,Solution).
:- doc(solve_general_form_rec_rel/5,
	"Solves the @var{Equation} for @var{F}(@var{Var}) and
         with @var{Boundary} conditions in @var{Solution}.
         The input and output must be in general form.").
solve_general_form_rec_rel(Equation,_F,_Var,_Boundary,Equation) :-
	num(Equation),!.
solve_general_form_rec_rel(Equation,F,Var,Boundary,Solution) :-
	normal_form(Equation,EqNormalForm),
	normal_form_each_boundary(Boundary,BoundaryNF),
	diff_equ_type(EqNormalForm,Var,F/1,A1n,A2n,Bn,T),
	solve_diff_equ(T,Var,A1n,A2n,Bn,BoundaryNF,SolutionNF),
	general_form(SolutionNF,Solution).

normal_form_each_boundary([],[]).
normal_form_each_boundary([val(Arg,Res)|R],[val(ArgN,ResN)|RN]) :-
	normal_form(Arg,ArgN),
	normal_form(Res,ResN),
	normal_form_each_boundary(R,RN).
% END Code to interface CiaoPP recurrence solver

% BEGIN Code to work with symbolic expressions
:- pred max_(X,Y,Z) : term * term * var => term * term * term.
:- doc(max_/3,
	"@var{Z} is an expression representing the
         maximum between @var{X} and @var{Y}.").
% Special cases
max_(A,B,A) :- A == B, !.
max_(Q,A,A) :- Q == '?', !.
max_(A,Q,A) :- Q == '?', !.
max_(I,_,inf) :- I == 'inf', !.
max_(_,I,inf) :- I == 'inf',!.
max_(Z,A,A) :- Z == 0, !.
max_(A,Z,A) :- Z == 0, !.
% Simplification
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
% Variables
max_(N,M,max(N,M)) :- var(N), var(M), N @=< M, !.
max_(N,M,max(M,N)) :- var(N), var(M), M @=< N, !.
% Syntactic
max_(N,M,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(M,N,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(N,M,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
max_(M,N,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
% Not ground elements
max_(N,M,max(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
max_(N,M,max(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
max_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N > M -> O = N ; O = M ).
max_(N,M,max(M,N)) :- N @>= M, !.
max_(N,M,max(N,M)) :- N @< M.

:- pred min_(X,Y,Z) : term * term * var => term * term * term.
:- doc(min_/3,
	"@var{Z} is an expression representing the
         minimum between @var{X} and @var{Y}.").
% Special cases
min_(A,B,A) :- A == B, !.
min_(Q,A,A) :- Q == '?', !.
min_(A,Q,A) :- Q == '?', !.
min_(I,A,A) :- I == 'inf', !.
min_(A,I,A) :- I == 'inf', !.
min_(Z,_,0) :- Z == 0, !.
min_(_,Z,0) :- Z == 0, !.
% Simplification
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
% Variables
min_(N,M,min(N,M)) :- var(N), var(M), M @=< N, !.
min_(N,M,min(M,N)) :- var(N), var(M), N @=< M, !.
% Syntactic
min_(N,M,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(M,N,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(N,M,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
min_(M,N,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
% Not ground elements
min_(N,M,min(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
min_(N,M,min(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
min_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N < M -> O = N ; O = M ).
min_(N,M,N) :- N @=< M, !.
min_(N,M,M) :- N @> M.

:- pred add_(X,Y,Z) : term * term * var => term * term * term.
:- doc(add_/3,
	"@var{Z} is an expression representing the
         sum of @var{X} and @var{Y}.").
add_(Q,_,'?') :- Q == '?', !.
add_(_,Q,'?') :- Q == '?', !.
add_(A,B,C) :-
	ground(A), arithexpression(A),
	ground(B), arithexpression(B),
	!, C is A + B.
add_(A,B,A+B).

:- pred subtract_(X,Y,Z) : term * term * var => term * term * term.
:- doc(subtract_/3,
	"@var{Z} is an expression representing the
         difference between @var{X} and @var{Y}.").
subtract_(Q,_,'?') :- Q == '?', !.
subtract_(_,Q,'?') :- Q == '?', !.
subtract_(A,B,C) :-
	ground(A), arithexpression(A),
	ground(B), arithexpression(B),
	!, C is A - B.
subtract_(A,B,A-B).
% END Code to work with symbolic expressions



%%%%%%%%BEGIN Code to work with a recurrence equation solver (mathematica and builtin ciaopp solver)%%%%%%%%.

:- doc(bug, "simplify/2, expand/2, greater_than/2, solve_rec_with_solver/3 could be defined in a new module that acts as an interface between @pred{eq_solver} and the recurrence equation solvers.").

:- pred solve_rec_with_solver(AllEqs,FinalVars,SolutionExpr) # "Tries to solve the recurrence equations in @var{AllEqs} on the variables on @var{FinalVars}, returning in @var{SolutionExpr} the solution.".

:- doc(bug, "There are some recurrences that cannot be solved by the builtin solver but are solvable by mathematica. For example some first order recurrences like f(X) = f(X-1)*f(X-1), f(0)=1. Some failing cases are given as test cases in @pred{solve_rec_with_solver_/4}").
solve_rec_with_solver(AllEqs,FinalVars,SolutionExpr):-
	current_pp_flag(req_solver, S),
	solve_rec_with_solver_(S,AllEqs,FinalVars,SolutionExpr).

:- test solve_rec_with_solver_(Solver, Eqs, Vars, R): (var(R),Solver=builtin, Eqs=[equation(f,[X],f(X-1)*f(X-1)),equation(f,[1],1)], Vars=[X]) => (R=1) # "first order non-linear recurrence (f(x) = f(x-1)*f(x-1)) using builtin.".

solve_rec_with_solver_(builtin,AllEqs,_,SolutionExpr):-
	solve_rec_ciaopp_solver(AllEqs,SolutionExpr).

greater_than_ciaopp(X,Y):-
	varset(X-Y,Vars),
	generate_general_form_var(Vars,GFVars),
	substitute_vars(X,Vars,GFVars,EqNFX),
	substitute_vars(Y,Vars,GFVars,EqNFY),
	exp_greater_than(EqNFX,EqNFY),!.

greater_or_equal_than_ciaopp(X,Y):-
	varset(X-Y,Vars),
	generate_general_form_var(Vars,GFVars),
	substitute_vars(X,Vars,GFVars,EqNFX),
	substitute_vars(Y,Vars,GFVars,EqNFY),
	exp_greater_eq_than(EqNFX,EqNFY),!.

equal_than_ciaopp(X,Y):-
	varset(X-Y,Vars),
	generate_general_form_var(Vars,GFVars),
	substitute_vars(X,Vars,GFVars,EqNFX),
	substitute_vars(Y,Vars,GFVars,EqNFY),
	exp_eq_than(EqNFX,EqNFY),!.




:- pred simplify_in_ciaopp(E,SE) # "This predicate simplifies an arithmetic expression, possibly containing variables, using the ciaopp builtin rec. solver.".
% Added cases for simplify factorial term.
simplify_in_ciaopp(Exp,SExp):-
	instance(Exp,A*fact(A-1)),!,
	Exp = A*fact(A-1),
	SExp = fact(A).
simplify_in_ciaopp(Exp,SExp):-
	instance(Exp,fact(A-1)*A),!,
	Exp = fact(A-1)*A,
	SExp = fact(A).
% Added basic cases for min and max expressions.
simplify_in_ciaopp(Exp,SExp):-
	instance(Exp,max(A,B)),
	Exp = max(A,B), A \== B,!,
	simplify_in_ciaopp(A,A0),
	simplify_in_ciaopp(B,B0),
	SExp = max(A0,B0).
simplify_in_ciaopp(Exp,SExp):-
	instance(Exp,min(A,B)),
	Exp = min(A,B), A \== B,!,
	simplify_in_ciaopp(A,A0),
	simplify_in_ciaopp(B,B0),
	SExp = min(A0,B0).
simplify_in_ciaopp(X,Y):-
	instance(X,min(A,A)),!,
	X = min(A,A),
	Y = A.
simplify_in_ciaopp(X,Y):-
	instance(X,max(A,A)),!,
	X = max(A,A),
	Y = A.
simplify_in_ciaopp(X,X):-
	instance(X,min(_A,_B)),!.
simplify_in_ciaopp(X,X):-
	instance(X,max(_A,_B)),!.




simplify_in_ciaopp(Exp,SExp):-
	instance(Exp,fact(A-1)*A),!,
	Exp = fact(A-1)*A,
	SExp = fact(A).

simplify_in_ciaopp(Exp,Exp):-
	var(Exp),!.
simplify_in_ciaopp(Exp,SExp):-
	varset(Exp,IVar),
	generate_general_form_var(IVar,GFVars),
	substitute_vars(Exp,IVar,GFVars,EqNF),
	simplification(EqNF,SExpGF),!,
	general_form_to_vars(SExpGF,GFVars,IVar,SExp).
simplify_in_ciaopp(Exp,Exp):-!.

order_of_function(X,Y):-
	varset(X,Vars),
	generate_general_form_var(Vars,GFVars),
	substitute_vars(X,Vars,GFVars,EqNFX),
	%substitute_vars(Y,Vars,GFVars,EqNFY),
	algebraic_res:order_of_function(EqNFX,Y0),
	general_form_to_vars(Y0,GFVars,Vars,Y).

	
:- pred generate_general_form_var(Vars, GFVars): list(var)*var => list(var)*list(term) # "Returns @var(GFVars), a list of general form variables, one for each variable in @var{Vars}. A general form variable is a term of the form $(I), where I is a natural number.".
:- test generate_general_form_var(Vars,GVars): (Vars = [A,B,C]) => (GVars = [$(1),$(2),$(3)]) # "Three variables".
generate_general_form_var(Vars,GFVars):-
	generate_general_form_var_(1,Vars,GFVars).
generate_general_form_var_(_,[],[]).
generate_general_form_var_(C,[_X|Y],[$(C)|R]):-
	C1 is C+1,
	generate_general_form_var_(C1,Y,R).

%%%%%%%%END Code to work with a recurrence equation solver (mathematica and builtin ciaopp solver)%%%%%%%%.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%END CODE ORIGINALLY IN eq_solver MODULE%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%TRANSLATIONS TO/FROM GENERAL FORM (BUILTIN SOLVER NOTATION)
% TODO: Complete translations
% TODO: Define type for general_form
% TODO: Translate recurrence relations to diff  equations in general form
translate_from_algebraic(Exp,NExp):-
	% \+ var(Exp),
	% list(Exp),
	algebraic2general(Exp,NExp).
translate_to_algebraic(Exp,NExp):-
	general2algebraic(Exp,NExp).

general2algebraic(Exp,NExp):-
	replace_term(Exp,[(exp,'**')],NExp0),
	wrap_math_fun(NExp0,NExp).

algebraic2general(Exp,NExp):-
	replace_term(Exp,[('**',exp)],NExp0),
	unwrap_fun(NExp0,NExp).

