:- module(solver_utils_ppl,_,
	% [
	%     solve_special_rec_rel_ppl/3,
	%     find_linear_rank_function/4
	% ],	
	[assertions,
	 nativeprops,
	 regtypes
	]).


:- doc(title, "Solver Utilities - Auxiliary predicates using PPL library").
:- doc(author, "Maximiliano Klemen").
:- doc(module, "Auxiliary predicates for solver using PPL library").

:- use_module(library(terms_vars)).
:- use_module(library(write),[numbervars/3]).
:- use_module(domain(resources/plai2infer), [replace_term/3]).
:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(library(lists),[length/2,append/3,nth/3]).
:- use_module(ppl_ops).
:- use_module(library(terms_check),[instance/2]).
:- use_module(domain(resources/eq_solver/solver/solver_utils),[simplify/3,greater_than/3]).


:- include(solver_utils_ppl_transitions).


% TODO: A predicate for inferring loop invariants



% :- regtype point_rf(RF).

% point_rf(point(Exp)):-
% 	arithexpression(Exp).
% point_rf(point(Exp,C)):-
% 	arithexpression(Exp),
% 	num(C).

:- pred find_linear_rank_function(Vars,AfterExp,Cons,Result):(list(Vars,var),list(AfterExp,arithexpression),list(Cons,arithexpression),var(Result)) => (point_rf(Result))
	# "Given the list @var{Vars} of variables of a loop; the (linear) expressions that adopt those variables after a single iteration of the loop (@var{AfterExp}); and the set of (linear) constraints that @var{Vars} U vars(@var{AfterExp}) verify inside the loop (@var{Cons}), this predicates returns in @var{Result} the (linear) ranking function of the loop, if it exists. Note that this predicate is limited to linear expressions.".

find_linear_rank_function(Vars,AfterExp,Cons,Result):-
	generate_primed_cons(AfterExp,Primed,ConsPrimed),
	append(ConsPrimed,Cons,AllCons),
	% We may verify the expressions sent to PPL before calling the
	% predicate, instead catching exception afterwards.
	catch(
		 ppl_find_linear_rank_function(Vars,Primed,AllCons,Result),
		 ppl_invalid_argument(found(_),expected(linear_expression_or_constraint),where('ppl_new_NNC_Polyhedron_from_constraints/2')),
		 fail
	).

generate_primed_cons([],[],[]).
generate_primed_cons([Exp|R],[X|V],[X=Exp|RR]):-
	generate_primed_cons(R,V,RR).

% ppl_find_linear_rank_function(Vars,PrimedVars,Constraints,Function):-
% 	append(PrimedVars,Vars,AllVars),
% 	copy_term(AllVars,NAllVars),
% 	numbervars(NAllVars,0,_),
% 	build_map(AllVars,NAllVars,MappingF),
% 	build_map(NAllVars,AllVars,MappingB),
% 	build_map(PrimedVars,Vars,VarMap),
% 	replace_term(Constraints, MappingF,Cs0),
% 	ppl_initialize,
% 	ppl_new_NNC_Polyhedron_from_constraints(Cs0,H0),
% 	ppl_one_affine_ranking_function_MS_C_Polyhedron(H0,H1),
% 	replace_term(H1,MappingB,Function0),
% 	replace_term(Function0,VarMap,Function),
% 	ppl_finalize.

% build_map([],[],[]).
% build_map([X|X1],[Y|Y1],[(X,Y)|Z]):-
% 	build_map(X1,Y1,Z).

% TODO: Implement a checker for checking that the recurrence fit in
% the form mentioned above.

% TODO: Support non-constant node cost by setting and solving a linear
% programming problem with the node cost expression and the
% constraints of the equation.

% solve_special_rec_rel_ppl(RecRel,upper,Sol):-
% 	obtain_parts(RecRel, Vars, VarsPrimed, Constraints,LocalCost,BaseCase),
% 	find_linear_rank_function(Vars,VarsPrimed,Constraints,RF),
% 	(RF = point(R) -> 
% 	 Sol = R*LocalCost + BaseCase 
% 	; 
% 	    (RF = point(R,D),
% 	     Sol = (R/D)*LocalCost + BaseCase
% 	    )
% 	).

:- use_module(domain(resources/term_utilities), [get_var_position/2,term_at_position/3,update_term/4]).


solve_special_rec_rel_ppl(RecRel,upper,Sol):-
	obtain_parts(RecRel, Vars, VarsPrimed, Constraints,LocalCost,BaseCase),
	find_linear_rank_function(Vars,VarsPrimed,Constraints,RF),
	(RF = point(R) -> RF0 = R; RF=point(R,D), RF0 = (R/D)),
	calculate_final_basecase(BaseCase, Vars, VarsPrimed, RF0, BaseCase0),
	% (RF = point(R) -> 
	 Sol = RF0*LocalCost + BaseCase0. 
	% ; 
	%     (RF = point(R,D),
	%      Sol = (R/D)*LocalCost + BaseCase
	%     ),
	    
	% ).
solve_special_rec_rel_ppl(RecRel,upper,Sol):-
	solve_special_rec_rel_ppl_transitions(RecRel,upper,Sol).

:- export(calculate_final_basecase/5).
calculate_final_basecase(Base, Vars, VarsPrimed, Rf, Base0):-
	var(Base),!,
	varset(Rf, VarsRank),
	\+ (get_var_position(VarsRank, (_,Var)), Var==Base),
	get_var_position(Vars, (Pos,V)), V==Base,!,
	term_at_position(Pos, VarsPrimed,ModBase),
	(ModBase = Base + C; ModBase = C + Base),
	 Base0 = Rf*C + Base.
calculate_final_basecase(Base, Vars, VarsPrimed, Rf, Solution):-
	varset(Base,[VBase]),!,
	varset(Rf, VarsRank),
	\+ membervar(VBase, VarsRank),
	% \+ (get_var_position(VarsRank, (_,Var)), Var==Base),
	get_var_position(Vars, (Pos,V)), V==VBase,!,
	term_at_position(Pos, VarsPrimed,ModBase),
	(ModBase = VBase + C; ModBase = C + VBase),
	 Base0 = (Rf*C + VBase),
	 get_var_position(Base, (PosExp,X)), X==VBase,!,
	 update_term(Base,PosExp,Base0,Solution).
calculate_final_basecase(Base, _,_,_,Base):-
	num(Base).

membervar(A, [B|_]):-
	A==B,!.
membervar(A, [_|Z]):-
	membervar(A,Z).

:- export(obtain_parts/6).
obtain_parts(RecRels,Vars,VarsP,Constraints,Cost,Base):-
	obtain_recursive_and_base_case(RecRels,equation(Name,Vars,Exp,Constraints),equation(Name,_,Base,_)),
	obtain_recursive_call(Exp,Name,fun(Name,VarsP)),
	obtain_node_cost(Exp, Cost),!.

obtain_recursive_call(Exp,_Name,_Exp):-
	var(Exp),!,fail.
obtain_recursive_call(Exp,Name,Exp):-
	nonvar(Exp),
	Exp = fun(Name,_Args),!.
obtain_recursive_call(Exp,Name,RExp):-
	Exp =.. [F|Args],
	F \== fun,
	obtain_recursive_call_args(Args,Name,RExp).

obtain_recursive_call_args([X|_N],Name,Exp):-
	obtain_recursive_call(X,Name,Exp),!.
obtain_recursive_call_args([_X|N],Name,Exp):-
	obtain_recursive_call_args(N,Name,Exp),!.

obtain_recursive_and_base_case([equation(Name,Args,Exp,Cons),Base],equation(Name,Args,Exp,Cons),Base):-
	obtain_recursive_call(Exp,Name,_),!.
obtain_recursive_and_base_case([Base,equation(Name,Args,Exp,Cons)],equation(Name,Args,Exp,Cons),Base):-
	obtain_recursive_call(Exp,Name,_),!.

obtain_node_cost(Exp,Cost):-
	nonvar(Exp),
	Exp =.. [Op|[Call,Cost]],
	(Op = '+';Op = '-'),
	instance(Call,fun(_,_)),
	copy_term(Cost,C),
	nonvar(Cost),
	arithexpression(C),!.
obtain_node_cost(Exp,0):-
	nonvar(Exp),
	Exp =.. [fun|[_Call,_Cost]],!.
obtain_node_cost(Exp,Cost):-
	nonvar(Exp),
	Exp =.. [Op|[Cost,Call]],
	(Op = '+';Op = '-'),
	instance(Call,fun(_,_)),
	arithexpression(Cost).
