% This predicates obtain an upper bound for recurrence equations with
% multiple "loops", or transitions. The technique followed is a
% simplified version of the one presented for Loopus in
% 
% A Simple and Scalable Static Analysis for Bound Analysis and
% Amortized Complexity Analysis
% Florian Zuleger Helmut Veith Moritz Sinn, CAV, 2014.
%
% It is assumed that there is only one function defined for the
% recurrence equations, and that all argument names (variables)
% are the same in all the equations
%
% Currently, the approach is limited to ranking functions consisting
% of a single variable, and with constant decrements/increments to
% variables in all the transitions.
%
% This is experimental and under development.

solve_special_rec_rel_ppl_transitions(RecRel,upper,Sol):-
	RecRel=[equation(Name,_,_,_)|_],
	obtain_transitions(RecRel,Name,Trans),
	bound_transitions(Trans,Trans,0,Sol0),
	simplify(builtin,Sol0,Sol).

obtain_transitions([],_Name,[]).
obtain_transitions([equation(Name,Args,Exp,Cons)|T],Name,[Trans|TT]):-
	obtain_recursive_call(Exp,Name,fun(Name,VarsP)),!,
	Trans = trans(Args,VarsP,Cons,Ind,Incs),
	find_linear_rank_function(Args,VarsP,Cons,RF),
	(RF = point(R) -> RF0 = R; RF=point(R,D), RF0 = (R/D)),
	 simplify(builtin,R,RF1),
	 rank_var_index(Args,RF1,1,Ind),
	 inc_indexes(Args,VarsP,1,Incs),
	obtain_transitions(T,Name,TT).
obtain_transitions([_|T],Name,TT):-
	obtain_transitions(T,Name,TT).

% bound_transitions([],_,B,B).
% bound_transitions([trans(Args,_,_,VarInd,_)|T],Trans,B,NB):-
% 	bound_transitions_incs(VarInd,Trans,0,Incs),
% 	nth(VarInd,Args,Var),
% 	NBound = B + Var + Incs,
% 	bound_transitions(T,Trans,NBound,NB).

bound_transitions([],_,B,B).
bound_transitions([Tran|T],Trans,B,NB):-
	bound_transition(Tran,Trans,B0),
	B1 = B0 + B,
	bound_transitions(T,Trans,B1,NB).

bound_transition(trans(Args,PArgs,_,VarInd,_),Trans,NBound):-
	bound_transitions_incs(VarInd,Trans,Trans,0,Incs),
	nth(VarInd,Args,Var),
	nth(VarInd,PArgs,VarD),
	simplify(builtin,Var-(VarD),Diff),
	number(Diff),
	NBound = (Var + Incs)/Diff.

bound_transitions_incs(_VarInd,[],_,Incs,Incs).
bound_transitions_incs(VarInd,[trans(Args,PArgs,Cons,BI,IncList)|T],Trans,Incs0,Incs):-
	member((VarInd,Inc),IncList),!,
	bound_transition(trans(Args,PArgs,Cons,BI,IncList),Trans,Bound),
	%nth(BI,Args,Bound),
	Incs1 = Incs0 + Bound*Inc,
	bound_transitions_incs(VarInd,T,Trans,Incs1,Incs).
bound_transitions_incs(VarInd,[_|T],Trans,Incs0,Incs):-
	bound_transitions_incs(VarInd,T,Trans,Incs0,Incs).

% bound_transitions_incs(_VarInd,[],Incs,Incs).
% bound_transitions_incs(VarInd,[trans(Args,_,_,BI,IncList)|T],Incs0,Incs):-
% 	member((VarInd,Inc),IncList),!,
% 	nth(BI,Args,Bound),
% 	Incs1 = Incs0 + Bound*Inc,
% 	bound_transitions_incs(VarInd,T,Incs1,Incs).
% bound_transitions_incs(VarInd,[_|T],Incs0,Incs):-
% 	bound_transitions_incs(VarInd,T,Incs0,Incs).

rank_var_index([X|_T],Var,Ind,Ind):-
	Var == X,!.
rank_var_index([_X|T],Var,Ind,Ind1):-
	Ind0 is Ind + 1,
	rank_var_index(T,Var,Ind0,Ind1).

inc_indexes([],[],_,[]).
inc_indexes([X|T],[Xp|TT],Ind,[(Ind,Inc)|Inds]):-
	greater_than(builtin,Xp,X),
	simplify(builtin,Xp-X,Inc),
	number(Inc),!,
	Ind1 is Ind + 1,
	inc_indexes(T,TT,Ind1,Inds).
inc_indexes([_X|T],[_Xp|TT],Ind,Inds):-
	Ind1 is Ind + 1,
	inc_indexes(T,TT,Ind1,Inds).