:- module(solver_utils_test,_,[assertions]).

%:- import %% It didn't work with import/2.
:- reexport(solver_utils,
	[
	    simplify/3,
	    simplify/2,
	    expand/3,
	    expand/2,
	    solve/4,
	    solve/3,
	    greater_than/3,
	    greater_than/2,
	    solve_rec_rel/5,
	    solve_rec_rel/4
	]
).


:- doc(bug, "The test cases involving calls to mathematica are not
	working, because they throw
      exception(predicate,error(mathematica_opening_link,mathematica:evaluate/2))).").

:- doc(bug, "The following test case: test expand(mathematica,0,Res)
	=> (Res = 0). causes that run_test_module/1 throws the error
 {ERROR: term_basic:=../2, arg 2 - expected a non-numeric atom,
  found 0}.").

%:- test simplify(mathematica,X+X,Res) => (Res=2*X).
%:- test simplify(mathematica,X*X,Res) => (Res = X**2).
%:- test simplify(mathematica,X+0,Res) => (Res = X).
:- test simplify(builtin,X+0,Res) => (Res = X).
:- test simplify(builtin,X+X,Res) => (Res=2*X).
:- test simplify(builtin,X*X,Res) => (Res = X**2).
%:- test simplify(mathematica,fun(f,[X])+fun(f,[X]),Res) => (Res=2*fun(f,[X])).
:- test simplify(builtin,fun(f,[X])+fun(f,[X]),Res) => (Res=2*fun(f,[X])).
:- test simplify(ppl,E,_) + exception(not_implemented(simplify,ppl,[E])).

%:- test expand(mathematica,(X+1)**2,Res) => (Res = X**2 + 2*X + 1).
%:- test expand(mathematica,(X+Y)**2,Res) => (Res = X**2 + 2*X*Y + Y**2).
:- test expand(builtin,(X+Y)**2,Res) + exception(not_implemented(expand,builtin,[(X+Y)**2])).
:- test expand(ppl,(X+Y)**2,Res) + exception(not_implemented(expand,ppl,[(X+Y)**2])).
%:- test expand(mathematica,0,Res) => (Res = 0).
%:- test expand(mathematica,X,Res) => (Res = X).

%:- test solve(mathematica,X=Y+Z,Z,Res) => (Res = X - Y).
:- test solve(builtin,X=Y+Z,Z,Res) + exception(not_implemented(solve,builtin,_)).
:- test solve(ppl,X=Y+Z,Z,Res) + exception(not_implemented(solve,builtin,_)).

:- test solve_rec_rel(builtin,[equation(f,[X],fun(f,[X-1]) + 1 ,[X>0]),equation(f,[0],0,[])],[X],upper,Sol) => (Sol=X).
%:- test solve_rec_rel(mathematica,[equation(f,[X],fun(f,[X-1]) + 1 ,[X>0]),equation(f,[0],0,[])],[X],upper,Sol) => (Sol=X).
:- test solve_rec_rel(ppl,[equation(f,[X],fun(f,[X-1]) + 1 ,[X>0]),equation(f,[0],0,[])],[X],upper,Sol) => (Sol=X).
:- test solve_rec_rel([equation(f,[X],fun(f,[X-1]) + 1 ,[X>0]),equation(f,[0],0,[])],[X],upper,Sol) => (Sol=X).
:- test solve_rec_rel(ppl,[equation(f,[X,N],fun(f,[X+1,N]) + 1 ,[X<N]),equation(f,[X,N],0,[X=N])],[X],upper,Sol) => (Sol=N-X).
:- test solve_rec_rel([equation(f,[X],fun(f,[X-1]) + 1 ,[X>0]),equation(f,[0],0,[])],[X],upper,Sol) => (Sol=X).
:- test solve_rec_rel([equation(f,[X,N],fun(f,[X+1,N]) + 1 ,[X<N]),equation(f,[X,N],0,[X=N])],[X],upper,Sol) => (Sol=N-X).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% WITNESS TEST CASES OF BUGS OR PENDING THINGS  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- test solve_rec_rel(ppl,[equation(inczznatszz1424z1z1z1,[X,Y],fun(inczznatszz1424z1z1z1,[X+1,Y])+X,[X<Y]),equation(inczznatszz1424z1z1z1,[X,Y],X,[X>=Y])],[X,Y],upper,Sol) => (Sol=Y**2-Y*X+X) # "Bug #1: local cost expression in the recursive case have to
be maximized (X, having [X<Y])".



:- test simplify(builtin,fact(X-1)*X,Res) => (Res = fact(X)) # "Imp #1: simplification of factorial expression using builtin solver".

:- test simplify(builtin,min(A,A),Res) => (Res = A).
:- test simplify(builtin,min(A+0,A),Res) => (Res = A).
:- test simplify(builtin,min(A,A+0),Res) => (Res = A).
:- test simplify(builtin,min(A,B),Res) => (Res = min(A,B);Res=min(B,A)).
:- test simplify(builtin,max(A,B),Res) => (Res = max(A,B);Res=max(B,A)).
:- test simplify(builtin,max(A,A),Res) => (Res = A).
:- test simplify(builtin,max(A+0,A),Res) => (Res = A).
:- test simplify(builtin,max(A,A+0),Res) => (Res = A).

%:- test simplify(mathematica,inf,L) => (L = inf).
:- test simplify(builtin,inf,L) => (L = inf).

%This case is some king of bug or missed case in the simplification
%predicate in the builtin.
:- test simplify(builtin,fun(f,[X])*fun(f,[X]),L) => (L = fun(f,[X])**2).



:- test solve_rec_rel(builtin,[equation(f,[1],1,[]),equation(f,[A],A,[])],[A],upper,Sol) => (Sol=A).

% This is for coupled
:- test solve_rec_rel(builtin,[equation(f,[0],1,[]),equation(f,[A],fun(g,[A-1])+1,[]),equation(g,[B],fun(f,[B-1])+1,[]),equation(g,[0],1,[])],[A,B],upper,Sol) => (Sol=A+1).

%this appears in erathostenes
:- test solve_rec_rel(builtin,[equation(f,[A,B],max(B,fun(f,[A-1,B])),[])],[A,B],upper,Sol) + fails.






%:- test solver_utils_builtin:translate_to_algebraic(inf,fun(inf,[])).
% solver_utils:solve_rec_rel(builtin,[equation(appendzzappendzz318z2z2z1z1z1,[1],1,[]),equation(appendzzappendzz318z2z2z1z1z1,[_748],_748,[])],[_748],lower,_141525).

%%% For fibonnacci (llvm - builtin)
% [f(_D)==f(_D-1)*f(_D-2),f(1)==1,f(0)==1]
