:- module(solver_utils_trans,[wrap_math_fun/2, unwrap_fun/2],	
	[assertions,
	 nativeprops
	]).


:- doc(title, "Solver Utilities - Translation utilities").
:- doc(author, "Maximiliano Klemen").
:- doc(module, "Auxiliary predicates for translation to/from algebraic expressions to/from solver-specific syntax").


:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    algebraic_builtin/1,
	    algebraic_expression/1
	]).

:- use_module(library(lists),[length/2]).



:- pred wrap_math_fun(Exp,WExp): (term(Exp),var(WExp)) =>
	(algebraic_expression(WExp)) # "Returns the expression with
      all the functions that are not algebraic_builtin wrapped with
      the functor fun/2.".

wrap_math_fun(Exp,Exp):-
	var(Exp);algebraic_builtin(Exp);Exp=='inf';Exp=='?',!.
wrap_math_fun(Func,fun(Name,NArgs)):-
	Func =.. [Name|Args],
	length(Args,N), length(FArgs,N),
	FFunc =.. [Name|FArgs],
	\+ arithexpression(FFunc), \+ algebraic_builtin(FFunc),!,
	wrap_math_args_(Args,NArgs),!.
wrap_math_fun(Exp,Res):-
	Exp =.. [ExpHead|Args],
	wrap_math_args_(Args,NArgs),!,
	Res =.. [ExpHead|NArgs].

wrap_math_args_([],[]).
wrap_math_args_([Arg|Args],[NArg|NArgs]):-
	wrap_math_fun(Arg,NArg),
	wrap_math_args_(Args,NArgs).


unwrap_fun(Exp,Exp):-
	var(Exp),!.
unwrap_fun(fun(Name,Args),Res):-
	replace_fun_args(Args,NArgs),
	Res =.. [Name|NArgs],!.
unwrap_fun(Exp,Res):-
	Exp =.. [ExpHead|Args],
	replace_fun_args(Args,NArgs),
	Res =.. [ExpHead|NArgs].

replace_fun_args([],[]).
replace_fun_args([Arg|Args],[NArg|NArgs]):-
	unwrap_fun(Arg,NArg),
	replace_fun_args(Args,NArgs).
