:- module(plai2infer,_,[assertions]).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).
:- use_module(library(terms_vars)).
:- use_module(ciaopp(infer), [type2measure/3]).
:- use_module(ciaopp(plai/domains)).
:- use_module(library(lists)).
:- use_module(ciaopp(preprocess_flags)).
:- use_module(domain(resources/eq_solver/eq_solver_mathematica), [expand_functions_in_mathematica/2]).
:- use_module(library(terms_check),[instance/2]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%First implementation of translation of the res_plai info, from
%plai_db form to infer_db form. This is done for integrating
%resource verification with res_plai analysis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%The main idea is to take a term containing analysis information
%from res_plai, as it is in plai_db: a term of the
%form complete(...), and translate to a term with
%the corresponding analysis information, but in the form of infer_db (inferred pred.)
%That is, a term of the form inferred(Prop,Goal,Info).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

translate_resource_info(Complete,Inferred):-
	translate_resource_info_(Complete,Inferred, ub).
translate_resource_info(Complete,Inferred):-
	translate_resource_info_(Complete,Inferred, lb).


translate_resource_info_(Complete,Inferred, Bound):-
	Complete = complete(_PredID,res_plai,Goal,res_plai(CallTypes,_CallResAbs),[res_plai(SuccTypes,SuccResAbs)|_],_Id,_Parents),
	Inferred = inferred(res_plai,Goal,Comp),
	term_variables(Goal,Gv), %predicate that returns the lists of vars appearing in a term.
	asub_to_info(etermsvar,CallTypes,Gv,InfoCallTypes,_),
	%asub_to_info(etermsvar,SuccTypes,Gv,InfoSuccTypes0,_),
	asub_to_info(res_plai,res_plai(SuccTypes,SuccResAbs),Gv,InfoSuccTypes0,Comps),
	split_sizetypes(InfoSuccTypes0,InfoSuccTypes,RSizes),
	Comp = complexity(InfoCallTypes,InfoSuccTypes,Modes,Measures,_MutExc,_Det,Sizes,inf,Bound,Resources,ResExps,_),
	types2modes(InfoCallTypes,Modes),
	type2measure(Goal,InfoSuccTypes,Measures),
	input_output_rsizes(RSizes,Gv,Modes,RSiz,RSizO),
	length(Gv,SGv),
	generate_vars(SGv,NVars),
	types2varmap(RSiz,NVars,Measures,VarMap),
	generate_sizes(Gv,Modes,VarMap,RSiz,RSizO,Sizes,Bound),
	resource2cost(VarMap,Bound,Comps,Resources,ResExps),!.
	
	% complexity(['basic_props:list'(_A,'basic_props:num'),'basic_props:list'(_B,'basic_props:num'),var(_C)],[rt4(_A),'basic_props:list'(_B,'basic_props:num'),'basic_props:list'(_C,'basic_props:num')],[+,+,-],[length,length,length],[[1],[2]],[1],[$(0,1),$(0,2),$(0,2)+ $(0,1)],inf,lb,[steps],[0],_),



:- test resource2cost([($(0,1),[X1,X2]),($(0,2),[Y1,Y2])], ub, ['native_props:costb'(_,steps,Y1*2,6*Y2 + 45)], Resources,Result) => (Resources = [steps],Result = [6* $(0,2)+45]).
resource2cost(VarMap,Bound,Comp,Resources,ResExps):-
	resource2cost_(VarMap,Bound,Comp,Resources,ResExps).

resource2cost_(_,_,[],[],[]).
resource2cost_(VarMap,Bound,['native_props:costb'(_,Res,EL,EU)|Rest], [Res|ResRest],[CostExp|CostRest]):-
	!,choose_exp(EL,EU,Bound,E),
	resource_exp2cost_exp(VarMap,E,CostExp),
	resource2cost_(VarMap,Bound,Rest,ResRest,CostRest),!.
resource2cost_(VarMap,Bound,[_|Rest], ResRest,CostRest):-
	resource2cost_(VarMap,Bound,Rest,ResRest,CostRest).



choose_exp(_E1,E2,ub,E2).
choose_exp(E1,_E2,lb,E1).




:- export(resource_exp2cost_exp/3).
:- pred resource_exp2cost_exp(MapVars,Exp,GExp)
	# "For all variables @{V} appearing in @var{Exp}: It replaces @var{V} for the term @var{T}, if (@var{T},@var{List}) is in
 @var{MapVars} and @var{V} is in @var{List}. It unifies the resulting term in @var{GExp}.".
:- test resource_exp2cost_exp([($(0,1),[X1,X2]),($(0,2),[Y1,Y2])], X1*5 + term(Y2,4),L) => L = term($(0,2),4)+ 5* $(0,1).
resource_exp2cost_exp(VarMap,E0,CostExp):-
	term_variables(E0,VE),
	replace_term(E0,[('**',exp)],E),
	% vars_position(E,PE),
	(current_pp_flag(req_solver,mathematica) ->
	 (expand_functions_in_mathematica(E,NE0),
	  replace_term(NE0,[('Power',exp),('Cos'('Pi'*A),exp(-1,A))],NE)
	 )
	  ;
	     NE=E
	 ),
	replace_vars_(VarMap,VE,NE,CostExp).


replace_vars_(_,[],E,E).
replace_vars_(VarMap,[V|Vars],E,NewE):-
	get_var(VarMap,V,CostVar),!,
	%var_position(E,V,Pos),
	substitute_vars(E,[V],[CostVar],NewE0),
	%change_terms_at_positions(E,Pos,CostVar,NewE0),
	replace_vars_(VarMap,Vars,NewE0,NewE).
replace_vars_(VarMap,[_V|Vars],E,NewE):-
	replace_vars_(VarMap,Vars,E,NewE).


% get_var([map(Var,VarList)],X,Var):-
% 	var_in_list(X,VarList).
get_var([(Var,VarList)|_Vars],X,Var):-
	var_in_list(X,VarList),!.
get_var([(_Var,_VarList)|Vars],C,RVar):-
	get_var(Vars,C,RVar).

var_in_list(X,[Y|_]):-
	X == Y,!.
var_in_list(X,[_|L]):-
	var_in_list(X,L).





% %generate the list of modes given the lists of types.
% % TODO: Add terms as an output mode.
:- export(types2modes/2).
types2modes([],[]).
types2modes(['term_typing:var'(_)|T],['-'|R]):-
	types2modes(T,R),!.
types2modes([_|T],['+'|R]):-
	types2modes(T,R),!.


generate_vars(C,Vars):-
	generate_vars_(1,C,Vars).
generate_vars_(N,C,[]):- N>C.
generate_vars_(N,C,[$(0,N)|D]):-
	N =< C,
	N1 is N + 1,
	generate_vars_(N1,C,D).



%:- export(split_sizetypes/3).
split_sizetypes([],[],[]).
split_sizetypes(['native_props:rsize'(Var,S)|T],R, ['native_props:rsize'(Var,S)|TT]):-
	split_sizetypes(T,R,TT),!.
split_sizetypes([Ty|T],[Ty|R],G):-
	split_sizetypes(T,R,G).

% TODO: If measure is lenght, then put var + 1.
types2varmap([],_,_,[]).
types2varmap(['native_props:rsize'(_Var,S)|T],[VP|VPS],[length|MT],[((VP+1),[LB,UB])|R]):- !,
	term_variables(S,[LB,UB|_]),
	types2varmap(T,VPS,MT,R).
types2varmap(['native_props:rsize'(_Var,S)|T],[VP|VPS],[_|MT],[(VP,[LB,UB])|R]):-
	term_variables(S,[LB,UB|_]),
	types2varmap(T,VPS,MT,R).



input_output_rsizes([],_Gv,_Measures,[],[]).
input_output_rsizes(['native_props:rsize'(Var,S)|RSizes],[Var|Gv],[+|Measures],['native_props:rsize'(Var,S)|RSiz],RSizO):-
	input_output_rsizes(RSizes,Gv,Measures,RSiz,RSizO).
input_output_rsizes(['native_props:rsize'(Var,S)|RSizes],[Var|Gv],[-|Measures],RSiz,['native_props:rsize'(Var,S)|RSizO]):-
	input_output_rsizes(RSizes,Gv,Measures,RSiz,RSizO).


generate_sizes([],_,_,_,_,[],_).
generate_sizes([V|Vars],[+|Modes],Varmap,['native_props:rsize'(V,S)|RSizI],RSizO,[VExp|Sizes],Bound):-
	S =.. [_,LB,UB|_],
	choose_exp(LB,UB,Bound,Exp),
	get_var(Varmap,Exp,VExp),
	generate_sizes(Vars,Modes,Varmap,RSizI,RSizO,Sizes,Bound).
generate_sizes([V|Vars],[-|Modes],Varmap,RSizI,['native_props:rsize'(V,S)|RSizO],[CostExp|Sizes],Bound):-
	S =.. [_,LB,UB|_],
	choose_exp(LB,UB,Bound,Exp),
	%get_var(Varmap,V,VarExp),
	resource_exp2cost_exp(Varmap,Exp,CostExp),
	generate_sizes(Vars,Modes,Varmap,RSizI,RSizO,Sizes,Bound).

:- export(replace_term/3). % Exported just for performing unit testing
:- pred replace_term(Exp,ListReplaces,Result) 
	# "For each par (@var{T},@var{R}) in @var{ListReplaces}, it replaces each subterm (or term head) in @var{Exp} that unifies with @var{T} with @var{R}.".
:- test replace_term(3 + exp(3,cos(pi*(3+D))),[(cos(pi*A),exp(-1,A))],L) => (L = 3 + exp(3,exp(-1,3+D))) # "cos(pi*A) ==> exp(-1,A)".
:- test replace_term('Power'(2,A),[('Power',exp)],L) => (L = exp(2,A)).

% TODO: Probably this predicate is already implemented somewhere else. This one should
% be removed if that is the case.
replace_term(Exp,[],Exp).
replace_term(Exp,[(Term,Rep)|Rest],Result):-
	replace_term_(Exp,(Term,Rep),Result0),!,
	replace_term(Result0,Rest,Result).

replace_term_(Exp,(Term,Rep),Rep):-
	var(Exp),
	Exp == Term.
replace_term_(Exp,(_Term,_Rep),Exp):-
	var(Exp).
replace_term_(Exp,(TExp,Rep),Rep):-
	Exp == TExp,!.
replace_term_(Exp,(Term,Rep),Res):-
	Exp =.. [ExpHead|Args],
	replace_term_args_(Args,(Term,Rep),NArgs),
	((ExpHead==Term) -> (Res =.. [Rep|NArgs]) ; (
						      Exp==Term ->
						      (
							  Res = Rep
						      );
						      Res =.. [ExpHead|NArgs]
						  )
	).

replace_term_args_([],_,[]).
replace_term_args_([Arg|Args],Tup,[NArg|NArgs]):-
	replace_term_(Arg,Tup,NArg),
	replace_term_args_(Args,Tup,NArgs).
