:- module(reachability,[
	perform_reachability_analysis/1,reachs/2% ,not_calling_any_cc/1,collect_not_calling_any_cc/0
   ],[assertions,datafacts]).

:- use_module(ciaopp(p_unit/program_keys), [predkey_from_sg/2]). % get_predkeys/2
:- use_module(library(formulae), [conj_to_list/2]).
:- use_module(ciaopp(p_unit), [predicate_names/1]).
:- import(resources_lib,[get_all_static_cost_centers/1]).

% calling(A,B): A reachs B.
%:- data not_calling_any_cc/1.

:- doc(title, "Predicate reachability information").
:- doc(author, "Maximiliano Klemen").
:- doc(module,"This module defines the predicate relation @tt{reachs}
	whose meaning is the following: for any p, q predicate keys
	(name/arity as an atom), @tt{reachs(p,q)} ifff a call to p
	could eventually cause a call to q.").


:- pred reachs/2 # "Defines the @tt{reachs} relation for a program,
	after calling @tt{perform_reachability_analysis/1}.".  
:- data	reachs/2.

:- pred perform_reachability_analysis(Clauses) # "Given the list of
	clauses of a module (as given in @tt{program(Clauses,Dics)},
	populates the @tt{reachs} relation with the transitive closure
	of the 'directly calls' relation. A predicate p 'directly
	calls' another predicate q ifff a literal q appears in the
	body of any clause defining p.)".

perform_reachability_analysis(Cls):-
	%retractall_fact(not_calling_any_cc(_)),
	retractall_fact(reachs(_,_)),!,
	perform_reachability_analysis_(Cls).
	%collect_not_calling_any_cc.	

perform_reachability_analysis_([clause(Head,Body):_ID|R]):-
	predkey_from_sg(Head,PredKey),
	conj_to_list(Body,LBody),
	perform_reachability_analysis_body(PredKey,LBody),
	perform_reachability_analysis_(R).
perform_reachability_analysis_([]):-
	closure.

perform_reachability_analysis_body(PredKey,['!'|R]):-
	perform_reachability_analysis_body(PredKey,R).
perform_reachability_analysis_body(PredKey,[Lit:_|R]):-
	predkey_from_sg(Lit,LitKey),
	(\+ current_fact(reachs(PredKey,LitKey)),LitKey \= 'true/0' -> asserta_fact(reachs(PredKey,LitKey));true),
	perform_reachability_analysis_body(PredKey,R).
perform_reachability_analysis_body(_PredKey,[]).

%:- export(closure/0).
%TODO  A bit inefficient. Improve the implementation.
closure:-
	current_fact(reachs(A,B)),
	current_fact(reachs(B,C)),
	\+ current_fact(reachs(A,C)),
	assertz_fact(reachs(A,C)),
	closure.
closure.

% collect_not_calling_any_cc:-
% 	predicate_names(Preds),
% 	get_predkeys(Preds,PredsKey),
% 	get_all_static_cost_centers(CCs),
% 	cc_keys(CCs,CCKs),
% 	collect_not_calling_any_cc_(PredsKey,CCKs).

% cc_keys([],[]).
% cc_keys([CC|T],[CCK|TT]):-
% 	CC =..[_Res,CCK],
% 	cc_keys(T,TT).


% collect_not_calling_any_cc_([],_).
% collect_not_calling_any_cc_([P|R],CCKs):-
% 	never_calling(P,CCKs),!,
% 	assertz_fact(not_calling_any_cc(P)),
% 	collect_not_calling_any_cc_(R,CCKs).
% collect_not_calling_any_cc_([_P|R],CCKs):-
% 	collect_not_calling_any_cc_(R,CCKs).
	
% :- export(never_calling/2).
% never_calling(_X,[]).
% never_calling(X,[Y|T]):-
% 	\+ current_fact(reachs(X,Y)),!,
% 	never_calling(X,T).
