:- module(res_abs_aux, _, [assertions,regtypes,basicmodes,hiord,nativeprops,datafacts]).
:- use_module(library(messages)).
%:- use_module(ciaopp(plai/static_profiling_info_tracker)).
:- use_module(domain(s_eqs)).
:- use_module(typeslib(typeslib),
        [ get_type_definition/2
        %, rule_type_symbol/1
        , dz_type_included/2
        , dz_equivalent_types/2
	, equiv_type/2
        ]).

:- use_module(domain(resources/eq_solver/eq_solver_graphs),[varmember/2]).
:- use_module(library(aggregates)).
:- use_module(library(lists)).
:- use_module(library(sets)).
:- use_module(library(sort)).
:- use_module(library(terms)).
:- use_module(domain(resources/eq_solver/solver/solver), [simplify/2]).
%:- use_module(argsize_utils).
:- use_module(domain(resources/eq_solver/eq_solver)).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).
:- use_module(extended_dz).
:- use_module(library(terms_vars), [varset/2]).
:- import(resources_lib,[get_all_static_cost_centers/1,get_all_resources/1,head_cost/3,literal_cost/3,default_cost/3, eval_cost_lit/3]).
:- import(resources_basic,[approx_to_bound/2]).
:- use_module(library(resdefs/res_litinfo), [get_litinfo/4]).

:- use_module(domain(resources/eq_solver/eq_solver_types), [equation/1]).

:- doc(bug,"In generate_resource_default_assigns, the argument passed to
	get_resource_default_cost is wrong, then it always returns 0. But if
      this is fixed and returns other default values different from 0, the
      glb operation in res_abs gives a wrong asub.").


:- doc(section, "Types").

% TODO: define concretely all the types.

:- regtype rsize_syntax_t(Size)
	# "@var{Size} is syntax for rsize property.". 

:- impl_defined(rsize_syntax_t/1).

:- regtype rsize_t(InternalSize)
	# "@var{InternalSize} is the internal represention of the rsize property.".

:- impl_defined(rsize_t/1). 

% TODO: regtype_t/1 should be defined in typelib or related modules. 
:- regtype regtype_t(Type) # "@var{Type} is a representation of a type". 

:- impl_defined(regtype_t/1).

% -----------------------------------------------------------
% Predicates defining sizes, ground sizes and undefined sizes
% They all work in a very similar way,
% by following the structure of the type
% -----------------------------------------------------------

get_vars_size_types([],_,[]).
get_vars_size_types([A|R],Types,[(A,SizeA)|SizeR]) :-
	find_variable_type(Types,A,TypeA),
	get_size_type(TypeA,SizeA),
	get_vars_size_types(R,Types,SizeR).

get_var_size_type([(A,SizeA)|_],B,SizeA) :- A == B, !.
get_var_size_type([_|R],A,SizeA) :- get_var_size_type(R,A,SizeA).

select_vars_size_types(_,[],[]).
select_vars_size_types(Mapping,[A|R],[(A,S)|SizesR]) :-
	get_var_size_type(Mapping,A,S),
	select_vars_size_types(Mapping,R,SizesR).

% It seems that simplify_types does not simplify
% all of them, so we keep an extra cache for
% equivalences between types, to have just one
% representative of each

:- data internal_equiv_type/2.

clean_internal_equiv_types :-
	retractall_fact(internal_equiv_type(_,_)).

find_in_internal_equiv_cache(Type,Type) :-
	current_fact(internal_equiv_type(Type,'$new')), !.
find_in_internal_equiv_cache(Type,Typ) :-
	current_fact(internal_equiv_type(Type,Typ)), !.
find_in_internal_equiv_cache(Type,Typ) :-
	current_fact(internal_equiv_type(Typ,'$new')),
	dz_equivalent_types(Type,Typ), !,
	assertz_fact(internal_equiv_type(Type,Typ)).
find_in_internal_equiv_cache(Type,Type) :-
	assertz_fact(internal_equiv_type(Type,'$new')).

get_type_definition_in_cache(Type,Defn) :-
	find_in_internal_equiv_cache(Type,Typ),
	get_type_definition(Typ,Defn).

% Get type from the list created by etermsvar
find_variable_type([],_,_) :- !, fail.
find_variable_type([V:(_,Type)|_],W,Typ) :- V == W, !,
	find_in_internal_equiv_cache(Type,Typ).
	%( equiv_type(Type,Typ) ; Typ = Type ).
	%Typ = Type.
find_variable_type([_|R],W,Typ) :- find_variable_type(R,W,Typ).

% Is type recursive?
recursive_type(Type) :-
	dz_type_included(Type,num), !,
	fail.
recursive_type(Type) :-
	dz_type_included(Type,vr), !,
	fail.
recursive_type(Type) :-
	dz_type_included(Type,atm), !,
	fail.
recursive_type(Type) :-
	% Any term can be given the term size
	dz_equivalent_types(Type,term), !,
	fail.
recursive_type(Type) :-
	get_type_definition_in_cache(Type,Defn),
	recursive_type_defn(Type,Defn).

recursive_type_defn(Ty,[[]|R]) :-
	!, recursive_type_defn(Ty,[^([])|R]).
recursive_type_defn(Ty,[.(A,B)|R]) :-
	!, recursive_type_defn(Ty,[^(.(A,B))|R]).
recursive_type_defn(Ty,[^(Functor)|_]) :-
	Functor =.. [_|Args],
	recursive_type_defn(Ty,Args).
recursive_type_defn(Ty,[T|_]) :-
	\+ functor(T,^,1), % It's not a rule
	dz_type_included(T,Ty).
	%dz_equivalent_types(Ty,T).
recursive_type_defn(Ty,[_|R]) :-
	recursive_type_defn(Ty,R).

% The sized type cache saves information about
% sized types that have already been constructed
:- data size_type_cache/2.

clean_size_type_cache :-
	retractall_fact(size_type_cache(_,_)).

get_size_type(Type,SType) :-
	current_fact(size_type_cache(Type,CachedType)),
	!, %found in cached sizes
	copy_term(CachedType,SType).
get_size_type(Type,SType) :-
	find_equivalent_in_cached(Type,CachedType),
	asserta_fact(size_type_cache(Type,CachedType)),
	!,
	copy_term(CachedType,SType).
get_size_type(Type,SType) :-
	get_size_type2(Type,NewType),
	asserta_fact(size_type_cache(Type,NewType)),
	!,
	copy_term(NewType,SType).

find_equivalent_in_cached(Type,CachedType) :-
	current_fact(size_type_cache(OType,OCached)),
	dz_equivalent_types(Type,OType),
	!,
	copy_term(OCached,CachedType).

% Generate the size related to some regular type
get_size_type2(Type,num((_Min,_Max))) :-
	% Numbers get numerical bounds
	dz_type_included(Type,num), !.
get_size_type2(Type,var) :-
	% Variables have no bounds
	dz_type_included(Type,vr), !.
get_size_type2(Type,gnd((_Min,_Max))) :-
	dz_equivalent_types(Type,gnd), !.
get_size_type2(Type,atom((1,1))) :-
	% Nor atoms
	dz_type_included(Type,atm), !.
get_size_type2(Type,term((_Min,_Max))) :-
	% Any term can be given the term size
	dz_equivalent_types(Type,term).
get_size_type2(Type,SizeType) :-
	get_type_definition_in_cache(Type,Defn),
	get_size_type_args(Type,Defn,SizeArgs),
	( recursive_type(Type)
	-> SizeType =.. [Type,(_Min,_Max)|SizeArgs]
	 ; SizeType =.. [Type|SizeArgs]
	).

get_size_type_args(_,[],[]).
get_size_type_args(T,[[]|Rest],SizeArgs) :-
	!, get_size_type_args(T,[^([])|Rest],SizeArgs).
get_size_type_args(T,[.(A,B)|Rest],SizeArgs) :-
	!, get_size_type_args(T,[^(.(A,B))|Rest],SizeArgs).
get_size_type_args(T,[^(Functor)|Rest],[^(FunctorSize)|SizeRest]) :-
	Functor =.. [F|FArgs],
	get_size_type_args(T,FArgs,FSizes),
	FunctorSize =.. [F|FSizes],
	get_size_type_args(T,Rest,SizeRest).
get_size_type_args(T,[Ty|Rest],Sizes) :-
	%( dz_equivalent_types(T,Ty)
	( dz_type_included(Ty,T)
	-> get_size_type_args(T,Rest,Sizes)
	 ; get_size_type(Ty,Size),
	   get_size_type_args(T,Rest,SizeRest),
	   Sizes = [Size|SizeRest]
	).

% Create the size type for a ground element
create_ground_size(N,Type,num((N,N))) :-
	dz_type_included(Type,num), !.
create_ground_size(_,^(F),^(F)) :-
	functor(F,_,0), !.
create_ground_size(_,Type,atom((1,1))) :-
	dz_type_included(Type,atm), !.
create_ground_size(T,Type,term((N,N))) :-
	dz_equivalent_types(Type,term), !,
	term_size(T,N).
create_ground_size(Ground,Type,Size) :-
	get_type_definition_in_cache(Type,Defn),
	create_ground_size_defn(Ground,Type,Defn,SizeArgs,Min,Max),
	( recursive_type(Type)
	-> Size =.. [Type,(Min,Max)|SizeArgs]
	 ; Size =.. [Type|SizeArgs]
	).

create_ground_size_defn(_,_,[],[],0,inf).
create_ground_size_defn(Term,Type,[[]|Rest],Size,Min,Max) :-
	create_ground_size_defn(Term,Type,[^([])|Rest],Size,Min,Max).
create_ground_size_defn(Term,Type,[.(A,B)|Rest],Size,Min,Max) :-
	create_ground_size_defn(Term,Type,[^(.(A,B))|Rest],Size,Min,Max).
create_ground_size_defn(Term,Type,[^(Rule)|Rest],[^(SizeT)|SizeR],Min,Max) :-
	functor(Term,F,N),
	( functor(Rule,F,N) % Same functor
	-> Rule =.. [F|RuleArgs],
	   Term =.. [F|TermArgs],
	   create_ground_size_defn_args(Type,TermArgs,RuleArgs,RSizes,Min,Max),
	   SizeT =.. [F|RSizes],
	   % We don't care about bounds of the rest
	   create_ground_size_defn(Term,Type,Rest,SizeR,_Min,_Max)
	 ; Rule =.. [G|Args],
	   create_undefined_size_defn(Type,Args,UndefSize),
	   SizeT =.. [G|UndefSize],
	   create_ground_size_defn(Term,Type,Rest,SizeR,Min,Max)
	).
create_ground_size_defn(Term,Type,[Ty|Rest],[SizeT|SizeR],Min,Max) :-
	create_undefined_size(Ty,SizeT),
	create_ground_size_defn(Term,Type,Rest,SizeR,Min,Max).

% At least we are applying the rule
create_ground_size_defn_args(_,[],[],[],1,1).
create_ground_size_defn_args(Type,[T|R],[TTy|RTy],[TSize|RSize],Min,Max) :-
	create_ground_size(T,TTy,TSize),
	create_ground_size_defn_args(Type,R,RTy,RSize,RMin,RMax),
	%( dz_equivalent_types(Type,TTy)
	( dz_type_included(TTy,Type)
	-> TSize =.. [_,(IMin,IMax)|_],
	   add_(IMin,RMin,Min), add_(IMax,RMax,Max)
	 ; Min = RMin, Max = RMax
	).

% Create the undefined size for a ground element
create_undefined_size(Type,num(('?','?'))) :-
	dz_type_included(Type,num), !.
create_undefined_size(Type,atom(('?','?'))) :-
	dz_type_included(Type,atm), !.
create_undefined_size(Type,term(('?','?'))) :-
	dz_equivalent_types(Type,term).
create_undefined_size(Type,Size) :-
	get_type_definition_in_cache(Type,Defn),
	create_undefined_size_defn(Type,Defn,SizeArgs),
	( recursive_type(Type)
	-> Size =.. [Type,('?','?')|SizeArgs]
	 ; Size =.. [Type|SizeArgs]
	).

create_undefined_sizes([],[]).
create_undefined_sizes([T|R],[UndT|UndR]) :-
	create_undefined_size(T,UndT),
	create_undefined_sizes(R,UndR).

create_undefined_size_defn(_,[],[]).
create_undefined_size_defn(Type,[[]|Rest],Size) :-
	create_undefined_size_defn(Type,[^([])|Rest],Size).
create_undefined_size_defn(Type,[.(A,B)|Rest],Size) :-
	create_undefined_size_defn(Type,[^(.(A,B))|Rest],Size).
create_undefined_size_defn(Type,[^(Rule)|Rest],[^(SizeT)|SizeR]) :-
	Rule =.. [F|Args],
	create_undefined_size_defn(Type,Args,Sizes),
	SizeT =.. [F|Sizes],
	create_undefined_size_defn(Type,Rest,SizeR).
create_undefined_size_defn(Type,[Ty|Rest],Sizes) :-
	%( dz_equivalent_types(Type,Ty)
	( dz_type_included(Ty,Type)
	-> create_undefined_size_defn(Type,Rest,Sizes)
	 ; create_undefined_size(Ty,Size),
	   create_undefined_size_defn(Type,Rest,SizeR),
	   Sizes = [Size|SizeR]
	).

% -----------------------------------------------
% Predicates for computing paths in call_to_entry
% -----------------------------------------------

% Use peel to generate unification equations between terms
get_unification_eqs(Eq1, Eq2, Eqs) :- peel(Eq1, Eq2, Eqs, []).

 :- pred paths(T1,T2,PE) # "Unify the terms @var{T1} and @var{T2} and get the path expressions in @var{PE}".
%%%% Tests %%%%
:- test paths(T1,T2,P): (T1=aterm(A,B),T2=aterm(t(R,T,Y),W)) => (P= [R=A//(t/3,1),T=A//(t/3,2),Y=A//(t/3,3),W=3]) + times(3) # "Example 1".
:- test paths(T1,T2,P): (T1=predicate1(InputVar1,OutputVar2),T2=predicate1(RealInput1,RealOutput2)) => (P=[RealInput1=InputVar1,RealOutput2=OutputVar2]).
%%%%%%%%%%%%%%%%

paths(E1, E2, Paths) :-
        get_unification_eqs(E1, E2, Eqs),
        move_to_var_and_selectors_list(Eqs, PathsNotSorted),
        sort(PathsNotSorted, Paths).

% Move from a list of unification equations to a list of variable and selectors
:- op(100, xfy, //).

initial_var(V,V) :- var(V), !.
initial_var(P//_,V) :- initial_var(P,V).

move_to_var_and_selectors(V=T, Selectors) :-
        move_to_var_and_selectors_aux(V, T, [], Selectors).

move_to_var_and_selectors_list([], []).
move_to_var_and_selectors_list([E|R], Selectors) :-
        move_to_var_and_selectors(E, ESelectors),
        move_to_var_and_selectors_list(R, RSelectors),
        append(ESelectors, RSelectors, Selectors).

% move_to_var_and_selectors_aux(V=G, CurrentPath, Selectors)
move_to_var_and_selectors_aux(V, GT, _, [V=GT]) :-
        ground(GT), !.
move_to_var_and_selectors_aux(V, Var, CurrentPath, [Var=Selector]) :-
        var(Var), !,
        reverse_to_selector(V, CurrentPath, Selector).
move_to_var_and_selectors_aux(V, Functor, CurrentPath, Selectors) :-
        functor(Functor, F, NArg), NArg > 0, !,
        move_to_var_and_selector_args(V, Functor, F, 1, NArg, CurrentPath, Selectors).

move_to_var_and_selector_args(V, Functor, F, I, NArg, CurrentPath, Selectors) :-
        I =< NArg,
        arg(I, Functor, Arg),
        move_to_var_and_selectors_aux(V, Arg, [(F/NArg, I)|CurrentPath], SelectorsArg),
        ( I < NArg
        -> I1 is I + 1,
           move_to_var_and_selector_args(V, Functor, F, I1, NArg, CurrentPath, SelectorsRest),
           append(SelectorsArg, SelectorsRest, Selectors)
         ; Selectors = SelectorsArg
        ).

reverse_to_selector(Var, [], Var).
reverse_to_selector(Var, [R|S], Var//Path) :-
        reverse_to_selector_acc(S, R, Path).
reverse_to_selector_acc([], CurrentPath, CurrentPath).
reverse_to_selector_acc([E|R], CurrentPath, EndPath) :-
        reverse_to_selector_acc(R, E//CurrentPath, EndPath).

selector_to_list(Var, []) :- var(Var), !.
selector_to_list((S,A), [(S,A)]) :- !.
%selector_to_list(P//T, List) :-
%	selector_to_list(P,ListP),
%	append(ListP,[T],List).
selector_to_list(P//T, List) :-
	selector_to_list(T,ListT),
	selector_to_list(P,ListP),
	append(ListP,ListT,List).

% -----------------------------------------------
% Main predicate for unification in call_to_entry
% -----------------------------------------------

% Get the type size for a path
% unify_path(Type,Size,Path,Out)
unify_path(Type,Size,Path,Out) :-
	selector_to_list(Path,PathL),
	unify_path_(Type,Size,PathL,Out).

unify_path_(_,Size,[],Size).
unify_path_(Type,Size,[(F,N)|R],Out) :-
	get_type_definition_in_cache(Type,Defn),
	( recursive_type(Type)
	-> Size =.. [_,(Min,Max)|Args],
	   unify_path_args(Type,Defn,Args,Args,(F,N),Min,Max,ArgT,StepSize)
	 ; Size =.. [_|Args],
	   unify_path_args(Type,Defn,Args,Args,(F,N),'?','?',ArgT,StepSize)
	),
	unify_path_(ArgT,StepSize,R,Out).

unify_path_args(T,[[]|RuleR],[I|ArgR],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize) :-
	unify_path_args(T,[^([])|RuleR],[I|ArgR],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize).
unify_path_args(T,[[A|B]|RuleR],[I|ArgR],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize) :-
	unify_path_args(T,[^(.(A,B))|RuleR],[I|ArgR],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize).
unify_path_args(T,[^(Rule)|_],[^(Arg)|_],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize) :-
	functor(Rule,F,FN), !,
	arg(N,Rule,ArgT),
	( dz_equivalent_types(T,ArgT)
	-> % We have a recursion, so one less
	   StepSize =.. [T,(Min-1,Max-1)|InitialArgs]
	 ; % We don't have recursion
	   Rule =.. [_|RArgs],
	   get_size_arg_number(T,N,RArgs,M),
	   arg(M,Arg,StepSize)
	).
unify_path_args(T,[_|RuleR],[_|ArgR],InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize) :-
	unify_path_args(T,RuleR,ArgR,InitialArgs,(F/FN,N),Min,Max,ArgT,StepSize).

get_size_arg_number(Ty,M,[RTy|RestArgs],N) :-
	dz_equivalent_types(Ty,RTy), !,
	get_size_arg_number(Ty,M,RestArgs,N).
get_size_arg_number(_,1,_,1).
get_size_arg_number(Ty,M,[_|RestArgs],N) :-
	M > 1,
	M1 is M - 1,
	get_size_arg_number(Ty,M1,RestArgs,N).

% -----------------------------------------------
% Main predicate for unification in exit_to_prime
% -----------------------------------------------

% Generate the size of a term from its components
% unify_back(Ty,Term,VarSize,VarType,OutSize)
% Special case, we are just doing a var with the same type
unify_back(Ty,Term,VarSizes,VarTypes,OutSize) :-
	var(Term), !,
	% If the variable is not there, it means that it is free
	( find_variable_type(VarTypes,Term,VarT)
	-> true
	 ; VarT = vr
	),
	( dz_equivalent_types(Ty,VarT)
	-> % It is the same type
	   ( get_var_size_type(VarSizes,Term,OutSize)
	   -> true
	    ; OutSize = var
	   )
	 ; % We have an inclusion
	   ( get_var_size_type(VarSizes,Term,SubSize)
	   -> true
	    ; SubSize = var
	   ),
	   get_size_type(Ty,SuperSize),
	   edz_type_included(VarT,SubSize,Ty,SuperSize,Relations),
	   substitute_relations(SuperSize,Relations,OutSize)
	).
%	 ; % Ty is a variant type
%	   % get_type_definition(Ty,Defn),
%	   get_var_size_type(VarSizes,Term,VarSize),
%	   create_undefined_but(Ty,VarT,VarSize,OutSize)
%	).
unify_back(Ty,Term,_VarSizes,_VarTypes,OutSize) :-
	ground(Term),
	create_ground_size(Term,Ty,OutSize).
unify_back(Ty,Term,VarSizes,VarTypes,^(OutSize)) :-
	Term =.. [Functor|Args],
	Ty = ^(Rule),
	Rule =.. [Functor|RuleArgs],
	unify_back_list(RuleArgs,Args,VarSizes,VarTypes,OutArgs),
	OutSize =.. [Functor|OutArgs].
unify_back(Ty,Term,VarSizes,VarTypes,OutSize) :-
	Term =.. [Functor|Args],
	Ty \= ^(_),
	get_type_definition_in_cache(Ty,Defn),
	find_rule_with_functor(Defn,Functor,Rule),
	Rule =.. [_|RuleArgs],
	unify_back_list(RuleArgs,Args,VarSizes,VarTypes,OutArgs),
	( typmember(Ty,RuleArgs)
	-> % The rule is recursive
	   generate_rec_size(Ty,RuleArgs,OutArgs,OutSize,Rule)
	 ; % The rule is not recursive
	   % 1. Create a undefined element
	   create_undefined_size(Ty,UndefS),
	   ThingToSmashWith =.. [Functor|OutArgs],
	   % 2. Change the part of the undefined size with the functor
	   ( recursive_type(Ty)
	   -> UndefS =.. [UnF,(UnMin,UnMax)|UnArgs],
	      replace_functor(UnArgs,ThingToSmashWith,SmArgs),
	      OutSize =.. [UnF,(UnMin,UnMax)|SmArgs]
	    ; UndefS =.. [UnF|UnArgs],
	      replace_functor(UnArgs,ThingToSmashWith,SmArgs),
	      OutSize =.. [UnF|SmArgs]
	   )
	).

find_rule_with_functor([[]|_],[],[]) :- !.
find_rule_with_functor([[A|B]|_],'.',.(A,B)) :- !.
find_rule_with_functor([^(Functor)|_],F,Functor) :-
	functor(Functor,F,_), !.
find_rule_with_functor([_|R],F,Rule) :-
	find_rule_with_functor(R,F,Rule).

typmember(Ty,[T|_]) :- dz_equivalent_types(Ty,T).
typmember(Ty,[_|R]) :- typmember(Ty,R).

substitute_relations(S,Relations,Out) :-
	split_equalities_in_two(Relations,Vars,Subst),
	substitute_vars(S,Vars,Subst,Out).

split_equalities_in_two([],[],[]).
split_equalities_in_two([A=B|R],[A|S],[B|T]) :-
	split_equalities_in_two(R,S,T).

unify_back_list([],[],_,_,[]).
unify_back_list([Ty|TyR],[Term|TermR],VarSizes,VarTypes,[Out|OutR]) :-
	unify_back(Ty,Term,VarSizes,VarTypes,Out),
	unify_back_list(TyR,TermR,VarSizes,VarTypes,OutR).

generate_rec_size(Ty,Ts,Sizes,Out,Rule) :-
	generate_rec_size_(Ty,Ts,Sizes,Step),
	Step =.. [F,(Min,Max)|Args],
	update_non_rec_param(Args,Ty,Ts,Sizes,Rule,NewArgs),
	add_(Min,1,Min1), add_(Max,1,Max1),
	Out =.. [F,(Min1,Max1)|NewArgs].

generate_rec_size_(Ty,[T|TR],[SizeT|SizeR],Size) :-
	dz_equivalent_types(Ty,T), !,
	( generate_rec_size_(Ty,TR,SizeR,RestSize)
	-> % There were more recursive applications
	   merge_rec_info(SizeT,RestSize,Size)
	 ; Size = SizeT
	).
generate_rec_size_(Ty,[_|TR],[_|SizeR],Size) :-
	generate_rec_size_(Ty,TR,SizeR,Size).

update_non_rec_param([],_,_,_,_,[]).
update_non_rec_param([^(F)|R],Ty,Ts,Sizes,Rule,[^(UpF)|R]) :-
	functor(F,Fun,_),
	functor(Rule,Fun,_), !,
	% It is the rule to use
	F =.. [Fun|Args],
	update_non_rec_param_args(Ty,Ts,Args,Sizes,NewArgs),
	UpF =.. [Fun|NewArgs].
update_non_rec_param([^(F)|R],Ty,Ts,Sizes,Rule,[^(F)|UpR]) :-
	functor(F,Fun,_),
	\+ functor(Rule,Fun,_),
	update_non_rec_param(R,Ty,Ts,Sizes,Rule,UpR).

update_non_rec_param_args(_,[],_,_,[]).
update_non_rec_param_args(Ty,[T|RestT],Args,Sizes,NewArgs) :-
	dz_equivalent_types(Ty,T), !,
	update_non_rec_param_args(Ty,RestT,Args,Sizes,NewArgs).
update_non_rec_param_args(Ty,[_|RestT],[A|Args],[S|Sizes],[NewA|NewArgs]) :-
	merge_rec_info_max_min([A],[S],[NewA]),
	update_non_rec_param_args(Ty,RestT,Args,Sizes,NewArgs).

merge_rec_info(Size1,Size2,JoinSize) :-
	Size1 =.. [F,(Min1,Max1)|Args1],
	Size2 =.. [F,(Min2,Max2)|Args2],
	merge_rec_info_max_min(Args1,Args2,JoinArgs),
	add_(Min1,Min2,JoinMin),
	add_(Max1,Max2,JoinMax),
	JoinSize =.. [F,(JoinMin,JoinMax)|JoinArgs].

merge_rec_info_max_min([],[],[]).
merge_rec_info_max_min([A1|R1],[A2|R2],[AJ|RJ]) :-
	( A1 =.. [F,(Min1,Max1)|Args1] % See whether it is recursive
	-> A2 =.. [F,(Min2,Max2)|Args2],
	   min_(Min1,Min2,JMin), max_(Max1,Max2,JMax),
	   merge_rec_info_max_min(Args1,Args2,ArgsJ),
	   AJ =.. [F,(JMin,JMax)|ArgsJ]
	 ; A1 =.. [F|Args1], A2 =.. [F|Args2],
	   merge_rec_info_max_min(Args1,Args2,ArgsJ),
	   AJ =.. [F|ArgsJ]
	),
	merge_rec_info_max_min(R1,R2,RJ).

replace_functor([],_,[]).
replace_functor([^(F)|R],G,[^(H)|SmR]) :-
	functor(F,Functor,N),
	( functor(G,Functor,N)
	-> H = G
	 ; H = F
	),
	replace_functor(R,G,SmR).

% Create the undefined size for a ground element
create_undefined_but(Type,_,_,num(('?','?'))) :-
	dz_type_included(Type,num), !.
create_undefined_but(Type,_,_,atom(('?','?'))) :-
	dz_type_included(Type,atm), !.
create_undefined_but(Type,_,_,term(('?','?'))) :-
	dz_equivalent_types(Type,term).
create_undefined_but(Type,VarT,VarSize,Size) :-
	get_type_definition_in_cache(Type,Defn),
	create_undefined_but_defn(Type,Defn,VarT,VarSize,SizeArgs),
	( recursive_type(Type)
	-> Size =.. [Type,('?','?')|SizeArgs]
	 ; Size =.. [Type|SizeArgs]
	).

create_undefined_buts([],_,_,[]).
create_undefined_buts([T|R],VarT,VarSize,[UndT|UndR]) :-
	create_undefined_but(T,VarT,VarSize,UndT),
	create_undefined_buts(R,VarT,VarSize,UndR).

create_undefined_but_defn(_,[],_,_,[]).
create_undefined_but_defn(Type,[[]|Rest],VarT,VarSize,Size) :-
	create_undefined_but_defn(Type,[^([])|Rest],VarT,VarSize,Size).
create_undefined_but_defn(Type,[.(A,B)|Rest],VarT,VarSize,Size) :-
	create_undefined_but_defn(Type,[^(.(A,B))|Rest],VarT,VarSize,Size).
create_undefined_but_defn(Type,[^(Rule)|Rest],VarT,VarSize,[^(SizeT)|SizeR]) :-
	Rule =.. [F|Args],
	create_undefined_but_defn(Type,Args,VarT,VarSize,Sizes),
	SizeT =.. [F|Sizes],
	create_undefined_but_defn(Type,Rest,VarT,VarSize,SizeR).
create_undefined_but_defn(Type,[Ty|Rest],VarT,VarSize,Sizes) :-
	( dz_equivalent_types(Type,Ty)
	-> create_undefined_but_defn(Type,Rest,VarT,VarSize,Sizes)
	 ; ( dz_equivalent_types(Type,VarT)
	   -> Size = VarSize
	    ; create_undefined_but(Ty,VarT,VarSize,Size)
	   ),
	   create_undefined_but_defn(Type,Rest,VarT,VarSize,SizeR),
	   Sizes = [Size|SizeR]
	).

% -------------------------------------------------------
% Special predicate for unification within is/2 predicate
% -------------------------------------------------------
% TODO: This is not working if VarL or VarR are not
% in the Mapping as numbers. This might be correct, but
% should not fail, and probably the mapping at this point
% should contain the types already inferred by the type
% analysis.

create_num_assignment(Var,Mapping,VarL,VarR) :-
	var(Var), !,
	get_var_size_type(Mapping,Var,num((VarL,VarR))).
create_num_assignment(Ground,_,Ground,Ground) :-
	ground(Ground), !.
create_num_assignment(mod(_,B),Mapping,Left,Right) :-
	% In a expressions mod(A,B)
	% - lower bound is always 0
	Left = 0,
	% - and upper bound is the B-1
	create_num_assignment(B,Mapping,_,RightB),
	Right = RightB - 1.
create_num_assignment(Fun,Mapping,Left,Right) :-
	Fun =.. [F|Args], F \= mod, !,
	create_num_assignment_list(Args,Mapping,ArgsLeft,ArgsRight),
	Left =.. [F|ArgsLeft],
	Right =.. [F|ArgsRight].

create_num_assignment_list([],_,[],[]).
create_num_assignment_list([E|R],Map,[EL|RL],[ER|RR]) :-
	create_num_assignment(E,Map,EL,ER),
	create_num_assignment_list(R,Map,RL,RR).

% ----------------------------------------
% Information about user-defined resources
% ----------------------------------------

% :- multifile 'resource:identifier'/1.
% :- multifile 'resource:head_cost'/4.
% :- multifile 'resource:literal_cost'/4.
% :- multifile 'resource:default_cost'/3.

%get_all_resources(Rs) :-
	% ( setof(R,'resource:identifier'(R),Rs)
	% -> true
	%  ; Rs = []
	% ).


get_resource_head_cost(Res,Bound,Sg,_CC,Cost):-
%	is_current_cc_for_resource(Res,CC),!,
%	error_message("Counting HEAD in ~q",[CC]),
	approx_to_bound(Approx,Bound),
	head_cost(Approx,Res,F),
	(callable(F) ->
	 (get_litinfo(Sg,Approx,_,Lit),
	  eval_cost_lit(F,Lit,Cost))
	; 
	    Cost = F
	).
% MH: all these default clauses which are commented out make sense
%     (with a warning to the user). However it is probably better to
%     implement this in resources/res_assrt_defs/resources_lib (as it
%     is done now).
%
% get_resource_head_cost(_Res,_Bound,_Sg,_CC,0).

get_resource_literal_cost(Res,Bound,Sg,_CC,Cost) :-
	%is_current_cc_for_resource(Res,CC),!,
%	error_message("Counting LIT in ~q",[CC]),
	approx_to_bound(Approx,Bound),
	literal_cost(Approx,Res,F),
	(callable(F) ->
	 (get_litinfo(Sg,Approx,_,Lit),
	  eval_cost_lit(F,Lit,Cost))
	; 
	    Cost = F
	).
%get_resource_literal_cost(_Res,_Bound,_Sg,_CC,0).

get_resource_default_cost(Res,Bound,_CC,Cost) :-
%	is_current_cc_for_resource(Res,CC),!,
%	error_message("Counting DEFAULT in ~q",[CC]),
	approx_to_bound(Approx,Bound),
	default_cost(Approx,Res,Cost),!.
%get_resource_default_cost(_Res,_Bound,_CC,0).



% get_resource_head_cost(Res,Bound,Sg,Cost) :-
% 	approx_to_bound(Approx,Bound),
% 	head_cost(Approx,Res,F),
% 	(callable(F) ->
% 	 (get_litinfo(Sg,Approx,_,Lit),
% 	  eval_cost_lit(F,Lit,Cost))
% 	; 
% 	    Cost = F
% 	).
% 	% ( 'resource:head_cost'(Res,Bound,Sg,Cost)
% 	% -> true
% 	%  ; Cost = 0
% 	% ).
% get_resource_literal_cost(Res,Bound,Sg,Cost) :-
% 	approx_to_bound(Approx,Bound),
% 	literal_cost(Approx,Res,F),
% 	(callable(F) ->
% 	 (get_litinfo(Sg,Approx,_,Lit),
% 	  eval_cost_lit(F,Lit,Cost))
% 	; 
% 	    Cost = F
% 	).

% 	% ( 'resource:literal_cost'(Res,Bound,Sg,Cost)
% 	% -> true
% 	%  ; Cost = 0
% 	% ).

% get_resource_default_cost(Res,Bound,Cost) :-
% 	approx_to_bound(Approx,Bound),
% 	default_cost(Approx,Res,Cost),!.
% get_resource_default_cost(_Res,_Bound,0).

% ( 'resource:default_cost'(Res,Bound,Cost)
	% -> true
	%  ; Cost = 0
	% ).

% get_all_resources(R):-
% 	get_all_resources(Res), % get_all_static_cost_centers(CCs),
% 	append(Res,CCs,R),!.

% Generates the variables for the resource analysis
generate_resource_variables(V) :-
	get_all_resources(R),
	generate_resource_variables_(R,V), !.

generate_resource_variables_([],[]).
generate_resource_variables_([R|Rs],[V|Vs]) :-
	\+ atom(R),!,
	R =.. [RN,CC],
	V =.. [RN,(CC,_,_)],
	generate_resource_variables_(Rs,Vs).
generate_resource_variables_([R|Rs],[V|Vs]) :-
	V =.. [R,(_,_)],
	generate_resource_variables_(Rs,Vs).

% Generates nil temporary for the resource analysis
generate_resource_nil_variables(V) :-
	get_all_resources(R),
%	get_all_resources(R),
	generate_resource_nil_variables_(R,V), !.

generate_resource_nil_variables_([],[]).
generate_resource_nil_variables_([R|Rs],[V|Vs]) :-
	\+ atom(R),!,
	R =.. [RN,CC],
	V =.. [RN,(CC,nil,nil)],
	generate_resource_nil_variables_(Rs,Vs).
generate_resource_nil_variables_([R|Rs],[V|Vs]) :-
	V =.. [R,(nil,nil)],
	generate_resource_nil_variables_(Rs,Vs).

% Generates the final expression for the number of solutions
% Basically, it selects 1 if there was no previous variable
% Or assigns the last solution variable
generate_final_solution_expr(PrevSolVar,IsDet,IsFail,OutSolVar,Expr):-
	generate_final_solution_expr_lower(IsFail,Lower),
	generate_final_solution_expr_upper(PrevSolVar,IsDet,Upper),
	Expr = assign(solution(OutSolVar),solution((Lower,Upper))).


generate_final_solution_expr_lower(true,0).
generate_final_solution_expr_lower(false,1).

generate_final_solution_expr_upper(PrevSolVar,_,1) :-
	PrevSolVar == (nil,nil), !.
generate_final_solution_expr_upper(_,true,1) :- !.
generate_final_solution_expr_upper((_,Up),false,Up) :- !.

% Generates the final expressions for resources
generate_final_resources_exprs([],_,_,_,[]).
generate_final_resources_exprs([P|Ps],Sg,IsFail,[O|Os],[Expr|Exprs]) :-
	P =.. [Resource,PrevVars],
	generate_final_resource_expr(Resource,PrevVars,Sg,IsFail,O,Expr),
	generate_final_resources_exprs(Ps,Sg,IsFail,Os,Exprs).

generate_final_resource_expr(Resource,Prevs,Sg,IsFail,OutStepVars,Expr) :-
	conj_to_list(Prevs,LPrevs),length(LPrevs,3),!,
	LPrevs = [CC,LowVar,UpVar],
	generate_final_res_expr_lower(Resource,CC,Sg,IsFail,LowVar,Lower),
	generate_final_res_expr_upper(Resource,CC,Sg,UpVar,Upper),
	R =.. [Resource,(CC,Lower,Upper)],
	Expr = assign(OutStepVars,R).
generate_final_resource_expr(Resource,Prevs,Sg,IsFail,OutStepVars,Expr) :-
	conj_to_list(Prevs,LPrevs),length(LPrevs,2),!,
	LPrevs = [LowVar,UpVar],
	generate_final_res_expr_lower(Resource,nil,Sg,IsFail,LowVar,Lower),
	generate_final_res_expr_upper(Resource,nil,Sg,UpVar,Upper),
	R =.. [Resource,(Lower,Upper)],
	Expr = assign(OutStepVars,R).

%%%%CHECKFORCC%%%%
%%MK: 2014-08-07: Maybe these clauses are mixing lower and upper headcosts.
% generate_final_res_expr_lower(_,_,_,true,_,0).
% generate_final_res_expr_lower(R,_,Sg,false,A,HeadCost) :-
% 	A == nil, !,
% 	get_resource_head_cost(R,lower,Sg,HeadCost).
% generate_final_res_expr_lower(R,_,Sg,false,A,A+HeadCost) :-
% 	A \== nil, !,
% 	get_resource_head_cost(R,upper,Sg,HeadCost).

% generate_final_res_expr_upper(R,_,Sg,A,HeadCost) :-
% 	A == nil, !,
% 	get_resource_head_cost(R,lower,Sg,HeadCost).
% generate_final_res_expr_upper(R,_,Sg,A,A+HeadCost) :-
% 	A \== nil, !,
% 	get_resource_head_cost(R,upper,Sg,HeadCost).

generate_final_res_expr_lower(_,_,_,true,_,0).
generate_final_res_expr_lower(R,CC,Sg,false,A,HeadCost) :-
	A == nil, !,
	get_resource_head_cost(R,lower,Sg,CC,HeadCost).
generate_final_res_expr_lower(R,CC,Sg,false,A,A+HeadCost) :-
	A \== nil, !,
	get_resource_head_cost(R,lower,Sg,CC,HeadCost).

generate_final_res_expr_upper(R,CC,Sg,A,HeadCost) :-
	A == nil, !,
	get_resource_head_cost(R,upper,Sg,CC,HeadCost).
generate_final_res_expr_upper(R,CC,Sg,A,A+HeadCost) :-
	A \== nil, !,
	get_resource_head_cost(R,upper,Sg,CC,HeadCost).



%%%%CHECKFORCC%%%%
% Different cases that may happen when extending resources info
% extend_res_first_predicate_fail(+TempVars,-NewVars,-Assigns)
extend_res_first_predicate_fail([],[],[]).
extend_res_first_predicate_fail([R|Rs],[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,3),!,
	LArgs = [CC,_,ST_U],
%	R =.. [Res,(CC,_,ST_U)],
	N =.. [Res,(CC,ZeroStep_L,ST_U)],
	A = assign(ZeroStep_L,0),
	% Recursive
	extend_res_first_predicate_fail(Rs,Ns,As).
extend_res_first_predicate_fail([R|Rs],[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,2),!,
	LArgs = [_,ST_U],


%	R =.. [Res,(_,ST_U)],
	N =.. [Res,(ZeroStep_L,ST_U)],
	A = assign(ZeroStep_L,0),
	% Recursive
	extend_res_first_predicate_fail(Rs,Ns,As).

%%%%CHECKFORCC%%%%
% extend_res_first_case(+TempVars,+YetFailed,+Sg,-NewVars,-Assign)
extend_res_first_case([],_,_,[],[]).
extend_res_first_case([R|Rs],YetFailed,Sg,[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,3),!,
	LArgs = [CC,ST_L,ST_U],
%	R =.. [Res,(CC,ST_L,ST_U)],
	N =.. [Res,(CC,_,_)],
	( YetFailed = true
	-> LowerInstr = 0
	 ; get_resource_literal_cost(Res,lower,Sg,CC,Lit_L),
	   LowerInstr = Lit_L + ST_L
	),
	get_resource_literal_cost(Res,upper,Sg,CC,Lit_U),
	UpperInstr = Lit_U + ST_U,
	Rhs =.. [Res,(CC,LowerInstr,UpperInstr)],
	A = assign(N,Rhs),
	% Recursive
	extend_res_first_case(Rs,YetFailed,Sg,Ns,As).

extend_res_first_case([R|Rs],YetFailed,Sg,[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,2),!,
	LArgs = [ST_L,ST_U],
%	R =.. [Res,(ST_L,ST_U)],
	N =.. [Res,(_,_)],
	( YetFailed = true
	-> LowerInstr = 0
	 ; get_resource_literal_cost(Res,lower,Sg,nil,Lit_L),
	   LowerInstr = Lit_L + ST_L
	),
	get_resource_literal_cost(Res,upper,Sg,nil,Lit_U),
	UpperInstr = Lit_U + ST_U,
	Rhs =.. [Res,(LowerInstr,UpperInstr)],
	A = assign(N,Rhs),
	% Recursive
	extend_res_first_case(Rs,YetFailed,Sg,Ns,As).

%%%%CHECKFORCC%%%% %%%VERYKEY%%%
% extend_res_usual_case(+TempVars,+PrevVars,+YetFailed,+SolVars,+Sg,-NewVars,-Assign)
extend_res_usual_case([],_,_,_,_,[],[]).
extend_res_usual_case([R|Rs],[P|Ps],YetFailed,(Sol_L,Sol_U),Sg,[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,3),!,
	LArgs = [CC,ST_L,ST_U],
%	R =.. [Res,(CC,ST_L,ST_U)],
	P =.. [Res|[PArgs]],
	conj_to_list(PArgs,PLArgs),
	length(PLArgs,3),
	PLArgs = [CC,SP_L,SP_U],
%	P =.. [_,(CC,SP_L,SP_U)],
	N =.. [Res,(CC,_,_)],
	( YetFailed = true
	-> LowerInstr = SP_L
	 ; get_resource_literal_cost(Res,lower,Sg,CC,Lit_L),
	   LowerInstr = SP_L + Sol_L*(Lit_L + ST_L)
	),
	get_resource_literal_cost(Res,upper,Sg,CC,Lit_U),
	UpperInstr = SP_U + Sol_U*(Lit_U + ST_U),
	Rhs =.. [Res,(CC,LowerInstr,UpperInstr)],
	A = assign(N,Rhs),
	% Recursive
	extend_res_usual_case(Rs,Ps,YetFailed,(Sol_L,Sol_U),Sg,Ns,As).
extend_res_usual_case([R|Rs],[P|Ps],YetFailed,(Sol_L,Sol_U),Sg,[N|Ns],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,2),!,
	LArgs = [ST_L,ST_U],


%	R =.. [Res,(ST_L,ST_U)],

	P =.. [Res|[PArgs]],
	conj_to_list(PArgs,PLArgs),
	length(PLArgs,2),
	PLArgs = [SP_L,SP_U],


%	P =.. [_,(SP_L,SP_U)],
	N =.. [Res,(_,_)],
	( YetFailed = true
	-> LowerInstr = SP_L
	 ; get_resource_literal_cost(Res,lower,Sg,nil,Lit_L),
	   LowerInstr = SP_L + Sol_L*(Lit_L + ST_L)
	),
	get_resource_literal_cost(Res,upper,Sg,nil,Lit_U),
	UpperInstr = SP_U + Sol_U*(Lit_U + ST_U),
	Rhs =.. [Res,(LowerInstr,UpperInstr)],
	A = assign(N,Rhs),
	% Recursive
	extend_res_usual_case(Rs,Ps,YetFailed,(Sol_L,Sol_U),Sg,Ns,As).

:- use_module(library(formulae),[conj_to_list/2]).

% Generate assignments from resources default costs
generate_resource_default_assigns([],[]).
generate_resource_default_assigns([R|Rs],[A|As]) :-
	R =.. [Res|[Args]],
	conj_to_list(Args,LArgs),
	length(LArgs,3),!,
	LArgs = [CC,_,_],
	get_resource_default_cost(Res,lower,CC,LowCost),
	get_resource_default_cost(Res,upper,CC,UpCost),
	Cost =.. [Res,(CC,LowCost,UpCost)],
	A = assign(R,Cost),
	generate_resource_default_assigns(Rs,As).
generate_resource_default_assigns([R|Rs],[A|As]) :-
	R =.. [Res|_],
	get_resource_default_cost(Res,lower,nil,LowCost),
	get_resource_default_cost(Res,upper,nil,UpCost),
	Cost =.. [Res,(LowCost,UpCost)],
	A = assign(R,Cost),
	generate_resource_default_assigns(Rs,As).

% ------------------------------------------------
% Conversions between internal and "beauty" syntax
% ------------------------------------------------

%% "Beauty" size descriptions with functor -> size syntax
beautify_syntax(Size,BeautifulSize) :-
	Size =.. [F,(L,U)|Args],!,
	beautify_args(F,Args,BeautifulArgs),
	choose_beauty_args(BeautifulArgs,FinalArgs),
	BeautifulSize =.. [F,L,U|FinalArgs].
beautify_syntax(Size,BeautifulSize) :-
	Size =.. [F|Args],!,
	beautify_args(F,Args,BeautifulArgs),
	choose_beauty_args(BeautifulArgs,FinalArgs),
	BeautifulSize =.. [F|FinalArgs].

beautify_args(_,[],[]).
beautify_args(T,[^(E)|R],Beauty) :-
	get_type_definition_in_cache(T,Defn),
	E =.. [F|Args],
	find_rule_with_functor(Defn,F,Rule),
	Rule =.. [F|RuleArgs],
	beautify_args_functor(T,F,1,RuleArgs,Args,BeautyArgs),
	beautify_args(T,R,BeautyR),
	append(BeautyArgs,BeautyR,Beauty).
beautify_args(T,[E|R],[p('$empty',S)|BeautyR]) :-
	E \== ^(_),
	beautify_syntax(E,S),
	beautify_args(T,R,BeautyR).

beautify_args_functor(_,_,_,[],_,[]).
beautify_args_functor(T,F,N,[Ty|RRule],Args,BeautyArgs) :-
	dz_equivalent_types(T,Ty), !,
	N1 is N + 1,
	beautify_args_functor(T,F,N1,RRule,Args,BeautyArgs).
beautify_args_functor(T,F,N,[_|RRule],[A|RArgs],[p(F/N,BeautyA)|BeautyR]) :-
	beautify_syntax(A,BeautyA), !,
	N1 is N + 1,
	beautify_args_functor(T,F,N1,RRule,RArgs,BeautyR).
beautify_args_functor(T,F,N,[_|RRule],[_|RArgs],BeautyR) :-
	% In some cases beutify fails
	N1 is N + 1,
	beautify_args_functor(T,F,N1,RRule,RArgs,BeautyR).

choose_beauty_args([],[]).
choose_beauty_args([p(_,Only)],[Only]).
choose_beauty_args([A,B|R],[[A,B|R]]).

%uglify_syntax(Ty,Size,Ugly)
uglify_syntax(Type,Size,num((L,U))) :-
	dz_type_included(Type,num), !,
	Size =.. [_,L,U].
uglify_syntax(Type,_,var) :-
	dz_type_included(Type,vr), !.
uglify_syntax(Type,_,atom((1,1))) :-
	dz_type_included(Type,atm),!.
uglify_syntax(Type,Size,Ugly) :-
	dz_equivalent_types(Type,term), !,
	( Size =.. [_,L,U|_]
	-> Ugly = term((L,U))
	; Ugly = term((0,inf))
	    %Ugly = term((0,'Infinity'))
	).
uglify_syntax(Type,Size,Ugly) :-
	get_type_definition_in_cache(Type,Defn),
	( recursive_type(Type)
	-> Size =.. [_,L,U,SArgs],
	   uglify_args(Type,Defn,SArgs,UArgs),
	   Ugly =.. [Type,(L,U)|UArgs]
	 ; Size =.. [_,SArgs],
	   uglify_args(Type,Defn,SArgs,UArgs),
	   Ugly =.. [Type|UArgs]
	).

uglify_args(_,[],_,[]).
uglify_args(T,[[]|Rest],SArgs,UArgs) :-
	!, uglify_args(T,[^([])|Rest],SArgs,UArgs).
uglify_args(T,[.(A,B)|Rest],SArgs,UArgs) :-
	!, uglify_args(T,[^(.(A,B))|Rest],SArgs,UArgs).
uglify_args(T,[^(Functor)|Rest],SArgs,[^(UFunctor)|URest]) :-
	Functor =.. [F|FArgs],
	uglify_args_each(T,F,FArgs,SArgs,1,UArgs),
	UFunctor =.. [F|UArgs],
	uglify_args(T,Rest,SArgs,URest).
uglify_args(T,[Ty|Rest],SArgs,[USize|URest]) :-
	( find_beauty_arg(SArgs,Ty,'$empty',Arg)
	-> uglify_syntax(Ty,Arg,USize)
	 ; create_undefined_size(Ty,USize)
	),
	uglify_args(T,Rest,SArgs,URest).

uglify_args_each(_,_,[],_,_,[]) :- !.
uglify_args_each(T,F,[^(A)|Rest],SArgs,N,[^(A)|URest]) :-
	!, N1 is N + 1,
	uglify_args_each(T,F,Rest,SArgs,N1,URest).
uglify_args_each(T,F,[E|Rest],SArgs,N,URest) :-
	dz_type_included(E,T), !,
	N1 is N + 1,
	uglify_args_each(T,F,Rest,SArgs,N1,URest).
uglify_args_each(T,F,[E|Rest],SArgs,N,[UElt|URest]) :-
	( find_beauty_arg(SArgs,E,F/N,BElt)
	-> uglify_syntax(E,BElt,UElt)
	 ; create_undefined_size(E,UElt)
	), !,
	N1 is N + 1,
	uglify_args_each(T,F,Rest,SArgs,N1,URest).

% find_beauty_arg(SArgs,Ty,Position,Arg)
find_beauty_arg([p(Position,Arg)|_],_,Position,Arg) :- !.
find_beauty_arg([E|_],Ty,_,E) :-
	E \= p(_,_), functor(E,TyE,_),
	dz_type_included(TyE,Ty), !.
find_beauty_arg([_|R],Ty,Position,Arg) :-
	!, find_beauty_arg(R,Ty,Position,Arg).
find_beauty_arg(A,Ty,_,A) :-
	functor(A,TyA,_),
	dz_type_included(TyA,Ty), !.

% --------------------------------------------
% Treat tests coming from non-failure analysis
% --------------------------------------------

expand_tests(ArithTests,Expanded) :-
	change_test_order(ArithTests,ArithTests2),
	sort(ArithTests2,ArithTests3),
	expand_tests_(ArithTests3,Expanded).

change_test_order([],[]).
change_test_order([B = A|R],[A =:= B,B =:= A|S]) :-
	var(A), var(B),
	!,change_test_order(R,S).
change_test_order([A = B|R],[A =:= B,B =:= A|S]) :-
	var(A), ground(B), num(B),
	!,change_test_order(R,S).
change_test_order([B = A|R],[A =:= B,B =:= A|S]) :-
	var(A), ground(B), num(B),
	!,change_test_order(R,S).
change_test_order([A =:= B|R],[A =:= B,B =:= A|S]) :-
	!,change_test_order(R,S).
change_test_order([A =< B|R],[A =< B|S]) :-
	!,change_test_order(R,S).
change_test_order([A < B|R],[A < B|S]) :-
	!,change_test_order(R,S).
change_test_order([A >= B|R],[B =< A|S]) :-
	!,change_test_order(R,S).
change_test_order([A > B|R],[B < A|S]) :-
	!,change_test_order(R,S).
change_test_order([_|R],S) :-
	!,change_test_order(R,S).

expand_tests_(T,E) :-
	expand_tests2(T,T,T,E1),
	( T == E1 -> E = E1 ; expand_tests_(E1,E) ).

expand_tests2([],_,H,H).
expand_tests2([E|R],T,H1,H3) :-
	expand_tests2_(E,T,H1,H2),
	!,expand_tests2(R,T,H2,H3).

expand_tests2_(_,[],H,H).
expand_tests2_(A =:= B,[C =:= D|R],H1,H4) :-
	(B == C -> insert(H1,A =:= D,H2) ; H2 = H1),
	(A == D -> insert(H2,C =:= B,H3) ; H3 = H2),
	!,expand_tests2_(A =:= B,R,H3,H4).
expand_tests2_(A =:= B,[C =< D|R],H1,H6) :-
	(A == C -> insert(H1,B =< D,H2) ; H2 = H1),
	(A == D -> insert(H2,C =< B,H3) ; H3 = H2),
	(B == C -> insert(H3,A =< D,H4) ; H4 = H3),
	(B == D -> insert(H4,C =< A,H5) ; H5 = H4),
	!,expand_tests2_(A =:= B,R,H5,H6).
expand_tests2_(A =:= B,[C < D|R],H1,H6) :-
	(A == C -> insert(H1,B < D,H2) ; H2 = H1),
	(A == D -> insert(H2,C < B,H3) ; H3 = H2),
	(B == C -> insert(H3,A < D,H4) ; H4 = H3),
	(B == D -> insert(H4,C < A,H5) ; H5 = H4),
	!,expand_tests2_(A =:= B,R,H5,H6).
expand_tests2_(A =< B,[C =< D|R],H1,H3) :-
	(B == C -> insert(H1,A =< D,H2) ; H2 = H1),
	!,expand_tests2_(A =< B,R,H2,H3).
expand_tests2_(A =< B,[C < D|R],H1,H3) :-
	(B == C -> insert(H1,A < D,H2) ; H2 = H1),
	!,expand_tests2_(A =< B,R,H2,H3).
expand_tests2_(A =< B,[_|R],H1,H2) :-
	!,expand_tests2_(A =< B,R,H1,H2).
expand_tests2_(A < B,[C < D|R],H1,H3) :-
	(B == C -> insert(H1,A < D,H2) ; H2 = H1),
	!,expand_tests2_(A < B,R,H2,H3).
expand_tests2_(A < B,[_|R],H1,H2) :-
	!,expand_tests2_(A < B,R,H1,H2).

find_wanted_tests(_,[],[]).
find_wanted_tests(Vars,[V =:= G|R], [V =:= G|S]) :-
	member(E,Vars), E == V, ground(G),
	!, find_wanted_tests(Vars,R,S).
find_wanted_tests(Vars,[V =< G|R], [V =< G|S]) :-
	member(E,Vars), E == V, ground(G),
	!, find_wanted_tests(Vars,R,S).
find_wanted_tests(Vars,[V < G|R], [V < G|S]) :-
	member(E,Vars), E == V, ground(G),
	!, find_wanted_tests(Vars,R,S).
find_wanted_tests(Vars,[_|R], S) :-
	!, find_wanted_tests(Vars,R,S).


% --------------------------------------------
% Predicates for pretty-print of equations 
% --------------------------------------------

% equation_name/1 stores the generated names of equations. This is used after in
% res_abs to distinguish which functors need to be replaced for pretty-print versions
% during the output.
:- export(equation_name/1).
:- data equation_name/1.

% counter_functions_name/1 is a counter used for generation of new equation names.
:- data counter_function_name/1.

% mapping_fn_names/1 stores a mapping from equation names to pretty-print names. 
:- data mapping_fn_names/1.



:- pred beautify_equations(Eq,EqB): (list(Eq,equation),var(EqB)) =>
	list(EqB) #
 "It receives a list of terms equation(Name,Args,Expr) and returns a list of pretty-print 
      versions of this equations (NewName(Args) == NewExpr), consistently replacing
      equations names with shorter (more readable) ones.".
beautify_equations(Eq,EqB):-
	retract_fact(mapping_fn_names(Map)),
	beautify_equations_(Eq,Map,EqB,Map2),!,
	asserta_fact(mapping_fn_names(Map2)).


beautify_equations_([],M,[],M).
beautify_equations_([equation(F,Args,Expr,T)|R],Map,[(H==E,T)|BeautyR],Map2):-
	get_function_name(Map,F,FName,NMap),
	beautify_head_args(Args,Expr,T,BArgs),
	H =.. [FName|BArgs],
	beautify_equations_expr(Expr,NMap,E0,NNMap),
	simplify(E0,E1),
	beautify_funcalls(E1,E),
	beautify_equations_(R,NNMap,BeautyR,Map2).

beautify_head_args(Args,Expr,T,BArgs):-
	varset((Expr,T),UsedVars),
	beautify_head_args_(Args,UsedVars,BArgs).
beautify_head_args_([],_,[]).
beautify_head_args_([V|Args],UsedVars,[V|BArgs]):-
	var(V),
	varmember(V,UsedVars),!,
	beautify_head_args_(Args,UsedVars,BArgs).
beautify_head_args_([V|Args],UsedVars,['-'|BArgs]):-
	var(V),!,
	beautify_head_args_(Args,UsedVars,BArgs).
beautify_head_args_(['?'|Args],UsedVars,['-'|BArgs]):-
	beautify_head_args_(Args,UsedVars,BArgs).
beautify_head_args_([A|Args],UsedVars,[A|BArgs]):-
	beautify_head_args_(Args,UsedVars,BArgs).


	
beautify_equations_expr(Expr,NMap,Expr,NMap):-
	var(Expr),!.
beautify_equations_expr(Expr,NMap,E,NNMap):-
	Expr = fun(F,Args),!,
	get_function_name(NMap,F,Name,Map1),
	beautify_equations_expr_args(Args,Map1,NArgs,NNMap),
	E =.. [Name|NArgs].
beautify_equations_expr(Expr,NMap,E,NNMap):-
	Expr =.. [F|Args],
	beautify_equations_expr_args(Args,NMap, NArgs,NNMap),
	E =.. [F|NArgs]. 

beautify_equations_expr_args([],M,[],M).
beautify_equations_expr_args([A|R],M,[NA|NR],NM):-
	beautify_equations_expr(A,M,NA,M0),!,
	beautify_equations_expr_args(R,M0,NR,NM).

beautify_funcalls(E,E):-
	var(E),!.
beautify_funcalls(E,E):-
	num(E),!.
beautify_funcalls(E,BE):-
	E = fun(Name,Args),!,
	BE =.. [Name|Args].
beautify_funcalls(E,BE):-
	E =.. [Op|Args],
	beautify_funcalls_args(Args,BArgs),
	BE =.. [Op|BArgs].

beautify_funcalls_args([],[]).
beautify_funcalls_args([A|Args],[B|BArgs]):-
	beautify_funcalls(A,B),!,
	beautify_funcalls_args(Args,BArgs).

get_function_name(Map,F,FName,Map):-
	member((F,FName),Map),!.
get_function_name(Map,F,FName,[(F,FName)|Map]):-
	get_new_function_name_verbose(F,FName),!.

get_new_function_name_verbose(F,FName):-
	atom_concat([_Mod,'zz',Pred,'zz',_Id,_],F),
	retract_fact(counter_function_name(Id)),
	C1 is Id + 1,
	asserta_fact(counter_function_name(C1)),
	atom_number(IdN,Id),
	atom_concat([Pred,'_',IdN],FName).

% get_new_function_name(F):-
% 	retract_fact(counter_function_name(C)),
% 	C1 is C + 1,
% 	asserta_fact(counter_function_name(C1)),
% 	number_to_function(C,F).

% Given a number, returns an atom that will be used as an equation/function name.
number_to_function(C,F):-
	C =< 20,!,
	C1 is C + 102,
	atom_codes(F,[C1]).
number_to_function(C,F):-
	Suf is C - 20,
	number_codes(Suf,CSuf),
	atom_codes(F,[102|CSuf]).


number_to_function_([],[]).
number_to_function_([N|R],[F|T]):-
	F is N + 54,
	number_to_function_(R,T).
 
% beautify_equations([],[]).
% beautify_equations([equation(F,Args,Expr)|R],[H==E|BeautyR]) :-
% 	H =.. [f|Args],
% 	update_functors(Expr,[(F,f)],E),
% 	beautify_equations(R,BeautyR).


clean_equation_names_cache:-
	retractall_fact(equation_name(_)),
	retractall_fact(counter_function_name(_)),
	retractall_fact(mapping_fn_names(_)),
	asserta_fact(mapping_fn_names([])),
	asserta_fact(counter_function_name(0)).


% ------------------------------------------------
% Conversions between internal and "beauty" syntax
% ------------------------------------------------

:- doc(section, "(De)Normalization").

:- use_module(typeslib(typeslib),
	[dz_type_included/2,
	 dz_equivalent_types/2,
	 param_type_symbol_renaming/2]).

:- export(normalize_rsize/3).

:- pred normalize_rsize(Type, SyntaxSize, InternalSize) :
	(regtype_t(Type), rsize_syntax_t(SyntaxSize)) => rsize_t(InternalSize)
        + (not_fails, is_det)
# "@var{InternalSize} is the internal representation of the syntactic
   rsize property @var{SyntaxSize}. The property applies on a term of type
   @var{Type}.".

 
normalize_rsize(Type,Size,num((L,U))) :-
	dz_type_included(Type,num), !,
	(Size =.. [_,L,U]; Size =.. [_,(L,U)]).
normalize_rsize(Type,_,var) :-
	dz_type_included(Type,vr), !.
normalize_rsize(Type,_,atom((1,1))) :-
	dz_type_included(Type,atm), !.
normalize_rsize(Type,Size,Ugly) :-
	dz_equivalent_types(Type,term),
	( (Size =.. [_,(L,U)];Size =.. [_,L,U|_])
	-> Ugly = term((L,U))
	 ; Ugly = term((0,'Infinity'))
	).
normalize_rsize(Type,Size,Ugly) :-
	get_type_definition_in_cache(Type,Defn),
	( recursive_type(Type)
	-> Size =.. [_,L,U,SArgs],
	   normalize_args(Type,Defn,SArgs,UArgs),
	   Ugly =.. [Type,(L,U)|UArgs]
	 ; Size =.. [_,SArgs],
	   normalize_args(Type,Defn,SArgs,UArgs),
	   Ugly =.. [Type|UArgs]
	).

normalize_args(_,[],_,[]).
normalize_args(T,[[]|Rest],SArgs,UArgs) :-
	!, normalize_args(T,[^([])|Rest],SArgs,UArgs).
normalize_args(T,[.(A,B)|Rest],SArgs,UArgs) :-
	!, normalize_args(T,[^(.(A,B))|Rest],SArgs,UArgs).
normalize_args(T,[^(Functor)|Rest],SArgs,[^(UFunctor)|URest]) :-
	Functor =.. [F|FArgs],
	normalize_args_each(T,F,FArgs,SArgs,1,UArgs),
	UFunctor =.. [F|UArgs],
	normalize_args(T,Rest,SArgs,URest).
normalize_args(T,[Ty|Rest],SArgs,[USize|URest]) :-
	( find_normalize_arg(SArgs,Ty,'$empty',Arg)
	-> normalize_rsize(Ty,Arg,USize)
	 ; create_undefined_size(Ty,USize)
	),
	normalize_args(T,Rest,SArgs,URest).

normalize_args_each(_,_,[],_,_,[]) :- !.
normalize_args_each(T,F,[^(A)|Rest],SArgs,N,[^(A)|URest]) :-
	!, N1 is N + 1,
	normalize_args_each(T,F,Rest,SArgs,N1,URest).
normalize_args_each(T,F,[E|Rest],SArgs,N,URest) :-
	dz_type_included(E,T), !,
	N1 is N + 1,
	normalize_args_each(T,F,Rest,SArgs,N1,URest).
normalize_args_each(T,F,[E|Rest],SArgs,N,[UElt|URest]) :-
	( find_normalize_arg(SArgs,E,F/N,BElt)
	-> normalize_rsize(E,BElt,UElt)
	 ; create_undefined_size(E,UElt)
	), !,
	N1 is N + 1,
	normalize_args_each(T,F,Rest,SArgs,N1,URest).


find_normalize_arg([p(Position,Arg)|_],_,Position,Arg) :- !.
find_normalize_arg([E|_],Ty,_,E) :-
	E \= p(_,_), functor(E,TyE,_),
	dz_type_included(TyE,Ty), !.
find_normalize_arg([_|R],Ty,Position,Arg) :-
	!, find_normalize_arg(R,Ty,Position,Arg).
find_normalize_arg(A,Ty,_,A) :-
	functor(A,TyA,_),
	dz_type_included(TyA,Ty), !.



:- pred denormalize_rsize(Size,SizeSyntax) : 
        rsize_t(Size) => rsize_syntax_t(SizeSyntax)
   + (not_fails, is_det)
# "@var{SizeSyntax} is the syntactical represantion of the internal
   representation @var{Size} for a rsize property.".

%% "Beauty" size descriptions with functor -> size rsize
denormalize_rsize(^(Size), ^(BeautifulSize)) :- !,
	Size =.. [F|Args], 
	denormalize_args(F,Args,BeautifulArgs),
	choose_denormalize_args(BeautifulArgs,FinalArgs),
	denormalize_lib_typesymbol(F,FN),
	BeautifulSize =.. [FN|FinalArgs].
denormalize_rsize(Size,BeautifulSize) :-
	Size =.. [_F,^(Args)],
	% NSize =.. [_F|Args],
	denormalize_rsize(^(Args),BeautifulSize).

denormalize_rsize(Size,BeautifulSize) :-
	Size =.. [F,(L,U)|Args],
	denormalize_args(F,Args,BeautifulArgs),
	choose_denormalize_args(BeautifulArgs,FinalArgs),
	% ( param_type_symbol_renaming(T,F) ->
	%     T =.. [G|_] 
	% ; G = F
	% ),
	type_symbol_rename(F,G),
	denormalize_lib_typesymbol(G,GN),
	BeautifulSize =.. [GN,L,U|FinalArgs].

denormalize_lib_typesymbol(T,N):-
	atom_concat([_,':',N],T),!.
denormalize_lib_typesymbol(T,T).

:- export(type_symbol_rename/2).
type_symbol_rename(F,Name):-
	 param_type_symbol_renaming(T,F),!,
	 T =.. [Name|_].
type_symbol_rename(F,Name):-
	(dz_equivalent_types(F,ET);equiv_type(F,ET)),
	param_type_symbol_renaming(T,ET),!,
	T =.. [Name|_].
type_symbol_rename(F,F).




denormalize_args(_,[],[]).
denormalize_args(T,[^(E)|R],Beauty) :-
	get_type_definition_in_cache(T,Defn),
	E =.. [F|Args],
	find_rule_with_functor(Defn,F,Rule),
	Rule =.. [F|RuleArgs],
	denormalize_args_functor(T,F,1,RuleArgs,Args,BeautyArgs),
	denormalize_args(T,R,BeautyR),
	append(BeautyArgs,BeautyR,Beauty).
denormalize_args(T,[E|R],[p('$empty',S)|BeautyR]) :-
	E \== ^(_),
	denormalize_rsize(E,S),
	denormalize_args(T,R,BeautyR).

denormalize_args_functor(_,_,_,[],_,[]).
denormalize_args_functor(T,F,N,[Ty|RRule],Args,BeautyArgs) :-
	dz_equivalent_types(T,Ty), !,
	N1 is N + 1,
	denormalize_args_functor(T,F,N1,RRule,Args,BeautyArgs).
denormalize_args_functor(T,F,N,[_|RRule],[A|RArgs],[p(F/N,BeautyA)|BeautyR]) :-
	denormalize_rsize(A,BeautyA), !,
	N1 is N + 1,
	denormalize_args_functor(T,F,N1,RRule,RArgs,BeautyR).
denormalize_args_functor(T,F,N,[_|RRule],[_|RArgs],BeautyR) :-
	% In some cases beutify fails
	N1 is N + 1,
	denormalize_args_functor(T,F,N1,RRule,RArgs,BeautyR).

choose_denormalize_args([],[]).
choose_denormalize_args([p(_,Only)],[Only]).
choose_denormalize_args([A,B|R],[[A,B|R]]).


