:- module(res_abs_stprf,
	[ res_abs_stprf_call_to_entry/15,
	  res_abs_stprf_exit_to_prime/9,
	  res_abs_stprf_project/5,
	  res_abs_stprf_extend/8,
	  res_abs_stprf_compute_lub/3,
	% res_abs_stprf_compute_clauses_lub/4,
	  res_abs_stprf_widen/3,
	  res_abs_stprf_glb/6,
	  res_abs_stprf_less_or_equal/2,
	  res_abs_stprf_identical_abstract/2,
	  res_abs_stprf_sort/2,
	  res_abs_stprf_call_to_success_fact/14,
	  res_abs_stprf_special_builtin/1,
	  res_abs_stprf_success_builtin/5,
	% res_abs_stprf_call_to_success_builtin/6,
	  res_abs_stprf_input_interface/4,
	  res_abs_stprf_input_user_interface/8,
	  res_abs_stprf_asub_to_native/5,
	  res_abs_stprf_unknown_call/3,
	  res_abs_stprf_unknown_entry/3,
	  res_abs_stprf_empty_entry/3,
	  clean_timers/0,
	  initialize/0,
	  profiling_timer/2,
	  clean_all_caches/0,
	  get_last_replacements/4
	],
	[assertions,regtypes,basicmodes,datafacts]).

:- use_module(library(formulae),[conj_to_list/2]).
%:- use_module(library(aggregates)).
:- use_module(library(lists)).
%:- use_module(library(sets)).
:- use_module(library(sort)).
:- use_module(library(system)).
:- use_module(typeslib(typeslib),
        [ dz_type_included/2, dz_equivalent_types/2, set_numeric_type/1 ]).
:- use_module(library(terms_check)).
:- use_module(library(terms_vars)).
:- use_module(ciaopp(preprocess_flags),[current_pp_flag/2]).
:- use_module(ciaopp(p_unit/program_keys), [predkey_from_sg/2, get_predkey/3]).
:- use_module(spec(s_simpspec), [make_atom/2]).
% For debugging purposes uncomment
%:- use_module(library(messages)).
%:- multifile issue_debug_messages/1.
%:- data issue_debug_messages/1.
%issue_debug_messages(res_abs_stprf).
:- use_module(library(messages), [warning_message/2,debug_message/2,error_message/2]).

:- use_module(res_abs_stprf_aux).
:- use_module(res_abs_utils).
%:- use_module(domain(resources/eq_solver/eq_solver_graphs)).
:- use_module(domain(resources/eq_solver/eq_solver)).
%:- use_module(domain(resources/eq_solver/eq_solver_mathematica)). %% MK 06-11-14: The predicate now is defined in eq_solver (expand).
:- use_module(domain(resources/eq_solver/pred_solver)).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).

:- use_module(domain(etermsvar), [etermsvar_project/3,etermsvar_call_to_success_builtin/6]).

initialize:-
	res_abs_stprf_aux:initialize_cost_centers.


% An abstract substitution is an 7-tuple:
% - Name is the name of the predicate (needed later to check for recursion)
% - Args is a tuple (IArgs,OArgs) where the input and output variables are specified
% - Clauses is a list of extended equations (as found in module pred_solve)
%   For a clause, it will be a singleton list, after the LUB is done, it is a list
% - Solution is the solution to the set of equations in Clauses, if already computed
% - AlreadySolved is a cache of the previous solutions to inner predicates
% - SizeMapping is a cache of the size types for each variable in the predicate
% - PrevCalled is used to implement widening: it saved information about
%   the calls in inner and the same predicates
% FOR NUMBER OF SOLUTIONS: #solutions is faked as an extra output argument solution((L,U))
% - OutSolVar is a pair of variables used to get the number of solutions
% - PrevSolVar is used to get the previous extended variable used for solutions
% - UseDet determines whether the set of solutions is deterministic
% - OutResVars is a list of resource functors with arguments pairs of variables
% - PrevResVars is used to get the previous extended variable used for resources
% - Tests is a series of arithmetic comparisons
%   NOTE: Currently, Tests are not propagated upwards

asub(res_abs_stprf(SgKey,Name,Args,Clauses,Solution,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	SgKey,Name,Args,Clauses,Solution,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests).

:- data profiling_timer/2.
profiling_timer(solve,0).
profiling_timer(call_to_entry,0).
profiling_timer(extend,0).
profiling_timer(exit_to_prime,0).
profiling_timer(widen,0).
profiling_timer(input_user_interface,0).


clean_timers :-
	retractall_fact(profiling_timer(_,_)),
	asserta_fact(profiling_timer(solve,0)),
	asserta_fact(profiling_timer(call_to_entry,0)),
	asserta_fact(profiling_timer(extend,0)),
	asserta_fact(profiling_timer(exit_to_prime,0)),
	asserta_fact(profiling_timer(widen,0)),
	asserta_fact(profiling_timer(input_user_interface,0)).

update_timer(T1,N) :-
	time(T2),
	current_fact(profiling_timer(N,Prev)),
	retract_fact(profiling_timer(N,Prev)),
	New is Prev + T2 - T1,
	asserta_fact(profiling_timer(N,New)).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PREDICATES FOR SETTING UP CLAUSES TESTS
:- pred set_clauses_tests(Clauses,SizeMapping,Paths,NewClauses) #
	"Given a list of clauses in @var{Clauses} and a size mapping in @var{SizeMapping},
	returns the list of clauses, with the tests of each clause in terms of lower and upper bound
         variables.".
set_clauses_tests([],_,_,[]).
set_clauses_tests([C|T],M,P,[NC|TT]):-
	set_clause_test(C,M,P,NC),!,
	set_clauses_tests(T,M,P,TT).
set_clause_test(clause(K,I,O,B,T),M,P,clause(K,I,O,B,TT)):-
	filter_app_tests(T,M,TT0),
	replace_head_variables(TT0,P,TT1),
	replace_variables_in_tests(TT1,M,TT).


replace_head_variables([],_,[]).
replace_head_variables([X|Y],P,[XX|YY]):-
	replace_head_variables_(X,P,XX),
	replace_head_variables(Y,P,YY).

replace_head_variables_(X,P,Y):-
	var(X),
	get_goal_var(X,P,Y),!.
replace_head_variables_(X,_P,X):-
	var(X).
replace_head_variables_(X,P,Y):-
	nonvar(X),
	X =.. [Op|Args],
	replace_head_variables(Args,P,NArgs),
	Y =.. [Op|NArgs].
% replace_head_variables_args([],_,[]).
% replace_head_variables_args([X|Y],P,[XX|YY]):-
% 	replace_head_variables

get_goal_var(X,[A1=Y|_R],Y):-
	X == A1,!.
get_goal_var(X,[_T|TT],Y):-
	get_goal_var(X,TT,Y),!.


filter_app_tests([],_,[]).
filter_app_tests([T|TT],M,TTR):-
	\+ is_input_test(T,M),!,
	filter_app_tests(TT,M,TTR).
filter_app_tests([T|TT],M,[T|TTR]):-
	filter_app_tests(TT,M,TTR).


is_input_test(T,M):-
	var(T),
	get_var_size_type(M,T,_ST).
is_input_test(T,_M):-
	var(T),!,fail.
is_input_test(T,M):-
	nonvar(T),
	T =.. [_Op|ListT],
	is_input_test_args(ListT,M).

is_input_test_args([],_).
is_input_test_args([R|T],M):-
	is_input_test(R,M),
	is_input_test_args(T,M).

	
replace_variables_in_tests([],_,[]).
replace_variables_in_tests([Test|T],M,TT):-
	generate_bound_tests(Test,M,LTest,lb),
	generate_bound_tests(Test,M,UTest,ub),
	replace_variables_in_tests(T,M,TT0),
	(LTest == UTest -> TT = [LTest|TT0];TT=[LTest,UTest|TT0]).
	 
generate_bound_tests(Test,M,LTest,B):-
	var(Test),
	get_var_size_type(M,Test,STest),
	% STest =.. [_|(LTest,_U)].
	STest =.. TTTest,
	TTTest = [_|[Tuple]],
	choose_bound_var(Tuple,B,LTest).
generate_bound_tests(Test,_M,Test,_B):-
	num(Test).
generate_bound_tests(Test,M,LTest,B):-
	nonvar(Test),
	Test =.. [Op|ListT],
	generate_bound_tests_args(ListT,M,LListT,B),
	LTest =.. [Op|LListT].
generate_bound_tests(Test,_M,Test,_).

choose_bound_var((L,_),lb,L):-!.
choose_bound_var((_,U),ub,U):-!.

	
generate_bound_tests_args([],_,[],_).
generate_bound_tests_args([V|T],M,[VV|TT],B):-
	generate_bound_tests(V,M,VV,B),
	generate_bound_tests_args(T,M,TT,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% END OF PREDICATES FOR SETTING UP CLAUSES TESTS




%-----------------------------%
% solve_asub(+,-)
% solve_asub(ASub,SolvedASub)
%-----------------------------%
% Try to get the closed form of the equations in the abstract substitution
solve_asub(ASub,TrySolve,SolvedASub) :-
	time(T1),
	asub(ASub,SgKey,F/N,(IArgs,OArgs),Clauses,_,Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	( Clauses = [] % Empty set of clauses
	-> SolvedASub = ASub
	; % Get the sizes
	  get_size_mappings(IArgs,SizeMapping,ISizes),
	  get_size_mappings(OArgs,SizeMapping,OSizes),
	  % Add extra argument for solution
	  append(OSizes,[solution(OutSolVar)|OutResVars],OSizesSolRes),
	  % Put everything in a predicate
%	(member(resolution(F/N,ISizes,OSizesSolRes,TheSolution),Already)
%	-> asub(SolvedASub,F/N,(IArgs,OArgs),Clauses,TheSolution,Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests) ;

	  %set_clauses_tests(Clauses0,SizeMapping,Clauses),
	  Pred = predicate(F/N,ISizes,OSizesSolRes,Clauses),
	  % Get number of equations to solve as resources
	  length(OutResVars,NbRes),
	  ( IsDet = true
	  -> NbResSol = NbRes
	   ; NbResSol is NbRes + 1 % When it is non deterministic
	  ),
	  % Solve the equations
	  debug_message("CALLING PRED_SOLVER>>>~q ::: ~q",[Pred,SizeMapping]),
	  pred_solve(Pred,Already,NbResSol,IsDet,IsFail,TrySolve,SolutionPre),
	  %debug_message("CALLING PRED_SOLVER AGAIN",[]),
	  %Solution = SolutionPre,
	  pred_solve(Pred,[SolutionPre|Already],NbResSol,IsDet,IsFail,TrySolve,Solution),
	  !,
	  % Take the solution
	  asub(SolvedASub,SgKey,F/N,(IArgs,OArgs),Clauses,Solution,Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	  % Include this as last replacement of that type
	  put_asub_last_replacement(SolvedASub,F/N)
	),
	update_timer(T1,solve).


% Extend each of the clause equations with new instructions at the end
extend_instructions([],_,[]).
extend_instructions([clause(K,ISizes,OSizes,Instr,T)|R],New,[clause(K,ISizes,OSizes,Ext,T)|ExtR]) :-
	append(Instr,New,Ext),
	extend_instructions(R,New,ExtR).

get_size_mappings([],_,[]).
get_size_mappings([E|R],Mappings,[SE|SR]) :-
	get_var_size_type(Mappings,E,SE),
	get_size_mappings(R,Mappings,SR).

%--------------------------------------------------------------------------------------------------%
% res_abs_stprf_call_to_entry(+,+,+,+,+,+,-,-,+)
% res_abs_stprf_call_to_entry(Sv,Sg,Hv,Head,Fv,Proj,Entry,ExtraInfo,CallTypes,PrimeTypes,PrimeDet,PrimeNf)
%--------------------------------------------------------------------------------------------------%

% Important!! This only works on destructuring predicates!!
res_abs_stprf_call_to_entry(_Sv,Sg,Hv,Head,K,_Fv,Proj,Entry,_Extra,CallTypes,EntryTypes,PrimeTypes,PrimeD,PrimeN,Tests):-
	time(T1),
	predkey_from_sg(Sg,SgKey),
	% 1. Find which arguments are input and which output
	( var(Proj) -> (PSizeMap = [],PrevResVars=[]) ; asub(Proj,_,_,_,_,_,_,PSizeMap,_,_,_,_,_,PrevResVars,_,_) ),

	% TODO: take out accumulating parameters
	term_variables(Sg,SubgoalVars),
	choose_input_output(SubgoalVars,CallTypes,IArgs,OArgs),
	choose_input_output(Hv,EntryTypes,HIArgs,_HOArgs),
	% 2. Generate the size types
	generate_initial_size_mappings(IArgs,CallTypes,OArgs,PrimeTypes,HIArgs,EntryTypes,ISizes,OSizes,HSizes,PSizeMap,SizeMapping),
	% 3. Generate deconstruction for this clause
	paths(Sg,Head,Paths),
	get_input_arguments(IArgs,CallTypes,Paths,ISizes,RealIArgs),
	( unify_paths(HIArgs,HSizes,ISizes,CallTypes,Paths,SizeUnifAssigns)
	; SizeUnifAssigns = [] % When called from unknown
	),
	% 4. Get the name of the predicate
        functor(Sg,Name,_), % Now we generate a new random name
	get_next_name(Name,FullName),
	% 5. Get output arguments and add extra for number of solutions and resources
	get_second_components(OSizes,RealOArgs),
	OutSolVar = (_,_),
	generate_resource_variables(SgKey,PrevResVars,OutResVars,ZeroResVars),
	generate_zero_assigns(ZeroResVars,ZeroAssigns),
	generate_resource_nil_variables(SgKey,PrevResVars,NilPrevResVars), !,
	append(RealOArgs,[solution(OutSolVar)|OutResVars],RealOArgsSolRes),
	% 6. Find if the predicate is deterministic or may fail
	PrimeD = det(_,_Covered,DFails), % Covered and Fails as in detabs:internal_value_to_native
	% ( ( Covered = mutex /* = mut_exclusive */ ; DFails = is_det /* = is_det */ )
	( (DFails = is_det /* = is_det */ )
	-> IsDet = true
	 ; IsDet = false
	),
	PrimeN = nf(_,_,Fails),
	( ( Fails = fails ; Fails = possibly_fails )
	-> IsFail = true
	 ; IsFail = false
	),
	append(ZeroAssigns,SizeUnifAssigns,AssignsZS),
	% 7. Initialize the clause
	set_clauses_tests([clause(K,RealIArgs,RealOArgsSolRes,AssignsZS,Tests)],SizeMapping,Paths,[Clause]),
	%Clause = clause(K,RealIArgs,RealOArgsSolRes,SizeUnifAssigns,Tests),
	debug_message("~q(~q)(~q)",[Clause,K,Tests]),
	% 8. Create abstract substitution and extra information
	!,
	asub(Entry,SgKey,FullName,(IArgs,OArgs),[Clause],'$unknown',[],SizeMapping,OutSolVar,(nil,nil),IsDet,IsFail,OutResVars,NilPrevResVars,false,[]),
	update_timer(T1,call_to_entry).

% A parameter is chosen as output if it has type vr
% If not, it is chosen as input
choose_input_output([],_,[],[]).
choose_input_output([Sv|Svars],Types,IArgs,OArgs) :-
	% Choose the rest
	choose_input_output(Svars,Types,IRest,ORest),
	% Find this type
	find_variable_type(Types,Sv,SvType),
	( dz_equivalent_types(SvType,vr)
	-> IArgs = IRest, OArgs = [Sv|ORest]
	 ; IArgs = [Sv|IRest], OArgs = ORest
	).

% Generate the initial mappings for the subgoal and head variables
generate_initial_size_mappings(IArgs,CallTypes,OArgs,PrimeTypes,HIArgs,EntryTypes,ISizes,OSizes,HSizes,PSizeMap,SizeMapping) :-
	get_vars_size_types(IArgs,CallTypes,ISizes0),
	get_vars_size_types(OArgs,PrimeTypes,OSizes0),
	get_vars_size_types(HIArgs,EntryTypes,HSizes0),
	% Merge the sizes
	merge_size_types(ISizes0,PSizeMap,PSizeMap2),
	merge_size_types(OSizes0,PSizeMap2,PSizeMap3),
	merge_size_types(HSizes0,PSizeMap3,SizeMapping),
	% Get the real ones
	get_size_mappings(IArgs,SizeMapping,ISizes1),
	zip(IArgs,ISizes1,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes1),
	zip(OArgs,OSizes1,OSizes),
	get_size_mappings(HIArgs,SizeMapping,HSizes1),
	zip(HIArgs,HSizes1,HSizes).

% Generate the input arguments for the clause
% They will be variables except for the case when a variable
% is unified with a ground argument, in which case a constant
% will be included as input
get_input_arguments([],_,_,_,[]).
get_input_arguments([IArg|IRest],Types,Paths,ISizes,[CArg|CRest]) :-
	get_var_size_type(ISizes,IArg,IArgSize),
	( find_ground_unification(Paths,IArg,GroundExpr)
	-> find_variable_type(Types,IArg,IType),
	   create_ground_size(GroundExpr,IType,CArg)
	 ; CArg = IArgSize
	),
	get_input_arguments(IRest,Types,Paths,ISizes,CRest).

find_ground_unification([X=Expr|_],Y,Expr) :-
	X == Y, !, ground(Expr).
find_ground_unification([_|R],Y,Expr) :-
	find_ground_unification(R,Y,Expr).

% Generates the equations needed to go from the variables in
% the goal to the variables in the head of the clause
unify_paths([],_,_,_,_,[]).

unify_paths([HArg|HRest],HISizes,ISizes,Types,Paths,[Assign|AssR]) :-
	% Get the variable in the subgoal for which the unification is done
	find_head_path(Paths,HArg,HPath),
	initial_var(HPath,SgVar),
	get_var_size_type(ISizes,SgVar,SgSize), !,
	find_variable_type(Types,SgVar,SgType),
	% Compute the path size type
	unify_path(SgType,SgSize,HPath,Out),
	% Create the call
	get_var_size_type(HISizes,HArg,HSize),
	Assign = assign(HSize,Out),
	% Do the rest
	unify_paths(HRest,HISizes,ISizes,Types,Paths,AssR).
unify_paths([HArg|HRest],HISizes,ISizes,Types,Paths,AssR) :-
	\+ find_head_path(Paths,HArg,_),
	unify_paths(HRest,HISizes,ISizes,Types,Paths,AssR).

find_head_path([X=Y|_],W,Y) :- W == X.
find_head_path([Y=X|_],W,Y) :- W == X.
find_head_path([X=Path|_],W,Path) :- W == X.
find_head_path([_|R],W,Path) :- find_head_path(R,W,Path).

%-------------------------------------------------------------------------------%
% res_abs_stprf_exit_to_prime(+,+,+,+,+,-,-,+,+)
% res_abs_stprf_exit_to_prime(Sg,Hv,Head,Sv,Exit,ExtraInfo,Prime,ExitTypes,PrimeTypes)
%-------------------------------------------------------------------------------%
% :- use_module(ciaopp(plai/static_profiling_info_tracker)).
res_abs_stprf_exit_to_prime(Sg,_Hv,Head,_Sv,Exit,_Extra,Prime,ExitTypes,PrimeTypes):-
	time(T1),
	asub(Exit,SgKey,Name,(IArgs,OArgs),Clauses,ByNowSol,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	% 1. Generate backwards unification
	get_unification_eqs(Sg,Head,Eqs),
	unify_back_s(OArgs,Eqs,SizeMapping,PrimeTypes,ExitTypes,AssignOut),
	extend_instructions(Clauses,AssignOut,Clauses2),
	% 2. Generate final solution and resources
	generate_final_solution_expr(PrevSolVar,IsDet,IsFail,OutSolVar,SolutionExpr),
%	get_current_cc(XX),
%	error_message("Exit to Prime: ~q",[SgKey]),
	generate_final_resources_exprs(PrevResVars,Sg,IsFail,OutResVars,ResExprs),
	extend_instructions(Clauses2,[SolutionExpr|ResExprs],QuasiFinalClauses),
	% 3. Take tests into account for constraining domain
	filter_unif_tests(Eqs,EqTests),
	append(EqTests,Tests,AllTests),
	obtain_tests(IArgs,AllTests,ETests),
	%FinalClauses = QuasiFinalClauses,
	( ETests = []
	-> FinalClauses = QuasiFinalClauses
	 ; explode_final_clauses_with_tests(PrimeTypes,SizeMapping,ETests,QuasiFinalClauses,FinalClauses)
	),
	%FinalClauses = QuasiFinalClauses,
	% PrimeD = det(_,_,det(_,_,DetAn)),
	% (DetAn=is_det -> TIsDet=true; TIsDet=false),

	% 4. Construct the final substitution
	asub(PrimeNotSolved,SgKey,Name,(IArgs,OArgs),FinalClauses,ByNowSol,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,AllTests),
%	Prime = PrimeNotSolved,
	update_timer(T1,exit_to_prime),
	
	% 5. Try to solve the equation
%	Prime = PrimeNotSolved.
	get_solver_behavior(TrySolve),
	solve_asub(PrimeNotSolved,TrySolve,Prime).

% Generates the final equations to move from the clause output variables
% to the output variables in the goal
unify_back_s([],_,_,_,_,[]).
unify_back_s([Arg|R],Eqs,SizeMapping,PrimeTypes,ExitTypes,[AssArg|AssR]) :-
	find_head_path(Eqs,Arg,Components),
	find_variable_type(PrimeTypes,Arg,ArgT),
	unify_back(ArgT,Components,SizeMapping,ExitTypes,AssSize),
	get_var_size_type(SizeMapping,Arg,ISize),
	AssArg = assign(ISize,AssSize),
	unify_back_s(R,Eqs,SizeMapping,PrimeTypes,ExitTypes,AssR).

filter_unif_tests([],[]).
filter_unif_tests([X=Y|R],[X =:= Y|S]) :-
	var(X), var(Y),
	!, filter_unif_tests(R,S).
filter_unif_tests([_|R],S) :-
	!, filter_unif_tests(R,S).

obtain_tests(Hv,ArithTests,Tests) :-
	expand_tests(ArithTests,ExpTests),
	find_wanted_tests(Hv,ExpTests,Tests).

explode_final_clauses_with_tests(_,_,[],_,[]).
explode_final_clauses_with_tests(Ty,Map,[E =:= G|R],Clauses,FinalClauses) :-
	explode_final_clauses_with_tests_(Ty,Map,E,G,Clauses,Clauses1),
	explode_final_clauses_with_tests(Ty,Map,R,Clauses,Clauses2),
	append(Clauses1,Clauses2,FinalClauses).
explode_final_clauses_with_tests(Ty,Map,[E =< G|R],Clauses,FinalClauses) :-
	range(E,G,S), append(S,R,T), % Convert into a list of equalities
	explode_final_clauses_with_tests(Ty,Map,T,Clauses,FinalClauses).
explode_final_clauses_with_tests(Ty,Map,[E < G|R],Clauses,FinalClauses) :-
	G1 is G-1, range(E,G1,S), append(S,R,T), % Convert into a list of equalities
	explode_final_clauses_with_tests(Ty,Map,T,Clauses,FinalClauses).


% :- export(range/3).
% range(E,N,[E =:= N]):-
% 	N<0,!.
% range(E,N,R):-
% 	range_(E,N,R).

% :- export(range_/3).
%range(E,N,[E =:= N])          :- N < 0, !.
range(_,N,[])          :- N < 0, !.
range(E,N,[E =:= N|R]) :- N >= 0, !, N1 is N - 1, range(E,N1,R).

explode_final_clauses_with_tests_(_,_,_,_,[],[]).
explode_final_clauses_with_tests_(Ty,Map,E,G,[clause(K,IArgs,OArgs,Body,TT)|R],[clause(K,NewIArgs,OArgs,NewBody,TT)|S]) :-
	get_var_size_type(Map,E,ESize),
	find_variable_type(Ty,E,EType),
	create_ground_size(G,EType,GSize),
	% Change in input args
	explode_input_args(IArgs,ESize,GSize,NewIArgs),
	% TODO: Change in body
	NewBody = Body,
	explode_final_clauses_with_tests_(Ty,Map,E,G,R,S).

explode_input_args([],_,_,[]).
explode_input_args([E|R],ESize,GSize,[GSize|S]) :-
	E == ESize, !,
	explode_input_args(R,ESize,GSize,S).
explode_input_args([E|R],ESize,GSize,[E|S]) :-
	E \== ESize, !,
	explode_input_args(R,ESize,GSize,S).

%------------------------------------------------------------------------%
% res_abs_stprf_project(+,+,-,+)
% res_abs_stprf_project(ASub,Vars,Proj,ProjTypes)
%------------------------------------------------------------------------%

res_abs_stprf_project(ASub,Sg,_Vars,Proj,ProjTypes) :-
	% 1. Generate name
	asub(ASub,SgKey,_,_,_,_,_,_,_,_,_,_,_,Prevs,_,_),
	functor(Sg,Name,_),
	get_next_name(Name,FullName), !,
	% 2. Choose input and output
	term_variables(Sg,Vars),
	choose_input_output(Vars,ProjTypes,IArgs,OArgs),
	% 3. Generate initial one
	generate_resource_variables(SgKey,Prevs,OutResVars,_),
	generate_resource_nil_variables(SgKey,Prevs,PrevResVars),
	!,
	asub(Proj,SgKey,FullName,(IArgs,OArgs),[],'$unknown',[],[],(_,_),(nil,nil),false,true,OutResVars,PrevResVars,false,[]).

%% project_arguments([],_Vars,[]).
%% project_arguments([V|R],Vars,[V|ProjR]) :-
%% 	varmember(V,Vars),
%% 	!, project_arguments(R,Vars,ProjR).
%% project_arguments([_|R],Vars,ProjR) :-
%% 	project_arguments(R,Vars,ProjR).
%% 
%% project_size_mapping([],_Vars,[]).
%% project_size_mapping([(V,S)|R],Vars,[(V,S)|ProjR]) :-
%% 	varmember(V,Vars),
%% 	!, project_size_mapping(R,Vars,ProjR).
%% project_size_mapping([_|R],Vars,ProjR) :-
%% 	project_size_mapping(R,Vars,ProjR).

%------------------------------------------------------------------------%
% res_abs_stprf_extend(+,+,+,-,+,+)
% res_abs_stprf_extend(Prime,Sv,Call,Succ,CallTypes,SuccTypes,NfInfo)
%------------------------------------------------------------------------%

res_abs_stprf_extend(Prime,Sg,Sv,Call,Succ,CallTypes,SuccTypes,PrimeNfInfo):-
	time(T1),
	
	% Extract information from abstract substitutions
	asub(Call,SgKey,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailedC,Tests),
	asub(Prime,_,_,(PIArgs,POArgs),_,PrimeSol,AlreadySolvedP,_,_,_,_,IsFailP,_,_,_,_),
	% 1. Extend size mapping
	get_vars_size_types(Sv,SuccTypes,SvSizes),
	merge_size_types(SvSizes,SizeMappingC,NewMapping),
	% 2. Generate a temporary variable for holding the number of solutions
	TempSolVar = (T_L,T_U),
	generate_resource_variables(SgKey,PrevResVars,TempResVars,_),
	% 2. Generate fake resolution if not available
	( PrimeSol = resolution(ResName,_,_,_)
	-> ResolutionToUse = PrimeSol, ResIArgs = PIArgs, ResOArgs = POArgs, IsFailToUse = IsFailP
	 ; debug_message("Fake resolution for ~q ~q",[Sg,CallTypes]),
	   generate_fake_resolution(Sg,CallTypes,NewMapping,ResIArgs,ResOArgs,TempSolVar,TempResVars,ResolutionToUse,ResName,PrimeNfInfo,IsFailToUse)
	),
	% 3. Append new call to list of instructions
	get_size_mappings(ResIArgs,NewMapping,SIArgs),
	get_size_mappings(ResOArgs,NewMapping,SOArgs),
	append(SOArgs,[solution(TempSolVar)|TempResVars],SOArgsSolRes),
	extend_instructions(Clauses,[call(ResName,SIArgs,SOArgsSolRes)],Clauses2),
	% 4. Extend the list of instructions for solutions if needed
	or_(YetFailedC,IsFailToUse,NewYetFailed),
	( PrevSolVar == (nil,nil)
	-> % No previous solution
	   NewSolVar = TempSolVar,
	   extend_res_first_case(TempResVars,NewYetFailed,Sg,NewResVars,FirstResClauses),
	   extend_instructions(Clauses2,FirstResClauses,NewClauses)
	 ; NewSolVar = (_,_), PrevSolVar = (P_L,P_U),
	   % Multiply with the previous number of solutions
	   extend_instructions(Clauses2,[assign(solution(NewSolVar),solution((P_L * T_L, P_U * T_U)))],Clauses3),
	   extend_res_usual_case(TempResVars,PrevResVars,NewYetFailed,PrevSolVar,Sg,NewResVars,UsualResClauses),
	   extend_instructions(Clauses3,UsualResClauses,NewClauses)
	),
	% 5. Extend the list of already solved equations and visited nodes
	unifiable_append(AlreadySolvedC,[ResolutionToUse|AlreadySolvedP],NewAlreadySolved),
	%insert(PrevCalledC,ResName,NewPrevCalledAux),
	%ord_union(NewPrevCalledAux,PrevCalledP,NewPrevCalled),
	% 6. Create new abstract substitution
	!,
	asub(Succ,SgKey,Name,Args,NewClauses,ByNowSol,NewAlreadySolved,NewMapping,OutSolVar,NewSolVar,IsDet,IsFail,OutResVars,NewResVars,NewYetFailed,Tests),
	% 7. Conflate the same equations -> now in widening
	% conflate_same_function(SuccNotConf,Succ).

	update_timer(T1,extend).

% Generate a call to a new empty equation when the information is unknown
% This is needed to create recursion
generate_fake_resolution(Sg,CallTypes,SizeTypes,IArgs,OArgs,TempSolVar,TempResVars,Resolution,NewName,PrimeNfInfo,IsFailToUse) :-
	term_variables(Sg,SubgoalVars),
	% Create input and output parameters
	choose_input_output(SubgoalVars,CallTypes,IArgs,OArgs),
	get_size_mappings(IArgs,SizeTypes,SizeI),
	get_size_mappings(OArgs,SizeTypes,SizeO),
	append(SizeO,[solution(TempSolVar)|TempResVars],SizeOSolRes),
	% Get a new name
	functor(Sg,F,_),
	get_next_name(F,NewName), !,
	% Fake-solve it
	Pred = predicate(NewName,SizeI,SizeOSolRes,[clause(_,SizeI,SizeOSolRes,[call(NewName,SizeI,SizeOSolRes)],[])]),
	get_solver_behavior(TrySolve),
	pred_solve(Pred,[],0,true,false,TrySolve,Resolution),
	% Also generate fake information for failure
	PrimeNfInfo = nf(_,_,Fails),
	( ( Fails = fails ; Fails = possibly_fails )
	-> IsFailToUse = true
	 ; IsFailToUse = false
	).

% Merge the size mapping of a clause with the success of a literal
merge_size_types([],V,V).
merge_size_types([(V,VSize)|R],Already,Merged) :-
	( get_var_size_type(Already,V,_)
	-> merge_size_types(R,Already,Merged)
	 ; merge_size_types(R,[(V,VSize)|Already],Merged)
	).

%------------------------------------------------------------------------%
% res_abs_stprf_compute_lub(+,-,+)
% res_abs_stprf_compute_lub(ListASub,Lub,LubTypes)
%------------------------------------------------------------------------%

res_abs_stprf_compute_lub([],'$bottom',_) :- !.
res_abs_stprf_compute_lub(ListASub,Lub,_LubTypes) :-
	time(T1),
	
	accumulate_asubs(ListASub,Accum),
	% Conflate into same function
	apply_replacements(Accum,LubNotSolved),
	% solve

	update_timer(T1,widen),
	get_solver_behavior(TrySolve),
	!, solve_asub(LubNotSolved,TrySolve,Lub).

get_solver_behavior(just_equations):-
	current_pp_flag(output_resources,OUTPUT),(OUTPUT=equations;OUTPUT=both),!.
get_solver_behavior(solve).
 

%------------------------------------------------------------------------%
% res_abs_stprf_widen(+,+,-)
% res_abs_stprf_widen(ASub0,ASub1,Widen)
%------------------------------------------------------------------------%

res_abs_stprf_widen(ASub0,ASub1,Widen) :-
	res_abs_stprf_compute_lub([ASub0,ASub1],Widen,_).

%------------------------------------------------------------------------%
% res_abs_stprf_glb(+,+,-)
% res_abs_stprf_glb(ASub0,ASub1,Glb)
%------------------------------------------------------------------------%

res_abs_stprf_glb(ASub0,ASub1,_,_,_,ASub0) :-
	asub(ASub1,_,_,_,[],_,_,_,_,_,_,_,_,_,_,_), !.
res_abs_stprf_glb(ASub0,ASub1,_,_,_,ASub1) :-
	asub(ASub0,_,_,_,[],_,_,_,_,_,_,_,_,_,_,_), !.
res_abs_stprf_glb(ASub0,ASub1,GlbTypes,GlbDet,GlbNf,Glb) :-
	asub(ASub0,SgKey,Name0/_,(I0,O0),Clauses0,R1,_,SMap0,_,_,_,_,OutRes0,PrevRes0,_,_),
	asub(ASub1,_,_Name1/_,(_I1,_O1),Clauses1,R2,_,_,_,_,_,_,_,_,_,_),
	%error_message("R1 >>> ~q",[R1]),
	%error_message("R2 >>> ~q",[R2]),
	Clauses0 \= [], Clauses1 \= [],
	% Find out about det and nf
	GlbDet = det(_,_Covered,DFails), % Covered and Fails as in detabs:internal_value_to_native
	% ( ( Covered = mutex /* = mut_exclusive */ ; DFails = is_det /* = is_det */ )
	( ( DFails = is_det /* = is_det */ )
	-> IsDet = true
	 ; IsDet = false
	),
	GlbNf = nf(_,_,Fails),
	( ( Fails = fails ; Fails = possibly_fails )
	-> IsFail = true
	 ; IsFail = false
	),
	% Check if set of arguments are equal
	( Clauses0 = [clause(_K,ClI0,ClO0,_,_)|_], Clauses1 = [clause(_,ClI1,ClO1,_,_)|_], variant((ClI0,ClO0),(ClI1,ClO1))
	-> % We have to accumulate clauses, deleting the empty ones
	((is_res(R1),
	  is_res(R2),
	  info_type(Clauses0,T1),
	  info_type(Clauses1,T2),
	  compute_resolution_glb(T1,T2,R1,R2,RGlb)
	 ) -> generate_new_clause(Clauses0,RGlb,Clauses); 
	    (
		RGlb = '$unknown',
		up_to_renaming_append(Clauses0,Clauses1,ClausesB),
		delete_empty_clauses(ClausesB,Clauses)
	    )),
	   % Generate new substitution
	   get_next_name(Name0,NewName), !,

	   asub(Glb0,SgKey,NewName,(I0,O0),Clauses,RGlb,[],SMap0,(_,_),(nil,nil),IsDet,IsFail,OutRes0,PrevRes0,false,[]),
	   ( get_solver_behavior(TrySolve),solve_asub(Glb0,TrySolve,Glb) -> true ; Glb = Glb0 )
	   % Glb = GlbPre
	 ; % Generate an empty entry
	   append(I0,O0,Vars0),
	   res_abs_stprf_empty_entry(Vars0,GlbTypes,Empty),
	   % Generate final substitution
	   asub(Empty,_,_,Args,Clauses,Solution,Already,SMap,OutSol,PrevSol,_,_,OutRes,PrevRes,YetFailed,_),
	   get_next_name(Name0,NewName), !,
	   asub(Glb,SgKey,NewName,Args,Clauses,Solution,Already,SMap,OutSol,PrevSol,IsDet,IsFail,OutRes,PrevRes,YetFailed,[])
	).


is_res(resolution(_,_,_,_)).

info_type([clause(Trust1,_,_,_,_)],trust):-
	Trust1 == trust,!.
info_type([clause(_Trust1,_,_,_,_)],notrust).

% :- export(is_trust_info/2).
% is_trust_info([clause(Trust1,_,_,_,_)],[clause(Trust2,_,_,_,_)]):-
% 	Trust1 == trust, Trust2 == trust.
% 	%make_atom([_,_,trust],Name),!.
% is_trust_info([clause(Trust1,_,_,_,_)|T],[clause(Trust2,_,_,_,_)|TT]):-
% 	Trust1 == trust, Trust2 == trust,
% 	is_trust_info(T,TT).

:- export(generate_new_clause/3).
generate_new_clause([],_,[]).
generate_new_clause([clause(N,I,O,B,T)|R],Map,[clause(N,I,O,NB,T)|RR]):-
	generate_new_clause_body(B,Map,NB),!,
	generate_new_clause(R,Map,RR).

:- export(generate_new_clause_body/3).
generate_new_clause_body([],_,[]):-!.
generate_new_clause_body(Tail,'$unknown',Tail):-!.
generate_new_clause_body([assign(L,_R)|Tail],Map,[assign(L,NR)|NTail]):-
	L =.. [Res,(Low,Up)],Map = resolution(_,_,_,MM),
	obtain_closed_form(Low,MM,CLow),
	obtain_closed_form(Up,MM,CUp),
	NR =.. [Res,(CLow,CUp)],
	generate_new_clause_body(Tail,Map,NTail).
generate_new_clause_body([assign(L,R)|Tail],Map,[assign(L,R)|NTail]):-
	generate_new_clause_body(Tail,Map,NTail).

obtain_closed_form(Var,[(V,closed_form(Exp))|_TT],Exp):-
	Var == V,!.
obtain_closed_form(Var,[_|TT],Exp):-!,
	obtain_closed_form(Var,TT,Exp).




% :-export(resolution_glb/3).
% resolution_glb('$unknown',A,A).
% resolution_glb(A,'$unknown',A).
resolution_glb(resolution(Name/N,IArgs1,OSol1,MapRes1),resolution(Name/_,IArgs1,OSol1,MapRes2),resolution(Name/N,IArgs1,OSol1,MapRes)):-
	resolution_glb_map(MapRes1,MapRes2,MapResPre),!,
	simplify_res(MapResPre,MapRes).

compute_resolution_glb(trust,trust,R1,R2,Glb):-!,
	resolution_glb(R1,R2,Glb).
compute_resolution_glb(trust,_notrust,R1,_R2,R1):-!.
compute_resolution_glb(_notrust,trust,_R1,R2,R2):-!.

resolution_glb_map([],[],[]).
resolution_glb_map([(V1,closed_form(Exp1)),(V2,closed_form(Exp2))|R],[(V1,closed_form(Exp3)),(V2,closed_form(Exp4))|T],[(V1,closed_form(max(Exp1,Exp3))),(V2,closed_form(min(Exp2,Exp4)))|TT]):-
	resolution_glb_map(R,T,TT).


:- use_module(domain(resources/eq_solver/solver/solver)).
simplify_res([],[]).
simplify_res([(V,closed_form(Exp))|T],[(V,closed_form(ExpS))|TR]):-
	simplify(Exp,ExpS),
	simplify_res(T,TR).


delete_empty_clauses([],[]).
delete_empty_clauses([clause(_,_,_,[],_)|R],DelR) :-
	!, delete_empty_clauses(R,DelR).
delete_empty_clauses([clause(K,I,O,E,T)|R],[clause(K,I,O,E,T)|DelR]) :-
	E \= [],
	delete_empty_clauses(R,DelR).

%------------------------------------------------------------------------%
% res_abs_stprf_less_or_equal(+,+)
% res_abs_stprf_less_or_equal(ASub0,ASub1)
%------------------------------------------------------------------------%

res_abs_stprf_less_or_equal(ASub0,ASub1) :-
	res_abs_stprf_sort(ASub0,ASub0S),
	res_abs_stprf_sort(ASub1,ASub1S),
	res_abs_stprf_less_or_equal_(ASub0S,ASub1S).

% A <= B <=> they refer to the same predicate and eq_A >= eq_B
res_abs_stprf_less_or_equal_(ASub0,ASub1) :-
	asub(ASub0,_SgKey,Name/_,Args0,Clauses0,Solution0,_,_,_,_,_,_,_,_,_,_),
	asub(ASub1,_SgKey1,Name/_,Args1,Clauses1,Solution1,_,_,_,_,_,_,_,_,_,_),
	% Check arguments
	variant(Args0,Args1),
	% If solutions are the same and none is unknown, we know they are equal
	( variant(Solution0,Solution1),
	  Solution0 \= '$unknown'
	; make_equal_same_functions(Clauses0,ConfClauses0,[],StepL),
	  make_equal_same_functions(Clauses1,ConfClauses1,StepL,_),
	  up_to_renaming_subset_extra(ConfClauses0,Args0,ConfClauses1,Args1)
	).

make_equal_same_functions([],[],L,L).
make_equal_same_functions([clause(K,IArgs,OArgs,Body,T)|R],[clause(K,IArgs,OArgs,UnifBody,T)|UnifR],InList,OutList) :-
	make_equal_same_functions_body(Body,UnifBody,InList,StepList),
	make_equal_same_functions(R,UnifR,StepList,OutList).

make_equal_same_functions_body([],[],L,L).
make_equal_same_functions_body([assign(X,Y)|R],[assign(X,Y)|UnifR],InL,OutL) :-
	make_equal_same_functions_body(R,UnifR,InL,OutL).
make_equal_same_functions_body([call(Pred,I,O)|R],[call(PredVar,I,O)|UnifR],InL,OutL) :-
	( find_map(InL,Pred,Var)
	-> PredVar = Var,
	   StepL = InL
	 ; PredVar = _,
	   StepL = [(Pred,PredVar)|InL]
	),
	make_equal_same_functions_body(R,UnifR,StepL,OutL).

find_map([(A,B)|_],A,B) :- !.
find_map([_|R],A,B) :- find_map(R,A,B).

%------------------------------------------------------------------------%
% res_abs_stprf_identical_abstract(+,+)
% res_abs_stprf_identical_abstract(ASub1,ASub2)
%------------------------------------------------------------------------%

% Define the equality as inclusion in both sides
res_abs_stprf_identical_abstract(ASub0,ASub1):-
	res_abs_stprf_less_or_equal(ASub0,ASub1), !,
	res_abs_stprf_less_or_equal(ASub1,ASub0), !.

%------------------------------------------------------------------------%
% res_abs_stprf_sort(+,-)
% res_abs_stprf_sort(ASub0,ASub1)
%------------------------------------------------------------------------%

res_abs_stprf_sort(ASub0,ASub1) :-
	asub(ASub0,SK,Name,Args,Clauses0,Solution,Already0,SizeMapping0,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	sort(Clauses0,Clauses1),
	sort(Already0,Already1),
	sort(SizeMapping0,SizeMapping1),
	asub(ASub1,SK,Name,Args,Clauses1,Solution,Already1,SizeMapping1,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests).

%------------------------------------------------------------------------%
% res_abs_stprf_call_to_success_fact(+,+,+,+,+,+,-,-,+,+)
% res_abs_stprf_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ,CallTypes,PrimeTypes)
%------------------------------------------------------------------------%

res_abs_stprf_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ,CallT,EntryT,PrimeT,SuccT,PrimeD,PrimeNf):-
	% "Fake" the call by doing the three stages by hand
	res_abs_stprf_call_to_entry(Sv,Sg,Hv,Head,_,[],Proj,Entry,ExtraInfo,CallT,EntryT,PrimeT,PrimeD,PrimeNf,[]),
	res_abs_stprf_exit_to_prime(Sg,Hv,Head,Sv,Entry,ExtraInfo,Prime,EntryT,PrimeT),
	res_abs_stprf_extend(Sg,Prime,Sv,Call,Succ,CallT,SuccT,nf(_,_,not_fails)).

%------------------------------------------------------------------------%
% res_abs_stprf_special_builtin(+)
% res_abs_stprf_special_builtin(SgKey)
%------------------------------------------------------------------------%

res_abs_stprf_special_builtin('is/2').
res_abs_stprf_special_builtin('true/0').
res_abs_stprf_special_builtin('fail/0').
res_abs_stprf_special_builtin('!/0').
res_abs_stprf_special_builtin('arithmetic:>/2').
res_abs_stprf_special_builtin('arithmetic:</2').
res_abs_stprf_special_builtin('arithmetic:=</2').
res_abs_stprf_special_builtin('arithmetic:>=/2').
res_abs_stprf_special_builtin('>/2').
res_abs_stprf_special_builtin('</2').
res_abs_stprf_special_builtin('=</2').
res_abs_stprf_special_builtin('>=/2').
res_abs_stprf_special_builtin('=/2').
res_abs_stprf_special_builtin('\\=/2').
res_abs_stprf_special_builtin('=:=/2').

%-------------------------------------------------------------------------
% res_abs_stprf_success_builtin(+,+,+,+,-)
% res_abs_stprf_success_builtin(Type,Sg,Call,CTypes,Succ)
%-------------------------------------------------------------------------


% TODO: REVISE THE MODIFICATION OF MKL 2014-11-13.
%%%%%%%%%%%%%%%%%%
%BASICALLY I CHANGE THE YETFAILED VALUE
%OF SOME BUILTINS, KEEPING THE SAME
%VALUE THAT THE ANALYSIS CARRIES, EVEN
%IF THAT BUILTIN COULD EVENTUALLY FAILS.
% UPDATED MKL 2015-08-25: I Rolled back this, because it needs revision. 
%%%%%%%%%%%%%%%%%%

% If we have true, a cut or a comparison, we get no new information
res_abs_stprf_success_builtin(_,true,Call,_,Call).
res_abs_stprf_success_builtin(_,!,Call,_,Call).
res_abs_stprf_success_builtin(_,fail,Call,_,Succ) :-
	% fail always fails
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,_,Tests),
	asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,true,Tests).
res_abs_stprf_success_builtin(_,A \= B,Call,_,Succ) :-
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests),
	ground(A), ground(B), !,
	( A \= B
	-> asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests)
	 ; % It will always fail, so don't add it as an equation
	   asub(Succ,SK,Name,Args,[],ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,true,Tests)
	).
res_abs_stprf_success_builtin(_,A =:= B,Call,_,Succ) :-
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests),
	ground(A), ground(B), !,
	( A =:= B
	-> asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests)
	 ; % It will always fail, so don't add it as an equation
	   asub(Succ,SK,Name,Args,[],ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,true,Tests)
	).
res_abs_stprf_success_builtin(_,A = B,Call,_,Succ) :-
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests),
	ground(A), ground(B), !,
	( A = B
	-> asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,Fail,Tests)
	 ; % It will always fail, so don't add it as an equation
	   asub(Succ,SK,Name,Args,[],ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,true,Tests)
	).

%%%MODIFY false in place of true. Wrong.
res_abs_stprf_success_builtin(_,_ \= _,Call,_,Succ) :-
	% \= may fail
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,FailI,Tests),
	% asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,FailI,Tests).
	asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,FailI,Tests).
	
% If we have is/2, we have to include a new assignment in the equation,
% which will be similar to the one in the right-hand side of is
res_abs_stprf_success_builtin(Sv,A =:= B,Call,CTypes,Succ) :-
	res_abs_stprf_success_builtin(Sv,A is B,Call,CTypes,Succ).
res_abs_stprf_success_builtin(Sv,Out is In,Call,CTypes,Succ):-
	ground(Out),
	find_variable_type(CTypes,In,T),
	dz_type_included(T,num),
	!,
	res_abs_stprf_success_builtin(Sv,In=Out,Call,CTypes,Succ).
res_abs_stprf_success_builtin(Sv,In is Out,Call,CTypes,Succ):-
	ground(Out),
	find_variable_type(CTypes,In,T),
	dz_type_included(T,num),
	!,
	res_abs_stprf_success_builtin(Sv,In=Out,Call,CTypes,Succ).
res_abs_stprf_success_builtin(_,Out is In,Call,CTypes,Succ) :-
	find_variable_type(CTypes,Out,T),
	dz_type_included(T,num), !, % It's already a number
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	( ( var(In) ; ground(In) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,NewTests).
res_abs_stprf_success_builtin(_,Out is In,Call,CTypes,Succ):-
	find_variable_type(CTypes,Out,T),
	dz_type_included(T,vr), % It's already a variable
	!, asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	% 1. Extend the size mappings with Out/num
	set_numeric_type(Num),
	get_size_type(Num,NumSizeType),
	merge_size_types([(Out,NumSizeType)],SizeMappingC,NewMapping),
	% 2. Add new instruction
	create_num_assignment(In,NewMapping,AssignL,AssignR),
	extend_instructions(Clauses,[assign(NumSizeType,num((AssignL,AssignR)))],NewClauses),
	% 3. Add test if possible
	( ( var(In) ; ground(In) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	% Create new substitution
	% Note: calling is/2 does not generate new solutions
	asub(Succ,SK,Name,Args,NewClauses,ByNowSol,AlreadySolvedC,NewMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,NewTests).
res_abs_stprf_success_builtin(Sv,Out=In,Call,CTypes,Succ):-
	% Handle it similarly to exit_to_prime
	asub(Call,SK,Name,(IArgs,OArgs),Clauses,ByNowSol,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	% 1. Get success types and extend mapping
	etermsvar_project(Sv,CTypes,PTypes),
	etermsvar_call_to_success_builtin('=/2',Out=In,Sv,CTypes,PTypes,STypes),
	find_variable_type(STypes,Out,T),
	get_size_type(T,SizeType),
	merge_size_types([(Out,SizeType)],SizeMapping,NewMapping),
	% 2. Generate backwards unification
	unify_back_s([Out],[Out=In],NewMapping,STypes,STypes,AssignOut),
	extend_instructions(Clauses,AssignOut,Clauses2),
	% 3. Add test if possible
	( ( var(Out) ; ( ground(Out), number(Out)) ) , ( var(In) ; ( ground(In), number(In)) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	% 3. Construct the final substitution
	asub(Succ,SK,Name,(IArgs,OArgs),Clauses2,ByNowSol,AlreadySolved,NewMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,NewTests).
%%MODIFIED. Instead of putting that fails we are keeping the previous value. Maybe is
% wrong.
res_abs_stprf_success_builtin(_,Sg,Call,_,Succ) :-
	Sg =.. [F,Out,In],
	( F = 'arithmetic:<' ; F = 'arithmetic:>'
	; F = 'arithmetic:=<' ; F = 'arithmetic:>='
	; F = '<' ; F = '>' ; F = '=<' ; F = '>=' ), !,
	% A comparison predicate may fail
	asub(Call,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,FailI,Tests),
	% Add test if possible
	( ( var(Out) ; ( ground(Out), number(Out)) ) , ( var(In) ; ( ground(In), number(In)) )
	-> NewTests = [Sg|Tests]
	 ; NewTests = Tests
	),
	asub(Succ,SK,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,FailI,NewTests).
	% asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,true,NewTests).


%------------------------------------------------------------------------%
% res_abs_stprf_input_interface(+,+,+,-)
% res_abs_stprf_input_interface(InputUser,Kind,StructI,StructO)
%------------------------------------------------------------------------%

% We don't support things right now
res_abs_stprf_input_interface(rsize(Arg,Size),perfect,(Sizes0,Sols,Res),(Sizes1,Sols,Res)) :-
	put_in_front(rsize(Arg,Size),Sizes0,Sizes1).
res_abs_stprf_input_interface(cardinality(Qv,L,U),perfect,(Sizes,Sols0,Res),(Sizes,Sols1,Res)) :-
	put_in_front(cardinality(Qv,L,U),Sols0,Sols1).
res_abs_stprf_input_interface(costb(Qv,R,L,U),perfect,(Sizes,Sols,Res0),(Sizes,Sols,Res1)) :-
	put_in_front(costb(Qv,R,L,U),Res0,Res1).

put_in_front(X,Xs,[X]) :-
	var(Xs), !.
put_in_front(X,Xs,[X|Xs]).

may_be_var(X,X):- ( X=[] ; true ), !.


%---------------------------------------------------------------------------------%
% res_abs_stprf_input_user_interface(+,+,-,+)
% res_abs_stprf_input_user_interface(InputUser,Qv,CallT,PrimeT,PrimeDet,PrimeNf,ASub,Sg)
%---------------------------------------------------------------------------------%
res_abs_stprf_input_user_interface(InputUser,Qv,CallT,PrimeT,PrimeD,PrimeN,ASub,Sg):-
	time(T1),	
	% Name,Args,Clauses,Solution,AlreadySolved,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed
	InputUser = (ArgInput0,SolInput0,ResInput0),
	may_be_var(ArgInput0,ArgInput),
	may_be_var(SolInput0,SolInput),
	may_be_var(ResInput0,ResInput),
	% 1. Generate a random name
	functor(Sg,Name,_),
	get_next_name(Name,FullName),
	predkey_from_sg(Sg, SgKey),
	!, % Just one name
	% 2a. Detect input and output
	% term_variables(Sg,Sv),
	choose_input_output(Qv,CallT,IArgs,OArgs),
	% 2b. Uglify input and output variables
	% normalize_vars(ArgInput,PrimeT,UglyVars),
	% normalize_vars(PrimeT,ArgInput,UglyVars),
	normalize_vars(CallT,ArgInput,UglyVars),
	select_vars_size_types(UglyVars,IArgs,ISizes),
	get_vars_size_types(OArgs,PrimeT,OSizes),
	merge_size_types(ISizes,OSizes,SizeMapping),
	% 2c. Create assignments
	input_args_assignments(OArgs,SizeMapping,UglyVars,Assigns),
	% 3. Detect determinism and non-failure
	PrimeD = det(_,_Covered,DFails), % Covered and Fails as in detabs:internal_value_to_native
	% ( Covered = mutex /* = mut_exclusive */ ; DFails = is_det /* = is_det */
	( DFails = is_det /* = is_det */
	-> IsDet = true
	 ; IsDet = false
	),
	PrimeN = nf(_,_,Fails),
	( Fails = fails ; Fails = possibly_fails
	-> IsFail = true
	 ; IsFail = false
	),
	% 4. Work on solutions
	OutSolVar = (_,_),
	( SolInput = [cardinality(_,L,U)|_]
	-> SolExpr = solution((L,U))
	; SolExpr = solution((0,inf)) % SolExpr = solution((1,1))
	
	    %SolExpr = solution((0,'Infinity'))
	),
	SolAssign = assign(solution(OutSolVar),SolExpr),
	% 5. Work on resources
	generate_resource_variables(SgKey,[],OutResVars,Zeros),
	generate_zero_assigns(Zeros,ZAssigns),
	generate_resource_nil_variables(SgKey,[],PrevResVars),
	input_res_assignments(OutResVars,ResInput,ResAssigns),
	append(ZAssigns,ResAssigns,ZResAssigns),
	% 6a. Generate clause
	get_second_components(ISizes,RealISizes),
	get_second_components(OSizes,RealOArgs),
	append(RealOArgs,[solution(OutSolVar)|OutResVars],RealOSizes),
	append(Assigns,[SolAssign|ZResAssigns],ClAssigns),
	functor(Sg,Functor,Arity),
	make_atom([Functor,Arity,'trust'],_Key),
	Clause = clause(trust,RealISizes,RealOSizes,ClAssigns,[]),
	% 6b. Generate substitution
	!,
	asub(ASubPre,SgKey,FullName,(IArgs,OArgs),[Clause],'$unknown',[],SizeMapping,OutSolVar,(nil,nil),IsDet,IsFail,OutResVars,PrevResVars,false,[]),

	update_timer(T1,input_user_interface),
	% ASub = ASubPre.
	get_solver_behavior(TrySolve),
	solve_asub(ASubPre,TrySolve,ASub).
res_abs_stprf_input_user_interface(InputUser,Qv,CallT,_PrimeT,_PrimeD,_PrimeN,_ASub,_Sg) :-
	warning_message("input user interface failed: ~q, ~q, ~q",[InputUser,Qv,CallT]),
	fail.

% normalize_vars([],_,[]).
% normalize_vars([rsize(Arg,Size)|R],Types,[(Arg,Ugly)|URest]) :-
% 	find_variable_type(Types,Arg,Ty),
% 	normalize_rsize(Ty,Size,Ugly),
% 	normalize_vars(R,Types,URest).

normalize_vars([],_,[]).
normalize_vars([V:(A,T)|Types],ArgInputs,[(V,Ugly)|URest]) :-
	find_variable_type([V:(A,T)|Types],V,Ty),
	get_size(V,Ty,ArgInputs,Size),
	% find_variable_type(Types,Arg,Ty),
	normalize_rsize(Ty,Size,Ugly),
	normalize_vars(Types,ArgInputs,URest).

% It obtains the size of the variable if exists. Creates a new one otherwise.
get_size(V,_Type,[rsize(D,Size)|_T],Size):-
	V == D,!.
get_size(V,Type,[rsize(D,_S)|T],Size):-
	V \== D,!,
	get_size(V,Type,T,Size).
get_size(_V,Type,[],Size):-
	get_size_type(Type,Size).


input_args_assignments([],_,_,[]).
input_args_assignments([A|R],SizeMapping,Ugly,[assign(S,U)|AssR]) :-
	get_var_size_type(SizeMapping,A,S),
	get_var_size_type(Ugly,A,U),
	input_args_assignments(R,SizeMapping,Ugly,AssR).

input_res_assignments([],_,[]).
input_res_assignments([R|Rest],ResInput,[AssR|AssRest]) :-
	R =.. [RName|_],
	( find_input_res_assignment(ResInput,RName,Lhs)
	-> AssR = assign(R,Lhs)
	 ; generate_resource_default_assigns([R],[AssR])
	),
	input_res_assignments(Rest,ResInput,AssRest).

find_input_res_assignment([costb(_,R,L,U)|_],R,Ass) :-
	!, Ass =.. [R,(L,U)].
find_input_res_assignment([_|L],R,Ass) :-
	find_input_res_assignment(L,R,Ass).

%------------------------------------------------------------------------%
% res_abs_stprf_asub_to_native(+,+,-,-,+)
% res_abs_stprf_asub_to_native(ASub,Qv,PostProps,CompProps,Types)
%------------------------------------------------------------------------%

% res_abs_stprf_asub_to_native(ASub,Qv,RSizes,CompProps,ATypes):-
% 	asub(ASub,_,_,_,Sol,_,_,_,_,_,_,_,_,_,_),
% 	Sol \= resolution(_,_,_,_), !, 
% 	%solve_asub is call in order to form the result of each pp (Kludge?).
% 	solve_asub(ASub,SASub),
% 	res_abs_stprf_asub_to_native_(SASub,Qv,RSizes,CompProps,ATypes).
res_abs_stprf_asub_to_native(ASub,Qv,RSizes,CompProps,ATypes):-
	res_abs_stprf_asub_to_native_(ASub,Qv,RSizes,CompProps,ATypes).

res_abs_stprf_asub_to_native_(ASub,_Qv,[],[],_ATypes):-
	asub(ASub,_,_,_,_,Sol,_,_,_,_,_,_,_,_,_,_),
	Sol \= resolution(_,_,_,_), !.
res_abs_stprf_asub_to_native_(ASub,Qv,RSizes,CompProps,_ATypes):-
	asub(ASub,_,_P,(IArgs,OArgs),_,Sol,_,SizeMapping,OutSolVar,_,IsDet,IsFail,OutResVars,_,_,_),
	Sol = resolution(_,_SIArgs,SOArgs,Solutions),
	% 1. Generate input arguments
	get_size_mappings(IArgs,SizeMapping,ISizes),
	args_to_native(IArgs,ISizes,[],[],RInput),
	% 2. Generate output arguments
	partition_closed_forms(Solutions,Vars,Expr),
	debug_message("RESULTS>>> ~q",[Expr]),
	args_to_native(OArgs,SOArgs,Vars,Expr,ROutput),
	% 3. Put arguments together
	append(RInput,ROutput,RSizes),
	% 3. Generate solutions
	substitute_vars(OutSolVar,Vars,Expr,(LSol,USol)),
	substitute_vars(OutResVars,Vars,Expr,ResComp),
	single_resources(ResComp,Qv,SingleResComp0),
	(current_pp_flag(output_resources,functions) ->
	 (%clean_unsolved_eq,
	 take_off_unsolved_eqs(SingleResComp0,SingleResComp,_UnsSet))
	;
	    SingleResComp=SingleResComp0
	),

%	error_message("Resources (~q) >>> ~q ((~q))",[P,SingleResComp,SizeMapping]),
	Props = [cardinality(Qv,LSol,USol)|SingleResComp],
	(IsDet = true -> CompPropsD = [is_det(Qv)|Props];CompPropsD = [non_det(Qv)|Props]),
	 (IsFail = true -> CompProps = [possibly_fails(Qv)|CompPropsD];CompProps=[not_fails(Qv)|CompPropsD]).

:- use_module(.(eq_solver/solver), [expand/2]).
:- use_module(domain(resources/eq_solver/solver/solver_messages),[show_user_message/1]).
% Generates a set of substitutions of variables to expressions
% If we have a closed form for the variable, we chose it
% Otherwise, we use ? as placeholder
partition_closed_forms([],[],[]).
partition_closed_forms([(V,closed_form(Expr))|R],[V|RestVars],[SExpr|RestExpr]) :-
	!,
	% Simplify
	simplify_min_max_expr(Expr,MExpr0),
	simplify(MExpr0,MExpr),
	%SExpr = MExpr,
	catch(
		 expand(MExpr,SExpr),
		 Ex,
		 (
		     show_user_message(Ex), 
		     (Ex = not_implemented(_Op,_Solver,_Exp) -> SExpr=MExpr; throw(Ex))		       )    
	     ), 
	%expand_in_mathematica(MExpr,SExpr), %%MK 06-11-14: Defined in eq_solver module.
	partition_closed_forms(R,RestVars,RestExpr).
%partition_closed_forms([(V,_)|R],[V|RestVars],['?'|RestExpr]) :-
partition_closed_forms([(V,unsolved_equations(_,Eqs))|R],[V|RestVars],[eqs(Unsolved)|RestExpr]) :-
	!,
	% Generate better form
	beautify_equations(Eqs,Unsolved),
	partition_closed_forms(R,RestVars,RestExpr).

% Converts the information of variables into size expressions
args_to_native([],_,_,_,[]).
args_to_native([V|R],[S|SizeR],Vars,Exprs,[rsize(V,Prop)|NativeR]) :-
	substitute_vars(S,Vars,Exprs,SubstS),
	denormalize_rsize(SubstS,Prop),
	args_to_native(R,SizeR,Vars,Exprs,NativeR).

single_resources([],_,[]).
% single_resources([R|Rest],Qv,SRest) :-
% 	R =.. [_RName,Args],
% 	conj_to_list(Args,LArgs),
% 	length(LArgs,4),
% 	LArgs = [_CC,_,L,U],L==0,U==0,!,
% 	%RCName =.. [RName,CC],
% 	single_resources(Rest,Qv,SRest).
single_resources([R|Rest],Qv,[costb(Qv,RCName,L,U)|SRest]) :-
	R =.. [RName,Args],
	conj_to_list(Args,LArgs),
	length(LArgs,4),!,
	LArgs = [CC,_,L,U],
	RCName =.. [RName,CC],
	single_resources(Rest,Qv,SRest).
single_resources([R|Rest],Qv,[costb(Qv,RName,L,U)|SRest]) :-
	R =.. [RName,(L,U)],!,
	single_resources(Rest,Qv,SRest).

:- export(unsolved_resource_eq/1).
:- data unsolved_resource_eq/1.
save_unsolved(E):-
	assertz_fact(unsolved_resource_eq(E)).

clean_unsolved_eq:-
	retractall_fact(unsolved_resource_eq(_)).


take_off_unsolved_eqs([],[],[]).
take_off_unsolved_eqs([costb(Qv,RName,L,U)|SRest],[costb(Qv,RName,NL,NU)|SSRest],T) :-
	replace_if_unsolved(0,L,NL),
	replace_if_unsolved(inf,U,NU),
	(nonvar(NL),NL=eqs(_) -> T=[NL|NN];T=NN),
	 (nonvar(NU),NU=eqs(_) -> NN=[NU|NNN];NN=NNN),
	take_off_unsolved_eqs(SRest,SSRest,NNN).

replace_if_unsolved(Default,Expr,Default):-
	nonvar(Expr),Expr = eqs(_),!,
	save_unsolved(Expr).
replace_if_unsolved(_Default,Expr,Expr).





%------------------------------------------------------------------------%
% res_abs_stprf_unknown_call(+,+,-)
% res_abs_stprf_unknown_call(Vars,Call,Succ)
%------------------------------------------------------------------------%

res_abs_stprf_unknown_call(_Vars,Call,Call):-!.

%------------------------------------------------------------------------%
% res_abs_stprf_unknown_entry(+,-)
% res_abs_stprf_unknown_entry(Vars,Entry)
%------------------------------------------------------------------------%

% When unknown, use the default solution and resources values
res_abs_stprf_unknown_entry(Vars,ETypes,Entry):-
	get_next_name('$unknown',NewName), !, % Only one name
	get_predkey('$unknown',0,SgKey),
	get_vars_size_types(Vars,ETypes,Mapping),
	get_size_mappings(Vars,Mapping,IArgs),
	% Solutions are between 0 and Infinity
	OutSolVar = (_,_),
	PrevSolVar = (nil,nil),
	SolAssign = assign(solution(OutSolVar),solution((0,inf))),
	%SolAssign = assign(solution(OutSolVar),solution((0,'Infinity'))),
	% For resources, generate the default values
	generate_resource_variables(SgKey,[],OutResVars,_),
	generate_resource_nil_variables(SgKey,[],PrevResVars),
	generate_resource_default_assigns(OutResVars,ResAssign), !,
	% Put together
	OArgs = [solution(OutSolVar)|OutResVars],
	asub(Temp,SgKey,NewName,(Vars,[]),[clause(_,IArgs,OArgs,[SolAssign|ResAssign],[])],'$unknwon',[],Mapping,OutSolVar,PrevSolVar,false,true,OutResVars,PrevResVars,false,[]),
	% Entry = Temp.
	get_solver_behavior(TrySolve),
	solve_asub(Temp,TrySolve,Entry).

%------------------------------------------------------------------------%
% res_abs_stprf_empty_entry(+,-)
% res_abs_stprf_empty_entry(Vars,Entry)
%------------------------------------------------------------------------%

% A bottom entry has '$unknown' as its resolution
res_abs_stprf_empty_entry(Vars,ETypes,Entry):-
	get_next_name('$unknown',NewName), !, % Only one name
	get_predkey('$unknown',0,SgKey),
	get_vars_size_types(Vars,ETypes,Mapping),
	generate_resource_variables(SgKey,[],OutResVars,_),
	generate_resource_nil_variables(SgKey,[],PrevResVars), !,
	asub(Entry,SgKey,NewName,(Vars,[]),[],'$unknown',[],Mapping,(_,_),(nil,nil),false,true,OutResVars,PrevResVars,false,[]).

% This counter is used internally to generate new names
:- data counter/1.
counter(0).

get_next_name(F,F/N) :-
	current_fact(counter(N)), !,
	retract_fact(counter(N)),
	N1 is N+1,
	asserta_fact(counter(N1)).

%--------------------------------------------
% Predicates for substituting predicate names
%--------------------------------------------

% Merge all the information in the clauses:
% - Make all the clauses have the same name
% - Merge the previous solutions
accumulate_asubs(['$top'],'$top') :- !.
accumulate_asubs([A],A) :- !.
accumulate_asubs([A|R],Accum) :-
	accumulate_asubs(R,RAccum),
	( A = '$top'
	-> Accum = RAccum
	 ; ( RAccum = '$top'
	   -> Accum = A
	    ; % Both are regular abstract substitutions
	      asub(RAccum,SgKey,Name,Args,RClauses,_,RAlready,RSizeMapping,ROutSol,RPrevSol,IsDet,IsFail,ROutRes,RPrevRes,RYetFailed,_),
	      asub(A,_,_AName,_AArgs,AClauses,_ASolution,AAlready,ASizeMapping,_AOutSol,_APrevSol,_AIsDet,_AIsFail,_AOutRes,_APrevRes,AYetFailed,_),
	      % Join clauses, already solved, types and previously called
	      up_to_renaming_append(AClauses,RClauses,AccumClauses),
	      unifiable_append(RAlready,AAlready,AccumAlready),
	      merge_size_types(RSizeMapping,ASizeMapping,AccumSizeMapping),
	      %ord_union(RPrev,ConfPrev,AccumPrev),
	      or_(RYetFailed,AYetFailed,AccumYetFailed),
	      % Put everything together
	      asub(Accum,SgKey,Name,Args,AccumClauses,'$unknown',AccumAlready,AccumSizeMapping,ROutSol,RPrevSol,IsDet,IsFail,ROutRes,RPrevRes,AccumYetFailed,[])
	   )
	).

or_(false,false,false) :- !.
or_(true,_,true) :- !.
or_(false,true,true) :- !.

:- data last_replacement/4.

clean_all_caches :-
	clean_equation_names_cache,
	retractall_fact(last_replacement(_,_,_,_)),
	clean_internal_equiv_types,
	clean_size_type_cache.

get_last_replacements(F,I,O,N) :-
	current_fact(last_replacement(F,I,O,N)).

put_asub_last_replacement(ASub,F/N) :-
	asub(ASub,_,_,(IArgs,OArgs),_,_,_,SizeMapping,OutSolVar,_,_,_,OutResVars,_,_,_),
	get_size_mappings(IArgs,SizeMapping,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes),
	append(OSizes,[solution(OutSolVar)|OutResVars],OSizesSolRes),
	put_last_replacement(F,ISizes,OSizesSolRes,N).

put_last_replacement(F,IArgs,OArgs,N) :-
	copy_term(last_replacement(F,IArgs,OArgs,_), C),
	retractall_fact(C),
	asserta_fact(last_replacement(F,IArgs,OArgs,N)).

find_last_replacement(F,IArgs,OArgs,N) :-
	copy_term(last_replacement(F,IArgs,OArgs,_), last_replacement(FP,IArgsP,OArgsP,N)),
	current_fact(last_replacement(FP,IArgsP,OArgsP,N)).

find_last_asub_replacement(ASub,N) :-
	asub(ASub,_,F/_,(IArgs,OArgs),_,_,_,SizeMapping,OutSolVar,_,_,_,OutResVars,_,_,_),
	get_size_mappings(IArgs,SizeMapping,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes),
	append(OSizes,[solution(OutSolVar)|OutResVars],OSizesSolRes),
	find_last_replacement(F,ISizes,OSizesSolRes,N).

apply_replacements(ASub,ASubR) :-
	find_last_asub_replacement(ASub,NewN),
	!,
	asub(ASub,SgKey,F/_,(IArgs,OArgs),Clauses,_,Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	update_clauses(Clauses,NewClauses),
	create_old_to_new_mapping(Already,Mapping),
	update_already_solved(Already,Mapping,NewAlready),
	asub(ASubR,SgKey,F/NewN,(IArgs,OArgs),NewClauses,'$unknown',NewAlready,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests).
apply_replacements(ASub,ASubR) :-
	asub(ASub,SgKey,F/N,(IArgs,OArgs),Clauses,_,Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests),
	update_clauses(Clauses,NewClauses),
	asub(ASubR,SgKey,F/N,(IArgs,OArgs),NewClauses,'$unknown',Already,SizeMapping,OutSolVar,PrevSolVar,IsDet,IsFail,OutResVars,PrevResVars,YetFailed,Tests).

% Change the clauses to point to new function
update_clauses([],[]).
update_clauses([clause(K,I,O,Instr,T)|R],Out) :-
	update_clauses(R,UpR),
	update_clause(Instr,UpInstr),
	( clause_member(clause(K,I,O,UpInstr,T),UpR)
	-> Out = UpR
	 ; Out = [clause(K,I,O,UpInstr,T)|UpR]
	).

update_clause([],[]).
update_clause([call(F/_,I,O)|R],[call(F/N,I,O)|UpR]) :-
	find_last_replacement(F,I,O,N),
	!, update_clause(R,UpR).
update_clause([E|R],[E|UpR]) :-
	!, update_clause(R,UpR).

clause_member(clause(K,I,O,Eqs,T),[clause(K,I0,O0,Eqs0,T)|_]) :-
	copy_term((I,O,Eqs),(I1,O1,Eqs1)),
	copy_term((I0,O0,Eqs0),(I2,O2,Eqs2)),
	varset((I1,O1),Vars),
	generate_aux_names(Vars),
	(I1,O1) = (I2,O2),
	variant((I1,O1,Eqs1),(I2,O2,Eqs2)), !.
clause_member(C,[_|R]) :-
	!, clause_member(C,R).

generate_aux_names([]).
generate_aux_names([V|R]) :-
	get_next_name(var,Var),
	V = Var, !,
	generate_aux_names(R).

create_old_to_new_mapping([],[]).
create_old_to_new_mapping([resolution(Name,RIArgs,ROArgs,Sols)|R],Mapping) :-
	create_old_to_new_mapping(R,RMapping),
	( Name = F/M, find_last_replacement(F,RIArgs,ROArgs,N), M \= N
	-> get_solution_fn_names(F/N,RIArgs,ROArgs,FnNames),
	   zip_names(Sols,FnNames,NewMapping),
	   append(NewMapping,RMapping,Mapping)
	 ; Mapping = RMapping
	).

zip_names([],[],[]).
zip_names([(_,unsolved_equations(_,[equation(F,_,_,_)|_]))|R],[Name|NameR],[(F,Name)|ZipR]) :-
	!,zip_names(R,NameR,ZipR).
zip_names([_|R],[_Name|NameR],ZipR) :-
	zip_names(R,NameR,ZipR).

% Get the names of the functions that will replace the old ones
get_solution_fn_names(Name,ISizes,OSizes,FnNames) :-
	copy_term(ISizes,ISizesN),
	copy_term(OSizes,OSizesN),
	FakePred = predicate(Name,ISizesN,OSizesN,[clause(_,ISizesN,OSizesN,[call(Name,ISizesN,OSizesN)],[])]), get_solver_behavior(TrySolve),
	pred_solve(FakePred,[],0,true,false,TrySolve,resolution(_,_,_,Sols)),
	get_solution_fn_names_vars(Sols,FnNames).

get_solution_fn_names_vars([],[]).
get_solution_fn_names_vars([(_,closed_form(_))|R],FR) :-
	get_solution_fn_names_vars(R,FR).
get_solution_fn_names_vars([(_,unsolved_equations(_,[equation(F,_,_,_)]))|R],[F|FR]) :-
	get_solution_fn_names_vars(R,FR).

% Update the equations and closed forms with the new mappings
update_already_solved([],_,[]).
%update_already_solved([resolution(F/_,_,_,_)|R],Mapping,UpR) :-
%	!, update_already_solved(R,F,Mapping,UpR).
update_already_solved([resolution(G,I,O,Sols)|R],Mapping,[resolution(G,I,O,NewSols)|UpR]) :-
	%member(G,ToUpdate), !,
	!, update_functors(Sols,Mapping,NewSols),
	update_already_solved(R,Mapping,UpR).
%update_already_solved([resolution(G,I,O,Sols)|R],F,ToUpdate,Mapping,[resolution(G,I,O,Sols)|UpR]) :-
%	update_already_solved(R,F,ToUpdate,Mapping,UpR).

update_functors(X,_,X) :-
	var(X), !.
update_functors(Functor,Mapping,NewFunctor) :-
	nonvar(Functor),
	Functor =.. [F|Args],
	update_functors_list(Args,Mapping,NewArgs),
	( member((F,NewF),Mapping)
	-> NewFunctor =.. [NewF|NewArgs]
	 ; NewFunctor =.. [F|NewArgs]
	).

update_functors_list([],_,[]).
update_functors_list([E|R],Mapping,[UpE|UpR]) :-
	update_functors(E,Mapping,UpE),
	update_functors_list(R,Mapping,UpR).
