:- module(res_abs_utils, _, [assertions,regtypes,basicmodes]).

:- doc(module,
	"Small utilities used in argument size analysis.").

:- use_module(library(lists)).
:- use_module(library(terms_check)).

% Get second components of a list of tuples
get_second_components([],[]).
get_second_components([(_,B)|R],[B|SR]) :-
	get_second_components(R,SR).

up_to_renaming_subset([],_).
up_to_renaming_subset([A|R],Z) :-
	member(E,Z), variant(E,A),
	up_to_renaming_subset(R,Z).

up_to_renaming_subset_extra([],_,_,_).
up_to_renaming_subset_extra([A|R],E,Z,F) :-
	member(B,Z), variant((A,E),(B,F)),
	up_to_renaming_subset_extra(R,E,Z,F).

up_to_renaming_append(S,[],S).
up_to_renaming_append(S,[E|R],T) :-
	( member(A,S), variant(E,A)
	-> up_to_renaming_append(S,R,T)
	 ; append(S,[E],SAndE),
	   up_to_renaming_append(SAndE,R,T)
	).

unifiable_subset([],_).
unifiable_subset([A|R],Z) :-
	member(E,Z), unifiable(A,E,_),
	unifiable_subset(R,Z).

unifiable_append(S,[],S).
unifiable_append(S,[E|R],T) :-
	( member(A,S), unifiable(E,A,_)
	-> unifiable_append(S,R,T)
	 ; append(S,[E],SAndE),
	   unifiable_append(SAndE,R,T)
	).

var_subset([],_).
var_subset([A|R],Z) :-
	member(E,Z), A == E,
	var_subset(R,Z).

var_append(S,[],S).
var_append(S,[E|R],T) :-
	( member(A,S), A == E
	-> var_append(S,R,T)
	 ; append(S,[E],SAndE),
	   var_append(SAndE,R,T)
	).

% Zips two lists
zip([],_,[]) :- !.
zip(_,[],[]) :- !.
zip([A|R],[B|S],[(A,B)|T]) :- zip(R,S,T).