:- module(res_plai,
	[ res_plai_call_to_entry/9,
	  res_plai_exit_to_prime/7,
	  res_plai_project/4,
	  res_plai_extend/5,
	  res_plai_widen/3,
	%  res_plai_widencall/3,
	  res_plai_compute_lub/2,
          res_plai_compute_clauses_lub/3,
	  res_plai_glb/3,
	  res_plai_less_or_equal/2,
	  res_plai_identical_abstract/2,
	  res_plai_sort/2,
	  res_plai_call_to_success_fact/9,
	  res_plai_split_combined_domain/3,
	  res_plai_special_builtin/4,
	%  res_plai_success_builtin/5,
	  res_plai_call_to_success_builtin/6,
	  res_plai_input_interface/4,
	  res_plai_input_user_interface/4,
	  res_plai_input_user_interface/5,
	  res_plai_asub_to_native/5,
	  res_plai_unknown_call/3,
	  res_plai_unknown_entry/2,
	  res_plai_empty_entry/2,
	  res_plai_collect_abstypes/3,
	  res_plai_rename_abs/4,
	  load_resources_modules/0,
	  init_abstract_domain/0
	],
	[assertions,regtypes,basicmodes,datafacts]).

:- use_module(ciaopp(p_unit/program_keys), [predkey_from_sg/2]).
%:- use_module(ciaopp(plai/domains), [identical_proj/5]).
:- use_module(ciaopp(plai/domains)).
:- use_module(library(lists)).
:- use_module(library(messages), [warning_message/2]).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(terms_vars)).
:- use_module(domain(etermsvar)).
:- use_module(domain(detplai)).
:- use_module(domain(detabs), [det_asub_to_native/3]).
:- use_module(domain(nfplai)).
:- use_module(domain(nfabs), [nf_asub_to_native/3]).
:- use_module(domain(resources/res_abs)).
:- use_module(domain(resources/eq_solver/solver),[init_solver/0]).
:- use_module(ciaopp(plai/plai_db)).
:- use_module(typeslib(typeslib), [dz_equivalent_types/2, dz_type_included/2, simplify_step2/0]).
:- use_module(typeslib(equiv_type), [equiv_type/2]).
:- import(resources_lib,[load_resources_modules/0]).
% Loading resource modules
:- use_module(library(aggregates), [findall/3]).
:- use_module(library(compiler), [use_module/1]).
:- use_module(ciaopp(p_unit/itf_base_db), [curr_file/2]).
:- use_module(ciaopp(p_unit/clause_db), [source_clause/3]).
:- use_module(ciaopp(infer),[get_absint/4]).

%:- use_module(library(messages), [warning_message/2]).

%:- reexport(resources(res_assrt_defs/resources_lib),[load_resources_modules/0]).

% TODO: max and min expressions of sizes in append are not working now
% It is giving correct but imprecise size bounds. Correct it.
% TODO: lower bounds of coupled are incorrect. Correct it.
% TODO: Check why '?' appears in expressions and avoid its use
% or handle it properly (e.g. listfact does not work because simplify
% fails with '?').
% TODO: Implement factorial example with accumulators and add it
% to examples directory. 

:- use_module(library(messages), [warning_message/2, debug_message/2]).
:- multifile issue_debug_messages/1.
:- data issue_debug_messages/1.
%%%%%%%%%%%%%%%%%%%%Uncomment for debugging a module%%%%%%%%%%%%%%%%%%%%%
% issue_debug_messages(res_abs).
% issue_debug_messages(res_plai).
% issue_debug_messages(eq_solver).
% issue_debug_messages(pred_solver).
% issue_debug_messages(eq_solver_mathematica).
% issue_debug_messages(diff_equ_res).
% issue_debug_messages(second_order_res).
% issue_debug_messages(fixpo_plai).

asub(res_plai(Types,Args),Types,Args).

%------------------------------------------------------------------------%
% res_plai_call_to_entry(+,+,+,+,+,+,-,-)
% res_plai_call_to_entry(Sv,Sg,Hv,Head,Fv,Proj,Entry,ExtraInfo)
%------------------------------------------------------------------------%
res_plai_call_to_entry(Sv,Sg,Hv,Head,K,Fv,Proj,Entry,(ProjT,PrimeT,ExtraInfoArgs)):-
	asub(Proj,PTypes,PArgs),
	append(Hv,Fv,HvFv),
	get_tests_from_det(K,HvFv,T),
	get_call_and_proj(Sg,etermsvar,PTypes,ProjT,PrimeT0),
	cannonical_types(PrimeT0,PrimeT),
	append(Hv,Fv,CVars),
	debug_message("CLAUSE>>>>~q  (~q) test: (~q) ~q ",[K,Sv,CVars,T]),
	% predkey_from_sg(Sg,SgKey),
	get_call_and_proj(Sg,det,PTypes,_Proj,PrimeD),
	get_call_and_proj(Sg,nf,PTypes,_Proj0,PrimeN),
	% get_info_from_db(det, SgKey, Sg, det(ProjT,_,_), Proj, PrimeD),
	% get_info_from_db(nf, SgKey, Sg, det(ProjT,_,_), Proj, PrimeN),

%	get_absint(pred,det,Sg,(_,PrimeD)),
%	get_absint(pred,nf,Sg,(_,PrimeN)),
	%get_call_and_proj(Sg,det,ProjT,_ProjD2,PrimeD),
	%det_sort(ProjD2,ProjD),
	%get_call_and_proj(Sg,nf,ProjT,_ProjNf,PrimeN),
	%nf_sort(ProjN2,ProjN),
	etermsvar_call_to_entry(Sg,Hv,Head,Fv,ProjT,ETypes0,_ExtraInfoTypes),
	cannonical_types(ETypes0,ETypes),
	( ETypes = '$bottom'
	-> Entry = '$bottom';
	 %; det_call_to_entry(Sv,Sg,Hv,Head,Fv,ProjD,EDet,ExtraInfoDet),
	 %  nf_call_to_entry(Sv,Sg,Hv,Head,Fv,ProjN,ENf,ExtraInfoNf),
	 %  % Get information for call_to_entry
	   PrimeD = det(_,_,DetInfo),
	   PrimeN = nf(_,_,NfInfo),
	   %upgrade_numbers(PTypes,PTypes0),
	   upgrade_numbers(ProjT,PTypes0),
	   res_abs_call_to_entry(Sv,Sg,Hv,Head,K,Fv,PArgs,EArgs,ExtraInfoArgs,PTypes0,ETypes,PrimeT,DetInfo,NfInfo,T),
	   asub(Entry,ETypes,EArgs)
	).
	%asub(ExtraInfo,ExtraInfoTypes,ExtraInfoDet,ExtraInfoNf,ExtraInfoArgs).



%:- export(get_tests_from_det)
:- pred get_tests_from_det(PPKey,Vars,Tests): (atom(PPKey),list(Vars,var),var(Tests)) => list(Tests,term) #
	"@var{Tests} is the list of (arithmetic) tests that was gathered 
		     by det analysis until the program point @var{PPKey}, on vars @var{Vars}.".
get_tests_from_det(K,HvFv,T):-
	get_absint(K,det,HvFv,ASub),!,
	ASub = det(_,_,det([t(_,_,T0,_)|_],_,_)), % TODO: This operation should be defined in det in order to encapsulate details.
	filter_tests(T0,T).
get_tests_from_det(no_provided,_,[]).
get_tests_from_det(_K,_HvFv,[]):-!.
% % TODO: Must this use types asub in order to get the right variant?. Because is returning the lub anyway...
% get_memo_table(AbsInt,K,HvFv,T):-
% 	get_memo_lub(K,HvFv,AbsInt,no,[T]).

:- use_module(domain(resources/eq_solver/eq_solver_types)).
filter_tests([],[]).
filter_tests([Comp|T], [Comp|TT]):-
	Comp =.. [Op,L,R],
	operator(Op),
	algebraic_expression(L),
	algebraic_expression(R),!,
	filter_tests(T,TT).
filter_tests([_Comp|T], TT):-
	filter_tests(T,TT).

operator('=').
operator('<').
operator('>').
operator('>=').
operator('=<').
operator('==').
operator('\==').
operator('\=').

%%Code for accessing the database
%%in a similar way that call_to_success
%%does in its first clause (fixpo_plai.pl)
%%NOT USED YET.
get_info_from_db(Sg,AbsInt,Call,Proj,Prime):-
	var(Prime),!,
	predkey_from_sg(Sg,SgKey),
	current_fact(complete(SgKey,AbsInt,Subg,Proj1,Prime1,_Id,_Fs),_R),
	% less_or_equal_proj(AbsInt,Sg,Call,Subg,Proj1),!,
	identical_proj(AbsInt,Sg,Call,Subg,Proj1),!,
	Prime1 = [Prime0|_],
	upgrade_numbers(Proj1,Proj),
	upgrade_numbers(Prime0,Prime).

% Following is commented out and separated for det and nf into two
% clauses since asub for determinacy analysis is changed to det(_)
% from nf(_).

% get_call_and_proj(Sg,Domain,PTypes,Proj,Prime) :-
% var(Prime), ( Domain = nf ; Domain = det ),
% current_fact(complete(_,Domain,Sg,Proj,Primes,_,_)), PTypes =
% nf(T1,_,_), Proj = nf(T2,_,_), compatible_projections(T1,T2), !,
% Primes = [Prime|_].
get_call_and_proj(Sg,etermsvar,PTypes,Proj,Prime) :-
	var(Prime), !,
	current_fact(complete(_,etermsvar,Sg,Proj0,Primes,_,_)),
	compatible_projections(PTypes,Proj0),
	!,
	Primes = [Prime0|_],
	upgrade_numbers(Proj0,Proj),
	upgrade_numbers(Prime0,Prime).
get_call_and_proj(Sg,Domain,PTypes,Proj,Prime) :-
	var(Prime), ( Domain = nf  ),
	current_fact(complete(_,Domain,Sg,Proj,Primes,_,_)),
	%PTypes = nf(T1,_,_),
	Proj = nf(T2,_,_),
	term_variables(T2,V),
	etermsvar_project(V,PTypes,PPTypes),
	compatible_projections(PPTypes,T2),
	Primes = [Prime|_],!.
%% In case the types (call pattern) does not match any
%% result for nf/det in the database, but there is
%% only one call pattern for the subgoal,
%% then we take that result without checking the
%% types.
get_call_and_proj(Sg,Domain,_PTypes,Proj,Prime) :-
	var(Prime), ( Domain = nf  ),
	findall((Proj0,Prime0),get_absint(pred,nf,Sg,(Proj0,Prime0)),[Res]),!,
	Res = (Proj, Prime).
get_call_and_proj(Sg,Domain,PTypes,Proj,Prime) :-
	var(Prime), (Domain = det ),
	current_fact(complete(_,Domain,Sg,Proj,Primes,_,_)),
	%PTypes = det(T1,_,_),
	Proj = det(T2,_,_),
	term_variables(T2,V),
	etermsvar_project(V,PTypes,PPTypes),
	compatible_projections(PPTypes,T2),
	Primes = [Prime|_],!.
get_call_and_proj(Sg,Domain,_PTypes,Proj,Prime) :-
	var(Prime), ( Domain = det  ),
	findall((Proj0,Prime0),get_absint(pred,det,Sg,(Proj0,Prime0)),[Res]),!,
	Res = (Proj, Prime).
get_call_and_proj(Sg,nf,PNf,PNf,Empty) :-
	var(Empty), !,
	% Create an empty substitution
	term_variables(Sg,Sv),
	nf_empty_entry(Sv,E),
	nf_unknown_call(Sv,E,Empty).
get_call_and_proj(Sg,det,PDet,PDet,Empty) :-
	var(Empty), !,
	% Create an empty substitution
	term_variables(Sg,Sv),
	det_empty_entry(Sv,E),
	det_unknown_call(Sv,E,Empty).
get_call_and_proj(Sg,Domain,_PTypes,Proj,Prime) :-
	nonvar(Prime), !,
	current_fact(complete(_,Domain,Sg,Proj0,[PrimeA|_],_,_)),
	compatible_projections(PrimeA,Prime),
	upgrade_numbers(Proj0,Proj).

compatible_projections([],_).
compatible_projections([V1:(_,T1)|R1],V2s) :-
	is_in_type_list(V1,T1,V2s),
	compatible_projections(R1,V2s).

is_in_type_list(V1,T1,[V2:(_,T2)|_]) :-
	V1 == V2, !,
	( dz_type_included(T1,T2)
	; % Hack for different numbers
	  dz_type_included(T1,num),
	  dz_type_included(T2,num)
	).
is_in_type_list(V1,T1,[_|R2]) :-
	!, is_in_type_list(V1,T1,R2).

upgrade_numbers('$bottom','$bottom').
upgrade_numbers([],[]).
upgrade_numbers([V:(N,T)|R],[V:(N,T2)|UR]) :-
	( (dz_type_included(T,num),T\=int,T\=nnegint)
	-> T2 = num
	 ; T2 = T
	),
	upgrade_numbers(R,UR).

%------------------------------------------------------------------------%
% res_plai_exit_to_prime(+,+,+,+,+,-,-)
% res_plai_exit_to_prime(Sg,Hv,Head,Sv,Exit,ExtraInfo,Prime)
%------------------------------------------------------------------------%

cannonical_types([],[]).
cannonical_types([V:(N,T)|R],[V:(N,CT)|RR]):-
	equiv_type(T,CT),!,
	cannonical_types(R,RR).
cannonical_types([V:(N,T)|R],[V:(N,T)|RR]):-
	cannonical_types(R,RR).



res_plai_exit_to_prime(_Sg,_Hv,_Head,_Sv,'$bottom',_ExtraInfo,'$bottom'):- !.
res_plai_exit_to_prime(_Sg,_Hv,_Head,_Sv,'$top',_ExtraInfo,'$top'):- !.
res_plai_exit_to_prime(Sg,Hv,Head,Sv,Exit,(_ProjT,PrimeT,ExtraInfoArgs),Prime):-
	asub(Exit,ETypes,EArgs),
	%asub(ExtraInfo,_ExtraInfoTypes,_ExtraInfoDet,_ExtraInfoNf,ExtraInfoArgs),
	%etermsvar_exit_to_prime(Sg,Hv,Head,Sv,ETypes,ExtraInfoTypes,PTypes),
	%( PTypes = '$bottom'
	%-> Prime = '$bottom'
	% ; resargs_exit_to_prime(Sg,Hv,Head,Sv,EArgs,ExtraInfoArgs,PArgs,PTypes),
	%   ( PArgs = '$bottom'
	%   -> Prime = '$bottom'
	%    ; asub(Prime,PTypes,PArgs) )
	%).
	% Reuse previously known information from etermsvar
	% Get information for exit_to_prime
	cannonical_types(ETypes,CETypes),
	cannonical_types(PrimeT,CPrimeT),
	res_abs_exit_to_prime(Sg,Hv,Head,Sv,EArgs,ExtraInfoArgs,PArgs,CETypes,CPrimeT),
	( PArgs = '$bottom'
	-> Prime = '$bottom'
	 ; asub(Prime,CPrimeT,PArgs)
	).

%------------------------------------------------------------------------%
% res_plai_project(+,+,-)
% res_plai_project(ASub,Vars,Proj)
%------------------------------------------------------------------------%

res_plai_project('$bottom',_Sg,_Vars,'$bottom'):- !.
res_plai_project('$top',_Sg,_Vars,'$top'):- !.
res_plai_project(ASub,Sg,Vars,Proj):-
	asub(ASub,ATypes,AArgs),
	etermsvar_project(Vars,ATypes,PTypes),
	%det_project(ADet,Vars,PDet),
	%nf_project(ANf,Vars,PNf),
	upgrade_numbers(PTypes,PTypes0),
	res_abs_project(AArgs,Sg,Vars,PArgs,PTypes0),
	asub(Proj,PTypes,PArgs).

%------------------------------------------------------------------------%
% res_plai_extend(+,+,+,-)
% res_plai_extend(Prime,Sv,Call,Succ)
%------------------------------------------------------------------------%

res_plai_extend(_Sg,'$bottom',_Sv,_Call,'$bottom'):- !.
res_plai_extend(_Sg,'$top',_Sv,_Call,'$top'):- !.
res_plai_extend(Sg,Prime,Sv,Call,Succ):-
	asub(Prime,PTypes1,PArgs),
	asub(Call,CTypes1,CArgs),

	% ULQ: For some reason the abstract substitution PTypes is not
	% sorted and that makes unify_back fail. The following two
	% lines are added that sorts PTypes and CTypes. (+MK)
	
	etermsvar_sort(PTypes1, PTypes),
	etermsvar_sort(CTypes1, CTypes),
	
	etermsvar_extend(PTypes,Sv,CTypes,STypes),
	%det_extend(PDet,Sv,CDet,SDet),
	%nf_extend(PNf,Sv,CNf,SNf),
	get_call_and_proj(Sg,nf,CTypes,_,PrimeNf),
	
	%get_absint(pred,nf,Sg,(_,PrimeNf)),
	PrimeNf = nf(_,_,NfInfo),
	upgrade_numbers(CTypes,CTypes0),
	upgrade_numbers(STypes,STypes0),
	res_abs_extend(PArgs,Sg,Sv,CArgs,SArgs,CTypes0,STypes0,NfInfo),
	asub(Succ,STypes,SArgs).

%------------------------------------------------------------------------%
% res_plai_widen(+,+,-)
% res_plai_widen(ASub1,ASub2,ASub)
%------------------------------------------------------------------------%

res_plai_widen('$bottom',ASub2,ASub2).
res_plai_widen(ASub1,'$bottom',ASub1).
res_plai_widen(ASub1,ASub2,ASub) :-
	asub(ASub1,Types1,Args1),
	asub(ASub2,Types2,Args2),
	etermsvar_widen(Types1,Types2,Types),
	( Types == '$bottom'
	-> ASub = '$bottom'
	 ; %term_variables(Types,Sv),
	   %det_empty_entry(Sv,Det),
	   %nf_empty_entry(Sv,Nf),
	   %det_compute_lub([Det1,Det2],DetL),
	   %det_compute_clauses_lub([DetL],DetL,[Det]),
	   %nf_compute_lub([Nf1,Nf2],NfL),
	   %nf_compute_clauses_lub([NfL],NfL,[Nf]),
	   res_abs_widen(Args1,Args2,Args),
	   asub(ASub,Types,Args)
	).

%------------------------------------------------------------------------%
% res_plai_widencall(+,+,-)
% res_plai_widencall(ASub1,ASub2,ASub)
%------------------------------------------------------------------------%
% No widening done

% res_plai_widencall(ASub0,ASub1,Result):-
% 	asub(ASub0,A0,_A1),
% 	asub(ASub1,B0,B1),
% 	etermsvar_widencall(A0,B0,T0),
% 	asub(Result,T0,B1).


%------------------------------------------------------------------------%
% res_plai_compute_lub(+,-)
% res_plai_compute_lub(ListASub,Lub)
%------------------------------------------------------------------------%

res_plai_compute_lub(ListASub,Lub):-
	remove_bottoms(ListASub,ListNoBottoms),
	( ListNoBottoms == []
	-> Lub = '$bottom'
	 ; split(ListNoBottoms,[L|LTypes],LArgs),
	   etermsvar_compute_lub([L|LTypes],LubTypes),
	   %det_compute_lub(LDets,LDet),
	   %nf_compute_lub(LNfs,LNf),
	   ( etermsvar_identical_abstract(L,LubTypes)
	   -> upgrade_numbers(LubTypes,LubTypes0),
	      res_abs_compute_lub(LArgs,LubArgs,LubTypes0),
	      asub(Lub,LubTypes,LubArgs)
	    ; asub(Lub,LubTypes,'$top')
	   )
	).

split([],[],[]).
split(['$bottom'|ListASub],['$bottom'|LTypes],['$bottom'|LArgs]):-
	split(ListASub,LTypes,LArgs).
split(['$top'|ListASub],['$bottom'|LTypes],['$top'|LArgs]):-
	split(ListASub,LTypes,LArgs).
split([ASub|ListASub],[ATypes|LTypes],[AArgs|LArgs]):-
	asub(ASub,ATypes,AArgs),
	split(ListASub,LTypes,LArgs).

res_plai_split_combined_domain(ListASub,[LTypes,LArgs],[etermsvar,res_plai]):-
	( var(LTypes)
	-> split(ListASub,LTypes,_LArgs),
	   LArgs=ListASub
	 ; split_back(ListASub,LTypes,LArgs)
	).

split_back([],[],[]).
split_back([ASub|ListASub],[ATypes|LTypes],[ASubArgs|LArgs]):-
	( ATypes == '$bottom' -> ASub = '$bottom'
	; asub(ASubArgs,_ATypes,AArgs),
	  asub(ASub,ATypes,AArgs)
	),
	split_back(ListASub,LTypes,LArgs).

remove_bottoms([],[]).
remove_bottoms(['$bottom'|R],RemR) :-
	!, remove_bottoms(R,RemR).
remove_bottoms([E|R],[E|RemR]) :-
	!, remove_bottoms(R,RemR).

%------------------------------------------------------------------------%
% res_plai_compute_clauses_lub(+,-)
% res_plai_compute_clauses_lub(ListASub,Lub)
%------------------------------------------------------------------------%

res_plai_compute_clauses_lub(['$top'],_Proj,[]):- !.
res_plai_compute_clauses_lub(['$bottom'],_Proj,['$bottom']):- !.
res_plai_compute_clauses_lub([ASub],_Proj,[ASub]).%:-
	% asub(ASub,ATypes,AArgs),
	% asub(Proj,_PTypes,_PArgs),
	% %det_compute_clauses_lub([ADet],PDet,[LubDet]),
	% %nf_compute_clauses_lub([ANf],PNf,[LubNf]),
	% asub(LubS,ATypes,AArgs),
	% res_plai_sort(LubS,Lub).

%------------------------------------------------------------------------%
% res_plai_glb(+,+,-)
% res_plai_glb(ASub0,ASub1,Glb)
%------------------------------------------------------------------------%

res_plai_glb('$bottom',_ASub2,'$bottom').
res_plai_glb(_ASub1,'$bottom','$bottom').
res_plai_glb(ASub1,ASub2,Glb) :-
	asub(ASub1,Types1,Args1),
	asub(ASub2,Types2,Args2),
	get_subgoal(ASub1,Sg),
	etermsvar_glb(Types1,Types2,Types),
	( Types = '$bottom'
	-> Glb = '$bottom'
	 ; %det_glb(Det1,Det2,Det),
	    %nf_glb(Nf1,Nf2,Nf),
	    
	   %get_absint(pred,nf,Sg,(_,Nf)),
 	   %get_absint(pred,det,Sg,(_,Det)),
	   get_call_and_proj(Sg,nf,Types1,_,Nf),
	   get_call_and_proj(Sg,det,Types1,_,Det),
	   Det = det(_,_,DetInfo),
	   Nf = nf(_,_,NfInfo),
	   upgrade_numbers(Types,Types0),
	   res_abs_glb(Args1,Args2,Types0,DetInfo,NfInfo,Args),
	   asub(Glb,Types,Args)
	).


:- pred get_subgoal(ASub,Sg) # "Get the subgoal corresponding of an abstract substitution".
get_subgoal(ASub,Sg):-
	asub(ASub,_,res_abs(Name,(IA,OA),_,_,_,_,_,_,_,_,_,_,_,_)),
	Name = Pred/_,
	append(IA,OA,Ar),
	Sg =.. [Pred|Ar].
	


%------------------------------------------------------------------------%
% res_plai_less_or_equal(+,+)
% res_plai_less_or_equal(ASub0,ASub1)
%------------------------------------------------------------------------%

res_plai_less_or_equal('$bottom',_) :- !.
res_plai_less_or_equal(X,'$bottom') :- !, X == '$bottom'.
res_plai_less_or_equal(_,'$top') :- !.
res_plai_less_or_equal('$top',X) :- !, X == '$top'.
res_plai_less_or_equal(ASub0,ASub1) :-
	asub(ASub0,ATypes0,AArgs0),
	asub(ASub1,ATypes1,AArgs1),
	etermsvar_less_or_equal(ATypes0,ATypes1),
	% Do nothing with det or nf
	res_abs_less_or_equal(AArgs0,AArgs1).

%------------------------------------------------------------------------%
% res_plai_identical_abstract(+,+)
% res_plai_identical_abstract(ASub1,ASub2)
%------------------------------------------------------------------------%

res_plai_identical_abstract('$bottom','$bottom'):- !.
res_plai_identical_abstract('$top','$top'):- !.
res_plai_identical_abstract(ASub0,ASub1):-
	asub(ASub0,ATypes0,AArgs0),
	asub(ASub1,ATypes1,AArgs1),
	etermsvar_identical_abstract(ATypes0,ATypes1),
	% Do nothing with det or nf
	res_abs_identical_abstract(AArgs0,AArgs1).

%------------------------------------------------------------------------%
% res_plai_sort(+,-)
% res_plai_sort(ASub0,ASub1)
%------------------------------------------------------------------------%

res_plai_sort('$bottom','$bottom'):- !.
res_plai_sort('$top','$top'):- !.
res_plai_sort(ASub0,ASub1):-
	asub(ASub0,ATypes0,AArgs0),
	etermsvar_sort(ATypes0,ATypes1),
	%det_sort(ADet0,ADet1),
	%nf_sort(ANf0,ANf1),
	res_abs_sort(AArgs0,AArgs1),
	asub(ASub1,ATypes1,AArgs1).

%------------------------------------------------------------------------%
% res_plai_call_to_success_fact(+,+,+,+,+,+,-,-)
% res_plai_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ)
%-------------------------------------------------------------------------

res_plai_call_to_success_fact(_,_,_,_K,_,'$bottom',_,'$bottom','$bottom').
res_plai_call_to_success_fact(_,_,_,_K,_,'$top',_,'$top','$top').
res_plai_call_to_success_fact(Sg,Hv,Head,K,Sv,Call,Proj,Prime,Succ):-
	asub(Call,CTypes,CArgs),
	asub(Proj,PTypes,PArgs),
	debug_message("CLAUSE(FACT)>>>~q",[K]),
	etermsvar_call_to_entry(Sg,Hv,Head,[],PTypes,ETypes,_ExtraInfoTypes),
	etermsvar_call_to_success_fact(Sg,Hv,Head,Sv,CTypes,PTypes,RTypes,STypes),
	%det_call_to_success_fact(Sg,Hv,Head,Sv,CDet,PDet,RDet,SDet),
	%nf_call_to_success_fact(Sg,Hv,Head,Sv,CNf,PNf,RNf,SNf),
	get_subgoal(Call,Sg),
	
	%get_absint(pred,nf,Sg,(_,RNf)),
	%get_absint(pred,det,Sg,(_,RDet)),
	get_call_and_proj(Sg,nf,CTypes,_,RDet),
	get_call_and_proj(Sg,det,CTypes,_,RNf),	
	RDet = det(_,_,DetInfo),
	RNf = nf(_,_,NfInfo),
	res_abs_call_to_success_fact(Sg,Hv,Head,Sv,CArgs,PArgs,RArgs,SArgs,PTypes,ETypes,RTypes,STypes,DetInfo,NfInfo),
	asub(Prime,RTypes,RArgs),
	asub(Succ,STypes,SArgs).

%-------------------------------------------------------------------------
% res_plai_special_builtin(+,+,-,-)
% res_plai_special_builtin(SgKey,Sg,SubgoalType,Condvars)
%-------------------------------------------------------------------------

res_plai_special_builtin(SgKey,_Sg,special(SgKey),_Condvars):-
	res_abs_special_builtin(SgKey).
        %etermsvar_special_builtin(SgKey,Sg,Subgoal,Type,Condvars).

%-------------------------------------------------------------------------
% res_plai_success_builtin(+,+,+,+,-)
% res_plai_success_builtin(Type,Sv_u,Condv,Call,Succ)
%-------------------------------------------------------------------------

%res_plai_success_builtin(Type,_Sv_u,Sg,Call,Succ):-
%	asub(Call,CTypes,CArgs),
%	res_abs_success_builtin(Type,Sg,CArgs,CTypes,SArgs),
%	asub(Succ,CTypes,SArgs).

%-------------------------------------------------------------------------
% res_plai_call_to_success_builtin(+,+,+,+,+,-)
% res_plai_call_to_success_builtin(SgKey,Sg,Sv,Call,Proj,Succ)
%-------------------------------------------------------------------------

res_plai_call_to_success_builtin(_,_,_,'$bottom',_,'$bottom').
res_plai_call_to_success_builtin(_,_,_,'$top',_,'$top').
res_plai_call_to_success_builtin(SgKey,Sg,Sv,Call,Proj,Succ):-
	asub(Call,CTypes,CArgs),
	asub(Proj,PTypes,_),
	( etermsvar_special_builtin(SgKey,Sg,Sv,_TKey,_TCondVars),
	  etermsvar_call_to_success_builtin(SgKey,Sg,Sv,CTypes,PTypes,STypes)
	-> true
	 ; STypes = CTypes
	),
	(STypes \== '$bottom' ->
	 res_abs_success_builtin(Sv,Sg,CArgs,CTypes,SArgs),
	 asub(Succ,STypes,SArgs)
	;
	    Succ = '$bottom'
	).


%------------------------------------------------------------------------%
% res_plai_input_interface(+,+,+,-)
% res_plai_input_interface(InputUser,Kind,StructI,StructO)
%------------------------------------------------------------------------%

res_plai_input_interface(InputUser,Kind,StructI,StructO):-
	( nonvar(Kind)
	-> KTypes=Kind, KRes=Kind
	 ; true ),
	asub(StructI,ITypes,IRes),
	etermsvar_input_interface_(InputUser,KTypes,ITypes,OTypes),
	%det_input_interface_(InputUser,KDet,IDet,ODet),
	%nf_input_interface_(InputUser,KNf,INf,ONf),
	res_abs_input_interface_(InputUser,KRes,IRes,ORes),
	!,
	asub(StructO,OTypes,ORes).

etermsvar_input_interface_(InputUser,Kind,ITypes,OTypes):-
	etermsvar_input_interface(InputUser,Kind,ITypes,OTypes), !.
etermsvar_input_interface_(_InputUser,_Kind,ITypes,ITypes).

% det_input_interface_(InputUser,Kind,ITypes,OTypes):-
% 	det_input_interface(InputUser,Kind,ITypes,OTypes), !.
% det_input_interface_(_InputUser,_Kind,ITypes,ITypes).

% nf_input_interface_(InputUser,Kind,ITypes,OTypes):-
% 	nf_input_interface(InputUser,Kind,ITypes,OTypes), !.
% nf_input_interface_(_InputUser,_Kind,ITypes,ITypes).

res_abs_input_interface_(InputUser,Kind,IArgs,OArgs):-
	res_abs_input_interface(InputUser,Kind,IArgs,OArgs), !.
res_abs_input_interface_(_InputUser,_Kind,IArgs,IArgs).

%------------------------------------------------------------------------%
% res_plai_input_user_interface(+,+,-,+)
% res_plai_input_user_interface(InputUser,Qv,ASub,Sg)
%------------------------------------------------------------------------%

res_plai_input_user_interface(Struct,Qv,ASub,Sg,CallInfo):-
	% Special version with specific calling information
	( var(Struct)
	-> !, Types = _, Res = _
	 ; asub(Struct,Types,Res)
	),
	etermsvar_input_user_interface(Types,Qv,ATypes),
	%det_input_user_interface(Det,Qv,ADet),
	%nf_input_user_interface(Nf,Qv,ANf),
	( simplify_step2 ; true ),
	% Two cases: call or success substitution
	( % If no Res or resources information is missing...
	  ( var(Res) ; ( Res = (A,B,C), var(A), var(B), var(C) ) )
	-> % ... then assume call
	   ( is_there_var(ATypes)
	   -> get_call_and_proj(Sg,etermsvar,ATypes,_ProjT,PrimeT0),
	      cannonical_types(PrimeT0,PrimeT)
	      %get_call_and_proj(Sg,det,ADet,_ProjD,PrimeD),
	      %get_call_and_proj(Sg,nf,ANf,_ProjN,PrimeN)
	    ; PrimeT = ATypes,
	      %get_subgoal(Struct,Sg),
	      %get_absint(pred,det,Sg,(_,PrimeD)),
	      %get_absint(pred,nf,Sg,(_,PrimeN))
	      get_call_and_proj(Sg,nf,ATypes,_,PrimeN),
	      get_call_and_proj(Sg,det,ATypes,_,PrimeD)
	   ),
	   PrimeD = det(_,_,DetInfo),
	   PrimeN = nf(_,_,NfInfo),
	   % term_variables(Qv,Sv),
	   res_abs_call_to_entry(Qv,Sg,Qv,Sg,_,_Fv,_Proj,ARes,_Extra,ATypes,ATypes,PrimeT,DetInfo,NfInfo,[])
	 ; % ... else assume a final substitution
	   asub(CallInfo,CallT0,_), % get_call_and_proj(Sg,etermsvar,_,CallT0,ATypes),
	   % Hack for ground substitutions
	   ( CallT0 = [] -> CallT = ATypes ; CallT = CallT0 ),
	   %ADet = det(_,_,DetInfo),
	   %ANf = nf(_,_,NfInfo),
	   res_abs_input_user_interface(Res,Qv,CallT,ATypes,DetInfo,NfInfo,ARes,Sg)
	),
	!, asub(ASub,ATypes,ARes).
res_plai_input_user_interface(Struct,Qv,ASub,Sg):-
	( var(Struct)
	-> !, Types = _,  Res = _
	 ; asub(Struct,Types,Res)
	),
	etermsvar_input_user_interface(Types,Qv,ATypes),
	%det_input_user_interface(Det,Qv,ADet),
	%nf_input_user_interface(Nf,Qv,ANf),
	( simplify_step2 ; true ),
	% Two cases: call or success substitution
	( % If no Res or resources information is missing...
	  ( var(Res) ; ( Res = (A,B,C), var(A), var(B), var(C) ) )
	-> % ... then assume call
	   ( is_there_var(ATypes)
	   -> get_call_and_proj(Sg,etermsvar,ATypes,_ProjT,PrimeT0),
	   cannonical_types(PrimeT0,PrimeT),
	   %get_absint(pred,det,Sg,(_,PrimeD)),
	   %get_absint(pred,nf,Sg,(_,PrimeN))
	      get_call_and_proj(Sg,det,ProjT,_ProjD,PrimeD),
	      get_call_and_proj(Sg,nf,ProjT,_ProjN,PrimeN)
	    ; PrimeT = ATypes,
	      PrimeD = ADet,
	      PrimeN = ANf
	   ),
	   PrimeD = det(_,_,DetInfo),
	   PrimeN = nf(_,_,NfInfo),
	   % term_variables(Qv,Sv),
	   res_abs_call_to_entry(Qv,Sg,Qv,Sg,_,_Fv,_Proj,ARes,_Extra,ATypes,ATypes,PrimeT,DetInfo,NfInfo,[])
	 ; % ... else assume a final substitution
	    get_call_and_proj(Sg,etermsvar,_,CallT0,ATypes0),
	    cannonical_types(ATypes0,ATypes),
	   % Hack for ground substitutions
	   ( CallT0 = [] -> CallT = ATypes ; CallT = CallT0 ),
	     %get_subgoal(Struct,Sg),
	     %get_absint(pred,nf,Sg,(_,ANf)),
	     %get_absint(pred,det,Sg,(_,ADet)),
	     get_call_and_proj(Sg,nf,CallT,_,ANf),
	     get_call_and_proj(Sg,det,CallT,_,ADet),
	     ADet = det(_,_,DetInfo),
	     ANf = nf(_,_,NfInfo),
	   res_abs_input_user_interface(Res,Qv,CallT,ATypes,DetInfo,NfInfo,ARes,Sg)
	),
	!, asub(ASub,ATypes,ARes).
res_plai_input_user_interface(_Struct,_Qv,'$bottom',_Sg) :-
	true. %warning_message("input user interface: ~q, ~q, ~q",[Struct,Qv,Sg]). %, fail.

is_there_var([_:(_,T)|_]) :-
	dz_equivalent_types(T,vr).
is_there_var([_|R]) :-
	is_there_var(R).

%------------------------------------------------------------------------%
% res_plai_asub_to_native(+,+,+,-,-)
% res_plai_asub_to_native(ASub,Qv,Flag,Stat,Comp)
%------------------------------------------------------------------------%

res_plai_asub_to_native('$bottom',_,_,[],[]).
res_plai_asub_to_native('$top',_,_,[],[]).
res_plai_asub_to_native(ASub,Qv,Flag,Props,Comp):-
	asub(ASub,ATypes,AArgs),
	%AArgs = res_abs(_,_,_,_,_,_,_,_,IsDet,IsFail,_,_,_,_),
	etermsvar_asub_to_native(ATypes,Flag,EtermsProps),
	%get_subgoal(ASub,Sg),
	%get_call_and_proj(Sg,nf,ATypes,_,ANf),
	%get_call_and_proj(Sg,det,ATypes,_,ADet),
	%ADet = det(_,_,DetInfo),
	%detabs:det_asub_to_native(DetInfo,Qv,DetProps),
	%ANf = nf(_,_,NfInfo),
	%nfabs:nf_asub_to_native(NfInfo,Qv,NfProps),
	%(IsDet = true -> DetProps = is_det(Qv))
	res_abs_asub_to_native(AArgs,Qv,ArgProps,Comp,EtermsProps),
	append(EtermsProps,ArgProps,Props).
	%append(DetProps,NfProps,DetNfProps),
	%append(DetNfProps,ArgComp,Comp).

%------------------------------------------------------------------------%
% res_plai_unknown_call(+,+,-)
% res_plai_unknown_call(Vars,Call,Succ)
%------------------------------------------------------------------------%

res_plai_unknown_call(_,'$bottom','$bottom').
res_plai_unknown_call(_,'$top','$top').
res_plai_unknown_call(Vars,Call,Succ):-
	asub(Call,CTypes,CArgs),
	etermsvar_unknown_call(CTypes,Vars,STypes),
	%det_unknown_call(Vars,CDet,SDet),
	%nf_unknown_call(Vars,CNf,SNf),
	res_abs_unknown_call(Vars,CArgs,SArgs),
	asub(Succ,STypes,SArgs).


%------------------------------------------------------------------------%
% res_plai_unknown_entry(+,-)
% res_plai_unknown_entry(Vars,Entry)
%------------------------------------------------------------------------%

res_plai_unknown_entry(Vars,Entry):-
	etermsvar_unknown_entry(Vars,ETypes),
	%det_unknown_entry(Vars,EDet),
	%nf_unknown_entry(Vars,ENf),
	res_abs_unknown_entry(Vars,ETypes,EArgs),
	asub(Entry,ETypes,EArgs).

%------------------------------------------------------------------------%
% res_plai_empty_entry(+,-)
% res_plai_empty_entry(Vars,Entry)
%------------------------------------------------------------------------%

res_plai_empty_entry(Vars,Entry):-
	etermsvar_empty_entry(Vars,ETypes),
	%det_empty_entry(Vars,EDet),
	%nf_empty_entry(Vars,ENf),
	res_abs_empty_entry(Vars,ETypes,EArgs),
	asub(Entry,ETypes,EArgs).

res_plai_collect_abstypes(ASub,Types0,Types):-
        asub(ASub,ATypes,_),
        etermsvar_collect_abstypes(ATypes,Types0,Types).

res_plai_rename_abs(ASub,Types,Names,RenASub):-
        asub(ASub,ATypes,AArgs),
        etermsvar_rename_abs(ATypes,Types,Names,RenATypes),
        asub(RenASub,RenATypes,AArgs).



% :- pred get_all_resources_modules(ResourceModules) :: list(gnd) # "This
% 	predicate returns all modules containing the modules that defines
% 	the cost functions dinamically.  We apply sort to avoid duplicates".

% get_all_resources_modules(ResourceModules) :-
% 	findall(R, source_clause(_Key, directive(load_resource_definition(R)),
% 		_Dict), ResourceModules0),
% 	sort(ResourceModules0, ResourceModules).


init_abstract_domain:-
	res_abs:set_rflags,
	solver:init_solver,
	res_plai:load_resources_modules.


:- pred load_resources_modules # "Loads dinamically all the resource
	modules specified in the module being analyzed, using the directive
	load_resource_module".

load_resources_modules :-
	clean_all_caches,
	resources_lib:load_resources_modules.
	% get_all_resources_modules(ResourceModules),
	% process_resource_modules(ResourceModules).

% process_resource_modules([]).
% process_resource_modules([A|R]) :-
% 	process_resource_module(A),
% 	process_resource_modules(R).

% process_resource_module(ResourceModule) :-
% 	curr_file(CurrFile, _),
% 	absolute_file_name(CurrFile, '_opt', '.pl', '.', _, _, BaseDir),
% 	absolute_file_name(ResourceModule, '_opt', '.pl', BaseDir, FileName,
% 	    _FileBase, _Dir),
% 	use_module(FileName).
% 	%atom_concat([Dir, '/', Module], FileBase),
% 	%(init_resource(Module) -> true ; true).
