:- module(sized_types,
	[ sized_types_call_to_entry/9,
	  sized_types_exit_to_prime/7,
	  sized_types_project/4,
	  sized_types_extend/5,
	  sized_types_widen/3,
	%  sized_types_widencall/3,
	  sized_types_compute_lub/2,
          sized_types_compute_clauses_lub/3,
	  sized_types_glb/3,
	  sized_types_less_or_equal/2,
	  sized_types_identical_abstract/2,
	  sized_types_sort/2,
	  sized_types_call_to_success_fact/9,
	  sized_types_split_combined_domain/3,
	  sized_types_special_builtin/4,
	%  sized_types_success_builtin/5,
	  sized_types_call_to_success_builtin/6,
	  sized_types_input_interface/4,
	  sized_types_input_user_interface/4,
	  sized_types_input_user_interface/5,
	  sized_types_asub_to_native/5,
	  sized_types_unknown_call/3,
	  sized_types_unknown_entry/2,
	  sized_types_empty_entry/2,
	  sized_types_collect_abstypes/3,
	  sized_types_rename_abs/4,
	  load_resources_modules/0,
	  init_abstract_domain/0
	],
	[ assertions,regtypes,basicmodes,datafacts
	]).

:- use_module(ciaopp(plai/domains)).
:- use_module(library(lists)).
:- use_module(library(messages), [warning_message/2]).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(terms_vars)).
:- use_module(domain(etermsvar)).
:- use_module(domain(resources/sized_types_abs)).
:- use_module(domain(resources/eq_solver/solver),[init_solver/0]).
:- use_module(typeslib(typeslib), [dz_type_included/2]).
:- import(resources_lib,[load_resources_modules/0]).
% Loading resource modules
:- use_module(library(compiler), [use_module/1]).
:- use_module(ciaopp(p_unit/itf_base_db), [curr_file/2]).
:- use_module(sized_types_aux_ai_info).


:- doc(module,
	"Sized types analysis. In order to run this analysis, it
	is necessary to run etermsvar and det analysis beforehand, and
	the vers transformation.").



%:- reexport(resources(res_assrt_defs/resources_lib),[load_resources_modules/0]).

% TODO: max and min expressions of sizes in append are not working now
% It is giving correct but imprecise size bounds. Correct it.
% TODO: lower bounds of coupled are incorrect. Correct it.
% TODO: Check why '?' appears in expressions and avoid its use
% or handle it properly (e.g. listfact does not work because simplify
% fails with '?').
% TODO: Implement factorial example with accumulators and add it
% to examples directory. 

:- use_module(library(messages), [warning_message/2, debug_message/2]).
:- multifile issue_debug_messages/1.
:- data issue_debug_messages/1.
%%%%%%%%%%%%%%%%%%%%Uncomment for debugging a module%%%%%%%%%%%%%%%%%%%%%
% issue_debug_messages(sized_types_abs).
% issue_debug_messages(sized_types).
% issue_debug_messages(eq_solver).
% issue_debug_messages(pred_solver).
% issue_debug_messages(eq_solver_mathematica).
% issue_debug_messages(diff_equ_res).
% issue_debug_messages(second_order_res).
% issue_debug_messages(fixpo_plai).

asub(sized_types(Types,Args),Types,Args).



%------------------------------------------------------------------------%
% sized_types_call_to_entry(+,+,+,+,+,+,-,-)
% sized_types_call_to_entry(Sv,Sg,Hv,Head,Fv,Proj,Entry,ExtraInfo)
%------------------------------------------------------------------------%
sized_types_call_to_entry(Sv,Sg,Hv,Head,K,Fv,Proj,Entry,(K,HvFv,ProjT,PrimeT,ExtraInfoArgs)):-
	asub(Proj,_,PArgs),
	append(Hv,Fv,HvFv),
	get_tests_from_det(K,HvFv,T),
	get_entry_types(K,Sg,Head,Hv,Fv,ETypes,ProjT,PrimeT),
	( ETypes = '$bottom'
	-> Entry = '$bottom';
	   upgrade_numbers(ProjT,ProjT0),
	   sized_types_abs_call_to_entry(Sv,Sg,Hv,Head,K,Fv,PArgs,EArgs,ExtraInfoArgs,ProjT0,ETypes,PrimeT,T),
	   asub(Entry,ETypes,EArgs)
	).

upgrade_numbers([],[]).
upgrade_numbers([V:(N,T)|R],[V:(N,T2)|UR]) :-
	( dz_type_included(T,num)
	-> T2 = num
	 ; T2 = T
	),
	upgrade_numbers(R,UR).

%------------------------------------------------------------------------%
% sized_types_exit_to_prime(+,+,+,+,+,-,-)
% sized_types_exit_to_prime(Sg,Hv,Head,Sv,Exit,ExtraInfo,Prime)
%------------------------------------------------------------------------%

sized_types_exit_to_prime(_Sg,_Hv,_Head,_Sv,'$bottom',_ExtraInfo,'$bottom'):- !.
sized_types_exit_to_prime(_Sg,_Hv,_Head,_Sv,'$top',_ExtraInfo,'$top'):- !.
sized_types_exit_to_prime(Sg,Hv,Head,Sv,Exit,(K,_Vars,_ProjT,PrimeT,ExtraInfoArgs),Prime):-
	asub(Exit,_,EArgs),
	get_exit_types(K,Head,Hv,_,ETypes),
	sized_types_abs_exit_to_prime(Sg,Hv,Head,Sv,EArgs,ExtraInfoArgs,PArgs,ETypes,PrimeT),
	( PArgs = '$bottom'
	-> Prime = '$bottom'
	 ; asub(Prime,PrimeT,PArgs)
	).

%------------------------------------------------------------------------%
% sized_types_project(+,+,-)
% sized_types_project(ASub,Vars,Proj)
%------------------------------------------------------------------------%

sized_types_project('$bottom',_Sg,_Vars,'$bottom'):- !.
sized_types_project('$top',_Sg,_Vars,'$top'):- !.
sized_types_project(ASub,Sg,Vars,Proj):-
	asub(ASub,ATypes,AArgs),
	etermsvar_sort(ATypes,ATypesS),
	etermsvar_project(Vars,ATypesS,PTypes),
	upgrade_numbers(PTypes,PTypes0),
	sized_types_abs_project(AArgs,Sg,Vars,PArgs,PTypes0),
	asub(Proj,PTypes,PArgs).

%------------------------------------------------------------------------%
% sized_types_extend(+,+,+,-)
% sized_types_extend(Prime,Sv,Call,Succ)
%------------------------------------------------------------------------%

sized_types_extend(_Sg,'$bottom',_Sv,_Call,'$bottom'):- !.
sized_types_extend(_Sg,'$top',_Sv,_Call,'$top'):- !.
sized_types_extend(Sg,Prime,Sv,Call,Succ):-
	asub(Prime,PTypes1,PArgs),
	asub(Call,CTypes1,CArgs),
	etermsvar_sort(PTypes1, PTypes),
	etermsvar_sort(CTypes1, CTypes),
	etermsvar_extend(PTypes,Sv,CTypes,STypes),
	upgrade_numbers(CTypes,CTypes0),
	upgrade_numbers(STypes,STypes0),
	sized_types_abs_extend(PArgs,Sg,Sv,CArgs,SArgs,CTypes0,STypes0),
	asub(Succ,STypes,SArgs).

%------------------------------------------------------------------------%
% sized_types_widen(+,+,-)
% sized_types_widen(ASub1,ASub2,ASub)
%------------------------------------------------------------------------%

sized_types_widen('$bottom',ASub2,ASub2).
sized_types_widen(ASub1,'$bottom',ASub1).
sized_types_widen(ASub1,ASub2,ASub) :-
	asub(ASub1,Types1,Args1),
	asub(ASub2,Types2,Args2),
	etermsvar_widen(Types1,Types2,Types),
	( Types == '$bottom'
	-> ASub = '$bottom'
	 ; 
	   sized_types_abs_widen(Args1,Args2,Args),
	   asub(ASub,Types,Args)
	).

%------------------------------------------------------------------------%
% sized_types_widencall(+,+,-)
% sized_types_widencall(ASub1,ASub2,ASub)
%------------------------------------------------------------------------%
% No widening done

%------------------------------------------------------------------------%
% sized_types_compute_lub(+,-)
% sized_types_compute_lub(ListASub,Lub)
%------------------------------------------------------------------------%

sized_types_compute_lub(ListASub,Lub):-
	remove_bottoms(ListASub,ListNoBottoms),
	( ListNoBottoms == []
	-> Lub = '$bottom'
	 ; split(ListNoBottoms,[L|LTypes],LArgs),
	   etermsvar_compute_lub([L|LTypes],LubTypes),
	   ( etermsvar_identical_abstract(L,LubTypes)
	   -> upgrade_numbers(LubTypes,LubTypes0),
	      sized_types_abs_compute_lub(LArgs,LubArgs,LubTypes0),
	      asub(Lub,LubTypes,LubArgs)
	    ; asub(Lub,LubTypes,'$top')
	   )
	).

split([],[],[]).
split(['$bottom'|ListASub],['$bottom'|LTypes],['$bottom'|LArgs]):-
	split(ListASub,LTypes,LArgs).
split(['$top'|ListASub],['$bottom'|LTypes],['$top'|LArgs]):-
	split(ListASub,LTypes,LArgs).
split([ASub|ListASub],[ATypes|LTypes],[AArgs|LArgs]):-
	asub(ASub,ATypes,AArgs),
	split(ListASub,LTypes,LArgs).

sized_types_split_combined_domain(ListASub,[LTypes,LArgs],[etermsvar,sized_types]):-
	( var(LTypes)
	-> split(ListASub,LTypes,_LArgs),
	   LArgs=ListASub
	 ; split_back(ListASub,LTypes,LArgs)
	).

split_back([],[],[]).
split_back([ASub|ListASub],[ATypes|LTypes],[ASubArgs|LArgs]):-
	( ATypes == '$bottom' -> ASub = '$bottom'
	; asub(ASubArgs,_ATypes,AArgs),
	  asub(ASub,ATypes,AArgs)
	),
	split_back(ListASub,LTypes,LArgs).

remove_bottoms([],[]).
remove_bottoms(['$bottom'|R],RemR) :-
	!, remove_bottoms(R,RemR).
remove_bottoms([E|R],[E|RemR]) :-
	!, remove_bottoms(R,RemR).

%------------------------------------------------------------------------%
% sized_types_compute_clauses_lub(+,-)
% sized_types_compute_clauses_lub(ListASub,Lub)
%------------------------------------------------------------------------%

sized_types_compute_clauses_lub(['$top'],_Proj,[]):- !.
sized_types_compute_clauses_lub(['$bottom'],_Proj,[]):- !.
sized_types_compute_clauses_lub([ASub],_Proj,[ASub]).

%------------------------------------------------------------------------%
% sized_types_glb(+,+,-)
% sized_types_glb(ASub0,ASub1,Glb)
%------------------------------------------------------------------------%

sized_types_glb('$bottom',_ASub2,'$bottom').
sized_types_glb(_ASub1,'$bottom','$bottom').
sized_types_glb(ASub1,ASub2,Glb) :-
	asub(ASub1,Types1,Args1),
	asub(ASub2,Types2,Args2),
	etermsvar_glb(Types1,Types2,Types),
	( Types = '$bottom'
	-> Glb = '$bottom'
	 ; 
	   upgrade_numbers(Types,Types0),
	   sized_types_abs_glb(Args1,Args2,Types0,Args),
	   asub(Glb,Types,Args)
	).

%------------------------------------------------------------------------%
% sized_types_less_or_equal(+,+)
% sized_types_less_or_equal(ASub0,ASub1)
%------------------------------------------------------------------------%

sized_types_less_or_equal('$bottom',_) :- !.
sized_types_less_or_equal(X,'$bottom') :- !, X == '$bottom'.
sized_types_less_or_equal(_,'$top') :- !.
sized_types_less_or_equal('$top',X) :- !, X == '$top'.
sized_types_less_or_equal(ASub0,ASub1) :-
	asub(ASub0,_ATypes0,AArgs0),
	asub(ASub1,_ATypes1,AArgs1),
	sized_types_abs_less_or_equal(AArgs0,AArgs1).

%------------------------------------------------------------------------%
% sized_types_identical_abstract(+,+)
% sized_types_identical_abstract(ASub1,ASub2)
%------------------------------------------------------------------------%

sized_types_identical_abstract('$bottom','$bottom'):- !.
sized_types_identical_abstract('$top','$top'):- !.
sized_types_identical_abstract(ASub0,ASub1):-
	asub(ASub0,_ATypes0,AArgs0),
	asub(ASub1,_ATypes1,AArgs1),
	sized_types_abs_identical_abstract(AArgs0,AArgs1).

%------------------------------------------------------------------------%
% sized_types_sort(+,-)
% sized_types_sort(ASub0,ASub1)
%------------------------------------------------------------------------%

sized_types_sort('$bottom','$bottom'):- !.
sized_types_sort('$top','$top'):- !.
sized_types_sort(ASub0,ASub1):-
	asub(ASub0,ATypes0,AArgs0),
	etermsvar_sort(ATypes0,ATypes1),
	sized_types_abs_sort(AArgs0,AArgs1),
	asub(ASub1,ATypes1,AArgs1).

%------------------------------------------------------------------------%
% sized_types_call_to_success_fact(+,+,+,+,+,+,-,-)
% sized_types_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ)
%-------------------------------------------------------------------------

sized_types_call_to_success_fact(_,_,_,_K,_,'$bottom',_,'$bottom','$bottom').
sized_types_call_to_success_fact(_,_,_,_K,_,'$top',_,'$top','$top').
sized_types_call_to_success_fact(Sg,Hv,Head,_K,Sv,Call,Proj,Prime,Succ):-
	asub(Call,CTypes,CArgs),
	asub(Proj,PTypes,PArgs),
	etermsvar_call_to_entry(Sg,Hv,Head,[],PTypes,ETypes,_ExtraInfoTypes),
	etermsvar_call_to_success_fact(Sg,Hv,Head,Sv,CTypes,PTypes,RTypes,STypes),
	sized_types_abs_call_to_success_fact(Sg,Hv,Head,Sv,CArgs,PArgs,RArgs,SArgs,PTypes,ETypes,RTypes,STypes),
	asub(Prime,RTypes,RArgs),
	asub(Succ,STypes,SArgs).

%-------------------------------------------------------------------------
% sized_types_special_builtin(+,+,-,-)
% sized_types_special_builtin(SgKey,Sg,SubgoalType,Condvars)
%-------------------------------------------------------------------------

sized_types_special_builtin(SgKey,_Sg,special(SgKey),_Condvars):-
	sized_types_abs_special_builtin(SgKey).

%-------------------------------------------------------------------------
% sized_types_call_to_success_builtin(+,+,+,+,+,-)
% sized_types_call_to_success_builtin(SgKey,Sg,Sv,Call,Proj,Succ)
%-------------------------------------------------------------------------

sized_types_call_to_success_builtin(_,_,_,'$bottom',_,'$bottom').
sized_types_call_to_success_builtin(_,_,_,'$top',_,'$top').
sized_types_call_to_success_builtin(SgKey,Sg,Sv,Call,Proj,Succ):-
	asub(Call,CTypes,CArgs),
	asub(Proj,PTypes,_),
	( etermsvar_special_builtin(SgKey,Sg,Sv,_TKey,_TCondVars),
	  etermsvar_call_to_success_builtin(SgKey,Sg,Sv,CTypes,PTypes,STypes)
	-> true
	 ; STypes = CTypes
	),
	sized_types_abs_success_builtin(Sv,Sg,CArgs,CTypes,SArgs),
	asub(Succ,STypes,SArgs).

%------------------------------------------------------------------------%
% sized_types_input_interface(+,+,+,-)
% sized_types_input_interface(InputUser,Kind,StructI,StructO)
%------------------------------------------------------------------------%

sized_types_input_interface(InputUser,Kind,StructI,StructO):-
	
	( nonvar(Kind)
	-> KRes=Kind
	 ; true ),
	asub(StructI,_ITypes,IRes),
	sized_types_abs_input_interface_(InputUser,KRes,IRes,ORes),
	asub(StructO,_OTypes,ORes).

sized_types_abs_input_interface_(InputUser,Kind,IArgs,OArgs):-
	sized_types_abs_input_interface(InputUser,Kind,IArgs,OArgs), !.
sized_types_abs_input_interface_(_InputUser,_Kind,IArgs,IArgs).

%------------------------------------------------------------------------%
% sized_types_input_user_interface(+,+,-,+)
% sized_types_input_user_interface(InputUser,Qv,ASub,Sg)
%------------------------------------------------------------------------%

sized_types_input_user_interface(Struct,Qv,ASub,Sg,CallInfo):-
	% Special version with specific calling information
	( var(Struct)
	-> !, Res = _
	 ; asub(Struct,_Types,Res)
	),
	% Two cases: call or success substitution
	( % If no Res or resources information is missing...
	  ( var(Res) ; ( Res = (A,B,C), var(A), var(B), var(C) ) )
	-> % ... then assume call
	get_complete_from_db(etermsvar,Sg,ATypes,PrimeT),
	sized_types_abs_call_to_entry(Qv,Sg,Qv,Sg,_,_Fv,_Proj,ARes,_Extra,ATypes,ATypes,PrimeT,[])
	 ; % ... else assume a final substitution
	   asub(CallInfo,CallT0,_), 
	   % Hack for ground substitutions
	   ( CallT0 = [] -> CallT = ATypes ; CallT = CallT0 ),
	   sized_types_abs_input_user_interface(Res,Qv,CallT,ATypes,ARes,Sg)
	),
	!, asub(ASub,ATypes,ARes).
sized_types_input_user_interface(Struct,Qv,ASub,Sg):-
	( var(Struct)
	-> !, Res = _
	 ; asub(Struct,_Types,Res)
	),
	% Two cases: call or success substitution
	( % If no Res or resources information is missing...
	  ( var(Res) ; ( Res = (A,B,C), var(A), var(B), var(C) ) )
	-> % ... then assume call
	get_complete_from_db(etermsvar,Sg,ATypes,PrimeT),
	sized_types_abs_call_to_entry(Qv,Sg,Qv,Sg,_,_Fv,_Proj,ARes,_Extra,ATypes,ATypes,PrimeT,[])
	 ; % ... else assume a final substitution
	    get_complete_from_db(etermsvar,Sg,CallT0,ATypes),
	   % Hack for ground substitutions
	   ( CallT0 = [] -> CallT = ATypes ; CallT = CallT0 ),
	   sized_types_abs_input_user_interface(Res,Qv,CallT,ATypes,ARes,Sg)
	),
	!, asub(ASub,ATypes,ARes).
sized_types_input_user_interface(_Struct,_Qv,'$bottom',_Sg) :-
	true. %warning_message("input user interface: ~q, ~q, ~q",[Struct,Qv,Sg]). %, fail.

%------------------------------------------------------------------------%
% sized_types_asub_to_native(+,+,+,-,-)
% sized_types_asub_to_native(ASub,Qv,Flag,Stat,Comp)
%------------------------------------------------------------------------%

sized_types_asub_to_native('$bottom',_,_,[],[]).
sized_types_asub_to_native('$top',_,_,[],[]).
sized_types_asub_to_native(ASub,Qv,Flag,Props,[]):-
	asub(ASub,ATypes,AArgs),
	etermsvar_asub_to_native(ATypes,Flag,EtermsProps),
	sized_types_abs_asub_to_native(AArgs,Qv,ArgProps,EtermsProps),
	append(EtermsProps,ArgProps,Props).

%------------------------------------------------------------------------%
% sized_types_unknown_call(+,+,-)
% sized_types_unknown_call(Vars,Call,Succ)
%------------------------------------------------------------------------%

sized_types_unknown_call(_,'$bottom','$bottom').
sized_types_unknown_call(_,'$top','$top').
sized_types_unknown_call(Vars,Call,Succ):-
	asub(Call,CTypes,CArgs),
	etermsvar_unknown_call(CTypes,Vars,STypes),
	sized_types_abs_unknown_call(Vars,CArgs,SArgs),
	asub(Succ,STypes,SArgs).

%------------------------------------------------------------------------%
% sized_types_unknown_entry(+,-)
% sized_types_unknown_entry(Vars,Entry)
%------------------------------------------------------------------------%

sized_types_unknown_entry(Vars,Entry):-
	etermsvar_unknown_entry(Vars,ETypes),
	sized_types_abs_unknown_entry(Vars,ETypes,EArgs),
	asub(Entry,ETypes,EArgs).

%------------------------------------------------------------------------%
% sized_types_empty_entry(+,-)
% sized_types_empty_entry(Vars,Entry)
%------------------------------------------------------------------------%

sized_types_empty_entry(Vars,Entry):-
	etermsvar_empty_entry(Vars,ETypes),
	sized_types_abs_empty_entry(Vars,ETypes,EArgs),
	asub(Entry,ETypes,EArgs).

sized_types_collect_abstypes(ASub,Types0,Types):-
        asub(ASub,ATypes,_),
        etermsvar_collect_abstypes(ATypes,Types0,Types).

sized_types_rename_abs(ASub,Types,Names,RenASub):-
        asub(ASub,ATypes,AArgs),
        etermsvar_rename_abs(ATypes,Types,Names,RenATypes),
        asub(RenASub,RenATypes,AArgs).

init_abstract_domain:-
	solver:init_solver,
	sized_types:load_resources_modules.


:- pred load_resources_modules # "Loads dinamically all the resource
	modules specified in the module being analyzed, using the directive
	load_resource_module".

load_resources_modules :-
	clean_all_caches,
	resources_lib:load_resources_modules.

