:- module(sized_types_abs,
	[ sized_types_abs_call_to_entry/13,
	  sized_types_abs_exit_to_prime/9,
	  sized_types_abs_project/5,
	  sized_types_abs_extend/7,
	  sized_types_abs_compute_lub/3,
	% sized_types_abs_compute_clauses_lub/4,
	  sized_types_abs_widen/3,
	  sized_types_abs_glb/4,
	  sized_types_abs_less_or_equal/2,
	  sized_types_abs_identical_abstract/2,
	  sized_types_abs_sort/2,
	  sized_types_abs_call_to_success_fact/12,
	  sized_types_abs_special_builtin/1,
	  sized_types_abs_success_builtin/5,
	% sized_types_abs_call_to_success_builtin/6,
	  sized_types_abs_input_interface/4,
	  sized_types_abs_input_user_interface/6,
	  sized_types_abs_asub_to_native/4,
	  sized_types_abs_unknown_call/3,
	  sized_types_abs_unknown_entry/3,
	  sized_types_abs_empty_entry/3,
	  clean_timers/0,
	  profiling_timer/2,
	  clean_all_caches/0,
	  get_last_replacements/4
	],
	[assertions,regtypes,basicmodes,datafacts]).

:- use_module(library(formulae),[conj_to_list/2]).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(system)).
:- use_module(typeslib(typeslib),
        [ dz_type_included/2, dz_equivalent_types/2, set_numeric_type/1 ]).
:- use_module(library(terms_check)).
:- use_module(library(terms_vars)).
:- use_module(ciaopp(preprocess_flags),[current_pp_flag/2]).
% For debugging purposes uncomment
%:- use_module(library(messages)).
%:- multifile issue_debug_messages/1.
%:- data issue_debug_messages/1.
%issue_debug_messages(sized_types_abs).
:- use_module(library(messages), [warning_message/2,debug_message/2,error_message/2]).
:- use_module(sized_types_aux_ai_info,[can_fail/2]). 

:- use_module(res_abs_aux,[
	paths/3,
	get_var_size_type/3,
	get_vars_size_types/3,
	find_variable_type/3,
	create_ground_size/3,
	unify_path/4,
	get_unification_eqs/3,
	initial_var/2,
	unify_back/5,
	expand_tests/2,
	find_wanted_tests/3,
	get_size_type/2,
	create_num_assignment/4,
	select_vars_size_types/3,
	normalize_rsize/3,
	beautify_equations/2,
	denormalize_rsize/2,
	clean_equation_names_cache/0,
	clean_internal_equiv_types/0,
	clean_size_type_cache/0
			  ]
	     ).

:- use_module(res_abs_utils,[
	get_second_components/2,
	zip/3,
	unifiable_append/3,
	up_to_renaming_append/3,
	up_to_renaming_subset_extra/4
			    ]).
:- use_module(domain(resources/eq_solver/eq_solver)).
:- use_module(domain(resources/eq_solver/pred_solver)).
:- use_module(domain(resources/eq_solver/eq_solver_term_position)).
:- use_module(domain(etermsvar), [etermsvar_project/3,etermsvar_call_to_success_builtin/6]).

% An abstract substitution is an 7-tuple:
% - Name is the name of the predicate (needed later to check for recursion)
% - Args is a tuple (IArgs,OArgs) where the input and output variables are specified
% - Clauses is a list of extended equations (as found in module pred_solve)
%   For a clause, it will be a singleton list, after the LUB is done, it is a list
% - AlreadySolved is a cache of the previous solutions to inner predicates
% - SizeMapping is a cache of the size types for each variable in the predicate
% - PrevCalled is used to implement widening: it saved information about
%   the calls in inner and the same predicates
% - Tests is a series of arithmetic comparisons
%   NOTE: Currently, Tests are not propagated upwards

asub(sized_types_abs(Name,Args,Clauses,Solution,AlreadySolved,SizeMapping,Tests),
	Name,Args,Clauses,Solution,AlreadySolved,SizeMapping,Tests).

:- data profiling_timer/2.
profiling_timer(solve,0).
profiling_timer(call_to_entry,0).
profiling_timer(extend,0).
profiling_timer(exit_to_prime,0).
profiling_timer(widen,0).
profiling_timer(input_user_interface,0).

clean_timers :-
	retractall_fact(profiling_timer(_,_)),
	asserta_fact(profiling_timer(solve,0)),
	asserta_fact(profiling_timer(call_to_entry,0)),
	asserta_fact(profiling_timer(extend,0)),
	asserta_fact(profiling_timer(exit_to_prime,0)),
	asserta_fact(profiling_timer(widen,0)),
	asserta_fact(profiling_timer(input_user_interface,0)).

update_timer(T1,N) :-
	time(T2),
	current_fact(profiling_timer(N,Prev)),
	retract_fact(profiling_timer(N,Prev)),
	New is Prev + T2 - T1,
	asserta_fact(profiling_timer(N,New)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PREDICATES FOR SETTING UP CLAUSES TESTS
:- pred set_clauses_tests(Clauses,SizeMapping,Paths,NewClauses) #
	"Given a list of clauses in @var{Clauses} and a size mapping in @var{SizeMapping},
	returns the list of clauses, with the tests of each clause in terms of lower and upper bound
         variables.".
set_clauses_tests([],_,_,[]).
set_clauses_tests([C|T],M,P,[NC|TT]):-
	set_clause_test(C,M,P,NC),!,
	set_clauses_tests(T,M,P,TT).
set_clause_test(clause(K,I,O,B,T),M,P,clause(K,I,O,B,TT)):-
	filter_app_tests(T,M,TT0),
	replace_head_variables(TT0,P,TT1),
	replace_variables_in_tests(TT1,M,TT).


replace_head_variables([],_,[]).
replace_head_variables([X|Y],P,[XX|YY]):-
	replace_head_variables_(X,P,XX),
	replace_head_variables(Y,P,YY).

replace_head_variables_(X,P,Y):-
	var(X),
	get_goal_var(X,P,Y),!.
replace_head_variables_(X,_P,X):-
	var(X).
replace_head_variables_(X,P,Y):-
	nonvar(X),
	X =.. [Op|Args],
	replace_head_variables(Args,P,NArgs),
	Y =.. [Op|NArgs].

get_goal_var(X,[A1=Y|_R],Y):-
	X == A1,!.
get_goal_var(X,[_T|TT],Y):-
	get_goal_var(X,TT,Y),!.


filter_app_tests([],_,[]).
filter_app_tests([T|TT],M,TTR):-
	\+ is_input_test(T,M),!,
	filter_app_tests(TT,M,TTR).
filter_app_tests([T|TT],M,[T|TTR]):-
	filter_app_tests(TT,M,TTR).


is_input_test(T,M):-
	var(T),
	get_var_size_type(M,T,_ST).
is_input_test(T,_M):-
	var(T),!,fail.
is_input_test(T,M):-
	nonvar(T),
	T =.. [_Op|ListT],
	is_input_test_args(ListT,M).

is_input_test_args([],_).
is_input_test_args([R|T],M):-
	is_input_test(R,M),
	is_input_test_args(T,M).

	
replace_variables_in_tests([],_,[]).
replace_variables_in_tests([Test|T],M,TT):-
	generate_bound_tests(Test,M,LTest,lb),
	generate_bound_tests(Test,M,UTest,ub),
	replace_variables_in_tests(T,M,TT0),
	(LTest == UTest -> TT = [LTest|TT0];TT=[LTest,UTest|TT0]).
	 
generate_bound_tests(Test,M,LTest,B):-
	var(Test),
	get_var_size_type(M,Test,STest),
	STest =.. TTTest,
	TTTest = [_|[Tuple]],
	choose_bound_var(Tuple,B,LTest).
generate_bound_tests(Test,_M,Test,_B):-
	num(Test).
generate_bound_tests(Test,M,LTest,B):-
	nonvar(Test),
	Test =.. [Op|ListT],
	generate_bound_tests_args(ListT,M,LListT,B),
	LTest =.. [Op|LListT].
generate_bound_tests(Test,_M,Test,_).

choose_bound_var((L,_),lb,L):-!.
choose_bound_var((_,U),ub,U):-!.

	
generate_bound_tests_args([],_,[],_).
generate_bound_tests_args([V|T],M,[VV|TT],B):-
	generate_bound_tests(V,M,VV,B),
	generate_bound_tests_args(T,M,TT,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% END OF PREDICATES FOR SETTING UP CLAUSES TESTS




%-----------------------------%
% solve_asub(+,-)
% solve_asub(ASub,SolvedASub)
%-----------------------------%
% Try to get the closed form of the equations in the abstract substitution
solve_asub(ASub,TrySolve,SolvedASub) :-
	time(T1),
	asub(ASub,F/N,(IArgs,OArgs),Clauses,_,Already,SizeMapping,Tests),
	( Clauses = [] % Empty set of clauses
	-> SolvedASub = ASub
	; % Get the sizes
	  get_size_mappings(IArgs,SizeMapping,ISizes),
	  get_size_mappings(OArgs,SizeMapping,OSizes),
	  % Add extra argument for solution
	  OSizesSolRes = OSizes,
	  Pred = predicate(F/N,ISizes,OSizesSolRes,Clauses),
	  % Solve the equations
	  NbResSol=0,
	  debug_message("CALLING PRED_SOLVER>>>~q ::: ~q",[Pred,SizeMapping]),
	  % TODO: The subgoal should be available here
	  append(IArgs,OArgs,Args),length(Args,Ar),
	  functor(Sg,F,Ar),
	  can_fail(Sg,CanFail),
	  pred_solve(Pred,Already,NbResSol,false,CanFail,TrySolve,SolutionPre),
	  pred_solve(Pred,[SolutionPre|Already],NbResSol,false,CanFail,TrySolve,Solution),
	  !,
	  % Take the solution
	  asub(SolvedASub,F/N,(IArgs,OArgs),Clauses,Solution,Already,SizeMapping,Tests),
	  % Include this as last replacement of that type
	  put_asub_last_replacement(SolvedASub,F/N)
	),
	update_timer(T1,solve).


% Extend each of the clause equations with new instructions at the end
extend_instructions([],_,[]).
extend_instructions([clause(K,ISizes,OSizes,Instr,T)|R],New,[clause(K,ISizes,OSizes,Ext,T)|ExtR]) :-
	append(Instr,New,Ext),
	extend_instructions(R,New,ExtR).

get_size_mappings([],_,[]).
get_size_mappings([E|R],Mappings,[SE|SR]) :-
	get_var_size_type(Mappings,E,SE),
	get_size_mappings(R,Mappings,SR).

%--------------------------------------------------------------------------------------------------%
% sized_types_abs_call_to_entry(+,+,+,+,+,+,-,-,+)
% sized_types_abs_call_to_entry(Sv,Sg,Hv,Head,Fv,Proj,Entry,ExtraInfo,CallTypes,PrimeTypes,PrimeDet,PrimeNf)
%--------------------------------------------------------------------------------------------------%

% Important!! This only works on destructuring predicates!!
sized_types_abs_call_to_entry(_Sv,Sg,Hv,Head,K,_Fv,Proj,Entry,_Extra,CallTypes,EntryTypes,PrimeTypes,Tests):-
	time(T1),
	
	% 1. Find which arguments are input and which output
	( var(Proj) -> PSizeMap = [] ; asub(Proj,_,_,_,_,_,PSizeMap,_) ),
	% TODO: take out accumulating parameters
	term_variables(Sg,SubgoalVars),
	choose_input_output(SubgoalVars,CallTypes,IArgs,OArgs),
	choose_input_output(Hv,EntryTypes,HIArgs,_HOArgs),
	% 2. Generate the size types
	generate_initial_size_mappings(IArgs,CallTypes,OArgs,PrimeTypes,HIArgs,EntryTypes,ISizes,OSizes,HSizes,PSizeMap,SizeMapping),
	% 3. Generate deconstruction for this clause
	paths(Sg,Head,Paths),
	get_input_arguments(IArgs,CallTypes,Paths,ISizes,RealIArgs),
	( unify_paths(HIArgs,HSizes,ISizes,CallTypes,Paths,SizeUnifAssigns)
	; SizeUnifAssigns = [] % When called from unknown
	),
	% 4. Get the name of the predicate
        functor(Sg,Name,_), % Now we generate a new random name
	get_next_name(Name,FullName),
	% 5. Get output arguments and add extra for number of solutions and resources
	get_second_components(OSizes,RealOArgs),
	RealOArgsSolRes = RealOArgs,
	% 7. Initialize the clause
	set_clauses_tests([clause(K,RealIArgs,RealOArgsSolRes,SizeUnifAssigns,Tests)],SizeMapping,Paths,[Clause]),
	debug_message("~q(~q)(~q)",[Clause,K,Tests]),
	% 8. Create abstract substitution and extra information
	!,
	asub(Entry,FullName,(IArgs,OArgs),[Clause],'$unknown',[],SizeMapping,[]),
	update_timer(T1,call_to_entry).

% A parameter is chosen as output if it has type vr
% If not, it is chosen as input
choose_input_output([],_,[],[]).
choose_input_output([Sv|Svars],Types,IArgs,OArgs) :-
	% Choose the rest
	choose_input_output(Svars,Types,IRest,ORest),
	% Find this type
	find_variable_type(Types,Sv,SvType),
	( dz_equivalent_types(SvType,vr)
	-> IArgs = IRest, OArgs = [Sv|ORest]
	 ; IArgs = [Sv|IRest], OArgs = ORest
	).

% Generate the initial mappings for the subgoal and head variables
generate_initial_size_mappings(IArgs,CallTypes,OArgs,PrimeTypes,HIArgs,EntryTypes,ISizes,OSizes,HSizes,PSizeMap,SizeMapping) :-
	get_vars_size_types(IArgs,CallTypes,ISizes0),
	get_vars_size_types(OArgs,PrimeTypes,OSizes0),
	get_vars_size_types(HIArgs,EntryTypes,HSizes0),
	% Merge the sizes
	merge_size_types(ISizes0,PSizeMap,PSizeMap2),
	merge_size_types(OSizes0,PSizeMap2,PSizeMap3),
	merge_size_types(HSizes0,PSizeMap3,SizeMapping),
	% Get the real ones
	get_size_mappings(IArgs,SizeMapping,ISizes1),
	zip(IArgs,ISizes1,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes1),
	zip(OArgs,OSizes1,OSizes),
	get_size_mappings(HIArgs,SizeMapping,HSizes1),
	zip(HIArgs,HSizes1,HSizes).

% Generate the input arguments for the clause
% They will be variables except for the case when a variable
% is unified with a ground argument, in which case a constant
% will be included as input
get_input_arguments([],_,_,_,[]).
get_input_arguments([IArg|IRest],Types,Paths,ISizes,[CArg|CRest]) :-
	get_var_size_type(ISizes,IArg,IArgSize),
	( find_ground_unification(Paths,IArg,GroundExpr)
	-> find_variable_type(Types,IArg,IType),
	   create_ground_size(GroundExpr,IType,CArg)
	 ; CArg = IArgSize
	),
	get_input_arguments(IRest,Types,Paths,ISizes,CRest).

find_ground_unification([X=Expr|_],Y,Expr) :-
	X == Y, !, ground(Expr).
find_ground_unification([_|R],Y,Expr) :-
	find_ground_unification(R,Y,Expr).

% Generates the equations needed to go from the variables in
% the goal to the variables in the head of the clause
unify_paths([],_,_,_,_,[]).

unify_paths([HArg|HRest],HISizes,ISizes,Types,Paths,[Assign|AssR]) :-
	% Get the variable in the subgoal for which the unification is done
	find_head_path(Paths,HArg,HPath),
	initial_var(HPath,SgVar),
	get_var_size_type(ISizes,SgVar,SgSize), !,
	find_variable_type(Types,SgVar,SgType),
	% Compute the path size type
	unify_path(SgType,SgSize,HPath,Out),
	% Create the call
	get_var_size_type(HISizes,HArg,HSize),
	Assign = assign(HSize,Out),
	% Do the rest
	unify_paths(HRest,HISizes,ISizes,Types,Paths,AssR).
unify_paths([HArg|HRest],HISizes,ISizes,Types,Paths,AssR) :-
	\+ find_head_path(Paths,HArg,_),
	unify_paths(HRest,HISizes,ISizes,Types,Paths,AssR).

find_head_path([X=Y|_],W,Y) :- W == X.
find_head_path([Y=X|_],W,Y) :- W == X.
find_head_path([X=Path|_],W,Path) :- W == X.
find_head_path([_|R],W,Path) :- find_head_path(R,W,Path).

%-------------------------------------------------------------------------------%
% sized_types_abs_exit_to_prime(+,+,+,+,+,-,-,+,+)
% sized_types_abs_exit_to_prime(Sg,Hv,Head,Sv,Exit,ExtraInfo,Prime,ExitTypes,PrimeTypes)
%-------------------------------------------------------------------------------%
%:- use_module(ciaopp(plai/static_profiling_info_tracker)).
sized_types_abs_exit_to_prime(Sg,_Hv,Head,_Sv,Exit,_Extra,Prime,ExitTypes,PrimeTypes):-
	time(T1),
	asub(Exit,Name,(IArgs,OArgs),Clauses,ByNowSol,AlreadySolved,SizeMapping,Tests),
	% 1. Generate backwards unification
	get_unification_eqs(Sg,Head,Eqs),
	unify_back_s(OArgs,Eqs,SizeMapping,PrimeTypes,ExitTypes,AssignOut),
	extend_instructions(Clauses,AssignOut,Clauses2),
	QuasiFinalClauses = Clauses2,
	% 3. Take tests into account for constraining domain
	filter_unif_tests(Eqs,EqTests),
	append(EqTests,Tests,AllTests),
	obtain_tests(IArgs,AllTests,ETests),
	( ETests = []
	-> FinalClauses = QuasiFinalClauses
	 ; explode_final_clauses_with_tests(PrimeTypes,SizeMapping,ETests,QuasiFinalClauses,FinalClauses)
	),
	asub(PrimeNotSolved,Name,(IArgs,OArgs),FinalClauses,ByNowSol,AlreadySolved,SizeMapping,AllTests),
	update_timer(T1,exit_to_prime),
	get_solver_behavior(TrySolve),
	solve_asub(PrimeNotSolved,TrySolve,Prime).

% Generates the final equations to move from the clause output variables
% to the output variables in the goal
unify_back_s([],_,_,_,_,[]).
unify_back_s([Arg|R],Eqs,SizeMapping,PrimeTypes,ExitTypes,[AssArg|AssR]) :-
	find_head_path(Eqs,Arg,Components),
	find_variable_type(PrimeTypes,Arg,ArgT),
	unify_back(ArgT,Components,SizeMapping,ExitTypes,AssSize),
	get_var_size_type(SizeMapping,Arg,ISize),
	AssArg = assign(ISize,AssSize),
	unify_back_s(R,Eqs,SizeMapping,PrimeTypes,ExitTypes,AssR).

filter_unif_tests([],[]).
filter_unif_tests([X=Y|R],[X =:= Y|S]) :-
	var(X), var(Y),
	!, filter_unif_tests(R,S).
filter_unif_tests([_|R],S) :-
	!, filter_unif_tests(R,S).

obtain_tests(Hv,ArithTests,Tests) :-
	expand_tests(ArithTests,ExpTests),
	find_wanted_tests(Hv,ExpTests,Tests).

explode_final_clauses_with_tests(_,_,[],_,[]).
explode_final_clauses_with_tests(Ty,Map,[E =:= G|R],Clauses,FinalClauses) :-
	explode_final_clauses_with_tests_(Ty,Map,E,G,Clauses,Clauses1),
	explode_final_clauses_with_tests(Ty,Map,R,Clauses,Clauses2),
	append(Clauses1,Clauses2,FinalClauses).
explode_final_clauses_with_tests(Ty,Map,[E =< G|R],Clauses,FinalClauses) :-
	range(E,G,S), append(S,R,T), % Convert into a list of equalities
	explode_final_clauses_with_tests(Ty,Map,T,Clauses,FinalClauses).
explode_final_clauses_with_tests(Ty,Map,[E < G|R],Clauses,FinalClauses) :-
	G1 is G-1, range(E,G1,S), append(S,R,T), % Convert into a list of equalities
	explode_final_clauses_with_tests(Ty,Map,T,Clauses,FinalClauses).


range(_,N,[])          :- N < 0, !.
range(E,N,[E =:= N|R]) :- N >= 0, !, N1 is N - 1, range(E,N1,R).

explode_final_clauses_with_tests_(_,_,_,_,[],[]).
explode_final_clauses_with_tests_(Ty,Map,E,G,[clause(K,IArgs,OArgs,Body,TT)|R],[clause(K,NewIArgs,OArgs,NewBody,TT)|S]) :-
	get_var_size_type(Map,E,ESize),
	find_variable_type(Ty,E,EType),
	create_ground_size(G,EType,GSize),
	% Change in input args
	explode_input_args(IArgs,ESize,GSize,NewIArgs),
	% TODO: Change in body
	NewBody = Body,
	explode_final_clauses_with_tests_(Ty,Map,E,G,R,S).

explode_input_args([],_,_,[]).
explode_input_args([E|R],ESize,GSize,[GSize|S]) :-
	E == ESize, !,
	explode_input_args(R,ESize,GSize,S).
explode_input_args([E|R],ESize,GSize,[E|S]) :-
	E \== ESize, !,
	explode_input_args(R,ESize,GSize,S).

%------------------------------------------------------------------------%
% sized_types_abs_project(+,+,-,+)
% sized_types_abs_project(ASub,Vars,Proj,ProjTypes)
%------------------------------------------------------------------------%

sized_types_abs_project(_ASub,Sg,_Vars,Proj,ProjTypes) :-
	% 1. Generate name
	functor(Sg,Name,_),
	get_next_name(Name,FullName), !,
	% 2. Choose input and output
	term_variables(Sg,Vars),
	choose_input_output(Vars,ProjTypes,IArgs,OArgs),
	!,
	asub(Proj,FullName,(IArgs,OArgs),[],'$unknown',[],[],[]).


%------------------------------------------------------------------------%
% sized_types_abs_extend(+,+,+,-,+,+)
% sized_types_abs_extend(Prime,Sv,Call,Succ,CallTypes,SuccTypes,NfInfo)
%------------------------------------------------------------------------%

sized_types_abs_extend(Prime,Sg,Sv,Call,Succ,CallTypes,SuccTypes):-
	time(T1),
	
	% Extract information from abstract substitutions
	asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	asub(Prime,_,(PIArgs,POArgs),_,PrimeSol,AlreadySolvedP,_,_),
	% 1. Extend size mapping
	get_vars_size_types(Sv,SuccTypes,SvSizes),
	merge_size_types(SvSizes,SizeMappingC,NewMapping),
	( PrimeSol = resolution(ResName,_,_,_)
	-> ResolutionToUse = PrimeSol, ResIArgs = PIArgs, ResOArgs = POArgs
	 ; debug_message("Fake resolution for ~q ~q",[Sg,CallTypes]),
	   generate_fake_resolution(Sg,CallTypes,NewMapping,ResIArgs,ResOArgs,ResolutionToUse,ResName)
	),
	% 3. Append new call to list of instructions
	get_size_mappings(ResIArgs,NewMapping,SIArgs),
	get_size_mappings(ResOArgs,NewMapping,SOArgs),
	SOArgs = SOArgsSolRes,
	extend_instructions(Clauses,[call(ResName,SIArgs,SOArgsSolRes)],Clauses2),
	Clauses2 = NewClauses,
	% 5. Extend the list of already solved equations and visited nodes
	unifiable_append(AlreadySolvedC,[ResolutionToUse|AlreadySolvedP],NewAlreadySolved),
	% 6. Create new abstract substitution
	!,
	asub(Succ,Name,Args,NewClauses,ByNowSol,NewAlreadySolved,NewMapping,Tests),
	update_timer(T1,extend).




% Generate a call to a new empty equation when the information is unknown
% This is needed to create recursion
generate_fake_resolution(Sg,CallTypes,SizeTypes,IArgs,OArgs,Resolution,NewName) :-
	term_variables(Sg,SubgoalVars),
	% Create input and output parameters
	choose_input_output(SubgoalVars,CallTypes,IArgs,OArgs),
	get_size_mappings(IArgs,SizeTypes,SizeI),
	get_size_mappings(OArgs,SizeTypes,SizeO),
	SizeOSolRes = SizeO,
	% Get a new name
	functor(Sg,F,_),
	get_next_name(F,NewName), !,
	% Fake-solve it
	Pred = predicate(NewName,SizeI,SizeOSolRes,[clause(_,SizeI,SizeOSolRes,[call(NewName,SizeI,SizeOSolRes)],[])]),
	get_solver_behavior(TrySolve),
	pred_solve(Pred,[],0,true,false,TrySolve,Resolution).

% Merge the size mapping of a clause with the success of a literal
merge_size_types([],V,V).
merge_size_types([(V,VSize)|R],Already,Merged) :-
	( get_var_size_type(Already,V,_)
	-> merge_size_types(R,Already,Merged)
	 ; merge_size_types(R,[(V,VSize)|Already],Merged)
	).
%------------------------------------------------------------------------%
% sized_types_abs_compute_lub(+,-,+)
% sized_types_abs_compute_lub(ListASub,Lub,LubTypes)
%------------------------------------------------------------------------%

sized_types_abs_compute_lub([],'$bottom',_) :- !.
sized_types_abs_compute_lub(ListASub,Lub,_LubTypes) :-
	time(T1),
	accumulate_asubs(ListASub,Accum),
	% Conflate into same function
	apply_replacements(Accum,LubNotSolved),
	% solve
	update_timer(T1,widen),
	get_solver_behavior(TrySolve),
	!, solve_asub(LubNotSolved,TrySolve,Lub).

get_solver_behavior(just_equations):-
	current_pp_flag(output_resources,OUTPUT),(OUTPUT=equations;OUTPUT=both),!.
get_solver_behavior(solve).
 

%------------------------------------------------------------------------%
% sized_types_abs_widen(+,+,-)
% sized_types_abs_widen(ASub0,ASub1,Widen)
%------------------------------------------------------------------------%

sized_types_abs_widen(ASub0,ASub1,Widen) :-
	sized_types_abs_compute_lub([ASub0,ASub1],Widen,_).

%------------------------------------------------------------------------%
% sized_types_abs_glb(+,+,-)
% sized_types_abs_glb(ASub0,ASub1,Glb)
%------------------------------------------------------------------------%

sized_types_abs_glb(ASub0,ASub1,_,ASub0) :-
	asub(ASub1,_,_,[],_,_,_,_), !.
sized_types_abs_glb(ASub0,ASub1,_,ASub1) :-
	asub(ASub0,_,_,[],_,_,_,_), !.
sized_types_abs_glb(ASub0,ASub1,GlbTypes,Glb) :-
	asub(ASub0,Name0/_,(I0,O0),Clauses0,R1,_,SMap0,_),
	asub(ASub1,_Name1/_,(_I1,_O1),Clauses1,R2,_,_,_),
	Clauses0 \= [], Clauses1 \= [],
	% Check if set of arguments are equal
	( Clauses0 = [clause(_K,ClI0,ClO0,_,_)|_], Clauses1 = [clause(_,ClI1,ClO1,_,_)|_], variant((ClI0,ClO0),(ClI1,ClO1))
	-> % We have to accumulate clauses, deleting the empty ones
	(is_trust_info(Clauses0,Clauses1) -> 
	 	resolution_glb(R1,R2,RGlb),!,generate_new_clause(Clauses0,RGlb,Clauses) ; 
	    (
		RGlb = '$unknown',
		up_to_renaming_append(Clauses0,Clauses1,ClausesB),
		delete_empty_clauses(ClausesB,Clauses)
	    )),
	   % Generate new substitution
	   get_next_name(Name0,NewName), !,

	   asub(Glb,NewName,(I0,O0),Clauses,RGlb,[],SMap0,[])
	 ; % Generate an empty entry
	   append(I0,O0,Vars0),
	   sized_types_abs_empty_entry(Vars0,GlbTypes,Empty),
	   % Generate final substitution
	   asub(Empty,_,Args,Clauses,Solution,Already,SMap,_),
	   get_next_name(Name0,NewName), !,
	   asub(Glb,NewName,Args,Clauses,Solution,Already,SMap,[])
	).



:- export(is_trust_info/2).
is_trust_info([clause(Trust1,_,_,_,_)],[clause(Trust2,_,_,_,_)]):-
	Trust1 == trust, Trust2 == trust.
is_trust_info([clause(Trust1,_,_,_,_)|T],[clause(Trust2,_,_,_,_)|TT]):-
	Trust1 == trust, Trust2 == trust,
	is_trust_info(T,TT).

%MKL: I am assuming just one trust assertion per predicate
:- export(generate_new_clause/3).
generate_new_clause([],_,[]).
generate_new_clause([clause(N,I,O,B,T)|R],Map,[clause(N,I,O,NB,T)|RR]):-
	generate_new_clause_body(B,Map,NB),!,
	generate_new_clause(R,Map,RR).

:- export(generate_new_clause_body/3).
generate_new_clause_body([],_,[]).
generate_new_clause_body([assign(L,_R)|Tail],Map,[assign(L,NR)|NTail]):-
	L =.. [Res,(Low,Up)],Map = resolution(_,_,_,MM),
	obtain_closed_form(Low,MM,CLow),
	obtain_closed_form(Up,MM,CUp),
	NR =.. [Res,(CLow,CUp)],
	generate_new_clause_body(Tail,Map,NTail).

obtain_closed_form(Var,[(V,closed_form(Exp))|_TT],Exp):-
	Var == V,!.
obtain_closed_form(Var,[_|TT],Exp):-
	obtain_closed_form(Var,TT,Exp).




:-export(resolution_glb/3).
resolution_glb('$unkown',A,A).
resolution_glb(A,'$unkown',A).
resolution_glb(resolution(Name/N,IArgs1,OSol1,MapRes1),resolution(Name/_,IArgs1,OSol1,MapRes2),resolution(Name/N,IArgs1,OSol1,MapRes)):-
	resolution_glb_map(MapRes1,MapRes2,MapResPre),!,
	simplify_res(MapResPre,MapRes).



resolution_glb_map([],[],[]).
resolution_glb_map([(V1,closed_form(Exp1)),(V2,closed_form(Exp2))|R],[(V1,closed_form(Exp3)),(V2,closed_form(Exp4))|T],[(V1,closed_form(max(Exp1,Exp3))),(V2,closed_form(min(Exp2,Exp4)))|TT]):-
	resolution_glb_map(R,T,TT).



:- use_module(domain(resources/eq_solver/solver/solver)).
simplify_res([],[]).
simplify_res([(V,closed_form(Exp))|T],[(V,closed_form(ExpS))|TR]):-
	simplify(Exp,ExpS),
	simplify_res(T,TR).


delete_empty_clauses([],[]).
delete_empty_clauses([clause(_,_,_,[],_)|R],DelR) :-
	!, delete_empty_clauses(R,DelR).
delete_empty_clauses([clause(K,I,O,E,T)|R],[clause(K,I,O,E,T)|DelR]) :-
	E \= [],
	delete_empty_clauses(R,DelR).

%------------------------------------------------------------------------%
% sized_types_abs_less_or_equal(+,+)
% sized_types_abs_less_or_equal(ASub0,ASub1)
%------------------------------------------------------------------------%

sized_types_abs_less_or_equal(ASub0,ASub1) :-
	sized_types_abs_sort(ASub0,ASub0S),
	sized_types_abs_sort(ASub1,ASub1S),
	sized_types_abs_less_or_equal_(ASub0S,ASub1S).

% A <= B <=> they refer to the same predicate and eq_A >= eq_B
sized_types_abs_less_or_equal_(ASub0,ASub1) :-
	asub(ASub0,Name/_,Args0,Clauses0,Solution0,_,_,_),
	asub(ASub1,Name/_,Args1,Clauses1,Solution1,_,_,_),
	% Check arguments
	variant(Args0,Args1),
	% If solutions are the same and none is unknown, we know they are equal
	( variant(Solution0,Solution1),
	  Solution0 \= '$unknown'
	; make_equal_same_functions(Clauses0,ConfClauses0,[],StepL),
	  make_equal_same_functions(Clauses1,ConfClauses1,StepL,_),
	  up_to_renaming_subset_extra(ConfClauses0,Args0,ConfClauses1,Args1)
	).

make_equal_same_functions([],[],L,L).
make_equal_same_functions([clause(K,IArgs,OArgs,Body,T)|R],[clause(K,IArgs,OArgs,UnifBody,T)|UnifR],InList,OutList) :-
	make_equal_same_functions_body(Body,UnifBody,InList,StepList),
	make_equal_same_functions(R,UnifR,StepList,OutList).

make_equal_same_functions_body([],[],L,L).
make_equal_same_functions_body([assign(X,Y)|R],[assign(X,Y)|UnifR],InL,OutL) :-
	make_equal_same_functions_body(R,UnifR,InL,OutL).
make_equal_same_functions_body([call(Pred,I,O)|R],[call(PredVar,I,O)|UnifR],InL,OutL) :-
	( find_map(InL,Pred,Var)
	-> PredVar = Var,
	   StepL = InL
	 ; PredVar = _,
	   StepL = [(Pred,PredVar)|InL]
	),
	make_equal_same_functions_body(R,UnifR,StepL,OutL).

find_map([(A,B)|_],A,B) :- !.
find_map([_|R],A,B) :- find_map(R,A,B).

%------------------------------------------------------------------------%
% sized_types_abs_identical_abstract(+,+)
% sized_types_abs_identical_abstract(ASub1,ASub2)
%------------------------------------------------------------------------%

% Define the equality as inclusion in both sides
sized_types_abs_identical_abstract(ASub0,ASub1):-
	sized_types_abs_less_or_equal(ASub0,ASub1), !,
	sized_types_abs_less_or_equal(ASub1,ASub0), !.

%------------------------------------------------------------------------%
% sized_types_abs_sort(+,-)
% sized_types_abs_sort(ASub0,ASub1)
%------------------------------------------------------------------------%

sized_types_abs_sort(ASub0,ASub1) :-
	asub(ASub0,Name,Args,Clauses0,Solution,Already0,SizeMapping0,Tests),
	sort(Clauses0,Clauses1),
	sort(Already0,Already1),
	sort(SizeMapping0,SizeMapping1),
	asub(ASub1,Name,Args,Clauses1,Solution,Already1,SizeMapping1,Tests).

%------------------------------------------------------------------------%
% sized_types_abs_call_to_success_fact(+,+,+,+,+,+,-,-,+,+)
% sized_types_abs_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ,CallTypes,PrimeTypes)
%------------------------------------------------------------------------%

sized_types_abs_call_to_success_fact(Sg,Hv,Head,Sv,Call,Proj,Prime,Succ,CallT,EntryT,PrimeT,SuccT):-
	% "Fake" the call by doing the three stages by hand
	sized_types_abs_call_to_entry(Sv,Sg,Hv,Head,_,[],Proj,Entry,ExtraInfo,CallT,EntryT,PrimeT,[]),
	sized_types_abs_exit_to_prime(Sg,Hv,Head,Sv,Entry,ExtraInfo,Prime,EntryT,PrimeT),
	sized_types_abs_extend(Sg,Prime,Sv,Call,Succ,CallT,SuccT).

%------------------------------------------------------------------------%
% sized_types_abs_special_builtin(+)
% sized_types_abs_special_builtin(SgKey)
%------------------------------------------------------------------------%

sized_types_abs_special_builtin('is/2').
sized_types_abs_special_builtin('true/0').
sized_types_abs_special_builtin('fail/0').
sized_types_abs_special_builtin('!/0').
sized_types_abs_special_builtin('arithmetic:>/2').
sized_types_abs_special_builtin('arithmetic:</2').
sized_types_abs_special_builtin('arithmetic:=</2').
sized_types_abs_special_builtin('arithmetic:>=/2').
sized_types_abs_special_builtin('>/2').
sized_types_abs_special_builtin('</2').
sized_types_abs_special_builtin('=</2').
sized_types_abs_special_builtin('>=/2').
sized_types_abs_special_builtin('=/2').
sized_types_abs_special_builtin('\\=/2').
sized_types_abs_special_builtin('=:=/2').

%-------------------------------------------------------------------------
% sized_types_abs_success_builtin(+,+,+,+,-)
% sized_types_abs_success_builtin(Type,Sg,Call,CTypes,Succ)
%-------------------------------------------------------------------------


% TODO: REVISE THE MODIFICATION OF MKL 2014-11-13.
%%%%%%%%%%%%%%%%%%
%BASICALLY I CHANGE THE YETFAILED VALUE
%OF SOME BUILTINS, KEEPING THE SAME
%VALUE THAT THE ANALYSIS CARRIES, EVEN
%IF THAT BUILTIN COULD EVENTUALLY FAILS.
% UPDATED MKL 2015-08-25: I Rolled back this, because it needs revision. 
%%%%%%%%%%%%%%%%%%

% If we have true, a cut or a comparison, we get no new information
sized_types_abs_success_builtin(_,true,Call,_,Call).
sized_types_abs_success_builtin(_,!,Call,_,Call).
sized_types_abs_success_builtin(_,fail,Call,_,Call).
sized_types_abs_success_builtin(_,_A \= _B,Call,_,Call). 
sized_types_abs_success_builtin(_,A =:= B,Call,_,Succ) :-
	asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	ground(A), ground(B), !,
	( A =:= B
	-> asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests)
	 ; % It will always fail, so don't add it as an equation
	   asub(Succ,Name,Args,[],ByNowSol,AlreadySolvedC,SizeMappingC,Tests)
	).
sized_types_abs_success_builtin(_,A = B,Call,_,Succ) :-
	asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	ground(A), ground(B), !,
	( A = B
	-> asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests)
	 ; % It will always fail, so don't add it as an equation
	   asub(Succ,Name,Args,[],ByNowSol,AlreadySolvedC,SizeMappingC,Tests)
	).

%%%MODIFY false in place of true. Wrong.
sized_types_abs_success_builtin(_,_ \= _,Call,_,Call).
	
% If we have is/2, we have to include a new assignment in the equation,
% which will be similar to the one in the right-hand side of is
sized_types_abs_success_builtin(Sv,A =:= B,Call,CTypes,Succ) :-
	sized_types_abs_success_builtin(Sv,A is B,Call,CTypes,Succ).
sized_types_abs_success_builtin(Sv,Out is In,Call,CTypes,Succ):-
	ground(Out),
	find_variable_type(CTypes,In,T),
	dz_type_included(T,num),
	!,
	sized_types_abs_success_builtin(Sv,In=Out,Call,CTypes,Succ).
sized_types_abs_success_builtin(Sv,In is Out,Call,CTypes,Succ):-
	ground(Out),
	find_variable_type(CTypes,In,T),
	dz_type_included(T,num),
	!,
	sized_types_abs_success_builtin(Sv,In=Out,Call,CTypes,Succ).
sized_types_abs_success_builtin(_,Out is In,Call,CTypes,Succ) :-
	find_variable_type(CTypes,Out,T),
	dz_type_included(T,num), !, % It's already a number
	asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	( ( var(In) ; ground(In) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,NewTests).
sized_types_abs_success_builtin(_,Out is In,Call,CTypes,Succ):-
	find_variable_type(CTypes,Out,T),
	dz_type_included(T,vr), % It's already a variable
	!, asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	% 1. Extend the size mappings with Out/num
	set_numeric_type(Num),
	get_size_type(Num,NumSizeType),
	merge_size_types([(Out,NumSizeType)],SizeMappingC,NewMapping),
	% 2. Add new instruction
	create_num_assignment(In,NewMapping,AssignL,AssignR),
	extend_instructions(Clauses,[assign(NumSizeType,num((AssignL,AssignR)))],NewClauses),
	% 3. Add test if possible
	( ( var(In) ; ground(In) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	% Create new substitution
	% Note: calling is/2 does not generate new solutions
	asub(Succ,Name,Args,NewClauses,ByNowSol,AlreadySolvedC,NewMapping,NewTests).
sized_types_abs_success_builtin(Sv,Out=In,Call,CTypes,Succ):-
	% Handle it similarly to exit_to_prime
	asub(Call,Name,(IArgs,OArgs),Clauses,ByNowSol,AlreadySolved,SizeMapping,Tests),
	% 1. Get success types and extend mapping
	etermsvar_project(Sv,CTypes,PTypes),
	etermsvar_call_to_success_builtin('=/2',Out=In,Sv,CTypes,PTypes,STypes),
	find_variable_type(STypes,Out,T),
	get_size_type(T,SizeType),
	merge_size_types([(Out,SizeType)],SizeMapping,NewMapping),
	% 2. Generate backwards unification
	unify_back_s([Out],[Out=In],NewMapping,STypes,STypes,AssignOut),
	extend_instructions(Clauses,AssignOut,Clauses2),
	% 3. Add test if possible
	( ( var(Out) ; ( ground(Out), number(Out)) ) , ( var(In) ; ( ground(In), number(In)) )
	-> NewTests = [Out =:= In|Tests]
	 ; NewTests = Tests
	),
	% 3. Construct the final substitution
	asub(Succ,Name,(IArgs,OArgs),Clauses2,ByNowSol,AlreadySolved,NewMapping,NewTests).
%%MODIFIED. Instead of putting that fails we are keeping the previous value. Maybe is
% wrong.
sized_types_abs_success_builtin(_,Sg,Call,_,Succ) :-
	Sg =.. [F,Out,In],
	( F = 'arithmetic:<' ; F = 'arithmetic:>'
	; F = 'arithmetic:=<' ; F = 'arithmetic:>='
	; F = '<' ; F = '>' ; F = '=<' ; F = '>=' ), !,
	% A comparison predicate may fail
	asub(Call,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,Tests),
	% Add test if possible
	( ( var(Out) ; ( ground(Out), number(Out)) ) , ( var(In) ; ( ground(In), number(In)) )
	-> NewTests = [Sg|Tests]
	 ; NewTests = Tests
	),
	asub(Succ,Name,Args,Clauses,ByNowSol,AlreadySolvedC,SizeMappingC,NewTests).


%------------------------------------------------------------------------%
% sized_types_abs_input_interface(+,+,+,-)
% sized_types_abs_input_interface(InputUser,Kind,StructI,StructO)
%------------------------------------------------------------------------%

% We don't support things right now
sized_types_abs_input_interface(rsize(Arg,Size),perfect,Sizes0,Sizes1) :-
	put_in_front(rsize(Arg,Size),Sizes0,Sizes1).

put_in_front(X,Xs,[X]) :-
	var(Xs), !.
put_in_front(X,Xs,[X|Xs]).

may_be_var(X,X):- ( X=[] ; true ), !.


%---------------------------------------------------------------------------------%
% sized_types_abs_input_user_interface(+,+,-,+)
% sized_types_abs_input_user_interface(InputUser,Qv,CallT,PrimeT,PrimeDet,PrimeNf,ASub,Sg)
%---------------------------------------------------------------------------------%

:- use_module(spec(s_simpspec), [make_atom/2]).
sized_types_abs_input_user_interface(InputUser,Qv,CallT,PrimeT,ASub,Sg):-
	time(T1),
	% InputUser = (ArgInput0),
	InputUser = ArgInput0,
	may_be_var(ArgInput0,ArgInput),
	% 1. Generate a random name
	functor(Sg,Name,_),
	get_next_name(Name,FullName),
	!, % Just one name
	% 2a. Detect input and output
	choose_input_output(Qv,CallT,IArgs,OArgs),
	% 2b. Uglify input and output variables
	normalize_vars(ArgInput,PrimeT,UglyVars),
	select_vars_size_types(UglyVars,IArgs,ISizes),
	get_vars_size_types(OArgs,PrimeT,OSizes),
	merge_size_types(ISizes,OSizes,SizeMapping),
	% 2c. Create assignments
	input_args_assignments(OArgs,SizeMapping,UglyVars,Assigns),
	% 6a. Generate clause
	get_second_components(ISizes,RealISizes),
	get_second_components(OSizes,RealOArgs),
	RealOArgs = RealOSizes,
	Assigns = ClAssigns,
	functor(Sg,Functor,Arity),
	make_atom([Functor,Arity,'trust'],_Key),
	Clause = clause(trust,RealISizes,RealOSizes,ClAssigns,[]),
	% 6b. Generate substitution
	!,
	asub(ASubPre,FullName,(IArgs,OArgs),[Clause],'$unknown',[],SizeMapping,[]),

	update_timer(T1,input_user_interface),
	% ASub = ASubPre.
	get_solver_behavior(TrySolve),
	solve_asub(ASubPre,TrySolve,ASub).
sized_types_abs_input_user_interface(InputUser,Qv,CallT,_PrimeT,_ASub,_Sg) :-
	warning_message("input user interface failed: ~q, ~q, ~q",[InputUser,Qv,CallT]),
	fail.

normalize_vars([],_,[]).
normalize_vars([rsize(Arg,Size)|R],Types,[(Arg,Ugly)|URest]) :-
	find_variable_type(Types,Arg,Ty),
	normalize_rsize(Ty,Size,Ugly),
	normalize_vars(R,Types,URest).

input_args_assignments([],_,_,[]).
input_args_assignments([A|R],SizeMapping,Ugly,[assign(S,U)|AssR]) :-
	get_var_size_type(SizeMapping,A,S),
	get_var_size_type(Ugly,A,U),
	input_args_assignments(R,SizeMapping,Ugly,AssR).


%------------------------------------------------------------------------%
% sized_types_abs_asub_to_native(+,+,-,-,+)
% sized_types_abs_asub_to_native(ASub,Qv,PostProps,CompProps,Types)
%------------------------------------------------------------------------%

sized_types_abs_asub_to_native(ASub,Qv,RSizes,ATypes):-
	sized_types_abs_asub_to_native_(ASub,Qv,RSizes,ATypes).

sized_types_abs_asub_to_native_(ASub,_Qv,[],_ATypes):-
	asub(ASub,_,_,_,Sol,_,_,_),
	Sol \= resolution(_,_,_,_), !.
sized_types_abs_asub_to_native_(ASub,_Qv,RSizes,_ATypes):-
	asub(ASub,_P,(IArgs,OArgs),_,Sol,_,SizeMapping,_),
	Sol = resolution(_,_SIArgs,SOArgs,Solutions),
	% 1. Generate input arguments
	get_size_mappings(IArgs,SizeMapping,ISizes),
	args_to_native(IArgs,ISizes,[],[],RInput),
	% 2. Generate output arguments
	partition_closed_forms(Solutions,Vars,Expr),
	%debug_message("RESULTS>>> ~q",[Expr]),
	args_to_native(OArgs,SOArgs,Vars,Expr,ROutput),
	% 3. Put arguments together
	append(RInput,ROutput,RSizes).
	% 3. Generate solutions

:- use_module(.(eq_solver/solver), [expand/2]).
:- use_module(domain(resources/eq_solver/solver/solver_messages),[show_user_message/1]).
% Generates a set of substitutions of variables to expressions
% If we have a closed form for the variable, we chose it
% Otherwise, we use ? as placeholder
partition_closed_forms([],[],[]).
partition_closed_forms([(V,closed_form(Expr))|R],[V|RestVars],[SExpr|RestExpr]) :-
	!,
	% Simplify
	simplify_min_max_expr(Expr,MExpr0),
	simplify(MExpr0,MExpr),
	catch(
		 expand(MExpr,SExpr),
		 Ex,
		 (
		     show_user_message(Ex), 
		     (Ex = not_implemented(_Op,_Solver,_Exp) -> SExpr=MExpr; throw(Ex))		       )    
	     ), 
	partition_closed_forms(R,RestVars,RestExpr).
partition_closed_forms([(V,unsolved_equations(_,Eqs))|R],[V|RestVars],[eqs(Unsolved)|RestExpr]) :-
	!,
	% Generate better form
	beautify_equations(Eqs,Unsolved),
	partition_closed_forms(R,RestVars,RestExpr).

% Converts the information of variables into size expressions
args_to_native([],_,_,_,[]).
args_to_native([V|R],[S|SizeR],Vars,Exprs,[rsize(V,Prop)|NativeR]) :-
	substitute_vars(S,Vars,Exprs,SubstS),
	denormalize_rsize(SubstS,Prop),
	args_to_native(R,SizeR,Vars,Exprs,NativeR).

%------------------------------------------------------------------------%
% sized_types_abs_unknown_call(+,+,-)
% sized_types_abs_unknown_call(Vars,Call,Succ)
%------------------------------------------------------------------------%

sized_types_abs_unknown_call(_Vars,Call,Call):-!.

%------------------------------------------------------------------------%
% sized_types_abs_unknown_entry(+,-)
% sized_types_abs_unknown_entry(Vars,Entry)
%------------------------------------------------------------------------%

% When unknown, use the default solution and resources values
sized_types_abs_unknown_entry(Vars,ETypes,Entry):-
	get_next_name('$unknown',NewName), !, % Only one name
	get_vars_size_types(Vars,ETypes,Mapping),
	get_size_mappings(Vars,Mapping,IArgs),
	% Solutions are between 0 and Infinity
	asub(Temp,NewName,(Vars,[]),[clause(_,IArgs,[],[],[])],'$unknwon',[],Mapping,[]),
	get_solver_behavior(TrySolve),
	solve_asub(Temp,TrySolve,Entry).

%------------------------------------------------------------------------%
% sized_types_abs_empty_entry(+,-)
% sized_types_abs_empty_entry(Vars,Entry)
%------------------------------------------------------------------------%

% A bottom entry has '$unknown' as its resolution
sized_types_abs_empty_entry(Vars,ETypes,Entry):-
	get_next_name('$unknown',NewName), !, % Only one name
	get_vars_size_types(Vars,ETypes,Mapping),
	asub(Entry,NewName,(Vars,[]),[],'$unknown',[],Mapping,[]).

% This counter is used internally to generate new names
:- data counter/1.
counter(0).

get_next_name(F,F/N) :-
	current_fact(counter(N)), !,
	retract_fact(counter(N)),
	N1 is N+1,
	asserta_fact(counter(N1)).

%--------------------------------------------
% Predicates for substituting predicate names
%--------------------------------------------

% Merge all the information in the clauses:
% - Make all the clauses have the same name
% - Merge the previous solutions
accumulate_asubs(['$top'],'$top') :- !.
accumulate_asubs([A],A) :- !.
accumulate_asubs([A|R],Accum) :-
	accumulate_asubs(R,RAccum),
	( A = '$top'
	-> Accum = RAccum
	 ; ( RAccum = '$top'
	   -> Accum = A
	    ; % Both are regular abstract substitutions
	      asub(RAccum,Name,Args,RClauses,_,RAlready,RSizeMapping,_),
	      asub(A,_AName,_AArgs,AClauses,_ASolution,AAlready,ASizeMapping,_),
	      % Join clauses, already solved, types and previously called
	      up_to_renaming_append(AClauses,RClauses,AccumClauses),
	      unifiable_append(RAlready,AAlready,AccumAlready),
	      merge_size_types(RSizeMapping,ASizeMapping,AccumSizeMapping),
	      % Put everything together
	      asub(Accum,Name,Args,AccumClauses,'$unknown',AccumAlready,AccumSizeMapping,[])
	   )
	).

or_(false,false,false) :- !.
or_(true,_,true) :- !.
or_(false,true,true) :- !.

:- data last_replacement/4.

clean_all_caches :-
	clean_equation_names_cache,
	retractall_fact(last_replacement(_,_,_,_)),
	clean_internal_equiv_types,
	clean_size_type_cache.

get_last_replacements(F,I,O,N) :-
	current_fact(last_replacement(F,I,O,N)).

put_asub_last_replacement(ASub,F/N) :-
	asub(ASub,_,(IArgs,OArgs),_,_,_,SizeMapping,_),
	get_size_mappings(IArgs,SizeMapping,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes),
	OSizesSolRes = OSizes,
	put_last_replacement(F,ISizes,OSizesSolRes,N).

put_last_replacement(F,IArgs,OArgs,N) :-
	copy_term(last_replacement(F,IArgs,OArgs,_), C),
	retractall_fact(C),
	asserta_fact(last_replacement(F,IArgs,OArgs,N)).

find_last_replacement(F,IArgs,OArgs,N) :-
	copy_term(last_replacement(F,IArgs,OArgs,_), last_replacement(FP,IArgsP,OArgsP,N)),
	current_fact(last_replacement(FP,IArgsP,OArgsP,N)).

find_last_asub_replacement(ASub,N) :-
	asub(ASub,F/_,(IArgs,OArgs),_,_,_,SizeMapping,_),
	get_size_mappings(IArgs,SizeMapping,ISizes),
	get_size_mappings(OArgs,SizeMapping,OSizes),
	OSizesSolRes = OSizes,
	find_last_replacement(F,ISizes,OSizesSolRes,N).

apply_replacements(ASub,ASubR) :-
	find_last_asub_replacement(ASub,NewN),
	!,
	asub(ASub,F/_,(IArgs,OArgs),Clauses,_,Already,SizeMapping,Tests),
	update_clauses(Clauses,NewClauses),
	create_old_to_new_mapping(Already,Mapping),
	update_already_solved(Already,Mapping,NewAlready),
	asub(ASubR,F/NewN,(IArgs,OArgs),NewClauses,'$unknown',NewAlready,SizeMapping,Tests).
apply_replacements(ASub,ASubR) :-
	asub(ASub,F/N,(IArgs,OArgs),Clauses,_,Already,SizeMapping,Tests),
	update_clauses(Clauses,NewClauses),
	asub(ASubR,F/N,(IArgs,OArgs),NewClauses,'$unknown',Already,SizeMapping,Tests).

% Change the clauses to point to new function
update_clauses([],[]).
update_clauses([clause(K,I,O,Instr,T)|R],Out) :-
	update_clauses(R,UpR),
	update_clause(Instr,UpInstr),
	( clause_member(clause(K,I,O,UpInstr,T),UpR)
	-> Out = UpR
	 ; Out = [clause(K,I,O,UpInstr,T)|UpR]
	).

update_clause([],[]).
update_clause([call(F/_,I,O)|R],[call(F/N,I,O)|UpR]) :-
	find_last_replacement(F,I,O,N),
	!, update_clause(R,UpR).
update_clause([E|R],[E|UpR]) :-
	!, update_clause(R,UpR).

clause_member(clause(K,I,O,Eqs,T),[clause(K,I0,O0,Eqs0,T)|_]) :-
	copy_term((I,O,Eqs),(I1,O1,Eqs1)),
	copy_term((I0,O0,Eqs0),(I2,O2,Eqs2)),
	varset((I1,O1),Vars),
	generate_aux_names(Vars),
	(I1,O1) = (I2,O2),
	variant((I1,O1,Eqs1),(I2,O2,Eqs2)), !.
clause_member(C,[_|R]) :-
	!, clause_member(C,R).

generate_aux_names([]).
generate_aux_names([V|R]) :-
	get_next_name(var,Var),
	V = Var, !,
	generate_aux_names(R).

create_old_to_new_mapping([],[]).
create_old_to_new_mapping([resolution(Name,RIArgs,ROArgs,Sols)|R],Mapping) :-
	create_old_to_new_mapping(R,RMapping),
	( Name = F/M, find_last_replacement(F,RIArgs,ROArgs,N), M \= N
	-> get_solution_fn_names(F/N,RIArgs,ROArgs,FnNames),
	   zip_names(Sols,FnNames,NewMapping),
	   append(NewMapping,RMapping,Mapping)
	 ; Mapping = RMapping
	).

zip_names([],[],[]).
zip_names([(_,unsolved_equations(_,[equation(F,_,_,_)|_]))|R],[Name|NameR],[(F,Name)|ZipR]) :-
	!,zip_names(R,NameR,ZipR).
zip_names([_|R],[_Name|NameR],ZipR) :-
	zip_names(R,NameR,ZipR).

% Get the names of the functions that will replace the old ones
get_solution_fn_names(Name,ISizes,OSizes,FnNames) :-
	copy_term(ISizes,ISizesN),
	copy_term(OSizes,OSizesN),
	FakePred = predicate(Name,ISizesN,OSizesN,[clause(_,ISizesN,OSizesN,[call(Name,ISizesN,OSizesN)],[])]), get_solver_behavior(TrySolve),
	pred_solve(FakePred,[],0,true,false,TrySolve,resolution(_,_,_,Sols)),
	get_solution_fn_names_vars(Sols,FnNames).

get_solution_fn_names_vars([],[]).
get_solution_fn_names_vars([(_,closed_form(_))|R],FR) :-
	get_solution_fn_names_vars(R,FR).
get_solution_fn_names_vars([(_,unsolved_equations(_,[equation(F,_,_,_)]))|R],[F|FR]) :-
	get_solution_fn_names_vars(R,FR).

% Update the equations and closed forms with the new mappings
update_already_solved([],_,[]).
update_already_solved([resolution(G,I,O,Sols)|R],Mapping,[resolution(G,I,O,NewSols)|UpR]) :-
	!, update_functors(Sols,Mapping,NewSols),
	update_already_solved(R,Mapping,UpR).

update_functors(X,_,X) :-
	var(X), !.
update_functors(Functor,Mapping,NewFunctor) :-
	nonvar(Functor),
	Functor =.. [F|Args],
	update_functors_list(Args,Mapping,NewArgs),
	( member((F,NewF),Mapping)
	-> NewFunctor =.. [NewF|NewArgs]
	 ; NewFunctor =.. [F|NewArgs]
	).

update_functors_list([],_,[]).
update_functors_list([E|R],Mapping,[UpE|UpR]) :-
	update_functors(E,Mapping,UpE),
	update_functors_list(R,Mapping,UpR).
