:- module(static_profiling_tr, [sentence/2], [dcg, assertions, datafacts]).

%:- use_module(library(aggregates), [findall/3]).
:- use_module(library(hiordlib), [maplist/3]).
:- use_module(library(lists), [append/3]).
:- use_module(library(terms), [atom_concat/2]).

% kludge
:- use_module(library(compiler/c_itf), [location/3]).



sentence(0, [0]):- !, reset. 
sentence(end_of_file, Output):- !,
	conclude(Output, [end_of_file]). 
sentence((:- Directive), Output):- !,
	directive(Directive, Output, []).
sentence((Head :- Body), Output):- !,
	clause(Head, Body, Output, []).
sentence((Fact), Output):-
	clause(Fact, true, Output, []).


:- data(clause/2).
:- data(defined/4).
:- data(cost_center/1).
:- data(resources/1). 
:- data(resource_pred/4).
:- data(sentence/1).


reset :-
	retractall_fact(clause(_, _)),
	retractall_fact(defined(_, _, _, _)),
	retractall_fact(cost_center(_)),
	retractall_fact(resources(_)),
	retractall_fact(resource_pred(_, _, _, _)),
	retractall_fact(sentence(_)),
	assertz_fact(resources([])).

directive(resource(R)) -->
	{ retract_fact(resources(L)), assertz_fact(resources([R|L])) }.
directive(cost_center(F/N)) -->
	{
	    current_fact(cost_center(F/N)) -> true;
	    assertz_fact(cost_center(F/N))
	}.
directive((trust comp F/N + costb(R, U, L))) -->
	{
	    assertz_fact(resource_pred(F/N, R, U, L))
	}.

clause(Head, Body) -->
	{ 
	    assertz_fact(clause(Head, Body)), 
	    functor(Head, F, N),
	    ( current_fact(defined(F, N, _, _)) ->
	        true
	    ; c_itf:location(_, Line, _),
	      atom_number(A, N),
	      atom_concat([F, '_', A, '__with_costs'], FTrans),
	      assertz_fact(defined(F, N, FTrans, Line))  )
	}.


conclude -->
	{ findall(F/N-Line, current_fact(defined(F, N, _, Line)), Preds) },
%	line_number(Preds),
	{
	    findall(Pred, current_fact(cost_center(Pred)), CostCenters),
	    retract_fact(resources(Res)),
	    product_list(CostCenters, Res, NewRes, Res),
	    assertz_fact(resources(NewRes))
	},
	[(:- static_profiling(resources(NewRes)))],
	bridge_preds(Preds),
	translate_clauses.  

line_number([]) --> [].
line_number([F/N-Line|T]) -->
	[ (:- static_profiling(line(F/N, Line))) ],
	line_number(T).

bridge_preds([]) --> [].
bridge_preds([F/N-_|T]) -->
	{
	    functor(Head, F, N),
	    varCosts(Costs),
	    translate_atom(Head, Costs, Body)
	},
	[(Head :- Body )],
	bridge_preds(T).

translate_clauses -->
	{
	    retract_fact(clause(Head, Body)),!,
	    functor(Head, F, N),
	    current_fact(resources(Rs)),
	    maplist(any2zero, Rs, ZeroCosts),
	    translate_goal(Body, ZeroCosts, Costs, Body_),
	    finalize_cost(Rs, F/N, Costs, Cost, G),
	    ( G = [] -> Body__ = Body_ ; list2goal(G, G_), Body__ = (Body_, G_) ),
	    translate_atom(Head, Cost, Head_)
	},
	[(Head_ :- Body__)], 
	translate_clauses.
translate_clauses --> [].

translate_atom(Atom, Costs, Atom_):-
	functor(Atom, F, N),
	Atom =.. [F|Args],
	append(Args, Costs, Args_),
	current_fact(defined(F, N, F_, _)),
	Atom_ =.. [F_|Args_].

translate_goal((A, B), InCosts, OutCosts, (A_, B_)):- !,
	translate_goal(A, InCosts, Costs_, A_),
	translate_goal(B, Costs_, OutCosts, B_).
translate_goal(G, Costs, Costs, G):-
	\+ (functor(G, F, N), current_fact(defined(F, N, _, _))), !.
translate_goal(G, InCosts, OutCosts, G_):-
	varCosts(GCosts),
	translate_atom(G, GCosts, G_),
	add_costs(InCosts, GCosts, OutCosts).


varCosts(L):-
	current_fact(resources(L_)),
	maplist(any2var, L_, L).
nullCosts(L):-
	current_fact(resources(L_)),
	maplist(any2zero, L_, L).

any2zero(_, 0).
any2var(_, _).

finalize_cost([], _, [], [], []).
finalize_cost([R|Rs], Pred, [H|T], [H_|T_], G):-
	( R = Pred:S -> 
	    replace2(S, H0, S, 0, Rs, T, Rs0, T0),
	    (  H == 0 ->
	       Sum = H
	    ;  H0 == 0 ->
	       Sum = H 
	    ;
		Sum = H + H0
	    ),
	    G = [(H_ is Sum)|G_],
	    finalize_cost(Rs0, Pred, T0, T_, G_)
	; ( resource_pred(Pred, R, U, L) ->
%	      G = [U_ = U, L_ = L, between(H_, U_, L_)]
	    G = [between(H_, U, L)]
%	  ; atomic(H) ->
%	      H = H_, G = G_
	  ; G = [(H_ is H)|G_]
	  ),
	  finalize_cost(Rs, Pred, T, T_, G_)
	).
	

add_costs([CostA|CostsA], [CostB|CostsB], [Cost|Costs]):-
	( CostA == 0 ->
	    Cost = CostB;
	  Cost = CostA + CostB),
	add_costs(CostsA, CostsB, Costs).
add_costs([], [], []).


product_list([], _) --> [].
product_list([A|As], Bs) --> 
	product_list_(A, Bs),
	product_list(As, Bs).

product_list_(_, []) --> [].
product_list_(A, [B|Bs]) -->
	[A:B],
	product_list_(A, Bs).

list2goal([A], A).
list2goal([H|T], (H, G)):-
	list2goal(T, G).

replace2(X, Y, X_, Y_, [X|T], [Y|S], [X_|T],[Y_| S]).
replace2(X, Y, X_, Y_, [HT|T], [HS|S], [HT|T_], [HS|S_]):-
	replace2(X, Y, X_, Y_, T, S, T_, S_).


%% kludge: work arround for wrong hight order in ciaopp %%%


:- use_module(library(lists), [reverse/2]).

:- data found/1.

findall(_, _, _):-
	set_fact(found([])),
	fail.
findall(Pattern, current_fact(F), _):-
	current_fact(F),
	current_fact(found(Tail)),
	set_fact(found([Pattern|Tail])),
	fail.
findall(_, _, List):-
	retract_fact(found(List_)),
	reverse(List_, List).
