% (included file)

% ---------------------------------------------------------------------------
% Printer for xC language

:- use_module(engine(io_basic)).
:- use_module(library(stream_utils), [get_line/1]).
:- use_module(library(dec10_io)).
:- use_module(library(write)).

:- use_module(ciaopp(p_unit), [pr_key_get/1]).
:- use_module(ciaopp(p_unit/p_printer), [with_mod_syntax/1, get_filtered_assertions/2]).
:- use_module(ciaopp(preprocess_flags), [current_pp_flag/2]).
% :- use_module(ciaopp(frontend_driver), [analysis_info_to_assertions/0]).

:- import(xc_assertionsxo, [pl2xc_translate_assertions/3]).

output_ext('.xc').
% TODO: why not? [JF]
% output_by_ext('.xc', Stream) :-
%	current_pp_flag(output_lang,intermediate),!,
%	output_by_ext('.pl', Stream). 
output_by_ext('.xc', Stream) :- !,
 %	curr_file(InputFile, _Module),
	remove_pragma_checks(BareInputFile),
	see(BareInputFile),
	( current_pp_flag(dump_ai, on) -> analysis_info_to_assertions ; true ),
	print_assertions_xc(Stream),
	% write_types(Stream),
	write_source_file(Stream),
	seen.

output_ext('.c').
output_by_ext('.c', Stream) :- !,
 %	curr_file(InputFile, _Module),
	remove_pragma_checks(BareInputFile),
	see(BareInputFile),
	( current_pp_flag(dump_ai, on) -> analysis_info_to_assertions ; true ),
	print_assertions_xc(Stream),
	% write_types(Stream),
	write_source_file(Stream),
	seen.

output_ext('.asm').
output_by_ext('.asm', Stream) :- !,
 	curr_file(InputFile, _Module),
	see(InputFile),
	( current_pp_flag(dump_ai, on) -> analysis_info_to_assertions ; true ),
	print_assertions_xc(Stream),
	write_types(Stream),
	write_source_file(Stream),
	seen.

:- use_module(library(process),[process_call/3]).

%% Removes from the xc source file the checks assertions, because they will be
%% either replaced by the results of the verification, or written back again.
% TODO: do not use external 'sed' program!
remove_pragma_checks(BareInputFile):-
	curr_file(InputFile,_M),
	atom_concat(InputFile,'.clean',BareInputFile),
	process_call(path(sed), ['/pragma check/d', InputFile], [status(0),stdout(file(BareInputFile))]).

write_source_file(WStream):-
	get_line(Line),
	(Line=end_of_file
	-> true
	; 
	    atom_codes(X,Line),
	    write(WStream, X),
	    nl(WStream),
	    write_source_file(WStream)
	).

print_assertions_xc(S) :-
	with_mod_syntax(print_assertions_xc_(S)).

print_assertions_xc_(S):-
	( % (failure-driven loop)
	  pr_key_get(Key),
	    print_from_key_xc(S, Key),
	    fail
	; true
	).

print_from_key_xc(S, Key) :- !,
	get_filtered_assertions(Key, AsFiltered),
	% Check if Key has an entry assertion
	( member(as${type => entry}, AsFiltered) -> true ),
	%
	pl2xc_translate_assertions(AsFiltered, XcAsList, []),
	( % (failure-driven loop)
          member(A, XcAsList),
	    xc_print_sentence(A, S),
	    nl(S),
	    fail
	; true
	).

xc_print_sentence(As,S) :- As = xc_assertion(Status, Scope, Hypothesis, Conclusion), !,
	numbervars(As, 0, _), 
%	display(S, 'energy consumption: '),
	display(S, '#pragma '),
	display(S, Status),
	display(S, ' '),
	write(S, Scope),
	write(S, ' : ('),
	( Hypothesis \= true ->
	    write_xc_constraint(S, Hypothesis,no),
	    write(S, ') ==> (')
	; true
	),
	write_xc_constraint(S, Conclusion,no),
	display(S, ')'),
	nl(S).	
	
%% MKL: new version considering disjunction
write_xc_constraint(S, (A;B),Parenthesized) :- !,
	(Parenthesized == yes -> display(S, '(') ; true),
	write_xc_constraint(S, A, yes),
	display(S, ' || '), 
	write_xc_constraint(S, B, yes),
	(Parenthesized == yes -> display(S, ')') ; true).
write_xc_constraint(S, (A,B),Parenthesized) :- !,
	(Parenthesized == yes -> display(S, '(') ; true),
	write_xc_constraint(S, A,yes),
	display(S, ' && '),
	write_xc_constraint(S, B,yes),
	(Parenthesized == yes -> display(S, ')') ; true).
write_xc_constraint(S, (A =< B),_) :-
	write(S, A), 
	display(S, ' <= '), 
	write(S, B).

% write_xc_constraint(S, (A,B)) :- !,
% 	write_xc_constraint(S, A),
% 	display(S, ' && '),
% 	write_xc_constraint(S, B).
% write_xc_constraint(S, (A,B)) :- !,
% 	write_xc_constraint(S, A),
% 	display(S, ' && '),
% 	write_xc_constraint(S, B).
% write_xc_constraint(S, (A =< B)) :-
% 	write(S, A), 
% 	display(S, ' <= '), 
% 	write(S, B).

