/*
 *  tabling.c
 *
 *  Support for tabling.
 *
 *  
 */


#include <stdlib.h>

#include <ciao_prolog.h>

#include <ciao/tabling.h>

#if defined(TABLING)

tagged_t tabling_adjust;
tagged_t tabling_overflow;
tagged_t tabling_increase;
#if defined(DEBUG)
tagged_t tabling_debug;
tagged_t tabling_debug_tries;
tagged_t tabling_trace;
tagged_t tabling_print;
tagged_t tabling_bypass;
#endif

tagged_t global_table_size;
tagged_t tabling_stack_size;    

CVOID__PROTO(heap_overflow_adjust_tabling, intmach_t heap_factor);
CVOID__PROTO(stack_overflow_adjust_tabling, intmach_t stack_factor);
CVOID__PROTO(choice_overflow_adjust_tabling, intmach_t choice_factor);
CVOID__PROTO(reinit_tabling);

void init_global_table(int size);
void dealloc_global_table(void);
void init_tabling_stack(int size);
void dealloc_tabling_stk(tagged_t* ptr);

stack_block_t *current_global_table = NULL;
stack_block_t *current_tabling_stack = NULL;
//stack_block_t *zombi_tabling_stack = NULL;

void init_global_table(int size) {
  stack_block_t *new_block;
					
  new_block = (stack_block_t*) malloc (size * sizeof(tagged_t*));

  if (new_block) {
    new_block->bck = current_global_table;				
    new_block->size = size;			
    current_global_table = new_block;
    
    global_table = (tagged_t*)&(current_global_table->mem_ptr);		       
    global_table_free = global_table;					
    global_table_end = (tagged_t *)current_global_table +  current_global_table->size;		
    
#if defined(DEBUG)
    if (tabling_debug == atom_on)
      printf("\t Global Table new block %10d size = %10d bck = %10d\n", 
	     (int)current_global_table, 
	     current_global_table->size, 
	     (int)current_global_table->bck);
#endif
  }
  else global_table = 0;
}

void init_tabling_stack(int size) {
  /*if (zombi_tabling_stack) {
#if defined(DEBUG)
    if (tabling_debug == atom_on)
      printf("\t Use zombi %10d\n", (int)zombi_tabling_stack);
    printf("|");
#endif
    current_tabling_stack = zombi_tabling_stack;
    zombi_tabling_stack = NULL;
  }
  else {*/
    stack_block_t *new_block;
					
    new_block = (stack_block_t*) malloc (size * sizeof(tagged_t*));

    if (new_block) {
      new_block->bck = current_tabling_stack;				
      new_block->size = size;		
      current_tabling_stack = new_block;
    }
    else {
      tabling_stack = 0;
      return;
    }
    //}

  tabling_stack = (tagged_t*)&(current_tabling_stack->mem_ptr);		       
  tabling_stack_free = tabling_stack;					
  tabling_stack_end = (tagged_t *)current_tabling_stack + current_tabling_stack->size;
  
#if defined(DEBUG)
  if (tabling_debug == atom_on)
    printf("\t Tabling Stack new block %10d size = %10d bck = %10d\n", 
	   (int)current_tabling_stack, 
	   current_tabling_stack->size, 
	   (int)current_tabling_stack->bck); 
#endif 
}

void dealloc_global_table(void) {
#if defined(DEBUG)
  if (tabling_debug == atom_on) {
    printf("dealloc global table\n");						
    printf("current global table = %10d bck = %10d\n",
	   (int)current_global_table, 
	   (int)current_global_table->bck); 
  }
#endif

  stack_block_t *aux = current_global_table;				

  while (current_global_table->bck) {					
    current_global_table = current_global_table->bck;
#if defined(DEBUG)
    if (tabling_debug == atom_on)			
      printf("Global table - free(%10d)\n", (int)aux);	
#endif				
    free((tagged_t*)aux);								
    aux = current_global_table;						
  }									

  global_table = (tagged_t *)&(current_global_table->mem_ptr);		
  global_table_free  = global_table;					
  global_table_end = (tagged_t *)current_global_table + current_global_table->size; 
  global_table_size = current_tabling_stack->size;
}

void dealloc_tabling_stk(tagged_t* ptr) {
#if defined(DEBUG)
  if (tabling_debug == atom_on) {
    printf("dealloc tabling stack ptr = %10d\n", (int)ptr);
    printf("current tablig stack = %10d bck = %10d\n",
	   (int)current_tabling_stack, 
	   (int)current_tabling_stack->bck); 
  }
#endif

  stack_block_t *aux = current_tabling_stack;

  while ( !(ptr >= &(aux->mem_ptr) && ptr <= ((tagged_t *)aux + aux->size)) ) {
    /*if (zombi_tabling_stack) {
      free((tagged_t*)zombi_tabling_stack);
#if defined(DEBUG)
      if (tabling_debug == atom_on)
	printf("Tabling stack - free(%10d)\n", (int)zombi_tabling_stack);
#endif
    }
    zombi_tabling_stack = current_tabling_stack;*/
    if (current_tabling_stack->bck)
      current_tabling_stack = current_tabling_stack->bck;
    else
      SERIOUS_FAULT(" Incorrect ptr for Tabling Stack Blocks");
    free((tagged_t *)aux);
    aux = current_tabling_stack;
  }

  tabling_stack = (tagged_t*)&(current_tabling_stack->mem_ptr);		       
  tabling_stack_free = ptr;					
  tabling_stack_end = (tagged_t *)current_tabling_stack + current_tabling_stack->size;
  tabling_stack_size = current_tabling_stack->size;
}

CVOID__PROTO(heap_overflow_adjust_tabling, intmach_t heap_factor) {


}
CVOID__PROTO(stack_overflow_adjust_tabling, intmach_t stack_factor) { 


}
CVOID__PROTO(choice_overflow_adjust_tabling, intmach_t choice_factor) { 

      
}

/*  To use c functions from tabling  */
CVOID__PROTO(reinit_tabling) {
  tagged_t *junk, key;
  definition_t *d = NULL;

  key = MakeString("tabling_rt:reinit_tabling");
  d = find_definition(predicates_location, key, &junk, FALSE);
  if (d != NULL) d->code.cinfo(Arg);
}
#endif

